/* 
 * tng_misc.h: miscellaneous Samba-TNG macros.
 * Copyright 2003 by Peter Samuelson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef _TNG_MISC_H
#define _TNG_MISC_H

#define IS_BITS_SET_ALL(var,bit) (((var)&(bit))==(bit))
#define IS_BITS_SET_SOME(var,bit) (((var)&(bit))!=0)
#define IS_BITS_CLR_ALL(var,bit) (((var)&(bit))==0)
#define IS_BITS_CLR_SOME(var,bit) (((var)&(bit))!=(bit))

/*
 * STRIP_CONST lets you initialise a non-const pointer from a const
 * pointer, suppressing the usual compiler warning.
 *
 * You should ONLY use this if:
 *
 * (a) You are SURE that the contents of the pointer will NOT be
 *     modified, but
 * (b) There is no clean way to let the compiler know this.
 *
 * In particular, please do not use this macro as a quick fix to
 * suppress legitimate warnings.  It might be best to add a
 * comment at each instance, briefly explaining why it is safe.
 *
 * Usage: char *ch = STRIP_CONST(some_const_char_ptr);
 */
#if defined(__GNUC__) && __GNUC__ >= 2
#define STRIP_CONST(var) ({ \
	union { void **newptr; typeof(var) *oldptr; } _strip_c; \
	_strip_c.oldptr = &(var); \
	*_strip_c.newptr; \
})
#else
#define STRIP_CONST(var) (var)
#endif


/* Detect overflow when adding 2 unsigned integers */
#define SIZE_OVERFLOW(a,b) ((size_t)(a) + (size_t)(b) < (size_t)(a))

/* Detect if the unsigned integer a+b overflows c */
#define SIZE_OVERFLOW3(a,b,c) (SIZE_OVERFLOW(a,b) || (size_t)(a) + (size_t)(b) > (size_t)(c))

/* Detect if a pointer offset will wrap around */
#define PTR_OVERFLOW(ptr, off) ((ptr) + (size_t)(off) < (ptr))


/*
 * Memory allocation helpers (idea and names from glib):
 *
 * g_malloc(count)              = malloc
 * g_new(type, count)           = g_malloc(count * sizeof type)
 * g_new0(type, count)          = g_new but using calloc
 * g_new1(type)                 = g_new(type, 1)
 * g_realloc(ptr, count)        = realloc
 * g_renew(type, ptr, count)    = g_realloc(ptr, count * sizeof type)
 * g_xrenew(type, &ptr, count)
 *     Returns bool and does not damage the pointer on failure.
 *     Thus you can either free the original pointer or keep it,
 *     your choice, without having to use a temp variable.
 *
 * Do not use the g__ names directly.
 */
void *g__new(size_t, size_t);
void *g__new0(size_t, size_t);
BOOL g__xrenew(size_t, void *, size_t);

#define g_new(type, count) \
	((type *) g__new(sizeof(type), count))
#define g_new0(type, count) \
	((type *) g__new0(sizeof(type), count))
#define g_new1(type) g_new(type, 1)
#define g_malloc(count) ((void *)g_new(char, count))
#define g_xrenew(type, memp, count) \
	((*memp) = (type *)(*memp), \
	 g__xrenew(sizeof(type), memp, count))
#define g_renew(type, mem, count) \
	((type *)(g_xrenew(type, &(mem), count) ? (mem) : NULL))
#define g_realloc(mem, count) g_renew(char, mem, count)


/* Debugging helpers */
#define g_return_val_if_fail(expr, val) \
	do { \
		if (! (expr)) \
		{ \
			DEBUG(1, ("%s:%d(%s): %s failed\n", \
				  __FILE__, __LINE__, __func__, \
				  #expr)); \
			return val; \
		} \
	} while(0)
#define g_return_if_fail(expr) \
	do { \
		if (! (expr)) \
		{ \
			DEBUG(1, ("%s:%d(%s): %s failed\n", \
				  __FILE__, __LINE__, __func__, \
				  #expr)); \
			return; \
		} \
	} while(0)


#endif /* _TNG_MISC_H */
