
/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-2001,
 *  Copyright (C) Elrond                            2000
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "includes.h"
#include "libsamba.h"
#include "tng_misc.h"
#include "rpc_parse.h"
#include "dlinklist.h"
#include "handle.h"
#include "others.h"


#undef  DBGC_CLASS
#define DBGC_CLASS DBGC_RPC


struct policy
{
	struct policy *next, *prev;
	int pnum;
	BOOL open;
	POLICY_HND pol_hnd;
	uint32 access_mask;
	uint16 vuid;

	Handle *handle;
};



static uint16 get_policy_vuid(struct policy_cache *cache, const POLICY_HND *hnd);



/****************************************************************************
  initialise policy handle states...
****************************************************************************/
static struct policy_cache *init_policy_cache(void)
{
	struct policy_cache *cache = g_new(struct policy_cache, 1);
	if (cache != NULL)
	{
		cache->Policy = NULL;
	}
	return cache;
}

/****************************************************************************
 free policy handle states...
****************************************************************************/
#ifndef NOONS_SMALL_BUILD
static void free_policy_cache(struct policy_cache *cache)
{
	free(cache);
}
#endif
/****************************************************************************
  i hate this.  a global policy handle cache.  yuk.
****************************************************************************/
struct policy_cache *get_global_hnd_cache(void)
{
	static struct policy_cache *cache = NULL;

	if (cache == NULL)
	{
		cache = init_policy_cache();
	}
	return cache;
}

/****************************************************************************
  find policy by handle
****************************************************************************/
static struct policy *find_policy(struct policy_cache *cache,
				  const POLICY_HND *hnd)
{
	struct policy *p;

	if (cache == NULL)
	{
		DEBUG(0, ("find_policy: NULL cache\n"));
		SMB_ASSERT(False);
	}

	if (hnd == NULL)
	{
		DEBUG(0, ("find_policy: NULL handle\n"));
		SMB_ASSERT(False);
		return NULL;
	}

	for (p = cache->Policy; p; p = p->next)
	{
		DEBUG(10, ("Compare policy hnd[%x] ", p->pnum));
		dump_data(10, (const char *)hnd, sizeof(*hnd));
		if (memcmp(&p->pol_hnd, hnd, sizeof(*hnd)) == 0)
		{
			DEBUG(4, ("Found policy hnd[%x] ", p->pnum));
			dump_data(4, (const char *)hnd, sizeof(*hnd));
			return p;
		}
	}

	DEBUG(4, ("cache->Policy not found: "));
	dump_data(4, (const char *)hnd, sizeof(*hnd));

	return NULL;
}

/****************************************************************************
  set the name of a POLICY_HND
****************************************************************************/
BOOL policy_hnd_set_name(struct policy_cache *cache,
			 POLICY_HND *hnd, const char *name)
{
	struct policy *p = find_policy(cache, hnd);
	if (!p)
	{
		DEBUG(3, ("Error setting name for policy\n"));
		return False;
	}
	handle_set_name(p->handle, name);

	return True;
}

/****************************************************************************
  get the name of a POLICY_HND
****************************************************************************/
static const char *pol_get_name(const struct policy *p)
{
	const char *n;
	if (!p)
	{
		return "(NULL)";
	}
	n = handle_get_name(p->handle);
	if (n)
	{
		return n;
	}
	return "";
}

/****************************************************************************
  get the name of a POLICY_HND, public interface
****************************************************************************/
const char *policy_hnd_get_name(struct policy_cache *cache,
				const POLICY_HND *hnd)
{
	const char *name;
	struct policy *p = find_policy(cache, hnd);

	if (!p)
	{
		DEBUG(3, ("Error getting name for policy\n"));
		return "(invalid POLICY_HND)";
	}
	name = pol_get_name(p);
	DEBUG(4, ("policy(pnum=%x %s): getting name\n", p->pnum, name));
	return name;
}


/****************************************************************************
  find first available policy slot.  copies a policy handle for you.
****************************************************************************/
BOOL dup_policy_hnd(struct policy_cache *cache,
		    POLICY_HND *hnd, const POLICY_HND *from)
{
	struct policy *p = find_policy(cache, from);

	if (!p || !p->open)
	{
		return False;
	}
	DEBUG(5, ("policy(pnum=%x %s): Duplicating policy\n",
		  p->pnum, pol_get_name(p)));
	return register_policy_hnd(cache, p->vuid, hnd, p->access_mask);
}

/****************************************************************************
  find first available policy slot.  creates a policy handle for you.
****************************************************************************/
BOOL register_policy_hnd(struct policy_cache *cache,
			 uint16 vuid,
			 POLICY_HND *hnd, uint32 access_mask)
{
	struct policy *p;
	static int count = 1;

	p = g_new(struct policy, 1);
	if (!p)
	{
		DEBUG(0, ("ERROR: out of memory!\n"));
		return False;
	}

	ZERO_STRUCTP(p);

	p->open = True;
	p->pnum = count++;
	p->access_mask = access_mask;
	p->vuid = vuid;
	p->handle = handle_new(NULL);

	DLIST_ADD(cache->Policy, p);

	DEBUG(4, ("Opened policy hnd[%x] ", p->pnum));
	DEBUG(10, ("register_policy_hnd: vuser [%x]\n", p->vuid));

	p->pol_hnd = *hnd;
	dump_data(4, (char *)hnd, sizeof(*hnd));

	return True;
}

/****************************************************************************
  find first available policy slot.  creates a policy handle for you.
****************************************************************************/
BOOL open_policy_hnd(struct policy_cache *cache,
		     uint16 vuid,
		     POLICY_HND *hnd, uint32 access_mask)
{
	create_pol_hnd(hnd);
	return register_policy_hnd(cache, vuid, hnd, access_mask);
}

/****************************************************************************
  find first available policy slot.  creates a policy handle for you.
****************************************************************************/
BOOL open_policy_hnd_link(struct policy_cache *cache,
			  const POLICY_HND *parent_hnd,
			  POLICY_HND *hnd, uint32 access_mask)
{
	uint16 vuid = get_policy_vuid(cache, parent_hnd);
	if (vuid == UID_FIELD_INVALID)
	{
		return False;
	}
	create_pol_hnd(hnd);
	return register_policy_hnd(cache, vuid, hnd, access_mask);
}

/****************************************************************************
  find policy index by handle
****************************************************************************/
int find_policy_by_hnd(struct policy_cache *cache, const POLICY_HND *hnd)
{
	struct policy *p = find_policy(cache, hnd);

	return p ? p->pnum : -1;
}


/****************************************************************************
  set pol state.
****************************************************************************/
BOOL set_policy_state(struct policy_cache *cache, POLICY_HND *hnd,
		      void (*fn) (void *), void *dev)
{
	struct policy *p = find_policy(cache, hnd);

	if (p && p->open)
	{
		DEBUG(5, ("policy(pnum=%x %s): Setting policy state\n",
			  p->pnum, pol_get_name(p)));

		if (! fn)
			fn = safe_free;

		handle_set_data(p->handle, "__util_hnd__/state(data)",
				dev, fn);
		return True;
	}

	DEBUG(3, ("Error setting policy state\n"));

	return False;
}

/****************************************************************************
  get pol state.
****************************************************************************/
void *get_policy_state_info(struct policy_cache *cache, const POLICY_HND *hnd)
{
	struct policy *p = find_policy(cache, hnd);

	if (p != NULL && p->open)
	{
		DEBUG(5, ("policy(pnum=%x %s): Getting policy state\n",
			  p->pnum, pol_get_name(p)));
		return handle_get_data(p->handle, "__util_hnd__/state(data)",
				       False);
	}

	DEBUG(3, ("Error getting policy state\n"));
	return NULL;
}

/****************************************************************************
  set the type of the state of a POLICY_HND
****************************************************************************/
BOOL policy_hnd_set_state_type(struct policy_cache *cache,
			       POLICY_HND *hnd, int type)
{
	struct policy *p = find_policy(cache, hnd);
	int *ptype;

	if (!p || !p->open)
	{
		DEBUG(3, ("Error setting type for policy state\n"));
		return False;
	}
	DEBUG(4, ("policy(pnum=%x %s): setting type to %d\n",
		  p->pnum, pol_get_name(p), type));

	ptype = g_new(int, 1);
	if (!ptype)
		return False;
	(* ptype) = type;

	handle_set_data(p->handle, "__util_hnd__/type(int)",
			ptype, safe_free);

	return True;
}

/****************************************************************************
  get the type of the state of a POLICY_HND
****************************************************************************/
int policy_hnd_get_state_type(struct policy_cache *cache,
			      const POLICY_HND *hnd)
{
	struct policy *p = find_policy(cache, hnd);
	int *ptype;

	if (!p || !p->open)
	{
		DEBUG(3, ("Error getting type for policy state\n"));
		return -1;
	}

	ptype = handle_get_data(p->handle, "__util_hnd__/type(int)",
				False);

	if (!ptype)
		return -1;

	DEBUG(4, ("policy(pnum=%x %s): getting type %d\n",
		  p->pnum, pol_get_name(p), *ptype));

	return *ptype;
}

/****************************************************************************
  check the type of the state of a POLICY_HND
****************************************************************************/
BOOL policy_hnd_check_state_type(struct policy_cache *cache,
				 const POLICY_HND *hnd, int exptype)
{
	struct policy *p = find_policy(cache, hnd);
	int *ptype;
	BOOL ret;

	if (!p || !p->open)
	{
		DEBUG(3, ("Error checking type for policy state\n"));
		return False;
	}

	ptype = handle_get_data(p->handle, "__util_hnd__/type(int)",
				False);

	if (!ptype)
		return -1;

	ret = (*ptype == exptype);

	if (ret)
	{
		DEBUG(4, ("policy(pnum=%x %s): checking if type %d is %d\n",
			  p->pnum, pol_get_name(p), *ptype, exptype));
	}
	else
	{
		DEBUG(3, ("policy(pnum=%x %s): type %d is not %d\n",
			  p->pnum, pol_get_name(p), *ptype, exptype));
	}

	return ret;
}

/****************************************************************************
  close an lsa policy
****************************************************************************/
BOOL close_policy_hnd(struct policy_cache *cache, POLICY_HND *hnd)
{
	struct policy *p = find_policy(cache, hnd);

	if (!p)
	{
		DEBUG(3, ("Error closing policy\n"));
		return False;
	}

	DEBUG(5, ("policy(pnum=%x %s): Closing\n", p->pnum, pol_get_name(p)));

	DLIST_REMOVE(cache->Policy, p);

	ZERO_STRUCTP(hnd);

	handle_close(p->handle);
	p->handle = NULL;

	free(p);

	DEBUG(10, ("policy closed\n"));

	return True;
}

/****************************************************************************
  get pol state.
****************************************************************************/
#ifndef NOONS_SMALL_BUILD
static
BOOL policy_link_key(struct policy_cache *cache, const POLICY_HND *hnd,
		     POLICY_HND *to)
{
	struct policy *p = find_policy(cache, hnd);
	struct policy *pto = find_policy(cache, to);

	if (p != NULL && p->open && pto != NULL && pto->open)
	{
		DEBUG(3, ("Linking policy key pnum=%x vuid=%x\n",
			  p->vuid, p->pnum));
		pto->vuid = p->vuid;
		return True;
	}

	DEBUG(3, ("Error getting policy link states\n"));
	return False;
}
#endif

/****************************************************************************
  get pol state.
****************************************************************************/
static
uint16 get_policy_vuid(struct policy_cache *cache,
				      const POLICY_HND *hnd)
{
	struct policy *p = find_policy(cache, hnd);

	if (p != NULL && p->open)
	{
		DEBUG(5, ("Getting policy vuid pnum=%x vuid=%x\n",
			  p->pnum, p->vuid));
		return p->vuid;
	}

	DEBUG(3, ("Error getting policy state\n"));
	return UID_FIELD_INVALID;
}

/****************************************************************************
  get user session key.
****************************************************************************/
BOOL pol_get_usr_sesskey(struct policy_cache *cache, const POLICY_HND *hnd,
			 uchar usr_sess_key[16])
{
	uint16 vuid = get_policy_vuid(cache, hnd);

	return vuid_get_usr_sesskey(vuid, usr_sess_key);
}


BOOL pol_hnd_set_sid(POLICY_HND *hnd, const char *sid_name, const DOM_SID *sid)
{
	struct policy *p = find_policy(get_global_hnd_cache(), hnd);
	pstring sidstr;
	fstring dataname;
	DOM_SID *dev;

	if (!p)
		return False;

	dev = sid_dup(sid);

	snprintf(dataname, sizeof(dataname), "__util_hnd__/sids/%s(DOM_SID)", sid_name);

	DEBUG(3, ("Handle %s:%s adding sid %s\n", pol_get_name(p),
		  sid_name, sid_to_string(sidstr, sid)));

	handle_set_data(p->handle, dataname,
			dev, free);

	return True;
}


const DOM_SID *pol_hnd_get_sid(const POLICY_HND *hnd, const char *sid_name)
{
	struct policy *p = find_policy(get_global_hnd_cache(), hnd);
	pstring sidstr;
	fstring dataname;
	const DOM_SID *sid;

	if (!p || !p->open)
		return NULL;

	snprintf(dataname, sizeof(dataname), "__util_hnd__/sids/%s(DOM_SID)", sid_name);

	sid = handle_get_data(p->handle, dataname, False);

	if (sid)
		DEBUG(4, ("Handle %s:%s sid=%s\n", pol_get_name(p),
			  sid_name, sid_to_string(sidstr, sid)));
	else
		DEBUG(1, ("Handle %s has no SID %s\n",
			  pol_get_name(p), sid_name));

	return sid;
}
