/* 
   Unix SMB/Netbios implementation.
   Functions useful in logging in people in the NT area
   Copyright (C) Elrond <elrond@samba-tng.org>  2002.
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#include "dlinklist.h"
#include "libsamba.h"
#include "tng_misc.h"
#include "rpc_parse.h"
#include "rpc_client.h"
#include "sids.h"
#include "others.h"

/****************************************************************************
  does _both_ nt->unix and unix->unix username remappings.
****************************************************************************/
const struct passwd *map_nt_and_unix_username(const char *domain,
					      const char *ntuser,
					      char *unix_user, char *nt_user)
{
	DOM_NAME_MAP gmep;
	fstring nt_username;
	if (nt_user == NULL)
	{
		nt_user = nt_username;
	}

	if (domain != NULL)
	{
		slprintf(nt_user, sizeof(fstring), "%s\\%s", domain, ntuser);
	}
	else
	{
		fstrcpy(nt_user, ntuser);
	}

	/*
	 * Pass the user through the NT -> unix user mapping
	 * function.
	 */

	if (lookupsmbpwntnam(nt_user, &gmep))
	{
		fstrcpy(unix_user, gmep.unix_name);
	}
	else
	{
		return NULL;
	}

	/*
	 * Pass the user through the unix -> unix user mapping
	 * function.
	 */

	(void)map_username(unix_user);
	/*
	 * Do any UNIX username case mangling.
	 */
	return Get_Pwnam(unix_user, True);
}


/****************************************************************************/

#ifndef NOONS_SMALL_BUILD
struct _sid_list;

typedef struct _sid_list
{
	struct _sid_list *prev, *next;
	DOM_SID sid;
} SID_LIST_ENTRY;

static void create_base(SID_LIST_ENTRY **base, const NT_USER_TOKEN *token)
{
	size_t i;

	for (i = 0; i < token->num_sids; i++)
	{
		SID_LIST_ENTRY *l = g_new(SID_LIST_ENTRY, 1);
		if (! l)
			return;
		sid_copy(&l->sid, &token->user_sids[i]);
		DLIST_ADD((*base), l);
	}
}

static SID_LIST_ENTRY *check_and_gen_sid(const SID_LIST_ENTRY *all,
					 const DOM_SID *sid)
{
	SID_LIST_ENTRY *new;

	for (; all; all = all->next)
	{
		if (sid_equal(&all->sid, sid))
			break;
	}

	if (all)
		return NULL;

	new = g_new(SID_LIST_ENTRY, 1);

	sid_copy(&new->sid, sid);
	
	return new;
}

/*
 * return False only for fatal errors
 */
static BOOL check_sid_in_one_dom(const DOM_SID *dom_sid,
				 const POLICY_HND *dom_hnd,
				 const DOM_SID *check_sid,
				 SID_LIST_ENTRY **list)
{
	uint32 ptr_sid;
	DOM_SID2 sid2;
	uint32 num_found_aliases, i;
	uint32 *found_rids;

	DEBUG(5, ("  in domain: %s\n", sid_string_static(dom_sid)));

	make_dom_sid2(&sid2, check_sid);
	ptr_sid = 1;

	if (! samr_query_useraliases(dom_hnd, &ptr_sid, &sid2,
				     &num_found_aliases, &found_rids))
		return False;

	for (i = 0; i < num_found_aliases; i++)
	{
		DOM_SID new_sid;
		SID_LIST_ENTRY *new_entry;

		sid_copy(&new_sid, dom_sid);
		sid_append_rid(&new_sid, found_rids[i]);

		DEBUGADD(5, ("  (possibly) adding: %s\n", sid_string_static(&new_sid)));

		new_entry = check_and_gen_sid(*list, &new_sid);
		if (new_entry)
			DLIST_ADD((*list), new_entry);
	}
	safe_free(found_rids);
	return True;
}

void tng_login_gen_aliassids(const NT_USER_TOKEN *base_sids)
{
	SID_LIST_ENTRY *list = NULL;
	SID_LIST_ENTRY *cur_entry;
	POLICY_HND sam_connect, hnd_dom, hnd_builtin;

	create_base(&list, base_sids);

	if (!list)
		/* No SIDs?! */
		return;

	if (! samr_connect("\\\\.", SEC_RIGHTS_MAXIMUM_ALLOWED, &sam_connect))
		return;
	if (! samr_open_domain(&sam_connect, SEC_RIGHTS_MAXIMUM_ALLOWED,
			       &global_sam_sid, &hnd_dom))
	{
		samr_close(&sam_connect);
		return;
	}
	if (! samr_open_domain(&sam_connect, SEC_RIGHTS_MAXIMUM_ALLOWED,
			       global_sid_builtin, &hnd_builtin))
	{
		samr_close(&hnd_dom);
		samr_close(&sam_connect);
		return;
	}

	cur_entry = list;
	/*
	 * We need to go to the end of the list, because
	 * DLIST_ADD prepends the new entry and we want to
	 * see the new entries in our loop.
	 */
	while(cur_entry->next)
		cur_entry = cur_entry->next;

	for(; cur_entry; cur_entry = cur_entry->prev)
	{
		DEBUG(4, ("Checking user's SID: %s\n",
			  sid_string_static(&cur_entry->sid)));
		if (! check_sid_in_one_dom(&global_sam_sid, &hnd_dom,
					   &cur_entry->sid, &list))
			break;
		if (! check_sid_in_one_dom(global_sid_builtin, &hnd_builtin,
					   &cur_entry->sid, &list))
			break;
	}

	samr_close(&hnd_builtin);
	samr_close(&hnd_dom);
	samr_close(&sam_connect);
}
#endif
