/* 
   Copyright (C) Andrew Tridgell 1992-1998
   Copyright (C) Elrond <elrond@samba-tng.org> 2004-2005
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: libsamba_proto.h		(do not remove this line) */

/**
 * \file
 * \brief Network/Socket abstraction layer
 */

#include "includes.h"
#include <errno.h>
#ifdef HAVE_WINSOCK2_H
#include <winsock2.h>
#include <winerror.h>
#endif
#include "libsamba.h"
#include "byteorder.h"


int smb_read_error = 0;



#ifndef SOCKET_ERROR
#define SOCKET_ERROR (-1)
#endif

#ifndef EINPROGRESS
#define EINPROGRESS WSAEINPROGRESS
#endif
#ifndef EALREADY
#define EALREADY WSAEALREADY
#endif

#ifndef WSAEISCONN
#define WSAEISCONN EISCONN
#endif

#if defined(HAVE_WINSOCK2_H) && defined(HAVE_OS_WIN32)
#define socket_error WSAGetLastError()
#else
#define socket_error errno
#endif


void tng_socket_close(int fd)
{
	int ret;
#if defined(HAVE_WINSOCK2_H) && defined(HAVE_OS_WIN32)
	ret = closesocket(fd);
#else
	ret = close(fd);
#endif
	if (ret != 0)
		DEBUG(1, ("tng_socket_close(%d) failed\n", fd));
}


#ifdef HAVE_GETNAMEINFO
static const char *debug_sock_addr(const struct sockaddr *sa, socklen_t salen)
{
	char host[1025];
	char port[32];
	char *hostport = NULL;

	if (getnameinfo(sa, salen,
			host, sizeof(host),
			port, sizeof(port),
			NI_NUMERICHOST | NI_NUMERICSERV) != 0)
	{
		return "getnameinfo failed";
	}
	asprintf(&hostport, "%s#%s", host, port);
	tss_addptr(hostport, strlen(hostport));
	return hostport;
}
#endif


/****************************************************************************
 Determine if a file descriptor is in fact a socket.
****************************************************************************/

BOOL is_a_socket(Socket fd)
{
	int v;
	socklen_t l = sizeof(int);
	return (getsockopt(fd, SOL_SOCKET, SO_TYPE, (char *)&v, &l) == 0);
}

enum SOCK_OPT_TYPES { OPT_BOOL, OPT_INT, OPT_ON };

typedef struct smb_socket_option
{
	char *name;
	int level;
	int option;
	int value;
	int opttype;
} smb_socket_option;

static const smb_socket_option socket_options[] = {
	{ "SO_KEEPALIVE", SOL_SOCKET, SO_KEEPALIVE, 0, OPT_BOOL},
	{ "SO_REUSEADDR", SOL_SOCKET, SO_REUSEADDR, 0, OPT_BOOL},
	{ "SO_BROADCAST", SOL_SOCKET, SO_BROADCAST, 0, OPT_BOOL},
#ifdef TCP_NODELAY
	{ "TCP_NODELAY", IPPROTO_TCP, TCP_NODELAY, 0, OPT_BOOL},
#endif
#ifdef IPTOS_LOWDELAY
	{ "IPTOS_LOWDELAY", IPPROTO_IP, IP_TOS, IPTOS_LOWDELAY, OPT_ON},
#endif
#ifdef IPTOS_THROUGHPUT
	{ "IPTOS_THROUGHPUT", IPPROTO_IP, IP_TOS, IPTOS_THROUGHPUT, OPT_ON},
#endif
#ifdef SO_REUSEPORT
	{ "SO_REUSEPORT", SOL_SOCKET, SO_REUSEPORT, 0, OPT_BOOL},
#endif
#ifdef SO_SNDBUF
	{ "SO_SNDBUF", SOL_SOCKET, SO_SNDBUF, 0, OPT_INT},
#endif
#ifdef SO_RCVBUF
	{ "SO_RCVBUF", SOL_SOCKET, SO_RCVBUF, 0, OPT_INT},
#endif
#ifdef SO_SNDLOWAT
	{ "SO_SNDLOWAT", SOL_SOCKET, SO_SNDLOWAT, 0, OPT_INT},
#endif
#ifdef SO_RCVLOWAT
	{ "SO_RCVLOWAT", SOL_SOCKET, SO_RCVLOWAT, 0, OPT_INT},
#endif
#ifdef SO_SNDTIMEO
	{ "SO_SNDTIMEO", SOL_SOCKET, SO_SNDTIMEO, 0, OPT_INT},
#endif
#ifdef SO_RCVTIMEO
	{ "SO_RCVTIMEO", SOL_SOCKET, SO_RCVTIMEO, 0, OPT_INT},
#endif
	{ NULL, 0, 0, 0, 0}
};



/****************************************************************************
 Set user socket options.
****************************************************************************/
void set_socket_options(int fd, char *options)
{
	fstring tok;

	while (next_token(&options, tok, " \t,", sizeof(tok)))
	{
		int ret = 0, i;
		int value = 1;
		char *p;
		BOOL got_value = False;

		if ((p = strchr(tok, '=')))
		{
			*p = 0;
			value = atoi(p + 1);
			got_value = True;
		}

		for (i = 0; socket_options[i].name; i++)
			if (strequal(socket_options[i].name, tok))
				break;

		if (!socket_options[i].name)
		{
			DEBUG(0, ("Unknown socket option %s\n", tok));
			continue;
		}

		switch (socket_options[i].opttype)
		{
			case OPT_BOOL:
			case OPT_INT:
				ret = setsockopt(fd, socket_options[i].level,
						 socket_options[i].option,
						 (char *)&value, sizeof(int));
				break;

			case OPT_ON:
				if (got_value)
					DEBUG(0,
					      ("syntax error - %s does not take a value\n",
					       tok));

				{
					int on = socket_options[i].value;
					ret =
						setsockopt(fd,
							   socket_options
							   [i].level,
							   socket_options
							   [i].option,
							   (char *)&on,
							   sizeof(int));
				}
				break;
		}

		if (ret != 0)
			DEBUG(0,
			      ("Failed to set socket option %s (Error %s)\n",
			       tok, strerror(errno)));
	}
}



/****************************************************************************
 Read from a socket.
****************************************************************************/


/*******************************************************************
 checks if read data is outstanding.
 ********************************************************************/
int read_data_outstanding(int fd, unsigned int time_out)
{
	int selrtn;
	fd_set fds;
	struct timeval timeout;

	FD_ZERO(&fds);
	FD_SET(fd, &fds);

	timeout.tv_sec = (time_t) (time_out / 1000);
	timeout.tv_usec = (long)(1000 * (time_out % 1000));

	selrtn = sys_select_intr(fd + 1, &fds, &timeout);

	if (selrtn <= 0)
	{
		return selrtn;
	}
	return FD_ISSET(fd, &fds) ? 1 : 0;
}

/**
 * \brief Read udp from a socket
 */
ssize_t read_udp_socket(Socket fd, char *buf, size_t len,
			struct in_addr *recv_ip, int *recv_port)
{
	ssize_t ret;
	struct sockaddr_in sock;
	socklen_t socklen;

	socklen = sizeof(sock);
	ZERO_OBJ(sock);
	ret = recvfrom(fd, buf, len, 0, (struct sockaddr *)&sock, &socklen);
	if (ret <= 0)
	{
		DEBUG(2, ("read socket failed. ERRNO=%s\n", strerror(errno)));
		return (0);
	}

	if (recv_ip)
		*recv_ip = sock.sin_addr;
	if (recv_port)
		*recv_port = ntohs(sock.sin_port);

	DEBUG(10, ("read_udp_socket: lastip %s lastport %d read: %d\n",
		   inet_ntoa(sock.sin_addr), ntohs(sock.sin_port), ret));

	return (ret);
}


/**
 * \brief send a udp packet on a already open socket
 */
BOOL send_udp(int fd, const void *buf, size_t len, struct in_addr ip, int port)
{
  BOOL ret;
  struct sockaddr_in sock_out;

  /* set the address and port */
  ZERO_OBJ(sock_out);
  putip((char *)&sock_out.sin_addr,(char *)&ip);
  sock_out.sin_port = htons( port );
  sock_out.sin_family = AF_INET;
  
  DEBUG( 5, ( "Sending a packet of len %d to (%s) on port %d\n",
	      len, inet_ntoa(ip), port ) );
  dump_data(100, buf, len);
	
  ret = (sendto(fd,buf,len,0,(struct sockaddr *)&sock_out,
		sizeof(sock_out)) >= 0);

  if (!ret)
    DEBUG(0,("Packet send failed to %s(%d) ERRNO=%s\n",
	     inet_ntoa(ip),port,strerror(errno)));

  return(ret);
}

/****************************************************************************
 Read data from a socket with a timout in msec.
mincount = if timeout, minimum to read before returning
maxcount = number to be read.
time_out = timeout in milliseconds
****************************************************************************/

ssize_t read_socket_with_timeout(int fd, char *buf, size_t mincnt,
				 size_t maxcnt, unsigned int time_out)
{
	ssize_t readret;
	size_t nread = 0;

	/* just checking .... */
	if (maxcnt <= 0)
		return (0);

	smb_read_error = 0;

	/* Blocking read */
	if (time_out <= 0)
	{
		if (mincnt == 0)
			mincnt = maxcnt;

		while (nread < mincnt)
		{
			readret = recv(fd, buf + nread, maxcnt - nread, 0);

			if (readret == 0)
			{
				DEBUG(5,
				      ("read_socket_with_timeout: blocking read. EOF from client.\n"));
				smb_read_error = READ_EOF;
				return -1;
			}

			if (readret == -1)
			{
				DEBUG(0,
				      ("read_socket_with_timeout: read error = %s.\n",
				       strerror(errno)));
				smb_read_error = READ_ERROR;
				return -1;
			}
			nread += readret;
		}
		return ((ssize_t) nread);
	}

	/* Most difficult - timeout read */
	/* If this is ever called on a disk file and 
	   mincnt is greater then the filesize then
	   system performance will suffer severely as 
	   select always returns true on disk files */

	for (nread = 0; nread < mincnt;)
	{
		int selrtn = read_data_outstanding(fd, time_out);

		/* Check if error */
		if (selrtn == -1)
		{
			/* something is wrong. Maybe the socket is dead? */
			DEBUG(0,
			      ("read_socket_with_timeout: timeout read. select error = %s.\n",
			       strerror(errno)));
			smb_read_error = READ_ERROR;
			return -1;
		}

		/* Did we timeout ? */
		if (selrtn == 0)
		{
			DEBUG(10,
			      ("read_socket_with_timeout: timeout read. select timed out.\n"));
			smb_read_error = READ_TIMEOUT;
			return -1;
		}

		readret = recv(fd, buf + nread, maxcnt - nread, 0);

		if (readret == 0)
		{
			/* we got EOF on the file descriptor */
			DEBUG(5,
			      ("read_socket_with_timeout: timeout read. EOF from client.\n"));
			smb_read_error = READ_EOF;
			return -1;
		}

		if (readret == -1)
		{
			/* the descriptor is probably dead */
			DEBUG(0,
			      ("read_socket_with_timeout: timeout read. read error = %s.\n",
			       strerror(errno)));
			smb_read_error = READ_ERROR;
			return -1;
		}
		nread += readret;
	}

	/* Return the number we got */
	return ((ssize_t) nread);
}

/****************************************************************************
 Read data from a fd with a timout in msec.
mincount = if timeout, minimum to read before returning
maxcount = number to be read.
time_out = timeout in milliseconds
****************************************************************************/

ssize_t read_with_timeout(int fd, char *buf, size_t mincnt, size_t maxcnt,
			  unsigned int time_out)
{
	ssize_t readret;
	size_t nread = 0;

	/* just checking .... */
	if (maxcnt <= 0)
		return (0);

	/* Blocking read */
	if (time_out <= 0)
	{
		if (mincnt == 0)
			mincnt = maxcnt;

		while (nread < mincnt)
		{
			readret = read(fd, buf + nread, maxcnt - nread);

			if (readret <= 0)
				return readret;

			nread += readret;
		}
		return ((ssize_t) nread);
	}

	/* Most difficult - timeout read */
	/* If this is ever called on a disk file and 
	   mincnt is greater then the filesize then
	   system performance will suffer severely as 
	   select always returns true on disk files */

	for (nread = 0; nread < mincnt;)
	{
		int selrtn = read_data_outstanding(fd, time_out);

		if (selrtn <= 0)
			return selrtn;

		readret = read(fd, buf + nread, maxcnt - nread);

		if (readret <= 0)
			return readret;

		nread += readret;
	}

	/* Return the number we got */
	return ((ssize_t) nread);
}


/****************************************************************************
 Read data from a socket, reading exactly N bytes. 
****************************************************************************/

ssize_t read_socket_data(int fd, char *buffer, size_t N)
{
	ssize_t ret;
	size_t total = 0;

	smb_read_error = 0;

	while (total < N)
	{
		ret = recv(fd, buffer + total, N - total, 0);

		if (ret == 0)
		{
			DEBUG(10,
			      ("read_socket_data: recv of %d returned 0. Error = %s\n",
			       (int)(N - total), strerror(errno)));
			smb_read_error = READ_EOF;
			return 0;
		}
		if (ret == -1)
		{
			DEBUG(0,
			      ("read_socket_data: recv failure for %d. Error = %s\n",
			       (int)(N - total), strerror(errno)));
			smb_read_error = READ_ERROR;
			return -1;
		}
		total += ret;
	}
	return (ssize_t) total;
}


/****************************************************************************
 Write data to a socket - use send rather than write.
****************************************************************************/

ssize_t write_socket_data(int fd, const char *buffer, size_t N)
{
	size_t total = 0;
	ssize_t ret;

	while (total < N)
	{
		ret = send(fd, buffer + total, N - total, 0);

		if (ret == -1)
		{
			DEBUG(0,
			      ("write_socket_data: write failure. Error = %s\n",
			       strerror(errno)));
			return -1;
		}
		if (ret == 0)
			return total;

		total += ret;
	}
	return (ssize_t) total;
}

/****************************************************************************
write to a socket
****************************************************************************/

ssize_t write_socket(int fd, char *buf, size_t len)
{
	ssize_t ret = 0;

	DEBUG(6, ("write_socket(%d,%d)\n", fd, (int)len));
	ret = write_socket_data(fd, buf, len);

	DEBUG(6, ("write_socket(%d,%d) wrote %d\n", fd, (int)len, (int)ret));
	if (ret <= 0)
		DEBUG(0,
		      ("write_socket: Error writing %d bytes to socket %d: ERRNO = %s\n",
		       (int)len, fd, strerror(errno)));

	return (ret);
}


/****************************************************************************
  send an null session message to a fd
****************************************************************************/
#ifndef NOONS_SMALL_BUILD
static
BOOL send_null_session_msg(int fd)
{
	ssize_t ret;
	uint32 blank = 0;
	size_t len = 4;
	size_t nwritten = 0;
	char *buffer = (char *)&blank;

	while (nwritten < len)
	{
		ret = write_socket(fd, buffer + nwritten, len - nwritten);
		if (ret <= 0)
		{
			DEBUG(0,
			      ("send_null_session_msg: Error writing %d bytes to client. %d. Exiting\n",
			       (int)len, (int)ret));
			exit(1);
		}
		nwritten += ret;
	}

	DEBUG(10, ("send_null_session_msg: sent 4 null bytes to client.\n"));
	return True;
}
#endif
/****************************************************************************
send a single packet to a port on another machine
****************************************************************************/
#ifndef NOONS_SMALL_BUILD
static
BOOL send_one_packet(char *buf, int len, struct in_addr ip, int port,
		     int type)
{
	BOOL ret;
	int out_fd;
	struct sockaddr_in sock_out;

	/* create a socket to write to */
	out_fd = socket(AF_INET, type, 0);
	if (out_fd == -1)
	{
		DEBUG(0, ("socket failed"));
		return False;
	}

	/* set the address and port */
	ZERO_OBJ(sock_out);
	putip((char *)&sock_out.sin_addr, (char *)&ip);
	sock_out.sin_port = htons(port);
	sock_out.sin_family = AF_INET;

	if (DEBUGLEVEL > 0)
		DEBUG(3,
		      ("sending a packet of len %d to (%s) on port %d of type %s\n",
		       len, inet_ntoa(ip), port,
		       type == SOCK_DGRAM ? "DGRAM" : "STREAM"));

	/* send it */
	ret =
		(sendto
	       (out_fd, buf, len, 0, (struct sockaddr *)&sock_out,
		sizeof(sock_out)) >= 0);

	if (!ret)
		DEBUG(0, ("Packet send to %s(%d) failed ERRNO=%s\n",
			  inet_ntoa(ip), port, strerror(errno)));

	tng_socket_close(out_fd);
	return (ret);
}
#endif
/****************************************************************************
open a socket of the specified type, port and address for incoming data
****************************************************************************/

int open_socket_in(int type, int port, int dlevel, uint32 socket_addr,
		   BOOL rebind)
{
	struct hostent *hp;
	struct sockaddr_in sock;
	pstring host_name;
	int res;

	/* get my host name */
	if (gethostname(host_name, sizeof(host_name)-1) == -1)
	{
		DEBUG(0, ("gethostname failed\n"));
		return -1;
	}

	/* get host info */
	if ((hp = Get_Hostbyname(host_name)) == 0)
	{
		DEBUG(0, ("Get_Hostbyname: Unknown host %s\n", host_name));
		return -1;
	}

	ZERO_OBJ(sock);
	memcpy((char *)&sock.sin_addr, (char *)hp->h_addr, hp->h_length);

#ifdef HAVE_SOCK_SIN_LEN
	sock.sin_len = sizeof(sock);
#endif
	sock.sin_port = htons(port);
	sock.sin_family = hp->h_addrtype;
	sock.sin_addr.s_addr = socket_addr;
	res = socket(hp->h_addrtype, type, 0);
	if (res == -1)
	{
		DEBUG(0, ("socket failed\n"));
		return -1;
	}

	{
		int val = 1;
		if (rebind)
			val = 1;
		else
			val = 0;
		if (setsockopt
		    (res, SOL_SOCKET, SO_REUSEADDR, (char *)&val,
		     sizeof(val)) == -1)
			DEBUG(dlevel,
			      ("setsockopt: SO_REUSEADDR=%d on port %d failed with error = %s\n",
			       val, port, strerror(errno)));
#ifdef SO_REUSEPORT
		if (setsockopt
		    (res, SOL_SOCKET, SO_REUSEPORT, (char *)&val,
		     sizeof(val)) == -1)
			DEBUG(dlevel,
			      ("setsockopt: SO_REUSEPORT=%d on port %d failed with error = %s\n",
			       val, port, strerror(errno)));
#endif /* SO_REUSEPORT */
	}

	/* now we've got a socket - we need to bind it */
	if (bind(res, (struct sockaddr *)&sock, sizeof(sock)) < 0)
	{
		if (port)
		{
			if (port == SMB_PORT || port == NMB_PORT)
				DEBUG(dlevel,
				      ("bind failed on port %d socket_addr=%s (%s)\n",
				       port, inet_ntoa(sock.sin_addr),
				       strerror(errno)));
			tng_socket_close(res);

			if (dlevel > 0 && port < 1000)
				port = 7999;

			if (port >= 1000 && port < 9000)
				return (open_socket_in
					(type, port + 1, dlevel, socket_addr,
					 rebind));
		}

		return (-1);
	}
	DEBUG(3, ("bind succeeded on port %d\n", port));

	return res;
}


/****************************************************************************
  create an outgoing socket. timeout is in milliseconds.
  **************************************************************************/
int open_socket_out(int type, struct in_addr *addr, int port, int timeout)
{
	struct sockaddr_in sock_out;
	int res, ret;
	int connect_loop = 250;	/* 250 milliseconds */
	int loops = (timeout) / connect_loop;

	/* create a socket to write to */
	res = socket(PF_INET, type, 0);
	if (res == -1)
	{
		DEBUG(0, ("socket error\n"));
		return -1;
	}

	if (type != SOCK_STREAM)
		return (res);

	ZERO_OBJ(sock_out);
	putip((char *)&sock_out.sin_addr, (char *)addr);

	sock_out.sin_port = htons(port);
	sock_out.sin_family = PF_INET;

	/* set it non-blocking */
	set_blocking(res, False);

	DEBUG(3, ("Connecting to %s at port %d\n", inet_ntoa(*addr), port));

	/* and connect it to the destination */
      connect_again:
	ret = connect(res, (struct sockaddr *)&sock_out, sizeof(sock_out));

	DEBUG(5, ("connect returned %d; errcode=%u\n",
		  ret, socket_error));

	/* Some systems return EAGAIN when they mean EINPROGRESS */
	if (ret < 0 
	    && (socket_error == EINPROGRESS 
		|| socket_error == EALREADY 
		|| socket_error == EAGAIN
#ifdef WSAEALREADY
		|| socket_error == WSAEALREADY
		|| socket_error == WSAEWOULDBLOCK
		|| socket_error == WSAEINVAL
#endif
		)
	    )
	{
		if (loops--)
		{
			msleep(connect_loop);
			goto connect_again;
		}

		DEBUG(1, ("timeout connecting to %s:%d\n",
			  inet_ntoa(*addr), port));
		tng_socket_close(res);
		return -1;
	}

	if (ret < 0 && socket_error == WSAEISCONN)
	{
		errno = 0;
		ret = 0;
	}

	if (ret < 0)
	{
		DEBUG(2, ("error connecting to %s:%d (%s)\n",
			  inet_ntoa(*addr), port, strerror(errno)));
		tng_socket_close(res);
		return -1;
	}

	/* set it blocking again */
	set_blocking(res, True);

	return res;
}


/*******************************************************************
 matchname - determine if host name matches IP address. Used to
 confirm a hostname lookup to prevent spoof attacks
 ******************************************************************/
static BOOL matchname(char *remotehost, struct in_addr addr)
{
	struct hostent *hp;
	int i;

	if ((hp = Get_Hostbyname(remotehost)) == 0)
	{
		DEBUG(0,
		      ("Get_Hostbyname(%s): lookup failure.\n", remotehost));
		return False;
	}

	/*
	 * Make sure that gethostbyname() returns the "correct" host name.
	 * Unfortunately, gethostbyname("localhost") sometimes yields
	 * "localhost.domain". Since the latter host name comes from the
	 * local DNS, we just have to trust it (all bets are off if the local
	 * DNS is perverted). We always check the address list, though.
	 */

	if (strcasecmp(remotehost, hp->h_name)
	    && strcasecmp(remotehost, "localhost"))
	{
		DEBUG(0, ("host name/name mismatch: %s != %s\n",
			  remotehost, hp->h_name));
		return False;
	}

	/* Look up the host address in the address list we just got. */
	for (i = 0; hp->h_addr_list[i]; i++)
	{
		if (memcmp(hp->h_addr_list[i], &addr, sizeof(addr))
		    == 0)
			return True;
	}

	/*
	 * The host name does not map to the original host address. Perhaps
	 * someone has compromised a name server. More likely someone botched
	 * it, but that could be dangerous, too.
	 */

	DEBUG(0, ("host name/address mismatch: %s != %s\n",
		  inet_ntoa(addr), hp->h_name));
	return False;
}

 
/*******************************************************************
 return the DNS name of the remote end of a socket
 ******************************************************************/
char *get_socket_name(int fd)
{
	static pstring name_buf;
	static fstring addr_buf;
	struct hostent *hp;
	struct in_addr addr;
	char *p;

	p = get_socket_addr(fd);

	/* it might be the same as the last one - save some DNS work */
	if (strcmp(p, addr_buf) == 0)
		return name_buf;

	pstrcpy(name_buf, "UNKNOWN");
	if (fd == -1)
		return name_buf;

	fstrcpy(addr_buf, p);

	addr = *interpret_addr2(p);

	/* Look up the remote host name. */
	if (
	    (hp =
	     gethostbyaddr((char *)&addr.s_addr, sizeof(addr.s_addr),
			   AF_INET)) == 0)
	{
		DEBUG(1, ("Gethostbyaddr failed for %s\n", p));
		pstrcpy(name_buf, p);
	}
	else
	{
		pstrcpy(name_buf, (char *)hp->h_name);
		if (!matchname(name_buf, addr))
		{
			DEBUG(0,
			      ("Matchname failed on %s %s\n", name_buf, p));
			pstrcpy(name_buf, "UNKNOWN");
		}
	}
	return name_buf;
}

/*******************************************************************
 return the IP addr of the remote end of a socket as a string 
 ******************************************************************/
char *get_socket_addr(int fd)
{
	struct sockaddr sa;
	struct sockaddr_in *sockin = (struct sockaddr_in *)(&sa);
	socklen_t length = sizeof(sa);
	static fstring addr_buf;

	fstrcpy(addr_buf, "0.0.0.0");

	if (fd == -1)
	{
		return addr_buf;
	}

	if (getpeername(fd, &sa, &length) < 0)
	{
		DEBUG(0,
		      ("getpeername failed. Error was %s\n",
		       strerror(errno)));
		return addr_buf;
	}

	fstrcpy(addr_buf, (char *)inet_ntoa(sockin->sin_addr));

	return addr_buf;
}

/*******************************************************************
 opens and connects to a unix pipe socket
 ******************************************************************/
int open_pipe_sock(char *path)
{
#ifndef HAVE_OS_WIN32
	int sock;
	struct sockaddr_un sa;

	sock = socket(AF_UNIX, SOCK_STREAM, 0);

	if (sock < 0)
	{
		DEBUG(0, ("unix socket open failed\n"));
		return sock;
	}

	ZERO_STRUCT(sa);
	sa.sun_family = AF_UNIX;
	safe_strcpy(sa.sun_path, path, sizeof(sa.sun_path));

	DEBUG(10, ("socket open succeeded.  file name: %s\n", sa.sun_path));

	if (connect(sock, (struct sockaddr *)&sa, sizeof(sa)) < 0)
	{
		DEBUG(0, ("socket connect to %s failed\n", sa.sun_path));
		close(sock);
		return -1;
	}

	return sock;
#else
	return -1;
#endif
}

int create_pipe_socket(char *dir, int dir_perms, char *path, int path_perms)
{
#ifndef HAVE_OS_WIN32
	int s;
	struct sockaddr_un sa;

	DEBUG(2, ("create_pipe_socket: %s 0%o %s 0%o\n",
		  dir, dir_perms, path, path_perms));

	DEBUG(3,
	      ("*** Please someone examine create_pipe_socket and fix it ***\n"));
	DEBUGADD(3, ("*** if used other than for exclusive root access ***\n"));
	DEBUGADD(3, ("*** (see perms, which should be 0700 and 0600) ***\n"));
	DEBUGADD(3, ("*** there is a race condition to be exploited. ***\n"));

	sys_mkdir(dir, dir_perms);

	if (chmod(dir, dir_perms) < 0)
	{
		DEBUG(0, ("chmod on %s failed\n", dir));
		return -1;
	}

	if (remove(path)!=0 && errno != ENOENT)
	{
		DEBUG(0, ("remove on %s failed\n", path));
	}

	/* start listening on unix socket */
	s = socket(AF_UNIX, SOCK_STREAM, 0);

	if (s < 0)
	{
		DEBUG(0, ("socket open failed\n"));
		return -1;
	}

	ZERO_STRUCT(sa);
	sa.sun_family = AF_UNIX;
	safe_strcpy(sa.sun_path, path, sizeof(sa.sun_path));

	if (bind(s, (struct sockaddr *)&sa, sizeof(sa)) < 0)
	{
		DEBUG(0, ("socket bind to %s failed\n", sa.sun_path));
		close(s);
		remove(path);
		return -1;
	}

	if (s == -1)
	{
		DEBUG(0, ("bind failed\n"));
		remove(path);
		return -1;
	}

	if (path_perms != 0)
	{
		chmod(path, path_perms);
	}

	if (listen(s, 5) == -1)
	{
		DEBUG(0, ("listen failed\n"));
		return -1;
	}

	DEBUG(5, ("unix socket opened: %s\n", path));

	return s;
#else
	return -1;
#endif
}

/****************************************************************************
Set a fd into blocking/nonblocking mode. Uses POSIX O_NONBLOCK if available,
else
if SYSV use O_NDELAY
if BSD use FNDELAY
****************************************************************************/
int set_blocking(int fd, BOOL set)
{
#ifdef HAVE_OS_WIN32
	u_long val;
	if (set)
		val = 0;
	else
		val = 1;
	if (ioctlsocket(fd, FIONBIO, &val) == 0)
		return 0;
	else
		return -1;
#else
	int val;
#ifdef O_NONBLOCK
#define FLAG_TO_SET O_NONBLOCK
#else
#ifdef SYSV
#define FLAG_TO_SET O_NDELAY
#else /* BSD */
#define FLAG_TO_SET FNDELAY
#endif
#endif

	if ((val = fcntl(fd, F_GETFL, 0)) == -1)
		return -1;
	if (set)		/* Turn blocking on - ie. clear nonblock flag */
		val &= ~FLAG_TO_SET;
	else
		val |= FLAG_TO_SET;
	return fcntl(fd, F_SETFL, val);
#undef FLAG_TO_SET
#endif
}


BOOL tng_socksys_init(void)
{
#ifdef HAVE_OS_WIN32
	WORD wVersionRequested;
	WSADATA wsaData;
	wVersionRequested = MAKEWORD(1, 1);
	if (WSAStartup(wVersionRequested, &wsaData) != 0)
	{
		DEBUG(0, ("WSAStartup failed\n"));
		return False;
	}
#endif
	return True;
}

void tng_socksys_close(void)
{
#ifdef HAVE_OS_WIN32
	WSACleanup();
#endif
}
