/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   Samba utility functions
   Copyright (C) Andrew Tridgell 1992-1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: libsamba_proto.h		(do not remove this line) */

#include "includes.h"
#include <errno.h>
#ifdef HAVE_GLIB
#include <glib/gunicode.h>
#endif
#include "tng_misc.h"
#include "libsamba.h"
#include "byteorder.h"


 smb_ucs2_t wchar_list_sep[] = { (smb_ucs2_t)' ', (smb_ucs2_t)'\t', (smb_ucs2_t)',',
								(smb_ucs2_t)';', (smb_ucs2_t)':', (smb_ucs2_t)'\n',
								(smb_ucs2_t)'\r', 0 };
#ifndef MAXUNI
#define MAXUNI 1024
#endif


/*******************************************************************
 Write a string in (little-endian) unicode format. src is in
 the current DOS codepage. len is the length in bytes of the
 string pointed to by dst.

 if null_terminate is True then null terminate the packet (adds 2 bytes)

 the return value is the length in bytes consumed by the string, including the
 null termination if applied
********************************************************************/

int dos_PutUniCode(char *dst,const char *src, ssize_t len, BOOL null_terminate)
{
	size_t src_len = strlen(src);

	if (null_terminate)
		src_len++;

	len &= ~1; /* ucs2 is always a multiple of 2 */

	return convert_string(CH_DOS, CH_UCS2, src, src_len, dst, len);
}

/*******************************************************************
 Write a string in (little-endian) unicode format. src is in
 the current UNIX codepage. len is the length in bytes of the
 string pointed to by dst.

 if null_terminate is True then null terminate the packet (adds 2 bytes)

 the return value is the length in bytes consumed by the string, including the
 null termination if applied
********************************************************************/

size_t unix_PutUniCode(char *dst,const char *src, ssize_t len, BOOL null_terminate)
{
	return push_ucs2(NULL, dst, src, len, 
			 STR_UNICODE|STR_NOALIGN | (null_terminate?STR_TERMINATE:0));
}


/*******************************************************************
 Pull a DOS codepage string out of a UNICODE array. len is in bytes.
 We assume, len is the size of the whole buffer,
 some callers assume, that len is the maximum length of the
 string, that can be written into the buffer (sizeof(buf) - 1).
 Our assumption is just "more secure".
********************************************************************/

void unistr_to_dos(char *dest, const char *src, size_t len)
{
	size_t src_len;

	src_len = strlen_w((const smb_ucs2_t *) src) + 1;
	src_len *= 2;

	convert_string(CH_UCS2, CH_DOS,
		       src, src_len,
		       dest, len);

	dest[len - 1] = '\0';
}

/*******************************************************************
 Skip a UNICODE string in a little endian buffer.
 ********************************************************************/

char *skip_unibuf(char *srcbuf, int len)
{
	uint16 *src = (uint16 *)srcbuf;
	uint16 *srcend = src + len/2;

	while ((src < srcend) && (*(src++) != 0))
	{
	}

	return (char *)src;
}

/*******************************************************************
 Return a DOS codepage version of a little-endian unicode string.
 len is the filename length (ignoring any terminating zero) in uin16
 units. Always null terminates.
 Hack alert: uses fixed buffer(s).
 len is in 2 byte (unicode) units.
********************************************************************/

const char *dos_unistrn2(const uint16 *src, size_t len)
{
	static char lbufs[8][MAXUNI];
	static int nexti;
	char *lbuf = lbufs[nexti];
	size_t written;

	nexti = (nexti+1)%8;

	/* MAXUNI - 1 to leave space for the terminating NUL */
	written = convert_string(CH_UCS2, CH_DOS,
				 src, len * 2,
				 lbuf, MAXUNI - 1);

	lbuf[written] = '\0';

	return lbuf;
}


/*******************************************************************
 Return a number stored in a buffer
 ********************************************************************/

uint32 buffer2_to_uint32(const BUFFER2 *str)
{
	if (str->buf_len == 4)
	{
		const uchar *src = str->buffer;
		return IVAL(src, 0);
	}
	else
	{
		return 0;
	}
}

/*******************************************************************
Return a DOS codepage version of a NOTunicode string
********************************************************************/


/*******************************************************************
 Create a null-terminated unicode string from a null-terminated DOS
 codepage string.
 Return number of unicode chars copied, excluding the null character.
 Unicode strings created are in little-endian format.
 max_len is in bytes.
********************************************************************/

size_t dos_struni2(char *dst, const char *src, size_t max_len)
{
	size_t len;

	len = dos_PutUniCode(dst, src, max_len, True);

	if (len >= 2)
		len -= 2; /* strip length of terminating NUL */

	len /= 2; /* Bytes -> Unicode chars */

	return len;
}


/*******************************************************************
 returns the length in number of wide characters 
 ******************************************************************/
int unistrlen(uint16 *s)
{
	int len;
	
	if (!s)
		return -1;
	
	for (len=0; *s; s++,len++);
	
	return len;
}

/*******************************************************************
 Strcpy for unicode strings.  returns length (in num of wide chars)
********************************************************************/

int unistrcpy(char *dst, char *src)
{
	int num_wchars = 0;
	uint16 *wsrc = (uint16 *)src;
	uint16 *wdst = (uint16 *)dst;

	while (*wsrc) {
		*wdst++ = *wsrc++;
		num_wchars++;
	}
	*wdst = 0;

	return num_wchars;
}



/*******************************************************************
 The following functions reproduce many of the non-UNICODE standard
 string functions in Samba.
********************************************************************/

/*******************************************************************
 Count the number of characters in a smb_ucs2_t string.
********************************************************************/

size_t strlen_w(const smb_ucs2_t *src)
{
  size_t len;

  for(len = 0; *src++; len++)
    ;

  return len;
}

/*******************************************************************
 Count up to max number of characters in a smb_ucs2_t string.
********************************************************************/
size_t strnlen_w(const smb_ucs2_t *src, size_t max)
{
	size_t len;

	for(len = 0; (len < max) && *src++; len++)
		;

	return len;
}

/*******************************************************************
 Safe wstring copy into a known length string. maxlength includes
 the terminating zero. maxlength is in ucs2 units.
********************************************************************/

smb_ucs2_t *safe_strcpy_w(smb_ucs2_t *dest,const smb_ucs2_t *src, size_t maxlength)
{
    size_t ucs2_len;

    if (!dest) {
        DEBUG(0,("ERROR: NULL dest in safe_strcpy_w\n"));
        return NULL;
    }

    if (!src) {
        *dest = 0;
        return dest;
    }

	ucs2_len = strlen_w(src);

	if (ucs2_len >= maxlength) {
		char *ascii_version = unicode_to_fancy_ascii(src, ucs2_len);

		DEBUG(0, ("ERROR: string overflow by %u bytes in safe_strcpy_w [%.50s]\n",
			  (unsigned int)((ucs2_len+1-maxlength)*sizeof(smb_ucs2_t)),
			  ascii_version));
		safe_free(ascii_version);
		ucs2_len = maxlength - 1;
	}

    memcpy(dest, src, ucs2_len*sizeof(smb_ucs2_t));
    dest[ucs2_len] = 0;
    return dest;
}

/*******************************************************************
 Safe string cat into a string. maxlength includes the terminating zero.
 maxlength is in ucs2 units.
********************************************************************/

smb_ucs2_t *safe_strcat_w(smb_ucs2_t *dest, const smb_ucs2_t *src, size_t maxlength)
{
    size_t ucs2_src_len, ucs2_dest_len;

    if (!dest) {
        DEBUG(0,("ERROR: NULL dest in safe_strcat_w\n"));
        return NULL;
    }

    if (!src)
        return dest;

    ucs2_src_len = strlen_w(src);
    ucs2_dest_len = strlen_w(dest);

	if (ucs2_src_len + ucs2_dest_len >= maxlength) {
		char *ascii_version = unicode_to_fancy_ascii(src, ucs2_src_len);
		int new_len = maxlength - ucs2_dest_len - 1;

		DEBUG(0, ("ERROR: string overflow by %u characters in safe_strcat_w [%.50s]\n",
			  (unsigned int)(sizeof(smb_ucs2_t)*(ucs2_src_len + ucs2_dest_len + 1 - maxlength)),
			  ascii_version));
		safe_free(ascii_version);
		ucs2_src_len = (size_t)(new_len > 0 ? new_len : 0);
	}

    memcpy(&dest[ucs2_dest_len], src, ucs2_src_len*sizeof(smb_ucs2_t));
    dest[ucs2_dest_len + ucs2_src_len] = 0;
    return dest;
}

/*******************************************************************
 Compare the two strings s1 and s2.
********************************************************************/

int strcmp_w(const smb_ucs2_t *s1, const smb_ucs2_t *s2)
{
	smb_ucs2_t c1, c2;

	for (;;) {
		c1 = *s1++;
		c2 = *s2++;

		if (c1 != c2)
			return c1 - c2;

		if (c1 == 0)
            return 0;
	}
	return 0;
}

/*******************************************************************
 Compare the first n characters of s1 to s2. len is in ucs2 units.
********************************************************************/

int strncmp_w(const smb_ucs2_t *s1, const smb_ucs2_t *s2, size_t len)
{
	smb_ucs2_t c1, c2;

	for (; len != 0; --len) {
		c1 = *s1++;
		c2 = *s2++;

		if (c1 != c2)
			return c1 - c2;

		if (c1 == 0)
			return 0;

	}
	return 0;
}

/*******************************************************************
 Search string s2 from s1.
********************************************************************/

smb_ucs2_t *strstr_w(const smb_ucs2_t *s1, const smb_ucs2_t *s2)
{
	size_t len = strlen_w(s2);

	if (!*s2)
		goto out;

	for(;*s1; s1++) {
		if (*s1 == *s2) {
			if (strncmp_w(s1, s2, len) == 0)
				goto out;
		}
	}
	return NULL; 
out:
	/* STRIP_CONST is ok - it is the *caller's* problem
	   if he uses this function to un-const-ify s1 */
	return STRIP_CONST(s1);
}

/*******************************************************************
 Search for ucs2 char c from the beginning of s.
********************************************************************/ 

smb_ucs2_t *strchr_w(const smb_ucs2_t *s, smb_ucs2_t c)
{
	do {
		if (*s == c)
			/* STRIP_CONST is ok - it is the *caller's* problem
			   if he uses this function to un-const-ify s */
			return STRIP_CONST(s);
	} while (*s++);

	return NULL;
}

/*******************************************************************
 Search for ucs2 char c from the end of s.
********************************************************************/ 

smb_ucs2_t *strrchr_w(const smb_ucs2_t *s, smb_ucs2_t c)
{
	smb_ucs2_t *retval = NULL;

	do {
		if (*s == c)
			/* STRIP_CONST is ok - it is the *caller's* problem
			   if he uses this function to un-const-ify s */
			retval = STRIP_CONST(s);
	} while (*s++);

	return retval;
}

/*******************************************************************
 Search token from s1 separated by any ucs2 char of s2.
********************************************************************/

smb_ucs2_t *strtok_w(smb_ucs2_t *s1, const smb_ucs2_t *s2)
{
	static smb_ucs2_t *s = NULL;
	smb_ucs2_t *q;

	if (!s1) {
		if (!s)
			return NULL;
		s1 = s;
	}

	for (q = s1; *s1; s1++) {
		smb_ucs2_t *p = strchr_w(s2, *s1);
		if (p) {
			if (s1 != q) {
				s = s1 + 1;
				*s1 = '\0';
				return q;
			}
			q = s1 + 1;
		}
	}

	s = NULL;
	if (*q)
		return q;

	return NULL;
}

/*******************************************************************
 Duplicate a ucs2 string.
********************************************************************/

smb_ucs2_t *strdup_w(const smb_ucs2_t *s)
{
    size_t newlen = strlen_w(s) + 1;
    smb_ucs2_t *newstr = g_new(smb_ucs2_t, newlen);
    if (newstr == NULL)
        return NULL;
    safe_strcpy_w(newstr, s, newlen);
    return newstr;
}

/*******************************************************************
 Mapping tables for UNICODE character. Allows toupper/tolower and
 isXXX functions to work.

 tridge: split into 2 pieces. This saves us 5/6 of the memory
 with a small speed penalty
 The magic constants are the lower/upper range of the tables two
 parts
********************************************************************/

#ifndef HAVE_GLIB
typedef struct {
	smb_ucs2_t lower;
	smb_ucs2_t upper;
	unsigned char flags;
} smb_unicode_table_t;

#define TABLE1_BOUNDARY 9450
#define TABLE2_BOUNDARY 64256

static const smb_unicode_table_t map_table1[] = {
#include "unicode_map_table1.h"
};

static const smb_unicode_table_t map_table2[] = {
#include "unicode_map_table2.h"
};

static unsigned char map_table_flags(smb_ucs2_t v)
{
	if (v < TABLE1_BOUNDARY) return map_table1[v].flags;
	if (v >= TABLE2_BOUNDARY) return map_table2[v - TABLE2_BOUNDARY].flags;
	return 0;
}

static smb_ucs2_t map_table_lower(smb_ucs2_t v)
{
	if (v < TABLE1_BOUNDARY) return map_table1[v].lower;
	if (v >= TABLE2_BOUNDARY) return map_table2[v - TABLE2_BOUNDARY].lower;
	return v;
}

static smb_ucs2_t map_table_upper(smb_ucs2_t v)
{
	if (v < TABLE1_BOUNDARY) return map_table1[v].upper;
	if (v >= TABLE2_BOUNDARY) return map_table2[v - TABLE2_BOUNDARY].upper;
	return v;
}
#endif

/*******************************************************************
 Is an upper case wchar.
********************************************************************/

int isupper_w( smb_ucs2_t val)
{
#ifdef HAVE_GLIB
	return g_unichar_isupper(val);
#else
	return (map_table_flags(val) & UNI_UPPER);
#endif
}

/*******************************************************************
 Is a lower case wchar.
********************************************************************/

int islower_w( smb_ucs2_t val)
{
#ifdef HAVE_GLIB
	return g_unichar_islower(val);
#else
	return (map_table_flags(val) & UNI_LOWER);
#endif
}

/*******************************************************************
 Is a digit wchar.
********************************************************************/

int isdigit_w( smb_ucs2_t val)
{
#ifdef HAVE_GLIB
	return g_unichar_isdigit(val);
#else
	return (map_table_flags(val) & UNI_DIGIT);
#endif
}

/*******************************************************************
 Is a hex digit wchar.
********************************************************************/

int isxdigit_w( smb_ucs2_t val)
{
#ifdef HAVE_GLIB
	return g_unichar_isxdigit(val);
#else
	return (map_table_flags(val) & UNI_XDIGIT);
#endif
}

/*******************************************************************
 Is a space wchar.
********************************************************************/

int isspace_w( smb_ucs2_t val)
{
#ifdef HAVE_GLIB
	return g_unichar_isspace(val);
#else
	return (map_table_flags(val) & UNI_SPACE);
#endif
}

/*******************************************************************
 Convert a wchar to upper case.
********************************************************************/

smb_ucs2_t toupper_w( smb_ucs2_t val )
{
#ifdef HAVE_GLIB
	return g_unichar_toupper(val);
#else
	return map_table_upper(val);
#endif
}

/*******************************************************************
 Convert a wchar to lower case.
********************************************************************/

smb_ucs2_t tolower_w( smb_ucs2_t val )
{
#ifdef HAVE_GLIB
	return g_unichar_tolower(val);
#else
	return map_table_lower(val);
#endif
}


static smb_ucs2_t *last_ptr = NULL;

void set_first_token_w(smb_ucs2_t *ptr)
{
	last_ptr = ptr;
}

/****************************************************************************
 Get the next token from a string, return False if none found
 handles double-quotes. 
 Based on a routine by GJC@VILLAGE.COM. 
 Extensively modified by Andrew.Tridgell@anu.edu.au
 bufsize is in bytes.
****************************************************************************/

static smb_ucs2_t sep_list[] = { (smb_ucs2_t)' ', (smb_ucs2_t)'\t',  (smb_ucs2_t)'\n',  (smb_ucs2_t)'\r', 0};
static smb_ucs2_t quotechar = (smb_ucs2_t)'\"';

BOOL next_token_w(smb_ucs2_t **ptr, smb_ucs2_t *buff, smb_ucs2_t *sep, size_t bufsize)
{
	smb_ucs2_t *s;
	BOOL quoted;
	size_t len=1;

	/*
	 * Convert bufsize to smb_ucs2_t units.
	 */

	bufsize /= sizeof(smb_ucs2_t);

	if (!ptr)
		ptr = &last_ptr;
	if (!ptr)
		return(False);

	s = *ptr;

	/*
	 * Default to simple separators.
	 */

	if (!sep)
		sep = sep_list;

	/*
	 * Find the first non sep char.
	 */

	while(*s && strchr_w(sep,*s))
		s++;

	/*
	 * Nothing left ?
	 */

	if (!*s)
		return(False);

	/*
	 * Copy over the token.
	 */

	for (quoted = False; len < bufsize && *s && (quoted || !strchr_w(sep,*s)); s++) {
		if (*s == quotechar) {
			quoted = !quoted;
		} else {
			len++;
			*buff++ = *s;
		}
	}

	*ptr = (*s) ? s+1 : s;  
	*buff = 0;
	last_ptr = *ptr;

	return(True);
}

/****************************************************************************
 Convert list of tokens to array; dependent on above routine.
 Uses last_ptr from above - bit of a hack.
****************************************************************************/

smb_ucs2_t **toktocliplist_w(int *ctok, smb_ucs2_t *sep)
{
	smb_ucs2_t *s=last_ptr;
	int ictok=0;
	smb_ucs2_t **ret, **iret;

	if (!sep)
		sep = sep_list;

	while(*s && strchr_w(sep,*s))
		s++;

	/*
	 * Nothing left ?
	 */

	if (!*s)
		return(NULL);

	do {
		ictok++;
		while(*s && (!strchr_w(sep,*s)))
			s++;
		while(*s && strchr_w(sep,*s))
			*s++=0;
	} while(*s);

	*ctok = ictok;
	s = last_ptr;

	if (!(ret = iret = g_new(smb_ucs2_t *, ictok)))
		return NULL;
  
	while(ictok--) {
		*iret++=s;
		while(*s++)
			;
		while(!*s)
			s++;
	}

	return ret;
}

/*******************************************************************
 Case insensitive string compararison.
********************************************************************/

int StrCaseCmp_w(const smb_ucs2_t *s, const smb_ucs2_t *t)
{
	/* 
	 * Compare until we run out of string, either t or s, or find a difference.
	 */

	while (*s && *t && RtlUpcaseUnicodeChar(*s) == RtlUpcaseUnicodeChar(*t)) {
		s++;
		t++;
	}

	return(RtlUpcaseUnicodeChar(*s) - RtlUpcaseUnicodeChar(*t));
}

/*******************************************************************
 Case insensitive string compararison, length limited.
 n is in ucs2 units.
********************************************************************/

int StrnCaseCmp_w(const smb_ucs2_t *s, const smb_ucs2_t *t, size_t n)
{
	/*
	 * Compare until we run out of string, either t or s, or chars.
	 */

	while (n && *s && *t && RtlUpcaseUnicodeChar(*s) == RtlUpcaseUnicodeChar(*t)) {
		s++;
		t++;
		n--;
	}

    /*
	 * Not run out of chars - strings are different lengths.
	 */

    if (n) 
      return(RtlUpcaseUnicodeChar(*s) - RtlUpcaseUnicodeChar(*t));

    /*
	 * Identical up to where we run out of chars, 
	 * and strings are same length.
	 */

	return(0);
}

/*******************************************************************
 Compare 2 strings.
********************************************************************/

BOOL strequal_w(const smb_ucs2_t *s1, const smb_ucs2_t *s2)
{
	if (s1 == s2)
		return(True);
	if (!s1 || !s2)
		return(False);
  
	return(StrCaseCmp_w(s1,s2)==0);
}

/*******************************************************************
 Compare 2 strings up to and including the nth char. n is in ucs2
 units.
******************************************************************/

BOOL strnequal_w(const smb_ucs2_t *s1,const smb_ucs2_t *s2,size_t n)
{
	if (s1 == s2)
		return(True);
	if (!s1 || !s2 || !n)
		return(False);
  
	return(StrnCaseCmp_w(s1,s2,n)==0);
}

/*******************************************************************
 Compare 2 strings (case sensitive).
********************************************************************/

BOOL strcsequal_w(const smb_ucs2_t *s1,const smb_ucs2_t *s2)
{
	if (s1 == s2)
		return(True);
	if (!s1 || !s2)
		return(False);
  
	return(strcmp_w(s1,s2)==0);
}

/*******************************************************************
 Convert a string to lower case.
 return True if any char is converted
********************************************************************/
BOOL strlower_w(smb_ucs2_t *s)
{
	BOOL ret = False;
	while (*s) {
		smb_ucs2_t v = tolower_w(*s);
		if (v != *s) {
			*s = v;
			ret = True;
		}
		s++;
	}
	return ret;
}

/*******************************************************************
 Convert a string to upper case.
 return True if any char is converted
********************************************************************/
BOOL strupper_w(smb_ucs2_t *s)
{
	BOOL ret = False;
	while (*s) {
		smb_ucs2_t v = RtlUpcaseUnicodeChar(*s);
		if (v != *s) {
			*s = v;
			ret = True;
		}
		s++;
	}
	return ret;
}

/*******************************************************************
 Convert a string to "normal" form.
********************************************************************/

void strnorm_w(smb_ucs2_t *s)
{
	extern int case_default;
	if (case_default == CASE_UPPER)
		strupper_w(s);
	else
		strlower_w(s);
}

/*******************************************************************
 Check if a string is in "normal" case.
********************************************************************/

BOOL strisnormal_w(smb_ucs2_t *s)
{
	extern int case_default;
	if (case_default == CASE_UPPER)
		return(!strhaslower_w(s));

	return(!strhasupper_w(s));
}

/****************************************************************************
 String replace.
****************************************************************************/

void string_replace_w(smb_ucs2_t *s, smb_ucs2_t oldc, smb_ucs2_t newc)
{
	while (*s) {
		if (oldc == *s)
			*s = newc;
		s++;
	}
}

/*******************************************************************
 Skip past some strings in a buffer. n is in bytes.
********************************************************************/

smb_ucs2_t *skip_string_w(smb_ucs2_t *buf,size_t n)
{
	while (n--)
		buf += (strlen_w(buf)*sizeof(smb_ucs2_t)) + 1;
	return(buf);
}

/*******************************************************************
 Count the number of characters in a string. Same as strlen_w in
 smb_ucs2_t string units.
********************************************************************/

size_t str_charnum_w(const smb_ucs2_t *s)
{
	return strlen_w(s);
}

/*******************************************************************
 Trim the specified elements off the front and back of a string.
********************************************************************/

BOOL trim_string_w(smb_ucs2_t *s,const smb_ucs2_t *front,const smb_ucs2_t *back)
{
	BOOL ret = False;
	size_t front_len = (front && *front) ? strlen_w(front) : 0;
	size_t back_len = (back && *back) ? strlen_w(back) : 0;
	size_t s_len;

	while (front_len && strncmp_w(s, front, front_len) == 0) {
		smb_ucs2_t *p = s;
		ret = True;

		while (1) {
			if (!(*p = p[front_len]))
				break;
			p++;
		}
	}

	if(back_len) {
		s_len = strlen_w(s);
		while ((s_len >= back_len) && 
			(strncmp_w(s + s_len - back_len, back, back_len)==0)) {
			ret = True;
			s[s_len - back_len] = 0;
			s_len = strlen_w(s);
		}
	}

	return(ret);
}

/****************************************************************************
 Does a string have any uppercase chars in it ?
****************************************************************************/

BOOL strhasupper_w(const smb_ucs2_t *s)
{
	while (*s) {
		if (isupper_w(*s))
			return(True);
		s++;
	}
	return(False);
}

/****************************************************************************
 Does a string have any lowercase chars in it ?
****************************************************************************/

BOOL strhaslower_w(const smb_ucs2_t *s)
{
	while (*s) {
		if (islower_w(*s))
			return(True);
		s++;
	}
	return(False);
}

/****************************************************************************
 Find the number of 'c' chars in a string.
****************************************************************************/

size_t count_chars_w(const smb_ucs2_t *s,smb_ucs2_t c)
{
	size_t count=0;

	while (*s) {
		if (*s == c)
			count++;
		s++;
	}
	return(count);
}

/*******************************************************************
 Return True if a string consists only of one particular character.
********************************************************************/

BOOL str_is_all_w(const smb_ucs2_t *s,smb_ucs2_t c)
{
	if(s == NULL)
		return False;
	if(!*s)
		return False;

	while (*s) {
		if (*s != c)
			return False;
		s++;
	}
	return True;
}

/*******************************************************************
 Paranoid strcpy into a buffer of given length (includes terminating
 zero. Strips out all but 'a-Z0-9' and replaces with '_'. Deliberately
 does *NOT* check for multibyte characters. Don't change it !
 maxlength is in ucs2 units.
********************************************************************/

smb_ucs2_t *alpha_strcpy_w(smb_ucs2_t *dest, const smb_ucs2_t *src, size_t maxlength)
{
	size_t len, i;

	if (!dest) {
		DEBUG(0,("ERROR: NULL dest in alpha_strcpy_w\n"));
		return NULL;
	}

	if (!src) {
		*dest = 0;
		return dest;
	}  

	len = strlen_w(src);
	if (len >= maxlength)
		len = maxlength - 1;

	for(i = 0; i < len; i++) {
		smb_ucs2_t val = src[i];
		if(isupper_w(val) ||islower_w(val) || isdigit_w(val))
			dest[i] = src[i];
		else
			dest[i] = (smb_ucs2_t)'_';
	}

	dest[i] = 0;

	return dest;
}

/****************************************************************************
 Like strncpy but always null terminates. Make sure there is room !
 The variable n should always be one less than the available size and is in
 ucs2 units.
****************************************************************************/

smb_ucs2_t *StrnCpy_w(smb_ucs2_t *dest,const smb_ucs2_t *src,size_t n)
{
	smb_ucs2_t *d = dest;
	if (!dest)
		return(NULL);
	if (!src) {
		*dest = 0;
		return(dest);
	}

	while (n-- && (*d++ = *src++))
		;
	*d = 0;
	return(dest);
}

/****************************************************************************
 Like strncpy but copies up to the character marker. Always null terminates.
 returns a pointer to the character marker in the source string (src).
 n is in ucs2 units.
****************************************************************************/

smb_ucs2_t *strncpyn_w(smb_ucs2_t *dest, const smb_ucs2_t *src,size_t n, smb_ucs2_t c)
{
	smb_ucs2_t *p;
	size_t str_len;

	p = strchr_w(src, c);
	if (p == NULL) {
		char *ascii_version = unicode_to_fancy_ascii(&c, 1);
		DEBUG(5, ("strncpyn_w: separator character (%s) not found\n",
			  ascii_version));
		safe_free(ascii_version);
		return NULL;
	}

	str_len = PTR_DIFF(p, src) / sizeof(smb_ucs2_t) + 1;
	safe_strcpy_w(dest, src, MIN(n, str_len));

	return p;
}

/*************************************************************
 Routine to get hex characters and turn them into a 16 byte array.
 The array can be variable length, and any non-hex-numeric
 characters are skipped.  "0xnn" or "0Xnn" is specially catered
 for. len is in bytes.
 Valid examples: "0A5D15"; "0x15, 0x49, 0xa2"; "59\ta9\te3\n"
**************************************************************/

static smb_ucs2_t hexprefix[] = { (smb_ucs2_t)'0', (smb_ucs2_t)'x', 0 };
static smb_ucs2_t hexchars[] = { (smb_ucs2_t)'0', (smb_ucs2_t)'1', (smb_ucs2_t)'2', (smb_ucs2_t)'3',
								(smb_ucs2_t)'4', (smb_ucs2_t)'5', (smb_ucs2_t)'6', (smb_ucs2_t)'7',
								(smb_ucs2_t)'8', (smb_ucs2_t)'9', (smb_ucs2_t)'A', (smb_ucs2_t)'B',
								(smb_ucs2_t)'C', (smb_ucs2_t)'D', (smb_ucs2_t)'E', (smb_ucs2_t)'F', 0 };

size_t strhex_to_str_w(char *p, size_t len, const smb_ucs2_t *strhex)
{
	size_t i;
	size_t num_chars = 0;
	unsigned char   lonybble, hinybble;
	smb_ucs2_t *p1 = NULL, *p2 = NULL;

	/*
	 * Convert to smb_ucs2_t units.
	 */

	len /= sizeof(smb_ucs2_t);

	for (i = 0; i < len && strhex[i] != 0; i++) {
		if (strnequal_w(hexchars, hexprefix, 2)) {
			i++; /* skip two chars */
			continue;
		}

		if (!(p1 = strchr_w(hexchars, RtlUpcaseUnicodeChar(strhex[i]))))
			break;

		i++; /* next hex digit */

		if (!(p2 = strchr_w(hexchars, RtlUpcaseUnicodeChar(strhex[i]))))
			break;

		/* get the two nybbles */
		hinybble = (PTR_DIFF(p1, hexchars)/sizeof(smb_ucs2_t));
		lonybble = (PTR_DIFF(p2, hexchars)/sizeof(smb_ucs2_t));

		p[num_chars] = (hinybble << 4) | lonybble;
		num_chars++;

		p1 = NULL;
		p2 = NULL;
	}
	return num_chars;
}

/****************************************************************************
 Check if a string is part of a list.
****************************************************************************/

BOOL in_list_w(smb_ucs2_t *s,smb_ucs2_t *list,BOOL casesensitive)
{
	wpstring tok;
	smb_ucs2_t *p=list;

	if (!list)
		return(False);

	while (next_token_w(&p,tok,LIST_SEP_W,sizeof(tok))) {
		if (casesensitive) {
			if (strcmp_w(tok,s) == 0)
				return(True);
		} else {
			if (StrCaseCmp_w(tok,s) == 0)
				return(True);
		}
	}
	return(False);
}

/* This is used to prevent lots of mallocs of size 2 */
static smb_ucs2_t null_string;

/****************************************************************************
 Set a string value, allocing the space for the string.
****************************************************************************/

BOOL string_init_w(smb_ucs2_t **dest,const smb_ucs2_t *src)
{
	size_t l;

	if (!src)     
		src = &null_string;

	l = strlen_w(src);

	if (l == 0)
		*dest = &null_string;
	else {
		(*dest) = g_new(smb_ucs2_t, l + 1);
		if ((*dest) == NULL) {
			DEBUG(0,("Out of memory in string_init_w\n"));
			return False;
		}
		memcpy(*dest, src, (l+1) * sizeof(smb_ucs2_t));
	}
	return(True);
}

/****************************************************************************
 Free a string value.
****************************************************************************/

void string_free_w(smb_ucs2_t **s)
{
	if (!s || !(*s))
		return;
	if (*s == &null_string)
		*s = NULL;
	SAFE_FREE(*s);
}

/****************************************************************************
 Set a string value, allocing the space for the string, and deallocating any 
 existing space.
****************************************************************************/

BOOL string_set_w(smb_ucs2_t **dest,const smb_ucs2_t *src)
{
	string_free_w(dest);

	return(string_init_w(dest,src));
}

/****************************************************************************
 Substitute a string for a pattern in another string. Make sure there is 
 enough room !

 This routine looks for pattern in s and replaces it with 
 insert. It may do multiple replacements.

 Any of " ; ' $ or ` in the insert string are replaced with _
 if len==0 then no length check is performed
 len is in ucs2 units.
****************************************************************************/

void string_sub_w(smb_ucs2_t *s,const smb_ucs2_t *pattern,const smb_ucs2_t *insert, size_t len)
{
	smb_ucs2_t *p;
	ssize_t ls,lp,li, i;

	if (!insert || !pattern || !s)
		return;

	ls = (ssize_t)strlen_w(s);
	lp = (ssize_t)strlen_w(pattern);
	li = (ssize_t)strlen_w(insert);

	if (!*pattern)
		return;
	
	while (lp <= ls && (p = strstr_w(s,pattern))) {
		if (len && (ls + (li-lp) >= len)) {
			char *ascii_version = unicode_to_fancy_ascii(pattern, lp);
			DEBUG(0,("ERROR: string overflow by %d in string_sub_w(%.50s, %d)\n", 
				 (int)(sizeof(smb_ucs2_t)*(ls + (li-lp) - len)),
				 ascii_version, (int)len*sizeof(smb_ucs2_t)));
			safe_free(ascii_version);
			break;
		}
		if (li != lp)
			memmove(p+li,p+lp,sizeof(smb_ucs2_t)*(strlen_w(p+lp)+1));

		for (i=0;i<li;i++) {
			switch (insert[i]) {
			case (smb_ucs2_t)'`':
			case (smb_ucs2_t)'"':
			case (smb_ucs2_t)'\'':
			case (smb_ucs2_t)';':
			case (smb_ucs2_t)'$':
			case (smb_ucs2_t)'%':
			case (smb_ucs2_t)'\r':
			case (smb_ucs2_t)'\n':
				p[i] = (smb_ucs2_t)'_';
				break;
			default:
				p[i] = insert[i];
			}
		}
		s = p + li;
		ls += (li-lp);
	}
}

void fstring_sub_w(smb_ucs2_t *s,const smb_ucs2_t *pattern,const smb_ucs2_t *insert)
{
	string_sub_w(s, pattern, insert, sizeof(wfstring));
}

void pstring_sub_w(smb_ucs2_t *s,const smb_ucs2_t *pattern,smb_ucs2_t *insert)
{
	string_sub_w(s, pattern, insert, sizeof(wpstring));
}

/****************************************************************************
 Similar to string_sub() but allows for any character to be substituted. 
 Use with caution !
 if len==0 then no length check is performed.
****************************************************************************/

void all_string_sub_w(smb_ucs2_t *s,const smb_ucs2_t *pattern,const smb_ucs2_t *insert, size_t len)
{
	smb_ucs2_t *p;
	ssize_t ls,lp,li;

	if (!insert || !pattern || !s)
		return;

	ls = (ssize_t)strlen_w(s);
	lp = (ssize_t)strlen_w(pattern);
	li = (ssize_t)strlen_w(insert);

	if (!*pattern)
		return;
	
	while (lp <= ls && (p = strstr_w(s,pattern))) {
		if (len && (ls + (li-lp) >= len)) {
			char *ascii_version = unicode_to_fancy_ascii(pattern, lp);
			DEBUG(0,("ERROR: string overflow by %d in all_string_sub_w(%.50s, %d)\n", 
				 (int)(sizeof(smb_ucs2_t)*(ls + (li-lp) - len)),
				 ascii_version, (int)len*sizeof(smb_ucs2_t)));
			safe_free(ascii_version);
			break;
		}
		if (li != lp)
			memmove(p+li,p+lp,sizeof(smb_ucs2_t)*(strlen_w(p+lp)+1));

		memcpy(p, insert, li*sizeof(smb_ucs2_t));
		s = p + li;
		ls += (li-lp);
	}
}

/****************************************************************************
 Splits out the front and back at a separator.
****************************************************************************/

void split_at_last_component_w(smb_ucs2_t *path, smb_ucs2_t *front, smb_ucs2_t sep, smb_ucs2_t *back)
{
	smb_ucs2_t *pend = path + strlen_w(path);
	smb_ucs2_t *p = strrchr_w(path, sep);

	if (p)
		*p = 0;
	if (front) {
		if (p)
			memcpy(front, path, PTR_DIFF(p + 1, front));
		else
			memcpy(front, path, PTR_DIFF(pend + 1, front));
	}
	if (back) {
		if (p)
			memcpy(back, p + 1, PTR_DIFF(pend, p));
		else
			back[0] = 0;
	}
	if (p)
		*p = sep;
}


/****************************************************************************
 Write an octal as a string.
****************************************************************************/
#ifndef NOONS_SMALL_BUILD
static
smb_ucs2_t *octal_string_w(int i)
{
	static smb_ucs2_t wret[64];
	char ret[64];

	if (i == -1)
		slprintf(ret, sizeof(ret)-1, "-1");
	else 
		slprintf(ret, sizeof(ret)-1, "0%o", i);

	convert_string(CH_UNIX, CH_UINT16,
		       ret, strlen(ret)+1,
		       wret, sizeof(wret));

	return wret;
}
#endif

/****************************************************************************
 Truncate a string at a specified length.
 length is in ucs2 units.
****************************************************************************/

smb_ucs2_t *string_truncate_w(smb_ucs2_t *s, size_t length)
{
	if (s && strlen_w(s) > length)
		s[length] = 0;

	return s;
}



/* const */
char *dos_to_unix_static(const char *src)
{
	size_t srclen;
	void *ret_val = NULL;

	if (! src)
		return NULL;

	srclen = strlen(src) + 1;
	convert_string_allocate(CH_DOS, CH_UNIX,
				src, srclen,
				&ret_val);

	if (! ret_val)
	{
		DEBUG(1, ("WARNING: dos_to_unix_static: conversion failed, "
			  "returning original [%.50s]\n", src));
		ret_val = strdup(src);
	}

	tss_addptr(ret_val, strlen(ret_val) + 1);

	return ret_val;
}


size_t unix_to_dos(char *str)
{
	size_t len, ret;

	if (! str)
		return 0;

	len = strlen(str) + 1;

	ret = convert_string(CH_UNIX, CH_DOS,
			     str, len,
			     str, len);

	str[len-1] = '\0'; /* nul terminate in any case */

	return ret;
}
