/* 
   Unix SMB/Netbios implementation.
   Version 3.0
   Locking functions
   Copyright (C) Andrew Tridgell 1992-2000
   Copyright (C) Jeremy Allison 1992-2000
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   Revision History:

   12 aug 96: Erik.Devriendt@te6.siemens.be
   added support for shared memory implementation of share mode locking

   May 1997. Jeremy Allison (jallison@whistle.com). Modified share mode
   locking to deal with multiple share modes per open file.

   September 1997. Jeremy Allison (jallison@whistle.com). Added oplock
   support.

   rewrtten completely to use new tdb code. Tridge, Dec '99

   Added POSIX locking support. Jeremy Allison (jeremy@valinux.com), Apr. 2000.
*/

#include "includes.h"
#include "libsamba.h"
#include "others.h"

/**
 * \brief Unpack a stored locking entry
 */
void locking_db_unpack(TDB_DATA *dbuf,
		       struct locking_data **data,
		       share_mode_entry **shares,
		       char **name)
{
	*data = (struct locking_data *)dbuf->dptr;
	if (shares)
		*shares = (share_mode_entry *)(dbuf->dptr + sizeof(struct locking_data));

	if (name)
		*name = dbuf->dptr + sizeof(struct locking_data)
			+ (*data)->u.num_share_mode_entries*sizeof(share_mode_entry);
}


/* the locking database handle */
static TDB_CONTEXT *tdb;


BOOL locking_db_init(int read_only)
{
	if (tdb)
		return True;

	tdb = tdb_open (tmp_path ("locking.tdb"),
			0, read_only ? TDB_DEFAULT : TDB_CLEAR_IF_FIRST,
			read_only?O_RDONLY:O_RDWR|O_CREAT,
			0644);

	if (!tdb) {
		DEBUG(0,("ERROR: Failed to initialise locking database\n"));
		return False;
	}

	return True;
}

/**
 * \brief Deinitialize the locking db.
 */
BOOL locking_db_end(void)
{
	if (tdb) {
		if (tdb_close(tdb) != 0)
			return False;
		tdb = NULL;
	}

	return True;
}

TDB_CONTEXT *locking_db_get_tdb(void)
{
	return tdb;
}


/****************************************************************************
traverse the whole database with this function, calling traverse_callback
on each share mode
****************************************************************************/

static int traverse_fn(TDB_CONTEXT *the_tdb, TDB_DATA kbuf, TDB_DATA dbuf, 
                       void* state)
{
	struct locking_data *data;
	share_mode_entry *shares;
	char *name;
	int i;

	SHAREMODE_FN(traverse_callback) = (SHAREMODE_FN_CAST())state;

	locking_db_unpack(&dbuf, &data, &shares, &name);

	for (i=0;i<data->u.num_share_mode_entries;i++) {
		traverse_callback(&shares[i], name);
	}
	return 0;
}

/*******************************************************************
 Call the specified function on each entry under management by the
 share mode system.
********************************************************************/

int share_mode_forall(SHAREMODE_FN(fn))
{
	if (!tdb)
		return 0;
	return tdb_traverse(tdb, traverse_fn, (void*)fn);
}
