/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Andrew Tridgell              1992-2000,
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-2000,
 *  Copyright (C) Paul Ashton                  1997-2000,
 *  Copyright (C) Jeremy Allison               1998-2000,
 *  Copyright (C) Sander Striker                    2000
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "includes.h"
#include "libsamba.h"
#include "tng_misc.h"
#include "libsmb.h"
#include "rpc_parse.h"
#include "rpc_client.h"
#include "nterr.h"
#include "winerror.h"
#include "sids.h"
#include "others.h"


extern pstring global_myname;


/******************************************************************
 gets a machine password entry.  checks access rights of the host.
 ******************************************************************/
static uint32 direct_samr_userinfo(const UNISTR2 *uni_user,
				   uint16 level,
				   SAM_USERINFO_CTR * ctr,
				   DOM_GID ** gids,
				   uint32 *num_grps, BOOL set)
{
	POLICY_HND sam_pol;
	POLICY_HND dom_pol;
	POLICY_HND usr_pol;
	uint32 user_rid = 0xffffffff;

	uint32 status_sam = NT_STATUS_NOPROBLEMO;
	uint32 status_dom = NT_STATUS_NOPROBLEMO;
	uint32 status_usr = NT_STATUS_NOPROBLEMO;
	uint32 status_pwd = NT_STATUS_NOPROBLEMO;
	uint32 status_grp = NT_STATUS_NOPROBLEMO;

	if (ctr == NULL)
	{
		return NT_STATUS_INVALID_PARAMETER;
	}

	if (!set)
	{
		ZERO_STRUCTP(ctr);
	}

	status_sam =
		_samr_connect(NULL, SEC_RIGHTS_MAXIMUM_ALLOWED, &sam_pol);
	if (status_sam == NT_STATUS_NOPROBLEMO)
	{
		status_dom = _samr_open_domain(&sam_pol,
					       SEC_RIGHTS_MAXIMUM_ALLOWED,
					       &global_sam_sid, &dom_pol);
	}
	if (status_dom == NT_STATUS_NOPROBLEMO)
	{
		uint32 type;
		uint32 num_rids;
		uint32 num_types;

		status_usr = _samr_lookup_names(&dom_pol, 1, 0x3e8,
						1, uni_user,
						&num_rids, &user_rid,
						&num_types, &type);
		if ((status_usr == NT_STATUS_NOPROBLEMO)
		    && (type != SID_NAME_USER))
		{
			status_usr = NT_STATUS_ACCESS_DENIED;
		}
		if (status_usr == NT_STATUS_NONE_MAPPED)
			status_usr = NT_STATUS_NO_SUCH_USER;
	}
	if (status_usr == NT_STATUS_NOPROBLEMO)
	{
		status_usr = _samr_open_user(&dom_pol,
					     SEC_RIGHTS_MAXIMUM_ALLOWED,
					     user_rid, &usr_pol);
	}
	DEBUG(10, ("_samr_open_user: status %x\n", status_usr));
	if (status_usr == NT_STATUS_NOPROBLEMO)
	{
		if (!set && gids != NULL && num_grps != NULL)
		{
			status_grp = _samr_query_usergroups(&usr_pol,
							    num_grps, gids);
		}
		DEBUG(10,
		      ("_samr_query_usergroups: status %x\n", status_grp));
		if (set)
		{
			status_pwd = _samr_set_userinfo(&usr_pol, level, ctr);
		}
		else
		{
			status_pwd =
				_samr_query_userinfo(&usr_pol, level, ctr);
		}
	}
	if (status_usr == NT_STATUS_NOPROBLEMO)
		_samr_close(&usr_pol);
	if (status_dom == NT_STATUS_NOPROBLEMO)
		_samr_close(&dom_pol);
	if (status_sam == NT_STATUS_NOPROBLEMO)
		_samr_close(&sam_pol);

	if (status_pwd != NT_STATUS_NOPROBLEMO)
	{
		return status_pwd;
	}

	if (status_usr != NT_STATUS_NOPROBLEMO)
	{
		return status_usr;
	}

	if (status_grp != NT_STATUS_NOPROBLEMO)
	{
		return status_grp;
	}

	if (ctr->info.id == NULL)
	{
		return NT_STATUS_NO_MEMORY;
	}

	return NT_STATUS_NOPROBLEMO;
}

/******************************************************************
 gets a machine password entry.  checks access rights of the host.
 ******************************************************************/
static BOOL get_md4pw(char *md4pw, const UNISTR2 *uni_trust_acct)
{
	SAM_USERINFO_CTR ctr;
	uint32 status_pwd = NT_STATUS_NOPROBLEMO;
	fstring trust_name;
	size_t namelen;

	ZERO_STRUCT(ctr);

	unistr2_to_ascii(trust_name, uni_trust_acct, sizeof(trust_name) - 1);

	strlower(trust_name);
	if (! *trust_name)
	{
		return False;
	}
	
	namelen = strlen(trust_name);
	if (trust_name[namelen - 1] != '$')
	{
		DEBUG(1, ("Security: trust account [%s] does not end in $\n",
			  trust_name));
		return False;
	}
	trust_name[namelen - 1] = '\0';

#if 0
	/*
	 * Currently this code is redundent as we already have a filter
	 * by hostname list. What this code really needs to do is to 
	 * get a hosts allowed/hosts denied list from the SAM database
	 * on a per user basis, and make the access decision there.
	 * I will leave this code here for now as a reminder to implement
	 * this at a later date. JRA.
	 */

	if (!allow_access(lp_domain_hostsdeny(), lp_domain_hostsallow(),
			  client_connection_name(), client_connection_addr()))
	{
		DEBUG(0,
		      ("get_md4pw: Workstation %s denied access to domain\n",
		       trust_acct));
		return False;
	}
#endif /* 0 */

	if (strequal(trust_name, global_myname)
	    || strequal(trust_name, "\\\\."))
	{
		BOOL ret;
		DEBUG(10, ("get_md4pw: loop-back, use $MACHINE.ACC\n"));
		become_system();
		ret = msrpc_lsa_query_trust_passwd("\\\\.", "$MACHINE.ACC",
						   md4pw, NULL);
		unbecome_system();
		return ret;
	}

	/*
	 * must do all this as root
	 */
	become_root();
	status_pwd = direct_samr_userinfo(uni_trust_acct, 0x12, &ctr,
					  NULL, NULL, False);
	unbecome_root();

	if (status_pwd == NT_STATUS_NOPROBLEMO)
	{
		memcpy(md4pw, ctr.info.id12->nt_pwd, 16);
		dump_data_pw("md4pw", md4pw, 16);
	}
	else
	{
		DEBUG(1, ("get_md4pw(%s): direct_samr_userinfo returned %s\n",
			  trust_name, get_nt_error_msg(status_pwd)));
	}

	free_samr_userinfo_ctr(&ctr);

	return status_pwd == NT_STATUS_NOPROBLEMO;
}

/*************************************************************************
 find out, if a domain is realy a remote one, which should be
 handled remotely
 *************************************************************************/

static BOOL is_our_canonical_domain(const char *domain)
{
	if ((domain == NULL)
	    || strequal(domain, global_sam_name)
	    || strequal(domain, "")
	    || strequal(domain, global_myname)
	    || strequal(domain, "Builtin"))
	{
		/* this is a "canonical" name for one of our domains */
		DEBUG(10, ("is_our_canonical_domain: %s -> True\n",
			   domain));
		return True;
	}
	DEBUG(10, ("is_our_canonical_domain: %s -> False\n",
		   domain));
	/* 
	 * Note: this can still be handled localy, it's
	 * just not a canonical name for any of our domains.
	 * See the next function.
	 */
	return False;
}


static BOOL is_realy_remote_domain(const char *domain)
{
	/* Some shortcuts */
	if (is_our_canonical_domain(domain))
	{
		/* this is local! */
		DEBUG(10, ("is_realy_remote_domain: %s -> False\n",
			   domain));
		return False;
	}

	switch (lp_server_role()) {
		case ROLE_STANDALONE:
		{
			/* for a stand alone box, everything is local */
			DEBUG(10, ("is_realy_remote_domain: %s -> False\n",
				   domain));
			return False;
		}
		case ROLE_DOMAIN_MEMBER:
		{
			/* 
			 * For a domain member, everything is remote,
			 * that can not be handled localy
			 */
			DEBUG(10, ("is_realy_remote_domain: %s -> True\n",
				   domain));
			return True;
		}
		default:
		{
			/*
			 * everything else is a domain controller
			 * and they have trusts and stuff
			 */
			break;
		}
	}
	/*
	 * Okay, this could be a trusted domain, check that.
	 */
	if (get_trusted_serverlist(domain))
	{
		/* looks like, it is a trusted domain */
		DEBUG(10, ("is_realy_remote_domain: %s -> True\n",
			   domain));
		return True;
	}
	/*
	 * not even a trusted domain, so we have
	 * to handle this localy
	 */
	DEBUG(10, ("is_realy_remote_domain: %s -> False\n",
		   domain));
	return False;
}


static uint32 remote_interactive(const NET_ID_INFO_1 * id1,
				 struct dcinfo *dc, NET_USER_INFO_3 * usr)
{
	const UNISTR2 *uni_samusr = &id1->uni_user_name;
	const UNISTR2 *uni_samnam = &id1->uni_domain_name;
	fstring user;
	fstring domain;
	uint32 status;
	char nt_pwd[16];
	char lm_pwd[16];
	uchar key[16];

	memset(key, 0, 16);
	memcpy(key, dc->sess_key, 8);

	memcpy(lm_pwd, id1->lm_owf.data, 16);
	memcpy(nt_pwd, id1->nt_owf.data, 16);

	dump_data_pw("key:", key, 16);

	dump_data_pw("lm owf password:", lm_pwd, 16);
	dump_data_pw("nt owf password:", nt_pwd, 16);

	SamOEMhash((uchar *) lm_pwd, key, 0);
	SamOEMhash((uchar *) nt_pwd, key, 0);

	dump_data_pw("decrypt of lm owf password:", lm_pwd, 16);
	dump_data_pw("decrypt of nt owf password:", nt_pwd, 16);

	unistr2_to_ascii(user, uni_samusr, sizeof(user) - 1);
	unistr2_to_ascii(domain, uni_samnam, sizeof(domain) - 1);

	DEBUG(5, ("BEGIN remote_interactive: %s\\%s\n", domain, user));

	become_system();
	status = check_domain_security(user, domain,
				       NULL, lm_pwd, 16, nt_pwd, 16, usr);
	unbecome_system();

	DEBUG((status ? 2 : 5), ("END remote_interactive: %s\n",
				 get_nt_error_msg(status)));

	if (status == NT_STATUS_NO_TRUST_LSA_SECRET)
		status = NT_STATUS_TRUSTED_DOMAIN_FAILURE;

	if (status != 0x0)
	{
		return status;
	}

	memset(key, 0, 16);
	memcpy(key, dc->sess_key, 8);
	dump_data_pw("key:", key, 16);
	dump_data_pw("user sess key:", usr->user_sess_key, 16);
	SamOEMhash((uchar *) usr->user_sess_key, key, 0);
	dump_data_pw("encrypt of user session key:", usr->user_sess_key, 16);

	return status;
}

/*************************************************************************
 net_login_interactive:
 *************************************************************************/
static uint32 net_login_interactive(const NET_ID_INFO_1 * id1,
				    struct dcinfo *dc)
{
	const UNISTR2 *uni_samusr = &id1->uni_user_name;
	uint32 status = NT_STATUS_NOPROBLEMO;
	SAM_USERINFO_CTR ctr;

	char nt_pwd[16];
	char lm_pwd[16];
	uchar key[16];

	memset(key, 0, 16);
	memcpy(key, dc->sess_key, 8);

	memcpy(lm_pwd, id1->lm_owf.data, 16);
	memcpy(nt_pwd, id1->nt_owf.data, 16);

	dump_data_pw("key:", key, 16);

	dump_data_pw("lm owf password:", lm_pwd, 16);
	dump_data_pw("nt owf password:", nt_pwd, 16);

	SamOEMhash((uchar *) lm_pwd, key, 0);
	SamOEMhash((uchar *) nt_pwd, key, 0);

	dump_data_pw("decrypt of lm owf password:", lm_pwd, 16);
	dump_data_pw("decrypt of nt owf password:", nt_pwd, 16);

	become_root();
	status = direct_samr_userinfo(uni_samusr, 0x12, &ctr,
				      NULL, NULL, False);
	unbecome_root();
	if (status != NT_STATUS_NOPROBLEMO)
	{
		free_samr_userinfo_ctr(&ctr);
		return status;
	}

	if (memcmp(ctr.info.id12->lm_pwd, lm_pwd, 16) != 0 ||
	    memcmp(ctr.info.id12->nt_pwd, nt_pwd, 16) != 0)
	{
		char *user_name = unistr2_to_ascii(NULL, uni_samusr, 0);
		DEBUG(2,
		      ("net_login_interactive: WRONG password for user %s\n",
		       user_name));
		safe_free(user_name);
		status = NT_STATUS_WRONG_PASSWORD;
	}

	free_samr_userinfo_ctr(&ctr);
	return status;
}

/*************************************************************************
 net_login_general:
 *************************************************************************/
static uint32 net_login_general(const NET_ID_INFO_4 * id4,
				struct dcinfo *dc, char usr_sess_key[16])
{
	fstring user;
	fstring domain;
	const char *general = id4->str_general.buffer;
	uint32 pw_len = id4->str_general.str_str_len;

	unistr2_to_ascii(user, &id4->uni_user_name, sizeof(user) - 1);
	unistr2_to_ascii(domain, &id4->uni_domain_name, sizeof(domain) - 1);

	if (pw_len >= sizeof(fstring))
	{
		return NT_STATUS_ACCESS_DENIED;
	}

	DEBUG(5, ("net_login_general: user:%s domain:%s", user, domain));
	dump_data_pw("password:\n", general, pw_len);

	DEBUG(0,
	      ("net_login_general: TODO - \"update encrypted\" disabled\n"));
	if (pass_check(user, general, pw_len, NULL,
#if 0
		       lp_update_encrypted()? update_smbpassword_file :
#endif
		       NULL))
	{
		uchar key[16];
		memset(key, 0, 16);
		memcpy(key, dc->sess_key, 8);
#ifdef DEBUG_PASSWORD
		DEBUG(100, ("key:"));
		dump_data(100, key, 16);
		DEBUG(100, ("user sess key:"));
		dump_data(100, usr_sess_key, 16);
#endif
		SamOEMhash((uchar *) usr_sess_key, key, 0);
#ifdef DEBUG_PASSWORD
		DEBUG(100, ("encrypt of user session key:"));
		dump_data(100, usr_sess_key, 16);
#endif
		return NT_STATUS_NOPROBLEMO;
	}

	return NT_STATUS_WRONG_PASSWORD;
}

/*************************************************************************
 net_login_network:
 *************************************************************************/
static uint32 remote_network(const NET_ID_INFO_2 * id2,
			     struct dcinfo *dc, NET_USER_INFO_3 * usr)
{
	const UNISTR2 *uni_samusr = &id2->uni_user_name;
	const UNISTR2 *uni_samnam = &id2->uni_domain_name;
	fstring user;
	fstring domain;
	uchar key[16];
	uint32 status;
	int nt_pw_len = id2->hdr_nt_chal_resp.str_str_len;
	int lm_pw_len = id2->hdr_lm_chal_resp.str_str_len;

	unistr2_to_ascii(user, uni_samusr, sizeof(user) - 1);
	unistr2_to_ascii(domain, uni_samnam, sizeof(domain) - 1);

	DEBUG(5, ("BEGIN remote_network: %s\\%s\n", domain, user));

	become_system();
	status = check_domain_security(user, domain,
				       id2->lm_chal,
				       (const uchar *)id2->
				       lm_chal_resp.buffer, lm_pw_len,
				       (const uchar *)id2->
				       nt_chal_resp.buffer, nt_pw_len, usr);
	unbecome_system();

	DEBUG((status ? 2 : 5), ("END remote_network: %s\n",
				 get_nt_error_msg(status)));

	if (status == NT_STATUS_NO_TRUST_LSA_SECRET)
		status = NT_STATUS_TRUSTED_DOMAIN_FAILURE;

	if (status != 0x0)
	{
		return status;
	}

	memset(key, 0, 16);
	memcpy(key, dc->sess_key, 8);
	dump_data_pw("key:", key, 16);
	dump_data_pw("user sess key:", usr->user_sess_key, 16);
	dump_data_pw("lm_pw8:", usr->padding, 16);
	SamOEMhash((uchar *) usr->padding, key, 3);
	SamOEMhash((uchar *) usr->user_sess_key, key, 0);
	dump_data_pw("encrypt of user session key:", usr->user_sess_key, 16);
	dump_data_pw("encrypt of lm_pw8:", usr->padding, 16);

	return status;
}

/*************************************************************************
 net_login_network:
 *************************************************************************/
static uint32 net_login_network(const NET_ID_INFO_2 * id2,
				uint16 acb_info,
				struct dcinfo *dc,
				char usr_sess_key[16], char lm_pw8[8])
{
	const UNISTR2 *uni_samusr = &id2->uni_user_name;
	fstring user;
	fstring domain;
	SAM_USERINFO_CTR ctr;
	uchar key[16];
	uint32 status;
	int nt_pw_len = id2->hdr_nt_chal_resp.str_str_len;
	int lm_pw_len = id2->hdr_lm_chal_resp.str_str_len;

	unistr2_to_ascii(user, uni_samusr, sizeof(user) - 1);
	unistr2_to_ascii(domain, &id2->uni_domain_name, sizeof(domain) - 1);

	become_root();
	status = direct_samr_userinfo(uni_samusr, 0x12, &ctr, NULL, NULL,
				      False);
	unbecome_root();

	if (status != NT_STATUS_NOPROBLEMO)
	{
		free_samr_userinfo_ctr(&ctr);
		return status;
	}

	DEBUG(5,
	      ("net_login_network: lm_len:%d nt_len:%d user:%s domain:%s\n",
	       lm_pw_len, nt_pw_len, user, domain));

	if (!smb_password_ok(acb_info, ctr.info.id12->lm_pwd,
			     ctr.info.id12->nt_pwd,
			     id2->lm_chal,
			     user, domain,
			     (const uchar *)id2->lm_chal_resp.buffer,
			     lm_pw_len,
			     (const uchar *)id2->nt_chal_resp.buffer,
			     nt_pw_len, usr_sess_key))
	{
		free_samr_userinfo_ctr(&ctr);
		return NT_STATUS_WRONG_PASSWORD;
	}

	memcpy(lm_pw8, ctr.info.id12->lm_pwd, 8);
	memset(key, 0, 16);
	memcpy(key, dc->sess_key, 8);
	dump_data_pw("key:", key, 16);
	dump_data_pw("user sess key:", usr_sess_key, 16);
	dump_data_pw("lm_pw8:", lm_pw8, 16);
	SamOEMhash((uchar *) lm_pw8, key, 3);
	SamOEMhash((uchar *) usr_sess_key, key, 0);
	dump_data_pw("encrypt of user session key:", usr_sess_key, 16);
	dump_data_pw("encrypt of lm_pw8:", lm_pw8, 16);

	free_samr_userinfo_ctr(&ctr);
	return NT_STATUS_NOPROBLEMO;
}


/*******************************************************************
creates a NETLOGON_INFO_3 structure.
********************************************************************/
static BOOL make_netinfo_3(NETLOGON_INFO_3 * info, uint32 flags,
			   uint32 logon_attempts)
{
	info->flags = flags;
	info->logon_attempts = logon_attempts;
	info->reserved_1 = 0x0;
	info->reserved_2 = 0x0;
	info->reserved_3 = 0x0;
	info->reserved_4 = 0x0;
	info->reserved_5 = 0x0;
	return True;
}


/*******************************************************************
creates a NETLOGON_INFO_1 structure.
********************************************************************/
static BOOL make_netinfo_1(NETLOGON_INFO_1 * info, uint32 flags,
			   uint32 pdc_status)
{
	info->flags = flags;
	info->pdc_status = pdc_status;
	return True;
}

/*******************************************************************
creates a NETLOGON_INFO_2 structure.
********************************************************************/
static BOOL make_netinfo_2(NETLOGON_INFO_2 * info, uint32 flags,
			   uint32 pdc_status, uint32 tc_status,
			   char *trusted_dc_name)
{
	int len_dc_name = strlen(trusted_dc_name);
	info->flags = flags;
	info->pdc_status = pdc_status;
	info->ptr_trusted_dc_name = 1;
	info->tc_status = tc_status;
	if (trusted_dc_name != NULL)
	{
		make_unistr2(&(info->uni_trusted_dc_name), trusted_dc_name,
			     len_dc_name + 1);
	}
	else
	{
		make_unistr2(&(info->uni_trusted_dc_name), "", 1);
	}

	return True;
}

/*************************************************************************
 _net_logon_ctrl2
 *************************************************************************/
uint32 _net_logon_ctrl2(PRS_HANDLE *con_hnd, const UNISTR2 *uni_server_name,
			uint32 function_code,
			uint32 query_level,
			uint32 switch_value,
			uint32 *reply_switch_value,
			NETLOGON_INFO * logon_info)
{
	/* lkclXXXX - guess what - absolutely no idea what these are! */
	uint32 flags = 0x0;
	uint32 pdc_status = 0x0;
	uint32 logon_attempts = 0x0;
	uint32 tc_status = ERROR_NO_LOGON_SERVERS;
	char *trusted_domain = "test_domain";

	DEBUG(1, ("_net_logon_ctrl2: %d %d %d\n", function_code, query_level,
		  switch_value));

	*reply_switch_value = query_level;
	switch (query_level)
	{
		case 1:
		{
			make_netinfo_1(&logon_info->info1, flags, pdc_status);
			break;
		}
		case 2:
		{
			make_netinfo_2(&logon_info->info2, flags, pdc_status,
				       tc_status, trusted_domain);
			break;
		}
		case 3:
		{
			make_netinfo_3(&logon_info->info3, flags,
				       logon_attempts); break;
		}
		default:
		{
			/* take a guess at an error code... */
			DEBUG(1, ("_net_logon_ctrl2: Unknown query %d\n",
				  query_level));
			return NT_STATUS_INVALID_INFO_CLASS;
			break;
		}
	}

	return NT_STATUS_NOPROBLEMO;
}

/*************************************************************************
 _net_trust_dom_list
 *************************************************************************/
uint32 _net_trust_dom_list(PRS_HANDLE *con_hnd, const UNISTR2 *uni_server_name,
			   BUFFER2 * uni_trust_dom_name)
{
	char **doms = NULL;
	uint32 num_doms = 0;
	uint32 i;

	enumtrustdoms(&doms, &num_doms);
	for (i = 0; i < num_doms; i++)
		strupper(doms[i]);
	make_buffer2_multi(uni_trust_dom_name, doms, num_doms);
	if (num_doms == 0)
	{
		uni_trust_dom_name->buf_max_len = 0x2;
		uni_trust_dom_name->buf_len = 0x2;
	}
	uni_trust_dom_name->undoc = 0x1;
	free_char_array(num_doms, doms);
	return NT_STATUS_NOPROBLEMO;
}

/*************************************************************************
 _net_req_chal
 *************************************************************************/
uint32 _net_req_chal(PRS_HANDLE *con_hnd, const UNISTR2 *uni_logon_server,
		     const UNISTR2 *uni_logon_client,
		     const DOM_CHAL * clnt_chal, DOM_CHAL * srv_chal)
{
	fstring trust_name;
	struct dcinfo dc;
	ZERO_STRUCT(dc);

	unistr2_to_ascii(trust_name, uni_logon_client,
			 sizeof(trust_name) - 1);
	strlower(trust_name);

	DEBUG(3, ("_net_req_chal: from [%s]\n", trust_name));


	/* copy the client credentials */
	memcpy(dc.clnt_chal.data, clnt_chal->data, sizeof(clnt_chal->data));
	memcpy(dc.clnt_cred.challenge.data, clnt_chal->data,
	       sizeof(clnt_chal->data));

	/* create a server challenge for the client */
	/* Set these to random values. */

	generate_random_buffer(srv_chal->data, sizeof(srv_chal->data), False);

	/* copy the server credentials */
	memcpy(dc.srv_chal.data, srv_chal->data, sizeof(srv_chal->data));
	memcpy(dc.srv_cred.challenge.data, srv_chal->data,
	       sizeof(srv_chal->data));

	/*
	 * Fill session key with random blob, just in case we miss
	 * to check dc.authenticated
	 */
	generate_random_buffer(dc.sess_key, sizeof(dc.sess_key), False);

	dc.authenticated = False;

	/* from client / server challenges and md4 password, generate sess key */
	if (!cred_store(global_sam_name, trust_name, &dc))
	{
		return NT_STATUS_ACCESS_DENIED;
	}

	return NT_STATUS_NOPROBLEMO;
}

/*************************************************************************
 get dc from the cred db, get appropiate md4pw, calc session key
 *************************************************************************/
static BOOL dc_md4pw_sesskey(struct dcinfo *dc, const char *trust_name,
			     const UNISTR2 *uni_acct_name)
{
	if (!cred_get(global_sam_name, trust_name, dc))
	{
		DEBUG(1, ("dc_md4pw_sesskey: cred_get(%s, %s) failed\n",
			  global_sam_name, trust_name));
		return False;
	}

	if (!get_md4pw(dc->md4pw, uni_acct_name))
	{
		fstring acct_name;
		unistr2_to_ascii(acct_name, uni_acct_name,
				 sizeof(acct_name) - 1);
		DEBUG(1, ("dc_md4pw_sesskey: get_md4pw(%s) failed\n",
			  acct_name));
		return False;
	}
	cred_session_key(&dc->clnt_chal, &dc->srv_chal,
			 dc->md4pw, dc->sess_key);
	return True;
}


/*************************************************************************
 _net_auth
 *************************************************************************/
uint32 _net_auth(PRS_HANDLE *con_hnd, const UNISTR2 *uni_logon_srv,
		 const UNISTR2 *uni_acct_name,
		 uint16 sec_chan,
		 const UNISTR2 *uni_comp_name,
		 const DOM_CHAL * clnt_chal, DOM_CHAL * srv_chal)
{
	UTIME srv_time;
	fstring trust_name;
	struct dcinfo dc;

	ZERO_STRUCT(dc);
	srv_time.time = 0;

	unistr2_to_ascii(trust_name, uni_comp_name, sizeof(trust_name) - 1);

	if (!dc_md4pw_sesskey(&dc, trust_name, uni_acct_name))
	{
		return NT_STATUS_ACCESS_DENIED;
	}

	/* check that the client credentials are valid */
	if (!cred_assert(clnt_chal, dc.sess_key,
			 &(dc.clnt_cred.challenge), srv_time))
	{
		DEBUG(2, ("_net_auth: cred_assert for %s failed. "
			  "Wrong trust account password?\n", trust_name));
		return NT_STATUS_ACCESS_DENIED;
	}

	/* create server challenge for inclusion in the reply */
	cred_create(dc.sess_key, &(dc.srv_cred.challenge), srv_time,
		    srv_chal);

	/* copy the received client credentials for use next time */
	memcpy(dc.clnt_cred.challenge.data, clnt_chal->data,
	       sizeof(clnt_chal->data));
	memcpy(dc.srv_cred.challenge.data, clnt_chal->data,
	       sizeof(clnt_chal->data));

	dc.authenticated = True;

	if (!cred_store(global_sam_name, trust_name, &dc))
	{
		return NT_STATUS_ACCESS_DENIED;
	}

	return NT_STATUS_NOPROBLEMO;
}

/*************************************************************************
 _net_auth_2
 *************************************************************************/
uint32 _net_auth_2(PRS_HANDLE *con_hnd, const UNISTR2 *uni_logon_srv,
		   const UNISTR2 *uni_acct_name,
		   uint16 sec_chan,
		   const UNISTR2 *uni_comp_name,
		   const DOM_CHAL * clnt_chal,
		   const NEG_FLAGS * clnt_flgs,
		   DOM_CHAL * srv_chal, NEG_FLAGS * srv_flgs)
{
	UTIME srv_time;
	fstring trust_name;
	fstring acct_name, logon_srv;
	struct dcinfo dc;

	ZERO_STRUCT(dc);
	srv_time.time = 0;

	unistr2_to_ascii(trust_name, uni_comp_name, sizeof(trust_name) - 1);
	unistr2_to_ascii(logon_srv, uni_logon_srv, sizeof(logon_srv) - 1);
	unistr2_to_ascii(acct_name, uni_acct_name, sizeof(acct_name) - 1);

	DEBUG(3, ("_net_auth_2: comp=[%s], acct=[%s], logon_srv=[%s], "
		  "sec_chan=%d, flags=0x%x\n",
		  trust_name, acct_name, logon_srv, sec_chan,
		  clnt_flgs->neg_flags));

	if (!dc_md4pw_sesskey(&dc, trust_name, uni_acct_name))
	{
		DEBUG(1, ("_net_auth_2: dc_md4pw_sesskey failed\n"));
		return NT_STATUS_ACCESS_DENIED;
	}

	/* check that the client credentials are valid */
	if (!cred_assert(clnt_chal, dc.sess_key,
			 &(dc.clnt_cred.challenge), srv_time))
	{
		DEBUG(2, ("_net_auth_2: cred_assert for %s failed. "
			  "Wrong trust account password?\n", trust_name));
		return NT_STATUS_ACCESS_DENIED;
	}

	/* create server challenge for inclusion in the reply */
	cred_create(dc.sess_key, &(dc.srv_cred.challenge), srv_time,
		    srv_chal);

	/* copy the received client credentials for use next time */
	memcpy(dc.clnt_cred.challenge.data, clnt_chal->data,
	       sizeof(clnt_chal->data));
	memcpy(dc.srv_cred.challenge.data, clnt_chal->data,
	       sizeof(clnt_chal->data));

	dc.authenticated = True;

	if (!cred_store(global_sam_name, trust_name, &dc))
	{
		DEBUG(1, ("_net_auth_2: cred_store failed\n"));
		return NT_STATUS_ACCESS_DENIED;
	}

	srv_flgs->neg_flags = clnt_flgs->neg_flags & NETLOGON_NEG_SUPPORTED;

	if (!IS_BITS_SET_ALL(srv_flgs->neg_flags, NETLOGON_NEG_MINIMUM))
	{
		DEBUG(1, ("_net_auth_2: Minimum flags not there: 0x%x\n",
			  srv_flgs->neg_flags));
		return NT_STATUS_ACCESS_DENIED;
	}

	/* secure channel NOT to be used */
	if (!lp_server_schannel())
	{
		srv_flgs->neg_flags &= ~NETLOGON_NEG_SCHANNEL;
	}
	else if (lp_server_schannel() == True)
	{
		/* secure channel MUST be used */
		if (IS_BITS_CLR_ALL(srv_flgs->neg_flags, NETLOGON_NEG_SCHANNEL))
		{
			DEBUG(1, ("ERROR: Client does not support "
				  "SChannel, but we require so.\n"));
			return NT_STATUS_ACCESS_DENIED;
		}
	}

	return NT_STATUS_NOPROBLEMO;
}

/*************************************************************************
 _net_srv_pwset
 *************************************************************************/
uint32 _net_srv_pwset(PRS_HANDLE *con_hnd, const UNISTR2 *uni_logon_srv,
		      const UNISTR2 *uni_acct_name,
		      uint16 sec_chan,
		      const UNISTR2 *uni_comp_name,
		      const DOM_CRED * clnt_cred,
		      const uint8 pwd[16], DOM_CRED * srv_cred)
{
	pstring trust_acct;
	uchar hash3_pwd[16];
	uint32 status_pwd;
	fstring trust_name;
	struct dcinfo dc;
	SAM_USERINFO_CTR ctr;

	ZERO_STRUCT(dc);

	unistr2_to_ascii(trust_name, uni_comp_name, sizeof(trust_name) - 1);
	unistr2_to_ascii(trust_acct, uni_acct_name, sizeof(trust_acct) - 1);
	DEBUG(2, ("_net_srv_pwset: Server Password Set from [%s] for account "
		  "[%s] (%d)\n", trust_name, trust_acct, sec_chan));

	if (!cred_get(global_sam_name, trust_name, &dc))
	{
		return NT_STATUS_ACCESS_DENIED;
	}

	if (! dc.authenticated)
	{
		DEBUG(1, ("ERROR: _net_srv_pwset: dc.authenticated == False\n"));
		return NT_STATUS_ACCESS_DENIED;
	}

	/* checks and updates credentials.  creates reply credentials */
	if (!deal_with_creds(dc.sess_key, &(dc.clnt_cred),
			     clnt_cred, srv_cred))
	{
		/* lkclXXXX take a guess at a sensible error code to return... */
		return NT_STATUS_ACCESS_DENIED;
	}

	memcpy(&(dc.srv_cred), &(dc.clnt_cred), sizeof(dc.clnt_cred));
	if (!cred_store(global_sam_name, trust_name, &dc))
	{
		return NT_STATUS_ACCESS_DENIED;
	}

	/* get info for trust account */
	ZERO_STRUCT(ctr);

	become_root();
	status_pwd = direct_samr_userinfo(uni_acct_name, 0x12, &ctr,
					  NULL, NULL, False);
	unbecome_root();
	if (status_pwd != NT_STATUS_NOPROBLEMO)
	{
		free_samr_userinfo_ctr(&ctr);
		return status_pwd;
	}

	/* Some debug output, needed an iterater variable */
	{
		int i;
		DEBUG(100, ("Server password set : new given value was :\n"));
		for (i = 0; i < 16; i++)
		{
			DEBUG(100, ("%02X ", pwd[i]));
		}
		DEBUG(100, ("\n"));
	}

	cred_hash3(hash3_pwd, pwd, dc.sess_key, 0);

	/* lies!  nt and lm passwords are _not_ the same: don't care */
	memcpy(ctr.info.id12->lm_pwd, hash3_pwd, sizeof(hash3_pwd));
	memcpy(ctr.info.id12->nt_pwd, hash3_pwd, sizeof(hash3_pwd));

	become_root();
	status_pwd = direct_samr_userinfo(uni_acct_name, 0x12, &ctr,
					  NULL, NULL, True);
	unbecome_root();

	if (status_pwd != NT_STATUS_NOPROBLEMO)
	{
		return status_pwd;
	}

	return NT_STATUS_NOPROBLEMO;
}

static uint32 TooMuchInformation(uint32 status)
{
	if (status != NT_STATUS_NOPROBLEMO)
	{
		DEBUG(2, ("_net_sam_logon/TooMuchInformation: %s\n",
			  get_nt_error_msg(status)));
	}
#ifdef SANITISE_FOR_SECURITY_REASONS
	switch (status)
	{
		case NT_STATUS_ACCOUNT_DISABLED:
		case NT_STATUS_INVALID_INFO_CLASS:
		case NT_STATUS_INVALID_PARAMETER:
		case NT_STATUS_ACCESS_DENIED:
		case NT_STATUS_NOPROBLEMO:
		case NT_STATUS_PASSWORD_EXPIRED:
		{
			return status;
		}
	}
	DEBUG(10,("TooMuchInformation: sanitising status %x\n", status));
	return NT_STATUS_ACCESS_DENIED;

#else /* oh well, who cares, anyway */
	return status;
#endif
}

static uint32 check_workstation(const UNISTR2 *uni_workstation,
				const UNISTR2 *uni_workstation_list)
{
	fstring workstation, workstation_list;
	const char *p;

	if (! uni_workstation)
		return NT_STATUS_INVALID_PARAMETER;
	if (! uni_workstation_list)
		return NT_STATUS_NOPROBLEMO;

	unistr2_to_ascii(workstation, uni_workstation, sizeof(workstation) - 1);
	unistr2_to_ascii(workstation_list, uni_workstation_list, sizeof(workstation_list) - 1);

	DEBUG(3, ("check_workstation: workstation [%s] in list [%s]\n",
		  workstation, workstation_list));

	if (!workstation_list[0])
		return NT_STATUS_NOPROBLEMO;

	fstrcat(workstation, ",");
	fstrcat(workstation_list, ",");

	p = strstr(workstation_list, workstation);

	if (! p)
		return NT_STATUS_INVALID_WORKSTATION;
	if (p == workstation_list)
		return NT_STATUS_NOPROBLEMO;
	p--;
	if (*p == ',')
		return NT_STATUS_NOPROBLEMO;

	return NT_STATUS_INVALID_WORKSTATION;
}


/*************************************************************************
 _net_sam_logon
 *************************************************************************/
uint32 _net_sam_logon(PRS_HANDLE *con_hnd, const UNISTR2 *uni_logon_srv,
		      const UNISTR2 *uni_comp_name,
		      const DOM_CRED * clnt_cred,
		      uint16 logon_level,
		      const NET_ID_INFO_CTR * id_ctr,
		      uint16 validation_level,
		      DOM_CRED * srv_creds,
		      NET_USER_INFO_CTR * uctr, uint32 *auth_resp)
{
	const UNISTR2 *uni_samusr = NULL;
	const UNISTR2 *uni_domain = NULL;
	const UNISTR2 *uni_wksta_name = NULL;
	fstring nt_username;
	fstring nt_samname;
	fstring nt_workstation;
	const char *logon_level_str ="NULL";
	char *enc_user_sess_key = NULL;
	char usr_sess_key[16];
	char lm_pw8[40];
	char *padding = NULL;
	uint32 status_pwd = 0x0;
	SAM_USERINFO_CTR ctr;
	NTTIME logon_time;
	NTTIME logoff_time;
	NTTIME kickoff_time;
	NTTIME pass_last_set_time;
	NTTIME pass_can_change_time;
	NTTIME pass_must_change_time;
	UNISTR2 *uni_nt_name = NULL;
	UNISTR2 *uni_full_name = NULL;
	UNISTR2 *uni_logon_script = NULL;
	UNISTR2 *uni_profile_path = NULL;
	UNISTR2 *uni_home_dir = NULL;
	UNISTR2 *uni_dir_drive = NULL;
	UNISTR2 *uni_workstations = NULL;
	uint32 user_rid;
	uint32 group_rid;
	int num_gids = 0;
	DOM_GID *gids = NULL;
	fstring trust_name;
	struct dcinfo dc;
	uint32 acb_info;
	UNISTR2 uni_myname;
	UNISTR2 uni_sam_name;
	uint32 status = NT_STATUS_NOPROBLEMO;
	uint32 lm_pw_len = 0;
	uint32 nt_pw_len = 0;
	uint32 param_ctrl;
	BOOL anonymouse = False;
	time_t pass_must_change_time_unix, now = time(NULL);

	ZERO_STRUCT(ctr);
	ZERO_STRUCT(lm_pw8);
	ZERO_STRUCT(usr_sess_key);

	/*
	 * checks and updates credentials.  creates reply credentials
	 */

	unistr2_to_ascii(trust_name, uni_comp_name, sizeof(trust_name) - 1);
	if (!cred_get(global_sam_name, trust_name, &dc))
	{
		return NT_STATUS_ACCESS_DENIED;
	}

	if (! dc.authenticated)
	{
		DEBUG(1, ("ERROR: _net_sam_logon: dc.authenticated == False\n"));
		return NT_STATUS_ACCESS_DENIED;
	}

	if (!deal_with_creds(dc.sess_key, &dc.clnt_cred,
			     clnt_cred, srv_creds))
	{
		return NT_STATUS_ACCESS_DENIED;
	}

	memcpy(&dc.srv_cred, &dc.clnt_cred, sizeof(dc.clnt_cred));
	if (!cred_store(global_sam_name, trust_name, &dc))
	{
		return NT_STATUS_ACCESS_DENIED;
	}

	/*
	 * find the username and domain name
	 */

	switch (logon_level)
	{
		case INTERACTIVE_LOGON_TYPE:
		{
			uni_samusr = &id_ctr->auth.id1.uni_user_name;
			uni_domain = &id_ctr->auth.id1.uni_domain_name;
			uni_wksta_name = &id_ctr->auth.id1.uni_wksta_name;
			nt_pw_len = 16;
			lm_pw_len = 16;
			param_ctrl = id_ctr->auth.id1.param_ctrl;
			logon_level_str = "Interactive";
			break;
		}
		case NETWORK_LOGON_TYPE:
		{
			uni_samusr = &id_ctr->auth.id2.uni_user_name;
			uni_domain = &id_ctr->auth.id2.uni_domain_name;
			uni_wksta_name = &id_ctr->auth.id2.uni_wksta_name;
			nt_pw_len = id_ctr->auth.id2.hdr_nt_chal_resp.str_str_len;
			lm_pw_len = id_ctr->auth.id2.hdr_lm_chal_resp.str_str_len;
			param_ctrl = id_ctr->auth.id2.param_ctrl;
			logon_level_str = "Network";
			break;
		}
		case GENERAL_LOGON_TYPE:
		{
			uni_samusr = &id_ctr->auth.id4.uni_user_name;
			uni_domain = &id_ctr->auth.id4.uni_domain_name;
			uni_wksta_name = &id_ctr->auth.id4.uni_wksta_name;
			lm_pw_len = id_ctr->auth.id4.hdr_general.str_str_len;
			nt_pw_len = 0;
			param_ctrl = id_ctr->auth.id4.param_ctrl;
			logon_level_str = "General";
			break;
		}
		default:
		{
			DEBUG(2, ("SAM Logon: unsupported logon level %d\n",
				  logon_level));
			return NT_STATUS_INVALID_INFO_CLASS;
		}
	}

	unistr2_to_ascii(nt_username, uni_samusr, sizeof(nt_username) - 1);
	unistr2_to_ascii(nt_samname, uni_domain, sizeof(nt_samname) - 1);
	unistr2_to_ascii(nt_workstation, uni_wksta_name, sizeof(nt_workstation) - 1);

	DEBUG(2, ("_net_sam_logon: User:[%s\\%s] for workstation [%s] "
		  "from [%s] at level %s (0x%x).\n",
		  nt_samname, nt_username, nt_workstation, trust_name,
		  logon_level_str, param_ctrl));

	/*
	 * check if we should remote-auth to a PDC.
	 */

#if 0
	if (lp_server_role() == ROLE_STANDALONE &&
			(!strequal(nt_samname, global_myname) &&
			 !strequal(nt_samname, "")))
	{
		DEBUG(1,("_net_sam_logon: stand-alone server cannot remote-auth domain users!\n"));
		return NT_STATUS_ACCESS_DENIED;
	}
#endif

	/*
	 * call up to a PDC if it's not our own SAM database.
	 */

	if (is_realy_remote_domain(nt_samname))
	{
		DEBUG(5,("remote-auth: SAM name: %s my name: %s\n",
					global_sam_name, global_myname));

		if (validation_level != 3)
		{
			/*
			 * see the g_new of the container, below?  well,
			 * that need to be id2 for AS/U, right.
			 * and i haven't the inclination to do it, right
			 * now.  needs to make a copy-level-two-to-three
			 * function.  lkclXXXX
			 */

			DEBUG(0, ("remote-auth, level %d not supported yet\n",
				  validation_level));
			return NT_STATUS_ACCESS_DENIED;
		}
		uctr->usr.id3 = g_new(NET_USER_INFO_3, 1);
		if (uctr->usr.id3 == NULL)
		{
			return NT_STATUS_INVALID_PARAMETER;
		}

		(*auth_resp) = 1;

		status = NT_STATUS_ACCESS_DENIED;

		switch (logon_level)
		{
			case NETWORK_LOGON_TYPE:
			{
				status = remote_network(&id_ctr->auth.id2, &dc,
						      uctr->usr.id3);
				break;
			}
			case INTERACTIVE_LOGON_TYPE:
			{
				status = remote_interactive(&id_ctr->auth.id1,
							  &dc, uctr->usr.id3);
				break;
			}
			case GENERAL_LOGON_TYPE:
			default:
			{
				return NT_STATUS_ACCESS_DENIED;
			}
		}
		return TooMuchInformation(status);
	}

	/*
	 * check anonymous status
	 */

	if (lm_pw_len <= 1 && nt_pw_len == 0 &&
	    strlen(nt_username) == 0 && strlen(nt_samname) == 0)
	{
		uchar key[16];
		char *guest_acct = lp_guestaccount(-1);
		anonymouse = guest_acct && guest_acct[0];
		DEBUG(5,("Anonymous Access allowed: %s\n",
					BOOLSTR(anonymouse)));

		memset(lm_pw8, 0, 8);
		memset(key, 0, 16);
		memcpy(key, dc.sess_key, 8);
		dump_data_pw("key:", key, 16);
		dump_data_pw("lm_pw8:", lm_pw8, 16);
		SamOEMhash((uchar *) lm_pw8, key, 3);
		dump_data_pw("encrypt of lm_pw8:", lm_pw8, 16);

		padding = lm_pw8;
	}

	/*
	 * nt_samname is the requested domain for the user
	 * if this domain can't be handled properly in a
	 * remote fashion (no trusted domain, etc., see
	 * is_realy_remote_domain), we handle this localy.
	 * nt_samname is currently ignored in the following
	 * code. If we ever have multiple local domains,
	 * it must be honored.
	 * Elrond
	 */

	/*
	 * IMPORTANT: do a General Login BEFORE the others,
	 * because "update encrypted" may be enabled, which
	 * will result in the smb password entry being added.
	 *
	 * calling general login AFTER the getsampwntname() is
	 * not guaranteed to deliver.
	 */

	if (logon_level == GENERAL_LOGON_TYPE)
	{
		/* general login.  cleartext password */
		status = net_login_general(&id_ctr->auth.id4, &dc,
					  usr_sess_key);
		enc_user_sess_key = usr_sess_key;
		if (status != NT_STATUS_NOPROBLEMO)
		{
			return TooMuchInformation(status);
		}
		(*auth_resp) = 1;
	}

	/*
	 * now obtain smb passwd entry, which MAY have just been
	 * added by "update encrypted" in general login.
	 * 
	 * this also obtains the user's groups.  all this is
	 * needed for the user profile.
	 */

	if (!anonymouse)
	{
		become_root();
		status_pwd = direct_samr_userinfo(uni_samusr, 21, &ctr,
						  &gids, &num_gids, False);
		unbecome_root();

		if (status_pwd != NT_STATUS_NOPROBLEMO)
		{
			free_samr_userinfo_ctr(&ctr);

			return TooMuchInformation(status_pwd);
		}
		acb_info = ctr.info.id21->acb_info;
	}
	else
	{
		acb_info = ACB_NORMAL | ACB_PWNOTREQ;
	}

	/*
	 * check the Account Control Bits
	 */

	if (IS_BITS_SET_ALL(acb_info, ACB_DISABLED))
	{
		return TooMuchInformation(NT_STATUS_ACCOUNT_DISABLED);
	}

	if (IS_BITS_SET_ALL(acb_info, ACB_DOMTRUST))
	{
		DEBUG(2, ("--> NT_STATUS_NOLOGON_INTERDOMAIN_TRUST_ACCOUNT\n"));
		return NT_STATUS_NOLOGON_INTERDOMAIN_TRUST_ACCOUNT;
	}

	if (IS_BITS_SET_ALL(acb_info, ACB_SVRTRUST))
	{
		DEBUG(2, ("--> NT_STATUS_NOLOGON_SERVER_TRUST_ACCOUNT\n"));
		return NT_STATUS_NOLOGON_SERVER_TRUST_ACCOUNT;
	}

	if (IS_BITS_SET_ALL(acb_info, ACB_WSTRUST))
	{
		DEBUG(2, ("--> NT_STATUS_NOLOGON_WORKSTATION_TRUST_ACCOUNT\n"));
		return NT_STATUS_NOLOGON_WORKSTATION_TRUST_ACCOUNT;
	}

	/*
	 * get some user profile info
	 */

	if (!anonymouse)
	{
		logon_time = ctr.info.id21->logon_time;
		logoff_time = ctr.info.id21->logoff_time;
		kickoff_time = ctr.info.id21->kickoff_time;
		pass_last_set_time = ctr.info.id21->pass_last_set_time;
		pass_can_change_time = ctr.info.id21->pass_can_change_time;
		pass_must_change_time = ctr.info.id21->pass_must_change_time;
		uni_nt_name = &ctr.info.id21->uni_user_name;
		uni_full_name = &ctr.info.id21->uni_full_name;
		uni_home_dir = &ctr.info.id21->uni_home_dir;
		uni_dir_drive = &ctr.info.id21->uni_dir_drive;
		uni_logon_script = &ctr.info.id21->uni_logon_script;
		uni_profile_path = &ctr.info.id21->uni_profile_path;
		uni_workstations = &ctr.info.id21->uni_workstations;
		user_rid = ctr.info.id21->user_rid;
		group_rid = ctr.info.id21->group_rid;
	}
	else
	{
		init_nt_time(&logon_time);
		init_nt_time(&logoff_time);
		init_nt_time(&kickoff_time);
		init_nt_time(&pass_last_set_time);
		init_nt_time(&pass_can_change_time);
		init_nt_time(&pass_must_change_time);
		user_rid = DOMAIN_USER_RID_GUEST;
		group_rid = DOMAIN_GROUP_RID_GUESTS;
	}

	/*
	 * validate password - if required
	 */

	if (!(IS_BITS_SET_ALL(acb_info, ACB_PWNOTREQ)))
	{
		switch (logon_level)
		{
			case INTERACTIVE_LOGON_TYPE:
			{
				/* interactive login. */
				uint32 new_status = NT_STATUS_NOPROBLEMO;

				status = net_login_interactive(&id_ctr->auth.
							      id1, &dc);
				/*
				 * Check for password expiring
				 */
				pass_must_change_time_unix = nt_time_to_unix(&pass_must_change_time);
				if ((status == NT_STATUS_OK)
				    && (!IS_BITS_SET_ALL(acb_info, ACB_PWNOEXP))
				    && (pass_must_change_time_unix != 0)
				    && (pass_must_change_time_unix != -1)
				    && (pass_must_change_time_unix < now))
				{
					DEBUG(2, ("Password expired: set time: %ld, current time: %ld\n",
						  pass_must_change_time_unix,
						  now));
					status = NT_STATUS_PASSWORD_EXPIRED;
				}
				if (status == NT_STATUS_NOPROBLEMO)
					new_status = check_workstation(uni_wksta_name, uni_workstations);
				if (new_status != NT_STATUS_NOPROBLEMO)
				{
					DEBUG(2, ("Workstation list disallowed user\n"));
					status = new_status;
				}
				(*auth_resp) = 1;
				break;
			}
			case NETWORK_LOGON_TYPE:
			{
				/* network login.  lm challenge and 24 byte responses */
				status = net_login_network(&id_ctr->auth.id2,
							  acb_info, &dc,
							  usr_sess_key,
							  lm_pw8);
				padding = lm_pw8;
				enc_user_sess_key = usr_sess_key;
				(*auth_resp) = 1;
				break;
			}
			case GENERAL_LOGON_TYPE:
			{
				/* general login type ALREADY been checked */
				break;
			}
		}
	}
	else
	{
		(*auth_resp) = 1;
		status = NT_STATUS_OK;
	}

	if (status != NT_STATUS_NOPROBLEMO)
	{
		safe_free(gids);
		free_samr_userinfo_ctr(&ctr);
		if (status == NT_STATUS_WRONG_PASSWORD
		    && !is_our_canonical_domain(nt_samname))
		{
			DEBUG(1, ("_net_sam_logon: Mapping %s\\%s from "
				  "NT_STATUS_WRONG_PASSWORD to NT_STATUS_NO_SUCH_USER\n", nt_samname, nt_username));
			status = NT_STATUS_NO_SUCH_USER;
		}
		return TooMuchInformation(status);
	}

	/* lkclXXXX this is the point at which, if the login was
	   successful, that the SAM Local Security Authority should
	   record that the user is logged in to the domain.
	 */

	/* return the profile plus other bits :-) */

	make_unistr2(&uni_myname, global_myname, strlen(global_myname));
	make_unistr2(&uni_sam_name, global_sam_name, strlen(global_sam_name));
	switch (validation_level)
	{
		case 2:
		{
			uctr->usr.id2 = g_new(NET_USER_INFO_2, 1);
			if (uctr->usr.id2 == NULL)
			{
				status = NT_STATUS_INVALID_PARAMETER;
				break;
			}
			make_net_user_info2W(uctr->usr.id2, &logon_time,
					     &logoff_time, &kickoff_time,
					     &pass_last_set_time,
					     &pass_can_change_time,
					     &pass_must_change_time,
					     uni_nt_name, uni_full_name,
					     uni_logon_script,
					     uni_profile_path, uni_home_dir,
					     uni_dir_drive, 0, 0, user_rid,
					     group_rid, num_gids, gids, 0x20,
					     enc_user_sess_key, &uni_myname,
					     &uni_sam_name, padding,
					     &global_sam_sid);
			break;
		}
		case 3:
		{
			uctr->usr.id3 = g_new(NET_USER_INFO_3, 1);
			if (uctr->usr.id3 == NULL)
			{
				status = NT_STATUS_INVALID_PARAMETER;
				break;
			}
			make_net_user_info3W(uctr->usr.id3, &logon_time,
					     &logoff_time, &kickoff_time,
					     &pass_last_set_time,
					     &pass_can_change_time,
					     &pass_must_change_time,
					     uni_nt_name, uni_full_name,
					     uni_logon_script,
					     uni_profile_path, uni_home_dir,
					     uni_dir_drive, 0,
					     0,
					     user_rid, group_rid, num_gids,
					     gids, 0x20,
					     enc_user_sess_key, &uni_myname,
					     &uni_sam_name, padding,
					     &global_sam_sid, NULL);
			break;
		}
		default:
		{
			status = NT_STATUS_INVALID_INFO_CLASS;
			break;
		}
	}

	/* Free any allocated groups array. */
	safe_free(gids);
	free_samr_userinfo_ctr(&ctr);
	if (status != NT_STATUS_NOPROBLEMO)
	{
		return TooMuchInformation(status);
	}

	return NT_STATUS_NOPROBLEMO;
}

/*************************************************************************
 _net_sam_logoff
 *************************************************************************/
uint32 _net_sam_logoff(PRS_HANDLE *con_hnd, const DOM_SAM_INFO * sam_id, DOM_CRED * srv_creds)
{
	fstring trust_name;
	struct dcinfo dc;

	ZERO_STRUCT(dc);
	unistr2_to_ascii(trust_name,
			 &(sam_id->client.login.uni_comp_name),
			 sizeof(trust_name) - 1);
	if (!cred_get(global_sam_name, trust_name, &dc))
	{
		return NT_STATUS_ACCESS_DENIED;
	}

	if (! dc.authenticated)
	{
		DEBUG(1, ("ERROR: _net_sam_logoff: dc.authenticated == False\n"));
		return NT_STATUS_ACCESS_DENIED;
	}

	/* checks and updates credentials.  creates reply credentials */
	if (!deal_with_creds(dc.sess_key, &(dc.clnt_cred),
			     &(sam_id->client.cred), srv_creds))
	{
		return NT_STATUS_ACCESS_DENIED;
	}

	memcpy(&(dc.srv_cred), &(dc.clnt_cred), sizeof(dc.clnt_cred));
	if (!cred_store(global_sam_name, trust_name, &dc))
	{
		return NT_STATUS_ACCESS_DENIED;
	}

	return NT_STATUS_NOPROBLEMO;
}

/*************************************************************************
 _net_sam_sync
 *************************************************************************/
uint32 _net_sam_sync(PRS_HANDLE *con_hnd, const UNISTR2 *uni_srv_name,
		     const UNISTR2 *uni_cli_name,
		     DOM_CRED * cli_creds,
		     DOM_CRED * srv_creds,
		     uint32 database_id,
		     uint32 restart_state,
		     uint32 *sync_context,
		     uint32 max_size,
		     uint32 *num_deltas,
		     uint32 *num_deltas2,
		     SAM_DELTA_HDR * hdr_deltas, SAM_DELTA_CTR * deltas)
{
	fstring trust_name;
	int i = 0;
	POLICY_HND sam_pol;
	POLICY_HND dom_pol;
	uint32 enum_status;
	SAM_ENTRY *sam = NULL;
	UNISTR2 *uni_acct_name = NULL;
	uint32 start_idx = 0x0;
	uint32 num_sam_users = 0;
	uint32 idx;
	struct dcinfo dc;

	unistr2_to_ascii(trust_name, uni_cli_name, sizeof(trust_name) - 1);
	if (!cred_get(global_sam_name, trust_name, &dc))
	{
		return NT_STATUS_ACCESS_DENIED;
	}

	if (! dc.authenticated)
	{
		DEBUG(1, ("ERROR: _net_sam_sync: dc.authenticated == False\n"));
		return NT_STATUS_ACCESS_DENIED;
	}

	/* checks and updates credentials.  creates reply credentials */
	if (!deal_with_creds(dc.sess_key, &dc.clnt_cred,
			     cli_creds, srv_creds))
	{
		return NT_STATUS_ACCESS_DENIED;
	}

	memcpy(&dc.srv_cred, &dc.clnt_cred, sizeof(dc.clnt_cred));
	if (!cred_store(global_sam_name, trust_name, &dc))
	{
		return NT_STATUS_ACCESS_DENIED;
	}

	(*sync_context) = 1;
	if (_samr_connect(NULL, SEC_RIGHTS_MAXIMUM_ALLOWED, &sam_pol) !=
	    NT_STATUS_NOPROBLEMO)
	{
		return NT_STATUS_ACCESS_DENIED;
	}
	if (_samr_open_domain(&sam_pol, SEC_RIGHTS_MAXIMUM_ALLOWED,
			      &global_sam_sid,
			      &dom_pol) != NT_STATUS_NOPROBLEMO)
	{
		_samr_close(&sam_pol);
		return NT_STATUS_ACCESS_DENIED;
	}

	/*
	 * enumerate users, first
	 */

	do
	{
		enum_status = _samr_enum_dom_users(&dom_pol,
						   &start_idx,
						   0x0, 0x0, 0x10000,
						   &sam,
						   &uni_acct_name,
						   &num_sam_users);
	}
	while (enum_status == STATUS_MORE_ENTRIES);

	for (idx = 0; idx < num_sam_users; idx++)
	{
		SAM_USERINFO_CTR ctr;
		POLICY_HND usr_pol;
		uint32 status_usr = NT_STATUS_NOPROBLEMO;
		ZERO_STRUCT(ctr);
		status_usr =
			_samr_open_user(&dom_pol, SEC_RIGHTS_MAXIMUM_ALLOWED,
					sam[idx].rid, &usr_pol);
		if (status_usr == NT_STATUS_NOPROBLEMO
		    && _samr_query_userinfo(&usr_pol, 21,
					    &ctr) == NT_STATUS_NOPROBLEMO)
		{
			SAM_USER_INFO_21 *usr = ctr.info.id21;
			make_sam_delta_hdr(&hdr_deltas[i], 5, usr->user_rid);
			make_sam_account_info(&deltas[i].account_info,
					      &usr->uni_user_name,
					      &usr->uni_full_name,
					      usr->user_rid, usr->group_rid,
					      &usr->uni_home_dir,
					      &usr->uni_dir_drive,
					      &usr->uni_logon_script,
					      &usr->uni_acct_desc,
					      usr->acb_info,
					      &usr->uni_profile_path,
					      &usr->uni_workstations,
					      &usr->uni_unknown_str,
					      &usr->uni_munged_dial);
			i++;
			free_samr_userinfo_ctr(&ctr);
		}

		if (status_usr == NT_STATUS_NOPROBLEMO)
		{
			_samr_close(&usr_pol);
		}
	}

	safe_free(sam);
	safe_free(uni_acct_name);

	sam = NULL;
	uni_acct_name = NULL;
	num_sam_users = 0;

	_samr_close(&dom_pol);
	_samr_close(&sam_pol);

	(*num_deltas2) = i;
	(*num_deltas) = i;

	return NT_STATUS_NOPROBLEMO;
}
