/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Andrew Tridgell              1992-1997,2000,
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-1997,2000,
 *  Copyright (C) Paul Ashton                       1997,2000,
 *  Copyright (C) Elrond                                 2000.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* MKPROTO: rpc_client_proto.h		(do not remove this line) */

#include "includes.h"
#include "libsamba.h"
#include "tng_misc.h"
#include "libsmb.h"
#include "rpc_parse.h"
#include "rpc_client.h"
#include "nterr.h"
#include "others.h"

#define DO_HND_PIPE_REQ(hnd, opcode) \
	rpc_hnd_pipe_req((hnd), (opcode), &buf, &rbuf)
#define report_error_code(opcodename, ntstatus) \
	DEBUG(2, (opcodename ": %s\n", \
		  get_nt_error_msg(ntstatus)))
#define clean_and_ret(val) \
	rpccli_close_prs(&buf, &rbuf); \
	\
	return val;



/****************************************************************************
do a LSA Open Policy
****************************************************************************/
BOOL lsa_open_policy(const char *system_name, POLICY_HND *hnd,
		     BOOL sec_qos, uint32 des_access)
{
	prs_struct buf, rbuf;
	LSA_Q_OPEN_POL q_q;
	LSA_R_OPEN_POL r_q;
	LSA_SEC_QOS qos;
	BOOL p;
	BOOL valid_pol = False;
	struct cli_connection *con = NULL;

	if (!cli_connection_init(system_name, PIPE_LSARPC, &con))
	{
		return False;
	}

	if (hnd == NULL)
		return False;

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("LSA Open Policy\n"));

	/* store the parameters */
	if (sec_qos)
	{
		make_lsa_sec_qos(&qos, 2, 1, 0, des_access);
		make_q_open_pol(&q_q, 0x5c, 0, des_access, &qos);
	}
	else
	{
		make_q_open_pol(&q_q, 0x5c, 0, des_access, NULL);
	}

	/* turn parameters into data stream */
	if (lsa_io_q_open_pol("", &q_q, &buf, 0) &&
	    rpc_con_pipe_req(con, LSA_OPENPOLICY, &buf, &rbuf))
	{
		lsa_io_r_open_pol("", &r_q, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_q.status != 0)
		{
			report_error_code("LSA_OPENPOLICY", r_q.status);
			p = False;
		}

		if (p)
		{
			/* ok, at last: we're happy. return the policy handle */
			*hnd = r_q.pol;

			valid_pol =
				register_policy_hnd(get_global_hnd_cache(),
						    cli_con_sec_ctx(con), hnd,
						    des_access)
				&& set_policy_con(get_global_hnd_cache(), hnd,
						  con, cli_connection_unlink);
			if (valid_pol)
			{
				policy_hnd_set_name(get_global_hnd_cache(),
						    hnd, "LSA_OPENPOL");
			}
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a LSA Open Policy2
****************************************************************************/
BOOL lsa_open_policy2(const char *system_name, POLICY_HND *hnd,
		      BOOL sec_qos, uint32 des_access)
{
	prs_struct buf, rbuf;
	LSA_Q_OPEN_POL2 q_q;
	LSA_R_OPEN_POL2 r_q;
	LSA_SEC_QOS qos;
	BOOL p;
	BOOL valid_pol = False;

	struct cli_connection *con = NULL;

	if (!cli_connection_init(system_name, PIPE_LSARPC, &con))
	{
		return False;
	}

	if (hnd == NULL)
		return False;

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("LSA Open Policy2\n"));

	/* store the parameters */
	if (sec_qos)
	{
		make_lsa_sec_qos(&qos, 2, 1, 0, des_access);
		make_q_open_pol2(&q_q, system_name, 0, des_access, &qos);
	}
	else
	{
		make_q_open_pol2(&q_q, system_name, 0, des_access, NULL);
	}

	/* turn parameters into data stream */
	if (lsa_io_q_open_pol2("", &q_q, &buf, 0) &&
	    rpc_con_pipe_req(con, LSA_OPENPOLICY2, &buf, &rbuf))
	{
		p = lsa_io_r_open_pol2("", &r_q, &rbuf, 0)
			&& (rbuf.offset != 0);

		if (p && r_q.status != 0)
		{
			report_error_code("LSA_OPENPOLICY2", r_q.status);
			p = False;
		}

		if (p)
		{
			/* ok, at last: we're happy. return the policy handle */
			*hnd = r_q.pol;

			valid_pol =
				register_policy_hnd(get_global_hnd_cache(),
						    cli_con_sec_ctx(con), hnd,
						    des_access)
				&& set_policy_con(get_global_hnd_cache(), hnd,
						  con, cli_connection_unlink);
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a LSA Create Secret
****************************************************************************/
BOOL lsa_create_secret(const POLICY_HND *hnd,
		       const char *secret_name,
		       uint32 des_access, POLICY_HND *hnd_secret)
{
	prs_struct buf, rbuf;
	LSA_Q_CREATE_SECRET q_q;
	LSA_R_CREATE_SECRET r_q;
	BOOL p;
	BOOL valid_pol = False;

	if (hnd == NULL)
		return False;

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("LSA Create Secret\n"));

	make_q_create_secret(&q_q, hnd, secret_name, des_access);

	/* turn parameters into data stream */
	if (lsa_io_q_create_secret("", &q_q, &buf, 0) &&
	    DO_HND_PIPE_REQ(hnd, LSA_CREATESECRET))
	{
		lsa_io_r_create_secret("", &r_q, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_q.status != 0)
		{
			report_error_code("LSA_CREATESECRET", r_q.status);
			p = False;
		}

		if (p)
		{
			/* ok, at last: we're happy. return the policy handle */
			*hnd_secret = r_q.pol;
			valid_pol = cli_pol_link(hnd_secret, hnd);
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a LSA Open Secret
****************************************************************************/
BOOL lsa_open_secret(const POLICY_HND *hnd,
		     const char *secret_name,
		     uint32 des_access, POLICY_HND *hnd_secret)
{
	prs_struct buf, rbuf;
	LSA_Q_OPEN_SECRET q_q;
	LSA_R_OPEN_SECRET r_q;
	BOOL p;
	BOOL valid_pol = False;

	if (hnd == NULL)
		return False;

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("LSA Open Secret\n"));

	make_q_open_secret(&q_q, hnd, secret_name, des_access);

	/* turn parameters into data stream */
	if (lsa_io_q_open_secret("", &q_q, &buf, 0) &&
	    DO_HND_PIPE_REQ(hnd, LSA_OPENSECRET))
	{
		lsa_io_r_open_secret("", &r_q, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_q.status != 0)
		{
			report_error_code("LSA_OPENSECRET", r_q.status);
			p = False;
		}

		if (p)
		{
			/* ok, at last: we're happy. return the policy handle */
			*hnd_secret = r_q.pol;
			valid_pol = cli_pol_link(hnd_secret, hnd);
			if (valid_pol)
				policy_hnd_set_name(get_global_hnd_cache(),
						    hnd_secret,
						    "LSA_OPENSECRET");
		}
	}

	clean_and_ret(valid_pol);
}

/****************************************************************************
do a LSA Set Secret
****************************************************************************/
uint32 lsa_set_secret(POLICY_HND *hnd, const STRING2 * secret)
{
	prs_struct buf, rbuf;
	LSA_Q_SET_SECRET q_q;
	LSA_R_SET_SECRET r_q;
	BOOL p;

	uchar sess_key[16];
#if 0
	char data[12] =
	{
		0x1a, 0x32, 0xb6, 0x63, 0xd7, 0x08, 0x79, 0x64,
		0x9c, 0x36, 0x94, 0x8b
	};
#endif
	uint32 status = NT_STATUS_NOPROBLEMO;

	if (hnd == NULL)
		return NT_STATUS_INVALID_PARAMETER;

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);
	ZERO_STRUCT(q_q);

	DEBUG(4, ("LSA Set Secret\n"));

	q_q.pol = *hnd;
	q_q.value.ptr_secret = 0x1;
	make_strhdr2(&q_q.value.hdr_secret, secret->str_str_len,
		     secret->str_max_len, 1);

	if (!cli_get_usr_sesskey(hnd, sess_key))
	{
		return NT_STATUS_INVALID_PARAMETER;
	}
	dump_data_pw("sess_key:", sess_key, 16);

#if 0
	memcpy(&q_q.value.enc_secret.buffer, data, sizeof(data));
	q_q.value.enc_secret.str_str_len = sizeof(data);
	q_q.value.enc_secret.str_max_len = sizeof(data);
	if (!nt_decrypt_string2(&q_q.value.enc_secret, secret, sess_key))
	{
		return NT_STATUS_INVALID_PARAMETER;
	}
#endif

	if (!nt_encrypt_string2(&q_q.value.enc_secret, secret, sess_key))
	{
		return NT_STATUS_INVALID_PARAMETER;
	}

	/* turn parameters into data stream */
	if (!lsa_io_q_set_secret("", &q_q, &buf, 0) ||
	    !DO_HND_PIPE_REQ(hnd, LSA_SETSECRET)) {
		status = NT_STATUS_INVALID_PARAMETER;
		goto out;
	}

	lsa_io_r_set_secret("", &r_q, &rbuf, 0);
	p = rbuf.offset != 0;

	if (p && r_q.status != 0)
	{
		report_error_code("LSA_SETSECRET", r_q.status);
		status = NT_STATUS_INVALID_PARAMETER;
	}
	else
	{
		status = r_q.status;
	}

 out:
	clean_and_ret(status);
}

/****************************************************************************
do a LSA Query Secret
****************************************************************************/
BOOL lsa_query_secret(POLICY_HND *hnd, STRING2 * secret, NTTIME * last_update)
{
	prs_struct buf, rbuf;
	LSA_Q_QUERY_SECRET q_q;
	LSA_R_QUERY_SECRET r_q;
	BOOL p;
	BOOL valid_info = False;

	if (hnd == NULL)
		return False;

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("LSA Query Secret\n"));

	make_q_query_secret(&q_q, hnd, secret, last_update);

	/* turn parameters into data stream */
	if (lsa_io_q_query_secret("", &q_q, &buf, 0) &&
	    DO_HND_PIPE_REQ(hnd, LSA_QUERYSECRET))
	{
		lsa_io_r_query_secret("", &r_q, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_q.status != 0)
		{
			report_error_code("LSA_QUERYSECRET", r_q.status);
			p = False;
		}

		if (p && (r_q.sec.curinfo.ptr_value != 0) &&
		    (r_q.sec.curinfo.value.ptr_secret != 0))
		{
			uchar sess_key[16];
			STRING2 enc_secret;
			memcpy(&enc_secret,
			       &(r_q.sec.curinfo.value.enc_secret),
			       sizeof(STRING2));
			if (!cli_get_usr_sesskey(hnd, sess_key))
			{
				return False;
			}
			dump_data_pw("sess key:", sess_key, 16);
			valid_info = nt_decrypt_string2(secret, &enc_secret,
							sess_key);
		}
		if (p && last_update != NULL &&
		    (r_q.sec.curinfo.ptr_value != 0) &&
		    (r_q.sec.curinfo.ptr_update != 0))
		{
			memcpy(last_update, &(r_q.sec.curinfo.last_update),
			       sizeof(NTTIME));
		}
	}

	clean_and_ret(valid_info);
}


/****************************************************************************
do a LSA Lookup Names
****************************************************************************/
BOOL lsa_lookup_names(POLICY_HND *hnd,
		      int num_names,
		      char **names,
		      DOM_SID ** sids, uint32 ** types, uint32 *num_sids)
{
	prs_struct buf, rbuf;
	LSA_Q_LOOKUP_NAMES q_q;
	LSA_R_LOOKUP_NAMES r_q;
	BOOL p;
	BOOL valid_response = False;

	if (hnd == NULL || num_sids == 0 || sids == NULL)
		return False;

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("LSA Lookup NAMEs\n"));

	/* store the parameters */
	make_q_lookup_names(&q_q, hnd, num_names, names);

	/* turn parameters into data stream */
	if (lsa_io_q_lookup_names("", &q_q, &buf, 0) &&
	    DO_HND_PIPE_REQ(hnd, LSA_LOOKUPNAMES))
	{
		DOM_R_REF ref;
		DOM_RID2 t_rids[MAX_LOOKUP_SIDS];

		ZERO_STRUCT(ref);
		ZERO_STRUCT(t_rids);

		r_q.dom_ref = &ref;
		r_q.dom_rid = t_rids;

		lsa_io_r_lookup_names("", &r_q, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_q.status != 0)
		{
			report_error_code("LSA_LOOKUP_NAMES", r_q.status);
			p = False;
		}

		if (p)
		{
			if (r_q.ptr_dom_ref != 0 && r_q.ptr_entries != 0)
			{
				valid_response = True;
			}
		}

		if (num_sids != NULL && valid_response)
		{
			(*num_sids) = r_q.num_entries;
		}
		if (valid_response)
		{
			uint32 i;
			for (i = 0; i < r_q.num_entries; i++)
			{
				if (t_rids[i].rid_idx >= ref.num_ref_doms_1 &&
				    t_rids[i].rid_idx != 0xffffffff)
				{
					DEBUG(0,
					      ("LSA_LOOKUP_NAMES: domain index %d out of bounds\n",
					       t_rids[i].rid_idx));
					valid_response = False;
					break;
				}
			}
		}

		if (types != NULL && valid_response && r_q.num_entries != 0)
		{
			(*types) = g_new(uint32, *num_sids);
		}

		if (sids != NULL && valid_response && r_q.num_entries != 0)
		{
			(*sids) = g_new(DOM_SID, *num_sids);
		}

		if (sids != NULL && (*sids) != NULL)
		{
			uint32 i;
			/* take each name, construct a SID */
			for (i = 0; i < (*num_sids); i++)
			{
				uint32 dom_idx = t_rids[i].rid_idx;
				uint32 dom_rid = t_rids[i].rid;
				DOM_SID *sid = &(*sids)[i];
				if (dom_idx != 0xffffffff)
				{
					sid_copy(sid,
						 &ref.ref_dom[dom_idx].
						 ref_dom.sid);
					if (dom_rid != 0xffffffff)
					{
						sid_append_rid(sid, dom_rid);
					}
					if (types != NULL && (*types) != NULL)
					{
						(*types)[i] = t_rids[i].type;
					}
				}
				else
				{
					ZERO_STRUCTP(sid);
					if (types != NULL && (*types) != NULL)
					{
						(*types)[i] = SID_NAME_UNKNOWN;
					}
				}
			}
		}
	}

	clean_and_ret(valid_response);
}

/****************************************************************************
do a LSA Lookup SIDs
****************************************************************************/
BOOL lsa_lookup_sids(const POLICY_HND *hnd,
		     int num_sids, DOM_SID **sids,
		     char ***names, uint32 ** types, int *num_names)
{
	prs_struct buf, rbuf;
	LSA_Q_LOOKUP_SIDS q_q;
	LSA_R_LOOKUP_SIDS r_q;
	BOOL p;
	BOOL valid_response = False;

	ZERO_STRUCT(q_q);

	if (hnd == NULL || num_sids == 0 || sids == NULL)
		return False;

	if (num_names != NULL)
	{
		*num_names = 0;
	}
	if (types != NULL)
	{
		*types = NULL;
	}
	if (names != NULL)
	{
		*names = NULL;
	}

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("LSA Lookup SIDs\n"));

	/* store the parameters */
	make_q_lookup_sids(&q_q, hnd, num_sids, sids, 1);

	/* turn parameters into data stream */
	if (lsa_io_q_lookup_sids("", &q_q, &buf, 0) &&
	    DO_HND_PIPE_REQ(hnd, LSA_LOOKUPSIDS))
	{
		DOM_R_REF ref;
		LSA_TRANS_NAME_ENUM t_names;

		r_q.dom_ref = &ref;
		r_q.names = &t_names;

		lsa_io_r_lookup_sids("", &r_q, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_q.status != 0 &&
		    r_q.status != 0x107 &&
		    r_q.status != (0xC0000000 | NT_STATUS_NONE_MAPPED))
		{
			report_error_code("LSA_LOOKUP_SIDS", r_q.status);
			p = False;
		}

		if (p)
		{
			if (t_names.ptr_trans_names != 0
			    && r_q.ptr_dom_ref != 0)
			{
				valid_response = True;
			}
		}

		if (num_names != NULL && valid_response)
		{
			(*num_names) = t_names.num_entries;
		}
		if (valid_response)
		{
			uint32 i;
			for (i = 0; i < t_names.num_entries; i++)
			{
				if ((t_names.name[i].domain_idx >=
				     ref.num_ref_doms_1)
				    && (t_names.name[i].domain_idx !=
					0xffffffff))
				{
					DEBUG(0,
					      ("LSA_LOOKUP_SIDS: domain index out of bounds\n"));
					valid_response = False;
					break;
				}
			}
		}

		if (types != NULL && valid_response && (*num_names) != 0)
		{
			(*types) = g_new(uint32, *num_names);
		}

		if (names != NULL && valid_response && (*num_names) != 0)
		{
			(*names) = g_new(char *, *num_names);
		}

		if (names != NULL && (*names) != NULL)
		{
			int i;
			/* take each name, construct a \DOMAIN\name string */
			for (i = 0; i < (*num_names); i++)
			{
				fstring name;
				fstring dom_name;
				fstring full_name;
				uint32 dom_idx = t_names.name[i].domain_idx;

				if (dom_idx != 0xffffffff)
				{
					unistr2_to_ascii(dom_name,
							 &ref.
							 ref_dom[dom_idx].
							 uni_dom_name,
							 sizeof(dom_name) -
							 1);
					unistr2_to_ascii(name,
							 &t_names.uni_name[i],
							 sizeof(name) - 1);

					ZERO_OBJ(full_name);

					slprintf(full_name,
						 sizeof(full_name) - 1,
						 "%s\\%s", dom_name, name);

					(*names)[i] = strdup(full_name);
					if (types != NULL && (*types) != NULL)
					{
						(*types)[i] = t_names.name[i].sid_name_use;
					}
				}
				else
				{
					(*names)[i] = NULL;
					if (types != NULL && (*types) != NULL)
					{
						(*types)[i] = SID_NAME_UNKNOWN;
					}
				}
			}
		}
	}

	clean_and_ret(valid_response);
}

/****************************************************************************
do a LSA Query Info Policy
****************************************************************************/
BOOL lsa_query_sec_obj(const POLICY_HND *hnd, uint32 sec_info,
		       SEC_DESC_BUF *sec_buf)
{
	prs_struct buf, rbuf;
	LSA_Q_QUERY_SEC_OBJ q_q;
	LSA_R_QUERY_SEC_OBJ r_q;
	BOOL p;
	BOOL valid_response = False;

	if (hnd == NULL || sec_buf == NULL)
		return False;

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("LSA Query Info Policy\n"));

	/* store the parameters */
	make_q_query_sec_obj(&q_q, hnd, sec_info);

	/* turn parameters into data stream */
	if (lsa_io_q_query_sec_obj("", &q_q, &buf, 0) &&
	    DO_HND_PIPE_REQ(hnd, LSA_QUERYSECOBJECT))
	{
		lsa_io_r_query_sec_obj("", &r_q, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_q.status != 0)
		{
			report_error_code("LSA_QUERY_SECOBJ", r_q.status);
			p = False;
		}

		if (p)
		{
			valid_response = True;
			sec_buf->sec = r_q.buf.sec;
		}
	}

	clean_and_ret(valid_response);
}

/****************************************************************************
do a LSA Query Info Policy
****************************************************************************/
uint32 lsa_query_info(const POLICY_HND *hnd, uint16 info_class,
		      LSA_INFO_UNION **info)
{
	prs_struct buf, rbuf;
	LSA_Q_QUERY_INFO q_q;
	LSA_R_QUERY_INFO r_q;
	BOOL p = True;
	uint32 status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL || info == NULL)
		return NT_STATUS_INVALID_PARAMETER;

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("LSA Query Info Policy\n"));

	/* store the parameters */
	make_q_query(&q_q, hnd, info_class);

	/* turn parameters into data stream */
	if (lsa_io_q_query("", &q_q, &buf, 0) &&
	    DO_HND_PIPE_REQ(hnd, LSA_QUERYINFOPOLICY)
	    && lsa_io_r_query("", &r_q, &rbuf, 0))
	{
		if (p)
			status = r_q.status;

		if (p && status != 0)
		{
			report_error_code("LSA_QUERYINFOPOLICY", status);
		}

		if (p && r_q.undoc_buffer && r_q.info_class != q_q.info_class)
		{
			/* report different info classes */
			DEBUG(0,
			      ("LSA_QUERYINFOPOLICY: error info_class (q,r) differ - (%x,%x)\n",
			       q_q.info_class, r_q.info_class));
			p = False;
		}

		if (p)
		{
			if (r_q.undoc_buffer)
				*info = r_q.dom;
			else
				*info = NULL;
		}
	}

	clean_and_ret(status);
}


/****************************************************************************
do a LSA Enumerate Trusted Domain 
****************************************************************************/
BOOL lsa_enum_trust_dom(POLICY_HND *hnd, uint32 * enum_ctx,
			uint32 * num_doms, char ***names, DOM_SID *** sids)
{
	prs_struct buf, rbuf;
	LSA_Q_ENUM_TRUST_DOM q_q;
	LSA_R_ENUM_TRUST_DOM r_q;
	BOOL p;
	BOOL valid_response = False;

	if (hnd == NULL || num_doms == NULL || names == NULL)
		return False;

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("LSA Enum Trusted Domains\n"));

	/* store the parameters */
	make_q_enum_trust_dom(&q_q, hnd, *enum_ctx, 0xffffffff);

	/* turn parameters into data stream */
	if (lsa_io_q_enum_trust_dom("", &q_q, &buf, 0) &&
	    DO_HND_PIPE_REQ(hnd, LSA_ENUMTRUSTDOM))
	{
		ZERO_STRUCT(r_q);

		lsa_io_r_enum_trust_dom("", &r_q, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_q.status != 0)
		{
			p = r_q.status == STATUS_NO_MORE_ENTRIES;
			if (!p) {
				report_error_code("LSA_ENUMTRUSTDOM", r_q.status);
			}
		}

		if (p)
		{
			uint32 i;
			uint32 num_sids = 0;
			valid_response = True;

			for (i = 0; i < r_q.num_domains; i++)
			{
				fstring tmp;
				unistr2_to_ascii(tmp, &r_q.uni_domain_name[i],
						 sizeof(tmp) - 1);
				add_chars_to_array(num_doms, names, tmp);
				add_sid_to_array(&num_sids, sids,
						 &r_q.domain_sid[i].sid);
			}

			if (r_q.status == NT_STATUS_NOPROBLEMO)
			{
				*enum_ctx = r_q.enum_context;
			}
			else
			{
				*enum_ctx = 0;
			}
		}

		lsa_free_r_enum_trust_dom(&r_q);
	}

	clean_and_ret(valid_response);
}

/****************************************************************************
do a LSA Close
****************************************************************************/
BOOL lsa_close(POLICY_HND *hnd)
{
	prs_struct buf, rbuf;
	LSA_Q_CLOSE q_q;
	LSA_R_CLOSE r_q;
	BOOL p;
	BOOL valid_close = False;

	if (hnd == NULL)
		return False;

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("LSA Close\n"));

	/* store the parameters */
	make_lsa_q_close(&q_q, hnd);

	/* turn parameters into data stream */
	if (lsa_io_q_close("", &q_q, &buf, 0) &&
	    DO_HND_PIPE_REQ(hnd, LSA_CLOSE))
	{
		lsa_io_r_close("", &r_q, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_q.status != 0)
		{
			report_error_code("LSA_CLOSE", r_q.status);
			p = False;
		}

		if (p)
		{
			valid_close = True;
		}
	}

	close_policy_hnd(get_global_hnd_cache(), hnd);

	clean_and_ret(valid_close);
}

/****************************************************************************
do a LSA Enum Privileges
****************************************************************************/
BOOL lsa_enum_privs(POLICY_HND *hnd, uint32 unk0, uint32 unk1,
		    uint32 *count, LSA_PRIV_ENTRY **entries)
{
	prs_struct buf, rbuf;
	LSA_Q_ENUM_PRIVS q_q;
	LSA_R_ENUM_PRIVS r_q;
	BOOL valid_resp = False;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return False;

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);
	ZERO_STRUCT(r_q);

	DEBUG(4, ("LSA Enum Privileges\n"));

	/* store the parameters */
	q_q.pol = *hnd;
	q_q.unk0 = unk0;
	q_q.unk1 = unk1;

	/* turn parameters into data stream */
	if (lsa_io_q_enum_privs("", &q_q, &buf, 0)
	    && DO_HND_PIPE_REQ(hnd, LSA_ENUM_PRIVS)
	    && lsa_io_r_enum_privs("", &r_q, &rbuf, 0))
	{
		status = r_q.status;
		if (! NT_STATUS_IS_OK(status))
			report_error_code("LSA_ENUM_PRIVS", status);
		if (r_q.ptr == 0)
			r_q.count = 0;
		if (count)
			*count = r_q.count;

		if (entries)
			*entries = r_q.privs;
		else
			safe_free(r_q.privs);

		valid_resp = True;
	}

	clean_and_ret(valid_resp);
}

/****************************************************************************
do a LSA Enum SIDs
****************************************************************************/
uint32 lsa_enum_sids(POLICY_HND *hnd, uint32 *start_idx, uint32 unk1,
		     uint32 *count, DOM_SID ***sids)
{
	prs_struct buf, rbuf;
	LSA_Q_ENUM_PRIVS q_q;
	LSA_R_ENUM_SIDS r_q;
	BOOL p;
	uint32 status = NT_STATUS_INVALID_PARAMETER;

	if (hnd == NULL || start_idx == NULL)
		return NT_STATUS_INVALID_PARAMETER;

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("LSA Enum SIDs\n"));

	/* store the parameters */
	q_q.pol = *hnd;
	q_q.unk0 = *start_idx;
	q_q.unk1 = unk1;

	/* turn parameters into data stream */
	if (lsa_io_q_enum_privs("", &q_q, &buf, 0) &&
	    DO_HND_PIPE_REQ(hnd, LSA_ENUM_SIDS))
	{
		ZERO_STRUCT(r_q);

		p = lsa_io_r_enum_sids("", &r_q, &rbuf, 0)
			&& rbuf.offset != 0;

		if (p)
		{
			(*start_idx) = r_q.start_idx;
			status = r_q.status;
			if (!make_sids_from_sid_enum(&r_q.sids, count, sids))
				status = NT_STATUS_INVALID_PARAMETER;
		}
	}

	clean_and_ret(status);
}

/****************************************************************************
do a LSA Privileges Info
****************************************************************************/
uint32 lsa_priv_get_dispname(const POLICY_HND *hnd, const UNISTR2 *name,
			     uint16 req_lang_id,
			     UNISTR2 **desc, uint16 *got_lang_id)
{
	prs_struct buf, rbuf;
	LSA_Q_PRIV_GET_DISPNAME q_q;
	LSA_R_PRIV_GET_DISPNAME r_q;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return NT_STATUS_INVALID_PARAMETER;

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);
	ZERO_STRUCT(r_q);

	DEBUG(4, ("LSA Privileges Info\n"));

	/* store the parameters */
	q_q.pol = *hnd;
	copy_unistr2(&q_q.name, name);
	q_q.lang_id = req_lang_id;
	q_q.lang_id_sys = req_lang_id;

	/* turn parameters into data stream */
	if (lsa_io_q_priv_get_dispname("", &q_q, &buf, 0)
	    && DO_HND_PIPE_REQ(hnd, LSA_PRIV_GET_DISPNAME)
	    && lsa_io_r_priv_get_dispname("", &r_q, &rbuf, 0))
	{
		status = r_q.status;
		if (! NT_STATUS_IS_OK(status))
			report_error_code("LSA_PRIV_GET_DISPNAME", status);
		if (r_q.ptr_info)
		{
			if (desc)
				(*desc) = unistr2_dup(&r_q.desc);
			if (got_lang_id)
				(*got_lang_id) = r_q.lang_id;
		}
	}

	clean_and_ret(status);
}

/****************************************************************************
do a LSA Get Privileges of a SID
****************************************************************************/
uint32 lsa_sid_get_privs(const POLICY_HND *hnd, const DOM_SID *sid,
			 /* [out] */ uint32 *num_privs,
			 /* [out] */ UNISTR2 ***privs)
{
	prs_struct buf, rbuf;
	LSA_Q_SID_GET_PRIVS q_q;
	LSA_R_SID_GET_PRIVS r_q;
	BOOL p;
	uint32 status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL || num_privs==NULL || privs==NULL)
		return NT_STATUS_INVALID_PARAMETER;

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("LSA SID get privs\n"));

	/* store the parameters */
	q_q.hnd = *hnd;
	make_dom_sid2(&q_q.sid, sid);

	(*num_privs) = 0;
	(*privs) = NULL;

	/* turn parameters into data stream */
	if (lsa_io_q_sid_get_privs("", &q_q, &buf, 0) &&
	    DO_HND_PIPE_REQ(hnd, LSA_SID_GET_PRIVS))
	{
		ZERO_STRUCT(r_q);

		p = lsa_io_r_sid_get_privs("", &r_q, &rbuf, 0)
			&& rbuf.offset != 0;

		if (p)
		{
			(*num_privs) = r_q.count;
			(*privs) = r_q.priv_names;
			status = r_q.status;
		}
	}

	if (status)
	{
		report_error_code("LSA_SID_GET_PRIVS", status);
	}

	clean_and_ret(status);
}

/****************************************************************************
****************************************************************************/
uint32 lsa_get_trust_info2(const POLICY_HND *hnd,
			   const DOM_SID *sid, uint16 level,
			   /* [out] */ void *ret_info)
{
	prs_struct buf, rbuf;
	LSA_Q_GET_TRUST_INFO2 q_q;
	uint32 status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL || sid == NULL)
		return NT_STATUS_INVALID_PARAMETER;

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("lsa_get_trust_info2\n"));

	/* store the parameters */
	q_q.hnd = *hnd;
	make_dom_sid2(&q_q.sid, sid);
	q_q.level = level;

	/* turn parameters into data stream */
	if (lsa_io_q_get_trust_info2("", &q_q, &buf, 0) &&
	    DO_HND_PIPE_REQ(hnd, LSA_GET_TRUST_INFO2))
	{
	}

	if (status)
	{
		report_error_code("LSA_GET_TRUST_INFO2", status);
	}

	clean_and_ret(status);
}


NTSTATUS lsa_create_trust_dom(const POLICY_HND *hnd,
			      const LSA_TRUST_INFO_SIMPLE *trust_info,
			      uint32 des_access,
			      POLICY_HND *hnd_domain)
{
	prs_struct buf, rbuf;
	LSA_Q_CREATE_TRUST_DOM q_q;
	MISC_R_HANDLE_STATUS r_q;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL || trust_info == NULL || hnd_domain == NULL)
		return NT_STATUS_INVALID_PARAMETER;

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("lsa_create_trust_dom\n"));

	q_q.hnd = *hnd;
	q_q.trust_info = *trust_info;
	q_q.des_access = des_access;

	/* turn parameters into data stream */
	if (lsa_io_q_create_trust_dom("", &q_q, &buf, 0)
	    && DO_HND_PIPE_REQ(hnd, LSA_CREATE_TRUST_DOM)
	    && misc_io_r_handle_status("", &r_q, &rbuf, 0))
	{
		status = r_q.status;
		if (status != 0)
			report_error_code("LSA_CREATE_TRUST_DOM", status);
	}

	if (status == 0x0)
	{
		/* ok, at last: we're happy. return the policy handle */
		*hnd_domain = r_q.hnd;
		if(cli_pol_link(hnd_domain, hnd))
			policy_hnd_set_name(get_global_hnd_cache(),
					    hnd_domain,
					    "LSA_CREATE_TRUST_DOM");
		else
			status = NT_STATUS_UNSUCCESSFUL;
	}

	clean_and_ret(status);
}


uint32 lsa_open_trust_dom(const POLICY_HND *hnd,
			  const DOM_SID *sid,
			  uint32 des_access, POLICY_HND *hnd_domain)
{
	prs_struct buf, rbuf;
	LSA_Q_OPEN_TRUSTED_DOM q_q;
	LSA_R_OPEN_TRUSTED_DOM r_q;
	uint32 status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return NT_STATUS_INVALID_PARAMETER;

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4, ("lsa_open_trusted_dom\n"));

	q_q.hnd = *hnd;
	make_dom_sid2(&q_q.sid, sid);
	q_q.des_access = des_access;

	/* turn parameters into data stream */
	if (lsa_io_q_open_trusted_dom("", &q_q, &buf, 0)
	    && DO_HND_PIPE_REQ(hnd, LSA_OPEN_TRUSTED_DOM)
	    && lsa_io_r_open_trusted_dom("", &r_q, &rbuf, 0))
	{
		status = r_q.status;
		if (status != 0)
			report_error_code("LSA_OPEN_TRUSTED_DOM", status);
	}

	if (status == 0x0)
	{
		/* ok, at last: we're happy. return the policy handle */
		*hnd_domain = r_q.hnd;
		if(cli_pol_link(hnd_domain, hnd))
			policy_hnd_set_name(get_global_hnd_cache(),
					    hnd_domain,
					    "LSA_OPEN_TRUSTED_DOM");
		else
			status = NT_STATUS_UNSUCCESSFUL;
	}

	clean_and_ret(status);
}
