/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Andrew Tridgell              1992-2000,
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-2000,
 *  Copyright (C) Paul Ashton                  1997-2000.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* MKPROTO: rpc_client_proto.h		(do not remove this line) */

#include "includes.h"
#include "libsamba.h"
#include "rpc_parse.h"
#include "rpc_client.h"
#include "nterr.h"
#include "winerror.h"
#include "others.h"


#define DECLARE_BASIC_VARS(MARSHALL_ST, UNMARSHALL_ST) \
	prs_struct buf, rbuf; \
	MARSHALL_ST q_o; \
	UNMARSHALL_ST r_o
#define INIT_BASIC_VARS() \
	rpccli_prs_open(&buf, &rbuf); \
	ZERO_STRUCT(r_o)
#define DO_MARSHALL(fn) \
	(fn("", &q_o, &buf, 0))
#define DO_CON_PIPE_REQ(opcode) \
	rpc_con_pipe_req(con, (opcode), &buf, &rbuf)
#define DO_HND_PIPE_REQ(hnd, opcode) \
	rpc_hnd_pipe_req((hnd), (opcode), &buf, &rbuf)
#define DO_UNMARSHALL(fn) \
	(fn("", &r_o, &rbuf, 0))
#define clean_and_ret(val) \
	rpccli_close_prs(&buf, &rbuf); \
	\
	return val;



/****************************************************************************
do a REG Open Policy
****************************************************************************/
static uint32 reg_open_hkXX_int(struct cli_connection *con,
				uint16 op_code, const char *op_name,
				uint16 unknown_0, uint32 access_mask,
				POLICY_HND *hnd)
{
	DECLARE_BASIC_VARS(REG_Q_OPEN_HKXX, REG_R_OPEN_HKXX);
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return status;

	INIT_BASIC_VARS();

	DEBUG(4,("REG Open (%s)\n", op_name));

	make_reg_q_open_hkXX(&q_o, unknown_0, access_mask);

	/* turn parameters into data stream */
	if (DO_MARSHALL(reg_io_q_open_hkXX)
	    && DO_CON_PIPE_REQ(op_code)
	    && DO_UNMARSHALL(reg_io_r_open_hkXX))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (!NT_STATUS_IS_OK(status))
		{
			DEBUG(0, ("%s: %s\n", op_name,
				  get_nt_error_msg(status)));
		}

		*hnd = r_o.pol;
	}

	clean_and_ret(status);
}

/****************************************************************************
do a REG Open Policy
****************************************************************************/
BOOL reg_connect(const char *srv_name,
				const char *full_keyname,
				char *key_name,
				uint32 access_mask,
				POLICY_HND *reg_hnd)
{
	NTSTATUS ret = NT_STATUS_UNSUCCESSFUL;
	uint32 reg_type = 0;

	struct cli_connection *con = NULL;

	if (full_keyname == NULL)
	{
		return False;
	}
	if (!reg_split_key(full_keyname, &reg_type, key_name))
	{
		DEBUG(0,("reg_connect: unrecognised key name %s\n",
		          full_keyname));	
		return False;
	}
	if (!cli_connection_init(srv_name, PIPE_WINREG, &con))
	{
		return False;
	}

	ZERO_STRUCTP(reg_hnd);

	if (reg_type == HKEY_CLASSES_ROOT)
	{
		ret = reg_open_hkXX_int(con,
					REG_OPEN_HKCR, "REG_OPEN_HKCR",
					0x5428, access_mask, reg_hnd);
	}
	else if (reg_type == HKEY_LOCAL_MACHINE)
	{
		ret = reg_open_hkXX_int(con,
					REG_OPEN_HKLM, "REG_OPEN_HKLM",
					0x84e0, access_mask, reg_hnd);
	}
	else if (reg_type == HKEY_CURRENT_USER)
	{
		ret = reg_open_hkXX_int(con,
					REG_OPEN_HKCU, "REG_OPEN_HKCU",
					0x84e0, access_mask, reg_hnd);
	}
	else if (reg_type == HKEY_USERS)
	{
		ret = reg_open_hkXX_int(con,
					REG_OPEN_HKU, "REG_OPEN_HKU",
					0x84e0, access_mask, reg_hnd);
	}
	else if (reg_type == HKEY_PERFORMANCE_DATA)
	{
		ret = reg_open_hkXX_int(con,
					REG_OPEN_HKPD, "REG_OPEN_HKPD",
					0x86e8, access_mask, reg_hnd);
	}
	else
	{
		DEBUG(0,("reg_connect: unrecognised hive key\n"));	
		return False;
	}

	if (NT_STATUS_IS_OK(ret))
	{
		struct policy_cache *cache = get_global_hnd_cache();
		if (!register_policy_hnd(cache, cli_con_sec_ctx(con),
		                         reg_hnd, access_mask) ||
		    !set_policy_con(cache, reg_hnd, con, 
				    cli_connection_unlink))
		{
			cli_connection_unlink(con);
			ret = NT_STATUS_UNSUCCESSFUL;
		}
	}

	return NT_STATUS_IS_OK(ret);
}

/****************************************************************************
do a REG Unknown 0xB command.  sent after a create key or create value.
this might be some sort of "sync" or "refresh" command, sent after
modification of the registry...
****************************************************************************/
BOOL reg_flush_key(POLICY_HND *hnd)
{
	DECLARE_BASIC_VARS(REG_Q_FLUSH_KEY, REG_R_FLUSH_KEY);
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS();

	DEBUG(4,("REG Unknown 0xB\n"));

	make_reg_q_flush_key(&q_o, hnd);

	/* turn parameters into data stream */
	if (DO_MARSHALL(reg_io_q_flush_key)
	    && DO_HND_PIPE_REQ(hnd, REG_FLUSH_KEY)
	    && DO_UNMARSHALL(reg_io_r_flush_key))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (!NT_STATUS_IS_OK(status))
		{
			DEBUG(0, ("REG_FLUSH_KEY: %s\n",
				  get_nt_error_msg(status)));
		}
	}

	clean_and_ret(NT_STATUS_IS_OK(status));
}

/****************************************************************************
do a REG Query Key
****************************************************************************/
BOOL reg_query_key(POLICY_HND *hnd,
				char *key_class, uint32 *class_len,
				uint32 *num_subkeys, uint32 *max_subkeylen,
				uint32 *max_subkeysize, uint32 *num_values,
				uint32 *max_valnamelen, uint32 *max_valbufsize,
				uint32 *sec_desc, NTTIME *mod_time)
{
	DECLARE_BASIC_VARS(REG_Q_QUERY_KEY, REG_R_QUERY_KEY);
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS();

	DEBUG(4,("REG Query Key\n"));

	make_reg_q_query_key(&q_o, hnd, *class_len);

	/* turn parameters into data stream */
	if (DO_MARSHALL(reg_io_q_query_key)
	    && DO_HND_PIPE_REQ(hnd, REG_QUERY_KEY)
	    && DO_UNMARSHALL(reg_io_r_query_key))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (!NT_STATUS_IS_OK(status))
		{
			DEBUG(0, ("REG_QUERY_KEY: %s\n",
				  get_nt_error_msg(status)));
		}
		else
		{
			*class_len      = r_o.hdr_class.uni_max_len;
			unistr2_to_ascii(key_class, &r_o.uni_class, sizeof(fstring)-1);
			*num_subkeys    = r_o.num_subkeys   ;
			*max_subkeylen  = r_o.max_subkeylen ;
			*max_subkeysize = r_o.max_subkeysize;
			*num_values     = r_o.num_values    ;
			*max_valnamelen = r_o.max_valnamelen;
			*max_valbufsize = r_o.max_valbufsize;
			*sec_desc       = r_o.sec_desc      ;
			*mod_time       = r_o.mod_time      ;
		}
	}

	clean_and_ret(NT_STATUS_IS_OK(status));
}

/****************************************************************************
do a REG Unknown 1A
****************************************************************************/
BOOL reg_unknown_1a(POLICY_HND *hnd, uint32 *unk)
{
	DECLARE_BASIC_VARS(REG_Q_UNK_1A, REG_R_UNK_1A);
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS();

	DEBUG(4,("REG Unknown 1a\n"));

	make_reg_q_unk_1a(&q_o, hnd);

	/* turn parameters into data stream */
	if (DO_MARSHALL(reg_io_q_unk_1a)
	    && DO_HND_PIPE_REQ(hnd, REG_UNK_1A)
	    && DO_UNMARSHALL(reg_io_r_unk_1a))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (!NT_STATUS_IS_OK(status))
		{
			DEBUG(0, ("REG_UNK_1A: %s\n",
				  get_nt_error_msg(status)));
		}
		else
		{
			if (unk)
			{
				*unk = r_o.unknown;
			}
			if (r_o.unknown != 5)
			{
				DEBUG(4,("REG Unknown 1a response: %x\n", r_o.unknown));
			}
		}
	}

	clean_and_ret(NT_STATUS_IS_OK(status));
}

/****************************************************************************
do a REG Query Info
****************************************************************************/
BOOL reg_query_info(POLICY_HND *hnd, const char *val_name,
		    uint32 *type, BUFFER2 *buffer)
{
	DECLARE_BASIC_VARS(REG_Q_INFO, REG_R_INFO);
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS();

	DEBUG(4,("REG Query Info\n"));

	make_reg_q_info(&q_o, hnd, val_name, 4, 0, buffer->buf_max_len);
	r_o.uni_val = buffer;

	/* turn parameters into data stream */
	if (DO_MARSHALL(reg_io_q_info)
	    && DO_HND_PIPE_REQ(hnd, REG_INFO)
	    && DO_UNMARSHALL(reg_io_r_info))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (!NT_STATUS_IS_OK(status))
		{
			DEBUG(0, ("REG_INFO: %s\n",
				  get_nt_error_msg(status)));
		}
		*type = r_o.type;
	}

	clean_and_ret(NT_STATUS_IS_OK(status));
}

/****************************************************************************
do a REG Set Key Security 
****************************************************************************/
BOOL reg_set_key_sec(POLICY_HND *hnd,
				uint32 sec_info,
				uint32 sec_buf_size, SEC_DESC *sec_buf)
{
	DECLARE_BASIC_VARS(REG_Q_SET_KEY_SEC, REG_R_SET_KEY_SEC);
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS();

	DEBUG(4,("REG Set Key security.\n"));

	make_reg_q_set_key_sec(&q_o, hnd, sec_info, sec_buf_size, sec_buf);

	/* turn parameters into data stream */
	if (DO_MARSHALL(reg_io_q_set_key_sec)
	    && DO_HND_PIPE_REQ(hnd, REG_SET_KEY_SEC)
	    && DO_UNMARSHALL(reg_io_r_set_key_sec))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (!NT_STATUS_IS_OK(status))
		{
			DEBUG(0, ("REG_SET_KEY_SEC: %s\n",
				  get_nt_error_msg(status)));
		}
	}

	clean_and_ret(NT_STATUS_IS_OK(status));
}


/****************************************************************************
do a REG Query Key Security 
****************************************************************************/
BOOL reg_get_key_sec(POLICY_HND *hnd,
				uint32 sec_info,
				uint32 *sec_buf_size, SEC_DESC_BUF *sec_buf)
{
	DECLARE_BASIC_VARS(REG_Q_GET_KEY_SEC, REG_R_GET_KEY_SEC);
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS();

	DEBUG(4,("REG query key security.  buf_size: %d\n", *sec_buf_size));

	make_reg_q_get_key_sec(&q_o, hnd, sec_info, *sec_buf_size, sec_buf);
	r_o.data = sec_buf;

	/* turn parameters into data stream */
	if (DO_MARSHALL(reg_io_q_get_key_sec)
	    && DO_HND_PIPE_REQ(hnd, REG_GET_KEY_SEC)
	    && DO_UNMARSHALL(reg_io_r_get_key_sec))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (r_o.status == ERROR_INSUFFICIENT_BUFFER)
		{
			/*
			 * get the maximum buffer size: it was too small
			 */
			(*sec_buf_size) = r_o.hdr_sec.buf_max_len;
			DEBUG(5, ("sec_buf_size too small.  use %d\n",
				  *sec_buf_size));
			status = NT_STATUS_OK;
		}
		else if (!NT_STATUS_IS_OK(status))
		{
			DEBUG(0, ("REG_GET_KEY_SEC: %s\n",
				  get_nt_error_msg(status)));
		}
		else
		{
			(*sec_buf_size) = r_o.data->len;
		}
	}

	clean_and_ret(NT_STATUS_IS_OK(status));
}

/****************************************************************************
do a REG Delete Value
****************************************************************************/
BOOL reg_delete_val(POLICY_HND *hnd, char *val_name)
{
	DECLARE_BASIC_VARS(REG_Q_DELETE_VALUE, REG_R_DELETE_VALUE);
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS();

	DEBUG(4,("REG Delete Value: %s\n", val_name));

	make_reg_q_delete_val(&q_o, hnd, val_name);

	/* turn parameters into data stream */
	if (DO_MARSHALL(reg_io_q_delete_val)
	    && DO_HND_PIPE_REQ(hnd, REG_DELETE_VALUE)
	    && DO_UNMARSHALL(reg_io_r_delete_val))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (!NT_STATUS_IS_OK(status))
		{
			DEBUG(0, ("REG_DELETE_VALUE: %s\n",
				  get_nt_error_msg(status)));
		}
	}

	clean_and_ret(NT_STATUS_IS_OK(status));
}

/****************************************************************************
do a REG Delete Key
****************************************************************************/
BOOL reg_delete_key(POLICY_HND *hnd, char *key_name)
{
	DECLARE_BASIC_VARS(REG_Q_DELETE_KEY, REG_R_DELETE_KEY);
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS();

	DEBUG(4,("REG Delete Key: %s\n", key_name));

	make_reg_q_delete_key(&q_o, hnd, key_name);

	/* turn parameters into data stream */
	if (DO_MARSHALL(reg_io_q_delete_key)
	    && DO_HND_PIPE_REQ(hnd, REG_DELETE_KEY)
	    && DO_UNMARSHALL(reg_io_r_delete_key))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (!NT_STATUS_IS_OK(status))
		{
			DEBUG(0, ("REG_DELETE_KEY: %s\n",
				  get_nt_error_msg(status)));
		}
	}

	clean_and_ret(NT_STATUS_IS_OK(status));
}

/****************************************************************************
do a REG Create Key
****************************************************************************/
BOOL reg_create_key(POLICY_HND *hnd,
				char *key_name, char *key_class,
				SEC_ACCESS *sam_access,
				POLICY_HND *key)
{
	DECLARE_BASIC_VARS(REG_Q_CREATE_KEY, REG_R_CREATE_KEY);
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;
	SEC_DESC sec;
	SEC_DESC_BUF sec_buf;
	int sec_len;

	ZERO_STRUCT(sec);	/* XXX - this is make_*()'s responsibility */
	ZERO_STRUCT(sec_buf);
	ZERO_STRUCT(q_o);

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS();

	DEBUG(4,("REG Create Key: %s %s 0x%08x\n", key_name, key_class,
		sam_access != NULL ? sam_access->mask : 0));

	sec_len = make_sec_desc(&sec, 1, SEC_DESC_SELF_RELATIVE,
	                        NULL, NULL, NULL, NULL);

	DEBUG(10,("make_sec_desc: len = %d\n", sec_len));

	make_reg_q_create_key(&q_o, hnd, key_name, key_class, sam_access,
	                      &sec_buf, sec_len, &sec);

	/* turn parameters into data stream */
	if (DO_MARSHALL(reg_io_q_create_key)
	    && DO_HND_PIPE_REQ(hnd, REG_CREATE_KEY)
	    && DO_UNMARSHALL(reg_io_r_create_key))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (!NT_STATUS_IS_OK(status))
		{
			DEBUG(0, ("REG_CREATE_KEY: %s\n",
				  get_nt_error_msg(status)));
		}
		else
		{
			status = NT_STATUS_UNSUCCESSFUL;
			*key = r_o.key_pol;
			if (cli_pol_link(key, hnd))
				status = NT_STATUS_OK;
		}
	}

	free_sec_desc(&sec);

	clean_and_ret(NT_STATUS_IS_OK(status));
}

/****************************************************************************
do a REG Enum Key
****************************************************************************/
BOOL reg_enum_key(POLICY_HND *hnd,
				int key_index, char *key_name,
				uint32 *unk_1, uint32 *unk_2,
				time_t *mod_time)
{
	DECLARE_BASIC_VARS(REG_Q_ENUM_KEY, REG_R_ENUM_KEY);
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS();

	DEBUG(4,("REG Enum Key\n"));

	make_reg_q_enum_key(&q_o, hnd, key_index);

	/* turn parameters into data stream */
	if (DO_MARSHALL(reg_io_q_enum_key)
	    && DO_HND_PIPE_REQ(hnd, REG_ENUM_KEY)
	    && DO_UNMARSHALL(reg_io_r_enum_key))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (!NT_STATUS_IS_OK(status))
		{
			DEBUG(0, ("REG_ENUM_KEY: %s\n",
				  get_nt_error_msg(status)));
		}
		else
		{
			if (unk_1)
			{
				(*unk_1) = r_o.unknown_1;
				(*unk_2) = r_o.unknown_2;
			}
			unistr_to_ascii(key_name, r_o.key_name.str.buffer,
					sizeof(fstring)-1);
			(*mod_time) = nt_time_to_unix(&r_o.time);
		}
	}

	clean_and_ret(NT_STATUS_IS_OK(status));
}

/****************************************************************************
do a REG Create Value
****************************************************************************/
BOOL reg_create_val(POLICY_HND *hnd,
				char *val_name, uint32 type, BUFFER3 *data)
{
	DECLARE_BASIC_VARS(REG_Q_CREATE_VALUE, REG_R_CREATE_VALUE);
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS();

	DEBUG(4,("REG Create Value: %s\n", val_name));

	make_reg_q_create_val(&q_o, hnd, val_name, type, data);

	/* turn parameters into data stream */
	if (DO_MARSHALL(reg_io_q_create_val)
	    && DO_HND_PIPE_REQ(hnd, REG_CREATE_VALUE)
	    && DO_UNMARSHALL(reg_io_r_create_val))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (!NT_STATUS_IS_OK(status))
		{
			DEBUG(0, ("REG_CREATE_VALUE: %s\n",
				  get_nt_error_msg(status)));
		}
	}

	clean_and_ret(NT_STATUS_IS_OK(status));
}

/****************************************************************************
do a REG Enum Value
****************************************************************************/
BOOL reg_enum_val(POLICY_HND *hnd,
				int val_index, int max_valnamelen, int max_valbufsize,
				fstring val_name,
				uint32 *val_type, BUFFER2 *value)
{
	DECLARE_BASIC_VARS(REG_Q_ENUM_VALUE, REG_R_ENUM_VALUE);
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS();

	DEBUG(4,("REG Enum Value\n"));

	make_reg_q_enum_val(&q_o, hnd, val_index, max_valnamelen, max_valbufsize);
	r_o.buf_value = value;

	/* turn parameters into data stream */
	if (DO_MARSHALL(reg_io_q_enum_val)
	    && DO_HND_PIPE_REQ(hnd, REG_ENUM_VALUE)
	    && DO_UNMARSHALL(reg_io_r_enum_val))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (r_o.status == ERROR_MORE_DATA)
		{
			status = NT_STATUS_OK;
		}
		if (!NT_STATUS_IS_OK(status))
		{
			DEBUG(0, ("REG_ENUM_VALUE: %s\n",
				  get_nt_error_msg(status)));
		}
		else
		{
			(*val_type) = r_o.type;
			unistr2_to_ascii(val_name, &r_o.uni_name, sizeof(fstring)-1);
		}
	}

	clean_and_ret(NT_STATUS_IS_OK(status));
}

/****************************************************************************
do a REG Open Key
****************************************************************************/
BOOL reg_open_entry(POLICY_HND *hnd,
		    const char *key_name, uint32 unk_0,
		    POLICY_HND *key_hnd)
{
	DECLARE_BASIC_VARS(REG_Q_OPEN_ENTRY, REG_R_OPEN_ENTRY);
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS();

	DEBUG(4,("REG Open Entry\n"));

	make_reg_q_open_entry(&q_o, hnd, key_name, unk_0);

	/* turn parameters into data stream */
	if (DO_MARSHALL(reg_io_q_open_entry)
	    && DO_HND_PIPE_REQ(hnd, REG_OPEN_ENTRY)
	    && DO_UNMARSHALL(reg_io_r_open_entry))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (!NT_STATUS_IS_OK(status))
		{
			DEBUG(0, ("REG_OPEN_ENTRY: %s\n",
				  get_nt_error_msg(status)));
		}
		else
		{
			status = NT_STATUS_UNSUCCESSFUL;
			*key_hnd = r_o.pol;
			if (cli_pol_link(key_hnd, hnd))
				status = NT_STATUS_OK;
		}
	}

	clean_and_ret(NT_STATUS_IS_OK(status));
}

/****************************************************************************
do a REG Close
****************************************************************************/
BOOL reg_close(POLICY_HND *hnd)
{
	DECLARE_BASIC_VARS(REG_Q_CLOSE, REG_R_CLOSE);
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS();

	DEBUG(4,("REG Close\n"));

	/* store the parameters */
	make_reg_q_close(&q_o, hnd);

	/* turn parameters into data stream */
	if (DO_MARSHALL(reg_io_q_close)
	    && DO_HND_PIPE_REQ(hnd, REG_CLOSE)
	    && DO_UNMARSHALL(reg_io_r_close))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (!NT_STATUS_IS_OK(status))
		{
			DEBUG(0, ("REG_CLOSE: %s\n",
				  get_nt_error_msg(status)));
		}
	}

	close_policy_hnd(get_global_hnd_cache(), hnd);

	clean_and_ret(NT_STATUS_IS_OK(status));
}

/****************************************************************************
do a REG Shutdown Server
****************************************************************************/
BOOL reg_shutdown(const char *srv_name,
			const char *msg, uint32 timeout, uint16 flags)
{
	DECLARE_BASIC_VARS(REG_Q_SHUTDOWN, REG_R_SHUTDOWN);
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	struct cli_connection *con = NULL;

	if (!cli_connection_init(srv_name, PIPE_WINREG, &con))
	{
		return False;
	}

	if (msg == NULL)
		return False;

	INIT_BASIC_VARS();

	DEBUG(4,("REG Shutdown: (timeout: %d secs) %s\n", timeout, msg));

	make_reg_q_shutdown(&q_o, msg, timeout, flags);

	/* turn parameters into data stream */
	if (DO_MARSHALL(reg_io_q_shutdown)
	    && DO_CON_PIPE_REQ(REG_SHUTDOWN)
	    && DO_UNMARSHALL(reg_io_r_shutdown))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (!NT_STATUS_IS_OK(status))
		{
			DEBUG(0, ("REG_SHUTDOWN: %s\n",
				  get_nt_error_msg(status)));
		}
	}

	cli_connection_unlink(con);

	clean_and_ret(NT_STATUS_IS_OK(status));
}


/****************************************************************************
do a REG Abort Shutdown
****************************************************************************/
BOOL reg_abort_shutdown(const char *srv_name)
{
	DECLARE_BASIC_VARS(REG_Q_ABORT_SHUTDOWN, REG_R_ABORT_SHUTDOWN);
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	struct cli_connection *con = NULL;

	if (!cli_connection_init(srv_name, PIPE_WINREG, &con))
	{
		return False;
	}

	INIT_BASIC_VARS();

	DEBUG(4,("REG Abort Shutdown\n"));

	make_reg_q_abort_shutdown(&q_o);

	/* turn parameters into data stream */
	if (DO_MARSHALL(reg_io_q_abort_shutdown)
	    && DO_CON_PIPE_REQ(REG_ABORT_SHUTDOWN)
	    && DO_UNMARSHALL(reg_io_r_abort_shutdown))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (!NT_STATUS_IS_OK(status))
		{
			DEBUG(0, ("REG_ABORT_SHUTDOWN: %s\n",
				  get_nt_error_msg(status)));
		}
	}

	cli_connection_unlink(con);

	clean_and_ret(NT_STATUS_IS_OK(status));
}
