/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   NT Domain Authentication SMB / MSRPC client
   Copyright (C) Andrew Tridgell 1994-1999
   Copyright (C) Luke Kenneth Casson Leighton 1996-1999
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: rpc_client_proto.h		(do not remove this line) */

/**
 * \file
 * \brief LSARPC helper functions on top of the client stubs
 */

#include "includes.h"
#include "libsamba.h"
#include "byteorder.h"
#include "libsmb.h"
#include "rpc_client.h"
#include "rpc_parse.h"
#include "nterr.h"


uint32 lookup_lsa_names(const char *srv_name,
			uint32 num_names, char **names,
			uint32 * num_sids, DOM_SID ** sids, uint32 ** types)
{
	BOOL res1 = True;
	BOOL res2 = True;
	POLICY_HND lsa_pol;

	if (srv_name == NULL)
	{
		srv_name = "\\\\.";
	}

	if (num_sids)
	{
		*num_sids = 0;
	}
	if (sids)
	{
		*sids = NULL;
	}
	if (types)
	{
		*types = NULL;
	}

	if (!num_sids || (!types && !sids))
	{
		/* Not sure, wether that's a good error-code */
		return NT_STATUS_NONE_MAPPED | 0xC0000000;
	}

	res1 = res1 ? lsa_open_policy(srv_name, &lsa_pol, True,
				      SEC_RIGHTS_MAXIMUM_ALLOWED) : False;

	res2 = res1 ? lsa_lookup_names(&lsa_pol,
				       num_names, names,
				       sids, types, num_sids) : False;

	res1 = res1 ? lsa_close(&lsa_pol) : False;

	if (!res2)
	{
		return NT_STATUS_NONE_MAPPED | 0xC0000000;
	}

	return 0x0;
}


uint32 lookup_lsa_name(const char *domain,
		       char *name, DOM_SID * sid, uint32 * type)
{
	fstring srv_name;
	BOOL res3 = True;
	BOOL res4 = True;
	char **names = NULL;
	uint32 *types = NULL;
	int num_names = 0;
	DOM_SID *sids = NULL;
	uint32 num_sids = 0;
	POLICY_HND lsa_pol;

	if (!get_any_dc_name(domain, srv_name))
	{
		return NT_STATUS_NONE_MAPPED | 0xC0000000;
	}

	num_names = 1;
	names = &name;

	/* lookup domain controller; receive a policy handle */
	res3 = res3 ? lsa_open_policy(srv_name, &lsa_pol, True,
				      SEC_RIGHTS_MAXIMUM_ALLOWED) : False;

	/* send lsa lookup sids call */
	res4 = res3 ? lsa_lookup_names(&lsa_pol,
				       num_names, names,
				       &sids, &types, &num_sids) : False;

	res3 = res3 ? lsa_close(&lsa_pol) : False;

	res4 = num_sids != 1 ? False : res4;

	if (!res4)
	{
		return NT_STATUS_NONE_MAPPED | 0xC0000000;
	}
	sid_copy(sid, &sids[0]);
	*type = types[0];

	if (types != NULL)
	{
		free(types);
	}

	if (sids != NULL)
	{
		free(sids);
	}

	return 0x0;
}

/****************************************************************************
lookup sids
****************************************************************************/
uint32 lookup_lsa_sid(const char *domain,
		      DOM_SID * sid, char *name, uint32 * type)
{
	POLICY_HND lsa_pol;
	fstring srv_name;
	DOM_SID **sids = NULL;
	uint32 num_sids = 0;
	char **names = NULL;
	int num_names = 0;
	uint32 *types = NULL;

	BOOL res = True;
	BOOL res1 = True;

	if (!get_any_dc_name(domain, srv_name))
	{
		return NT_STATUS_NONE_MAPPED;
	}

	add_sid_to_array(&num_sids, &sids, sid);

	/* lookup domain controller; receive a policy handle */
	res = res ? lsa_open_policy(srv_name, &lsa_pol, True,
				    SEC_RIGHTS_MAXIMUM_ALLOWED) : False;

	/* send lsa lookup sids call */
	res1 = res ? lsa_lookup_sids(&lsa_pol,
				     num_sids, sids,
				     &names, &types, &num_names) : False;

	res = res ? lsa_close(&lsa_pol) : False;

	if (!res1 || names == NULL || types == NULL)
	{
		return NT_STATUS_NONE_MAPPED | 0xC0000000;
	}

	fstrcpy(name, names[0]);
	*type = types[0];

	free_sid_array(num_sids, sids);
	free_char_array(num_names, names);

	if (types != NULL)
	{
		free(types);
	}

	return 0x0;
}

/****************************************************************************
nt lsa create secret
****************************************************************************/
BOOL msrpc_lsa_create_secret(const char *srv_name, const char *secret_name,
			     uint32 access_rights)
{
	BOOL res = True;
	BOOL res1;

	POLICY_HND pol_sec;
	POLICY_HND lsa_pol;

	/* lookup domain controller; receive a policy handle */
	res = res ? lsa_open_policy(srv_name,
				    &lsa_pol, True, 
                                    SEC_RIGHTS_MAXIMUM_ALLOWED) : False;

	/* lookup domain controller; receive a policy handle */
	res1 = res ? lsa_create_secret(&lsa_pol,
				       secret_name, access_rights,
				       &pol_sec) : False;

	res1 = res1 ? lsa_close(&pol_sec) : False;

	res = res ? lsa_close(&lsa_pol) : False;

	return res1;
}

/****************************************************************************
put data into secret buffer.
****************************************************************************/
void secret_store_data(STRING2 * secret, const char *data, int len)
{
	ZERO_STRUCTP(secret);

	secret->str_max_len = len + 8;
	secret->undoc = 0;
	secret->str_str_len = len + 8;

	SIVAL(secret->buffer, 0, len);
	SIVAL(secret->buffer, 4, 0x01);
	memcpy(secret->buffer + 8, data, len);
}


BOOL msrpc_lsa_set_secret(const char *srv_name,
			  const char *secret_name, const char *data, int len)
{
	BOOL res = True;
	BOOL res1;
	BOOL res2;

	POLICY_HND pol_sec;
	POLICY_HND lsa_pol;
	STRING2 secret;

	secret_store_data(&secret, data, len);

	/* lookup domain controller; receive a policy handle */
	res = res ? lsa_open_policy(srv_name,
				    &lsa_pol, True, 
                                    SEC_RIGHTS_MAXIMUM_ALLOWED) : False;

	/* lookup domain controller; receive a policy handle */
	res1 = res ? lsa_open_secret(&lsa_pol,
				     secret_name, SEC_RIGHTS_MAXIMUM_ALLOWED,
				     &pol_sec) : False;

	res2 = res1 ? (lsa_set_secret(&pol_sec, &secret) ==
		       NT_STATUS_NOPROBLEMO) : False;

	res1 = res1 ? lsa_close(&pol_sec) : False;

	res = res ? lsa_close(&lsa_pol) : False;

	return res2;
}

/****************************************************************************
nt lsa query secret
****************************************************************************/
BOOL msrpc_lsa_query_secret(const char *srv_name,
			    const char *secret_name,
			    STRING2 * secret, NTTIME * last_update)
{
	BOOL res = True;
	BOOL res1;
	BOOL res2;

	POLICY_HND pol_sec;
	POLICY_HND lsa_pol;

	/* lookup domain controller; receive a policy handle */
	res = res ? lsa_open_policy2(srv_name,
				     &lsa_pol, False, 
                                     SEC_RIGHTS_MAXIMUM_ALLOWED) : False;

	/* lookup domain controller; receive a policy handle */
	res1 = res ? lsa_open_secret(&lsa_pol,
				     secret_name, SEC_RIGHTS_MAXIMUM_ALLOWED,
				     &pol_sec) : False;

	res2 = res1 ? lsa_query_secret(&pol_sec, secret, last_update) : False;

	res1 = res1 ? lsa_close(&pol_sec) : False;

	res = res ? lsa_close(&lsa_pol) : False;

	return res2;
}

/****************************************************************************
obtains a trust account password
****************************************************************************/
BOOL secret_get_data(const STRING2 * secret, uchar * data, uint32 * len)
{
	(*len) = IVAL(secret->buffer, 0);
	if (secret->str_str_len != (*len) + 8)
	{
		return False;
	}
	if (IVAL(secret->buffer, 4) != 0x1)
	{
		return False;
	}
	if ((*len) > 1024)
	{
		DEBUG(0,("secret_get_data: length too long!\n"));
	}
	memcpy(data, secret->buffer + 8, *len);
	return True;
}

/****************************************************************************
obtains a trust account password
****************************************************************************/
BOOL secret_to_nt_owf(uchar trust_passwd[16], const STRING2 * secret)
{
#if 0
	UNISTR2 uni_pwd;
	uint32 i;
#endif
	uint32 len = 0;
	pstring data;

	if (!secret_get_data(secret, data, &len))
	{
		return False;
	}
#if 0
	for (i = 0; i < len; i++)
	{
		uni_pwd.buffer[i] = SVAL(data, i * 2);
	}
	uni_pwd.uni_str_len = len / 2;
	uni_pwd.uni_max_len = len / 2;
	nt_owf_genW(&uni_pwd, trust_passwd);
#else
	mdfour(trust_passwd, data, len);
#endif

	return True;
}

/****************************************************************************
obtains a trust account password
****************************************************************************/
BOOL msrpc_lsa_query_trust_passwd(const char *srv_name,
				  const char *secret_name,
				  uchar trust_passwd[16],
				  NTTIME * last_update)
{
	STRING2 secret;

	if (!msrpc_lsa_query_secret(srv_name, secret_name, &secret,
				    last_update))
	{
		return False;
	}
	return secret_to_nt_owf(trust_passwd, &secret);
}

/****************************************************************************
 obtain the sid from the PDC.  do some verification along the way...
****************************************************************************/
BOOL get_domain_sids(const char *domain, DOM_SID * sid3, DOM_SID * sid5)
{
	POLICY_HND pol;
	fstring srv_name;
	BOOL res = True;
	BOOL res1 = True;
	fstring dom3;
	fstring dom5;

	extern struct ntuser_creds *usr_creds;
	struct ntuser_creds null_usr;

	copy_nt_creds(&null_usr, NULL);
	usr_creds = &null_usr;

	if (sid3 == NULL && sid5 == NULL)
	{
		/* don't waste my time... */
		return False;
	}

	if (!get_any_dc_name(domain, srv_name))
	{
		return False;
	}

	/*
	 * Ok - we have an anonymous connection to the IPC$ share.
	 * Now start the NT Domain stuff :-).
	 */

	fstrcpy(dom3, "");
	fstrcpy(dom5, "");
	if (sid3 != NULL)
	{
		ZERO_STRUCTP(sid3);
	}
	if (sid5 != NULL)
	{
		ZERO_STRUCTP(sid5);
	}

	/* lookup domain controller; receive a policy handle */
	res = res ? lsa_open_policy(srv_name, &pol, False,
				    SEC_RIGHTS_MAXIMUM_ALLOWED) : False;

	if (sid3 != NULL)
	{
		/* send client info query, level 3.  receive domain name and sid */
		res1 = res ? lsa_query_info_pol(&pol, 3, dom3, sid3) : False;
	}

	if (sid5 != NULL)
	{
		/* send client info query, level 5.  receive domain name and sid */
		res1 = res1 ? lsa_query_info_pol(&pol, 5, dom5, sid5) : False;
	}

	/* close policy handle */
	res = res ? lsa_close(&pol) : False;

	if (res1)
	{
		pstring sid;
		DEBUG(2, ("LSA Query Info Policy\n"));
		if (sid3 != NULL)
		{
			sid_to_string(sid, sid3);
			DEBUG(2,
			      ("Domain Member     - Domain: %s SID: %s\n",
			       dom3, sid));
		}
		if (sid5 != NULL)
		{
			sid_to_string(sid, sid5);
			DEBUG(2,
			      ("Domain Controller - Domain: %s SID: %s\n",
			       dom5, sid));
		}
	}
	else
	{
		DEBUG(1, ("lsa query info failed\n"));
	}

	return res;
}

#if 0
/****************************************************************************
 obtain a sid and domain name from a Domain Controller.  
****************************************************************************/
BOOL get_trust_sid_and_domain(const char *myname, char *server,
			      DOM_SID * sid, char *domain, size_t len)
{
	POLICY_HND pol;
	fstring srv_name;
	struct cli_connection *con = NULL;
	BOOL res = True;
	BOOL res1 = True;
	DOM_SID sid3;
	DOM_SID sid5;
	fstring dom3;
	fstring dom5;

	if (!cli_connection_init_list(server, PIPE_LSARPC, &con))
	{
		DEBUG(0,
		      ("get_trust_sid: unable to initialise client connection.\n"));
		return False;
	}

	fstrcpy(dom3, "");
	fstrcpy(dom5, "");
	ZERO_STRUCT(sid3);
	ZERO_STRUCT(sid5);

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, myname);
	strupper(srv_name);

	/* lookup domain controller; receive a policy handle */
	res = res ? lsa_open_policy(srv_name, &pol, False,
				    SEC_RIGHTS_MAXIMUM_ALLOWED) : False;

	/* send client info query, level 3.  receive domain name and sid */
	res1 = res ? lsa_query_info_pol(&pol, 3, dom3, &sid3) : False;

	/* send client info query, level 5.  receive domain name and sid */
	res1 = res1 ? lsa_query_info_pol(&pol, 5, dom5, &sid5) : False;

	/* close policy handle */
	res = res ? lsa_close(&pol) : False;

	if (res1)
	{
		pstring sid_str;
		DEBUG(2, ("LSA Query Info Policy\n"));
		sid_to_string(sid_str, &sid3);
		DEBUG(2, ("Domain Member     - Domain: %s SID: %s\n",
			  dom3, sid_str));
		sid_to_string(sid_str, &sid5);
		DEBUG(2, ("Domain Controller - Domain: %s SID: %s\n",
			  dom5, sid_str));

		if (dom5[0] != 0 && sid_equal(&sid3, &sid5))
		{
			safe_strcpy(domain, dom5, len);
			sid_copy(sid, &sid5);
		}
		else
		{
			DEBUG(2, ("Server %s is not a PDC\n", server));
			return False;
		}

	}
	else
	{
		DEBUG(1, ("lsa query info failed\n"));
	}

	return res1;
}
#endif

/****************************************************************************
do a LSA Query Info Policy
****************************************************************************/
BOOL lsa_query_info_pol(const POLICY_HND *hnd, uint16 info_class,
			fstring domain_name, DOM_SID * domain_sid)
{
	BOOL valid_response = False;
	LSA_INFO_UNION *info;

	if (hnd == NULL || domain_name == NULL || domain_sid == NULL)
		return False;

	if ((info_class != 3) && (info_class != 5))
	{
		DEBUG(0, ("lsa_query_info_pol: Please use lsa_query_info "
			  "for level %d! Only 3 and 5 are valid.\n",
			  info_class));
		return False;
	}

	ZERO_STRUCTP(domain_sid);
	domain_name[0] = 0;

	if (!NT_STATUS_IS_OK(lsa_query_info(hnd, info_class, &info)))
		return False;

	{
		fstring sid_str;
		/* ok, at last: we're happy. */
		switch (info_class)
		{
			case 3:
			{
				if (info->id3.buffer_dom_name != 0)
				{
					unistr2_to_ascii(domain_name,
							 &info->id3.
							 uni_domain_name,
							 sizeof
							 (fstring) -
							 1);
				}
				if (info->id3.buffer_dom_sid != 0)
				{
					*domain_sid =
						info->id3.dom_sid.sid;
				}

				valid_response = True;
				break;
			}
			case 5:
			{
				if (info->id5.buffer_dom_name != 0)
				{
					unistr2_to_ascii(domain_name,
							 &info->id5.
							 uni_domain_name,
							 sizeof
							 (fstring) -
							 1);
				}
				if (info->id5.buffer_dom_sid != 0)
				{
					*domain_sid =
						info->id5.
						dom_sid.sid;
				}

				valid_response = True;
				break;
			}
			default:
			{
				DEBUG(3,
				      ("LSA_QUERYINFOPOLICY: unknown info class\n"));
				domain_name[0] = 0;

				break;
			}
		}

		sid_to_string(sid_str, domain_sid);
		DEBUG(3,
		      ("LSA_QUERYINFOPOLICY (level %x): domain:%s  domain sid:%s\n",
		       info_class, domain_name, sid_str));
	}

	lsa_free_info_union(info);

	return valid_response;
}
