/* 
 *  Samba TNG
 *  DCE/RPC Handle (un)marshalling
 *  Copyright (C) Elrond <elrond@samba-tng.org>     2002-2003.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* MKPROTO: rpc_parse_proto.h		(do not remove this line) */


#include "includes.h"
#include "libsamba.h"
#include "tng_misc.h"
#include "byteorder.h"
#include "dlinklist.h"
#include "rpc_parse.h"
#include "handle.h"


/*******************************************************************
 Reads or writes an POLICY_HND structure.
********************************************************************/
BOOL smb_io_pol_hnd(char *desc,  POLICY_HND *pol, prs_struct *ps, int depth)
{
	if (pol == NULL)
		return False;

	prs_debug(ps, depth, desc, "smb_io_pol_hnd");
	depth++;

	prs_align(ps);
	
	prs_uint32("ptr", ps, depth, &pol->ptr); 
	return smb_io_rpc_uuid("uuid", &pol->uuid, ps, depth);
}

/****************************************************************************
  create a unique policy handle
****************************************************************************/
void create_pol_hnd(POLICY_HND *hnd)
{
	static uint32 pol_hnd_low = 0;
	NTTIME ntt;

	if (hnd == NULL)
		return;

	ZERO_STRUCTP(hnd);

	pol_hnd_low++;

	unix_to_nt_time(&ntt, time(NULL));

	hnd->ptr = 0;
	hnd->uuid.time_low = ntt.low;
	hnd->uuid.time_mid = (ntt.high & 0xffff);
	hnd->uuid.time_hi_and_version = ((ntt.high >> 16) & 0xffff);
	SIVAL(hnd->uuid.remaining, 0, sys_getuniqid());
	SIVAL(hnd->uuid.remaining, 4, pol_hnd_low);
}

#ifndef NOONS_SMALL_BUILD
static BOOL wire_handle_compare(const POLICY_HND *a, const POLICY_HND *b)
{
	return (memcmp(a, b, sizeof(POLICY_HND))==0);
}


/*******************************************************************/

static void rpc_handle_unregister(void *hnd);


static void *rpc_handle_allocate_default(void)
{
	Handle *hnd = handle_new(NULL);

	handle_set_data(hnd, "__rpc__/self(hnd)", hnd, rpc_handle_unregister);

	return hnd;
}


static void rpc_handle_close_default(void *hnd)
{
	handle_close(hnd);
}
#endif

/*******************************************************************/

typedef struct _HandleMapping
{
	struct _HandleMapping *next, *prev;
	POLICY_HND wire_handle;
	void *voidp;
	void *context;
} HandleMapping;

#ifndef NOONS_SMALL_BUILD
static HandleMapping *map = NULL;
static void * (* rpc_handle_allocate_fn) (void) = rpc_handle_allocate_default;
static void (* rpc_handle_close_fn) (void *) = rpc_handle_close_default;


static void rpc_handle_unregister(void *hnd)
{
	HandleMapping *m;

	for (m = map; m; m = m->next)
	{
		if (m->voidp != hnd)
			continue;
		DLIST_REMOVE(map, m);
		free(m);
		return;
	}
}
#endif
#ifndef NOONS_SMALL_BUILD
static void rpc_handle_register(const POLICY_HND *wire_handle, void *voidp,
				void *context)
{
	HandleMapping *m;

	m = g_new(HandleMapping, 1);
	m->wire_handle = *wire_handle;
	m->voidp = voidp;
	m->context = context;
	DLIST_ADD(map, m);
}
#endif

#ifndef NOONS_SMALL_BUILD
static
void rpc_handle_close_context(void *context)
{
	HandleMapping *m, *next;

	for (m = map; m; m = next)
	{
		next = m->next;
		if (m->context != context)
			continue;
		rpc_handle_unregister(m->voidp);
		rpc_handle_close_fn(m->voidp);
	}
}
#endif
#ifndef NOONS_SMALL_BUILD
static
BOOL rpc_io_handle(const char *desc, void **voidpp, prs_struct *ps, int depth)
{
	POLICY_HND wire_handle;
	HandleMapping *m;

	if (! voidpp)
		return False;

	prs_debug(ps, depth, desc, "rpc_io_handle");
	depth++;

	if (UNMARSHALLING(ps))
	{
		if (! smb_io_pol_hnd("wire_handle", &wire_handle, ps, depth))
			return False;
		for (m = map; m; m = m->next)
		{
			if (wire_handle_compare(&m->wire_handle, &wire_handle))
				break;
		}
		if (m)
		{
			(* voidpp) = m->voidp;
			return True;
		}
		/* new handle on the wire */
		if (PRS_IS_SERVER(ps))
		{
			DEBUG(1, ("WARNING: rpc_io_handle(%s: %s): "
				  "Server side received unknown handle\n",
				  prs_last_level0(ps), desc));
			(* voidpp) = NULL;
			/* XXX - should we fail? Elrond */
			return False;
		}
		/* Client */
		(* voidpp) = rpc_handle_allocate_fn();
		if (! (* voidpp))
		{
			DEBUG(1, ("WARNING: rpc_io_handle(%s: %s): "
				  "rpc_handle_allocate_fn returns NULL\n",
				  prs_last_level0(ps), desc));
			return False;
		}
		rpc_handle_register(&wire_handle, (* voidpp),
				    NULL);
	}
	else
	{
		for (m = map; m; m = m->next)
		{
			if (m->voidp == (* voidpp))
				break;
		}
		if (m)
		{
			wire_handle = m->wire_handle;
		}
		else if (PRS_IS_CLIENT(ps))
		{
			DEBUG(1, ("WARNING: rpc_io_handle(%s: %s): "
				  "Client side wants to marshall unknown handle\n",
				  prs_last_level0(ps), desc));
			return False;
		}
		else
		{
			create_pol_hnd(&wire_handle);
			rpc_handle_register(&wire_handle, (* voidpp),
					    NULL);
		}

		if (! smb_io_pol_hnd("wire_handle", &wire_handle, ps, depth))
			return False;
	}

	return True;
}
#endif

/*******************************************************************/

BOOL misc_io_r_handle_status(const char *desc, MISC_R_HANDLE_STATUS *lds,
			     prs_struct *ps, int depth)
{
	if (lds == NULL)
		return False;

	prs_debug(ps, depth, desc, "lsa_io_r_open_trusted_dom");
	depth++;

	if (!smb_io_pol_hnd("", &lds->hnd, ps, depth))
		return False;

	prs_uint32("status", ps, depth, &lds->status);

	return True;
}
