/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-2001
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* MKPROTO: rpc_parse_proto.h		(do not remove this line) */

#include "includes.h"
#include "libsamba.h"
#include "tng_misc.h"
#include "rpc_parse.h"


/*******************************************************************
makes a user struct.
********************************************************************/
BOOL make_vuid_user_struct(user_struct * r_u,
			   uid_t uid, gid_t gid,
			   const char *name,
			   const char *requested_name,
			   const char *real_name,
			   BOOL guest,
			   uint32 n_groups, const gid_t * groups)
{
	int i;

	if (r_u == NULL)
		return False;

	DEBUG(5, ("make_user_struct\n"));

	r_u->uid = uid;
	r_u->gid = gid;

	fstrcpy(r_u->unix_name, name);
	fstrcpy(r_u->requested_name, requested_name);
	fstrcpy(r_u->real_name, real_name);
	r_u->guest = guest;

	r_u->n_groups = n_groups;
	r_u->groups = g_new(gid_t, r_u->n_groups);
	if (r_u->groups == NULL && n_groups != 0)
	{
		return False;
	}
	for (i = 0; i < n_groups; i++)
	{
		r_u->groups[i] = (gid_t) groups[i];
	}

	return True;
}

static RPC_IO_DECLARE(vuid_io_user_struct, user_struct, r_u)
{
	uint32 tmp_uint32;
	int i;

	uint32 uid = (uint32)r_u->uid;
	uint32 gid = (uint32)r_u->gid;
	uint32 *groups = NULL;

	RPC_MARSHALLER_INTRO(r_u);

	prs_uint32("uid", ps, depth, &uid);
	prs_uint32("gid", ps, depth, &gid);

	r_u->uid = (uid_t)uid;
	r_u->gid = (uid_t)gid;

	prs_align(ps);
	prs_string("name", ps, depth, r_u->unix_name, strlen(r_u->unix_name),
		   sizeof(r_u->unix_name));
	prs_align(ps);
	prs_string("requested_name", ps, depth, r_u->requested_name,
		   strlen(r_u->requested_name), sizeof(r_u->requested_name));
	prs_align(ps);
	prs_string("real_name", ps, depth, r_u->real_name,
		   strlen(r_u->real_name), sizeof(r_u->real_name));
	prs_align(ps);
	tmp_uint32 = r_u->guest;
	prs_uint32("guest", ps, depth, &(tmp_uint32));
	r_u->guest = tmp_uint32;

	prs_uint32("n_groups", ps, depth, &(r_u->n_groups));

	if (r_u->n_groups != 0)
	{
		groups = g_new(uint32, r_u->n_groups);
		if (groups == NULL)
		{
			vuid_free_user_struct(r_u);
			return False;
		}
		if (UNMARSHALLING(ps))
		{
			/* reading */
			r_u->groups = g_new(gid_t, r_u->n_groups);
		}
		if (r_u->groups == NULL)
		{
			vuid_free_user_struct(r_u);
			safe_free(groups);
			return False;
		}
	}
	else if (UNMARSHALLING(ps))
	{
		r_u->groups = NULL;
	}

	if (MARSHALLING(ps))
	{
		for (i = 0; i < r_u->n_groups; i++)
		{
			groups[i] = (uint32)r_u->groups[i];
		}
	}
	for (i = 0; i < r_u->n_groups; i++)
	{
		prs_uint32("", ps, depth, &groups[i]);
	}
	if (UNMARSHALLING(ps))
	{
		for (i = 0; i < r_u->n_groups; i++)
		{
			r_u->groups[i] = (gid_t)groups[i];
		}
	}

	safe_free(groups);

	return True;
}


/*******************************************************************
frees a structure.
********************************************************************/
void vuid_free_user_struct(user_struct * r_u)
{
	if (r_u != NULL)
	{
		safe_free(r_u->groups);
		r_u->groups = NULL;
	}
	safe_free(r_u);
}

/*******************************************************************
makes a vuser struct.
********************************************************************/
BOOL make_vuid_vuser_struct(vuser_struct * r_u,
			   const user_struct * usr,
			   const NET_USER_INFO_3 * ntusr)
{
	if (r_u == NULL)
		return False;

	DEBUG(5, ("make_vuser_struct\n"));

	r_u->usr = NULL;
	r_u->ntusr = NULL;

	if (usr != NULL)
	{
		r_u->usr = g_new0(user_struct, 1);
		memcpy(&r_u->usr, usr, sizeof(r_u->usr));
	}
	if (ntusr != NULL)
	{
		r_u->ntusr = g_new0(NET_USER_INFO_3,1);
		memcpy(&r_u->ntusr, ntusr, sizeof(r_u->ntusr));
	}

	return True;
}

/*******************************************************************
reads or writes a structure.
********************************************************************/
RPC_IO_DECLARE(vuid_io_vuser_struct, vuser_struct, r_u)
{
	uint32 ptr_usr;
	uint32 ptr_ntusr;

	RPC_MARSHALLER_INTRO(r_u);

	ptr_usr = (r_u->usr != NULL ? 1 : 0);
	prs_uint32("ptr_usr", ps, depth, &ptr_usr);
	ptr_ntusr = (r_u->ntusr != NULL ? 1 : 0);
	prs_uint32("ptr_ntusr", ps, depth, &ptr_ntusr);
	prs_uint32("flags", ps, depth, &r_u->flags);

	if (ptr_usr)
	{
		if (UNMARSHALLING(ps))
		{
			r_u->usr = g_new0(user_struct, 1);
			if(r_u->usr == NULL)
				return False;
		}

		vuid_io_user_struct("usr", r_u->usr, ps, depth);
	}
	if (ptr_ntusr)
	{
		if (UNMARSHALLING(ps))
		{
			r_u->ntusr = g_new0(NET_USER_INFO_3,1);
			if(r_u->ntusr == NULL)
				return False;
		}

		net_io_user_info3("usr", r_u->ntusr, ps, depth);
	}

	return True;
}


/*******************************************************************
frees a structure including its components.
********************************************************************/
void vuid_free_vuser_struct(vuser_struct * r_u)
{
	if (r_u == NULL)
		return;

	if (r_u->usr != NULL)
	{
		vuid_free_user_struct(r_u->usr);
	}
	if (r_u->ntusr != NULL)
	{
		free(r_u->ntusr);
	}
	safe_free(r_u);
}

