
/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Andrew Tridgell              1992-1998,2000,
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-1998,2000,
 *  Copyright (C) Paul Ashton                  1997-1998,2000,
 *  Copyright (C) Elrond <elrond@samba-tng.org>     2000-2001.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 *
 *

 this module provides nt user / nt rid lookup functions.
 users, local groups, domain groups.

 no unix / samba functions should be called in this module:
 it should purely provide a gateway to the password database API,
 the local group database API or the domain group database API,
 but first checking built-in rids.

 did i say rids?  oops, what about "S-1-1" the "Everyone" group
 and other such well-known sids...

 speed is not of the essence: no particular optimisation is in place.

 *
 *
 */

#include "includes.h"
#include "libsamba.h"
#include "tng_misc.h"
#include "rpc_client.h"
#include "nterr.h"
#include "sids.h"
#include "samdb.h"
#include "others.h"


/*******************************************************************
 lookup_wk_group_sid
 ********************************************************************/
static uint32 lookup_wk_group_sid(const DOM_SID *sid, char *group_name,
				  uint32 *type)
{
	uint32 rid;
	DOM_SID tmp;
	char *mapped;

	sid_copy(&tmp, sid);
	sid_split_rid(&tmp, &rid);

	if (!sid_equal(global_sid_builtin, &tmp))
	{
		return NT_STATUS_NONE_MAPPED;
	}

	DEBUG(5, ("lookup_wk_group_sid: rid: %d", rid));

	/* look up the well-known domain group rids first */
	mapped = lookup_wk_group_rid(rid);
	if (mapped == NULL)
	{
		DEBUG(5, (" none mapped\n"));
		return NT_STATUS_NONE_MAPPED;
	}
	fstrcpy(group_name, mapped);
	(*type) = SID_NAME_DOM_GRP;
	DEBUG(5, (" = %s\n", group_name));
	return 0x0;
}

/*******************************************************************
 lookup_group_sid
 ********************************************************************/
static uint32 lookup_group_sid(const DOM_SID *sid, char *group_name, uint32 *type)
{
	pstring sid_str;
	uint32 rid;
	DOM_SID tmp;
	DOMAIN_GRP *grp = NULL;
	uint32 status = NT_STATUS_NONE_MAPPED;

	sid_to_string(sid_str, sid);
	DEBUG(5, ("lookup_group_sid: sid: %s", sid_str));

	sid_copy(&tmp, sid);
	sid_split_rid(&tmp, &rid);

	if (!sid_equal(&global_sam_sid, &tmp))
	{
		DEBUG(5, ("not our SID\n"));
		return NT_STATUS_NONE_MAPPED;
	}

	become_root();
	grp = getgrouprid(rid, NULL, NULL);
	unbecome_root();

	if (grp != NULL)
	{
		fstrcpy(group_name, grp->name);
		(*type) = SID_NAME_DOM_GRP;
		DEBUG(5, (" = %s\n", group_name));
		return 0x0;
	}

	DEBUG(5, (" none mapped\n"));
	return status;
}

/*******************************************************************
 lookup_wk_alias_sid
 ********************************************************************/
static uint32 lookup_wk_alias_sid(const DOM_SID *sid, char *alias_name,
				  uint32 *type)
{
	uint32 rid;
	DOM_SID tmp;
	char *mapped;

	sid_copy(&tmp, sid);
	sid_split_rid(&tmp, &rid);

	if (!sid_equal(global_sid_builtin, &tmp))
	{
		return NT_STATUS_NONE_MAPPED;
	}

	DEBUG(5, ("lookup_wk_alias_sid: rid: %d", rid));

	/* look up the well-known alias group rids first */
	mapped = lookup_wk_alias_rid(rid);
	if (mapped == NULL)
	{
		DEBUG(5, (" none mapped\n"));
		return NT_STATUS_NONE_MAPPED;
	}
	fstrcpy(alias_name, mapped);
	(*type) = SID_NAME_ALIAS;
	DEBUG(5, (" = %s\n", alias_name));
	return 0x0;
}

/*******************************************************************
 lookup_alias_sid
 ********************************************************************/
static uint32 lookup_alias_sid(const DOM_SID *sid, char *alias_name, uint32 *type)
{
	pstring sid_str;
	uint32 rid;
	DOM_SID tmp;
	LOCAL_GRP *als = NULL;
	uint32 status = NT_STATUS_NONE_MAPPED;

	sid_to_string(sid_str, sid);
	DEBUG(5, ("lookup_alias_sid: sid: %s", sid_str));

	sid_copy(&tmp, sid);
	sid_split_rid(&tmp, &rid);

	if (!sid_equal(&global_sam_sid, &tmp))
	{
		DEBUG(5, ("not our SID\n"));
		return NT_STATUS_NONE_MAPPED;
	}

	become_root();
	als = getaliasrid(rid, NULL, NULL);
	unbecome_root();

	if (als != NULL)
	{
		fstrcpy(alias_name, als->name);
		(*type) = SID_NAME_ALIAS;
		DEBUG(5, (" = %s\n", alias_name));
		return 0x0;
	}

	DEBUG(5, (" none mapped\n"));
	return status;
}

/*******************************************************************
 lookup well-known user name
 ********************************************************************/
static uint32 lookup_wk_user_sid(const DOM_SID *sid, char *user_name, uint32 *type)
{
	uint32 rid;
	DOM_SID tmp;
	char *mapped;

	sid_copy(&tmp, sid);
	sid_split_rid(&tmp, &rid);

	if (!sid_equal(global_sid_builtin, &tmp))
	{
		return NT_STATUS_NONE_MAPPED;
	}

	DEBUG(5, ("lookup_wk_user_sid: rid: %d", rid));

	/* look up the well-known domain user rids first */
	mapped = lookup_wk_user_rid(rid);
	if (mapped == NULL)
	{
		DEBUG(5, (" none mapped\n"));
		return NT_STATUS_NONE_MAPPED;
	}
	fstrcpy(user_name, mapped);
	(*type) = SID_NAME_USER;
	DEBUG(5, (" = %s\n", user_name));
	return 0x0;
}

/*******************************************************************
 lookup user name
 ********************************************************************/
static uint32 lookup_user_sid(const DOM_SID *sid, char *user_name, uint32 *type)
{
	struct sam_passwd *sam_pass;
	uint32 rid;
	DOM_SID tmp;

	sid_copy(&tmp, sid);
	sid_split_rid(&tmp, &rid);

	if (sid_equal(&global_sam_sid, &tmp))
	{
		DEBUG(5, ("lookup_user_sid in SAM %s: rid: %d",
			  global_sam_name, rid));

		/* find the user account */
		become_root();
		sam_pass = getsam21pwrid(rid);
		unbecome_root();

		if (sam_pass != NULL)
		{
			fstrcpy(user_name, sam_pass->nt_name);
			(*type) = SID_NAME_USER;
			DEBUG(5, (" = %s\n", user_name));
			return 0x0;
		}

		DEBUG(5, (" none mapped\n"));
	}

	return NT_STATUS_NONE_MAPPED;
}

/*******************************************************************
 lookup_builtin_sid
 ********************************************************************/
static uint32 lookup_builtin_sid(const DOM_SID *sid, char *name, uint32 *type)
{
	uint32 status = NT_STATUS_NONE_MAPPED;

	if (status != 0x0)
		status = lookup_wk_user_sid(sid, name, type);
	if (status != 0x0)
		status = lookup_wk_group_sid(sid, name, type);
	if (status != 0x0)
		status = lookup_wk_alias_sid(sid, name, type);

	return status;
}

/*******************************************************************
 lookup_added_sid - names that have been added to the SAM database by admins.
 ********************************************************************/
static uint32 lookup_added_sid(const DOM_SID *sid, char *name, uint32 *type)
{
	uint32 status = NT_STATUS_NONE_MAPPED;

	if (status != 0x0)
		status = lookup_user_sid(sid, name, type);
	if (status != 0x0)
		status = lookup_group_sid(sid, name, type);
	if (status != 0x0)
		status = lookup_alias_sid(sid, name, type);

	return status;
}


/*******************************************************************
 lookup_sid
 ********************************************************************/
uint32 lookup_sid(const DOM_SID *sid, char *name, uint32 *type)
{
	uint32 status = NT_STATUS_NONE_MAPPED;

	if (status != 0x0)
		status = lookup_added_sid(sid, name, type);
	if (status != 0x0)
		status = lookup_builtin_sid(sid, name, type);

	return status;
}


/*******************************************************************
 lookup_group_rid
 ********************************************************************/
static uint32 lookup_added_group_name(const char *grp_name,
				      const char *domain,
				      DOM_SID *sid, uint32 *type)
{
	DOMAIN_GRP *grp = NULL;

	DEBUG(5, ("lookup_added_group_name: name: %s", grp_name));

	if (!strequal(domain, global_sam_name))
	{
		DEBUG(5, (" not our domain\n"));
		return NT_STATUS_NONE_MAPPED;
	}

	become_root();
	grp = getgroupntnam(grp_name);
	unbecome_root();

	if (grp != NULL)
	{
		sid_copy(sid, &global_sam_sid);
		sid_append_rid(sid, grp->rid);
		(*type) = SID_NAME_DOM_GRP;

		DEBUG(5, (" = 0x%x\n", grp->rid));
		return 0x0;
	}

	DEBUG(5, (" none mapped\n"));
	return NT_STATUS_NONE_MAPPED;
}

/*******************************************************************
 lookup_added_alias_name
 ********************************************************************/
static uint32 lookup_added_alias_name(const char *als_name,
				      const char *domain,
				      DOM_SID *sid, uint32 *type)
{
	LOCAL_GRP *als = NULL;

	DEBUG(5, ("lookup_added_alias_name: name: %s\\%s", domain, als_name));

	if (!strequal(domain, global_sam_name))
	{
		DEBUG(5, (" not our domain\n"));
		return NT_STATUS_NONE_MAPPED;
	}

	become_root();
	als = getaliasntnam(als_name, NULL, NULL);
	unbecome_root();

	if (als != NULL)
	{
		sid_copy(sid, &global_sam_sid);
		sid_append_rid(sid, als->rid);
		(*type) = SID_NAME_ALIAS;

		DEBUG(5, (" = 0x%x\n", als->rid));
		return 0x0;
	}

	DEBUG(5, (" none mapped\n"));
	return NT_STATUS_NONE_MAPPED;
}

/*******************************************************************
 lookup_added_user_name
 ********************************************************************/
static uint32 lookup_added_user_name(const char *nt_name, const char *domain,
				     DOM_SID *sid, uint32 *type)
{
	struct sam_passwd *sam_pass;

	if (!strequal(domain, global_sam_name))
	{
		return NT_STATUS_NONE_MAPPED;
	}

	/* find the user account */
	become_root();
	sam_pass = getsam21pwntnam(nt_name);
	unbecome_root();

	if (sam_pass != NULL)
	{
		DEBUG(10, ("lookup_added_user_name: nt name: %s rid: %x\n",
			   nt_name, sam_pass->user_rid));
		sid_copy(sid, &global_sam_sid);
		sid_append_rid(sid, sam_pass->user_rid);
		(*type) = SID_NAME_USER;

		return 0x0;
	}

	return NT_STATUS_NONE_MAPPED;
}

/*******************************************************************
 lookup_grp_name
 ********************************************************************/
static uint32 lookup_grp_name(const char *name, const char *domain,
			      DOM_SID *sid, uint32 *type)
{
	uint32 status = NT_STATUS_NONE_MAPPED;

	if (status != 0x0)
		status = lookup_added_group_name(name, domain, sid, type);
	if (status != 0x0)
		status = lookup_added_alias_name(name, domain, sid, type);

	return status;
}

/*******************************************************************
 lookup_user_name
 ********************************************************************/
static uint32 lookup_user_name(const char *name, const char *domain,
			       DOM_SID *sid, uint32 *type)
{
	uint32 status = NT_STATUS_NONE_MAPPED;

	if (status != 0x0)
		status = lookup_added_user_name(name, domain, sid, type);

	return status;
}

/*******************************************************************
 lookup_name
 ********************************************************************/
uint32 lookup_name(const char *name, DOM_SID *sid, uint32 *type)
{
	uint32 status = NT_STATUS_NONE_MAPPED;
	fstring domain;
	fstring user;

	split_domain_name(name, domain, user);

	if (status != 0x0)
		status = lookup_user_name(user, domain, sid, type);
	if (status != 0x0)
		status = lookup_grp_name(user, domain, sid, type);
#if 0
	if (status != 0x0)
		status = lookup_domain_name(domain, sid, type);
#endif

	if (! NT_STATUS_IS_OK(status))
		DEBUG(2, ("lookup_name(%s) [%s][%s]: %s\n", name,
			  domain, user, get_nt_error_msg(status)));

	return status;
}
