/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   NT Domain Authentication SMB / MSRPC client
   Copyright (C) Andrew Tridgell 1994-2000
   Copyright (C) Luke Kenneth Casson Leighton 1996-2000
   Copyright (C) Elrond 2000
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#include "libsamba.h"
#include "rpc_parse.h"
#include "rpc_client.h"
#include "rpcclient.h"
#include "nterr.h"

extern FILE *out_hnd;

#if 0
#include "uuid/uuid.h"

static void rpc_uuid_to_uuid_t(uuid_t *uuid, const RPC_UUID *rpc_uuid)
{
	if (!uuid || !rpc_uuid)
		return;
	uuid->time_low = rpc_uuid->time_low;
	uuid->time_mid = rpc_uuid->time_mid;
	uuid->time_hi_and_version = rpc_uuid->time_hi_and_version;
	uuid->clock_seq_hi_and_reserved = rpc_uuid->remaining[0];
	uuid->clock_seq_low = rpc_uuid->remaining[1];

	/* unrolled version... */
	uuid->node[0] = rpc_uuid->remaining[2];
	uuid->node[1] = rpc_uuid->remaining[3];
	uuid->node[2] = rpc_uuid->remaining[4];
	uuid->node[3] = rpc_uuid->remaining[5];
	uuid->node[4] = rpc_uuid->remaining[6];
	uuid->node[5] = rpc_uuid->remaining[7];
}
#endif


/****************************************************************************
nt enumerate trusted domains
****************************************************************************/
NTSTATUS cmd_lsa_enum_trust_dom(struct client_info *info, int argc, char *argv[])
{
	NTSTATUS ret = NT_STATUS_OK;

	fstring srv_name;
	uint32 num_doms = 0;
	char **domains = NULL;
	DOM_SID **sids = NULL;
	POLICY_HND lsa_pol;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	DEBUG(4, ("cmd_lsa_enum_trust_dom: server:%s\n", srv_name));

	/* lookup domain controller; receive a policy handle */
	if (!lsa_open_policy(srv_name, &lsa_pol, False,
			     SEC_RIGHTS_MAXIMUM_ALLOWED))
	{
		ret = NT_STATUS_UNSUCCESSFUL;
		goto out;
	}

	while (NT_STATUS_IS_OK(ret))
	{
		uint32 enum_ctx;
		if(!lsa_enum_trust_dom(&lsa_pol,
				       &enum_ctx, &num_doms, &domains, &sids))
			ret = NT_STATUS_UNSUCCESSFUL;
		if (!enum_ctx)
			break;
	}

	if (!lsa_close(&lsa_pol))
	{
		ret = NT_STATUS_UNSUCCESSFUL;
	}
out:
	if (NT_STATUS_IS_OK(ret))
	{
		uint32 i;
		DEBUG(5, ("cmd_lsa_enum_trust_dom: query succeeded\n"));

		report(out_hnd, "LSA Enumerate Trusted Domains\n");
		for (i = 0; i < num_doms; i++)
		{
			fstring sid;
			sid_to_string(sid, sids[i]);
			report(out_hnd, "Domain:\t%s\tSID:\t%s\n",
			       domains[i], sid);
		}
	}
	else
	{
		DEBUG(5, ("cmd_lsa_enum_trust_dom: query failed\n"));
	}

	free_char_array(num_doms, domains);
	free_sid_array(num_doms, sids);

	return ret;
}


static void display_lsa_info(const LSA_INFO_UNION *info, uint16 level)
{
	switch(level)
	{
		case 12:
		case 13:
		{
			fstring sid_str;
			uchar *uuid_str;

			report(out_hnd, "\tDNS Domain Information:\n");
			show_unistr2("\tName:          [%s]\n", &info->id12.Name);
			show_unistr2("\tDNS Name:      [%s]\n", &info->id12.DnsDomainName);
			show_unistr2("\tDNS Tree Name: [%s]\n", &info->id12.DnsForestName);
			uuid_str = rpc_uuid_to_str(&info->id12.DomainGuid);
			if (uuid_str)
			{
				report(out_hnd, "\tGUID:          %s\n", uuid_str);
				free(uuid_str);
			}
			if (info->id12.Sid)
				sid_to_string(sid_str, &info->id12.Sid->sid);
			else
				fstrcpy(sid_str, "(NULL)");
			report(out_hnd, "\tSid:           %s\n", sid_str);
			break;
		}
		case 3:
		case 5:
		{
			report(out_hnd,
			       "for level 3 and 5 use lsaquery "
			       "without parameter\n");
			break;
		}
		default:
		{
			report(out_hnd,
			       "No display functionality for level "
			       "%d implemented yet\n", level);
			break;
		}
	}
}


static NTSTATUS cmd_lsa_query_info_3_5(struct client_info *info,
				     const POLICY_HND *lsa_pol)
{
	BOOL domain_something = False;

	fstrcpy(info->dom.level3_dom, "");
	fstrcpy(info->dom.level5_dom, "");
	ZERO_STRUCT(info->dom.level3_sid);
	ZERO_STRUCT(info->dom.level5_sid);

	if (!lsa_query_info_pol(lsa_pol, 0x03, info->dom.level3_dom,
				&info->dom.level3_sid))
	{
		report(out_hnd, "cmd_lsa_query_info: query level 3 failed\n");
		return NT_STATUS_UNSUCCESSFUL;
	}
	if (!lsa_query_info_pol(lsa_pol, 0x05, info->dom.level5_dom,
				&info->dom.level5_sid))
	{
		report(out_hnd, "cmd_lsa_query_info: query level 5 failed\n");
		return NT_STATUS_UNSUCCESSFUL;
	}

	DEBUG(5, ("cmd_lsa_query_info: query succeeded\n"));
	report(out_hnd, "LSA Query Info Policy\n");

	if (info->dom.level3_dom[0])
	{
		fstring sid;
		sid_to_string(sid, &info->dom.level3_sid);
		report(out_hnd, "Domain Member     - Domain: %s SID: %s\n",
		       info->dom.level3_dom, sid);
		domain_something = True;
	}
	if (info->dom.level5_dom[0])
	{
		fstring sid;
		sid_to_string(sid, &info->dom.level5_sid);
		report(out_hnd, "Domain Controller - Domain: %s SID: %s\n",
		       info->dom.level5_dom, sid);
		domain_something = True;
	}
	if (!domain_something)
	{
		report(out_hnd, "%s is not a Domain Member or Controller\n",
		       info->dest_host);
	}

	return NT_STATUS_OK;
}


/****************************************************************************
nt lsa query
****************************************************************************/
NTSTATUS cmd_lsa_query_info(struct client_info *info, int argc, char *argv[])
{
	fstring srv_name;
	POLICY_HND lsa_pol;
	uint16 info_level = 0;
	LSA_INFO_UNION *lsa_info = NULL;

	NTSTATUS status;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	if (argc >= 2)
	{
		info_level = atol(argv[1]);

		if (info_level==0)
		{
			report(out_hnd, "lsaquery [infolevel]\n");
			return NT_STATUS_INVALID_PARAMETER;
		}
	}

	DEBUG(4, ("cmd_lsa_query_info: server:%s\n", srv_name));

	/* lookup domain controller; receive a policy handle */
	if (!lsa_open_policy(srv_name, &lsa_pol,
			     False, SEC_RIGHTS_MAXIMUM_ALLOWED))
	{
		report(out_hnd, "lsa_open_policy failed\n");
		return NT_STATUS_UNSUCCESSFUL;
	}

	if (argc <= 1)
	{
		status = cmd_lsa_query_info_3_5(info, &lsa_pol);
	}
	else
	{
		status = lsa_query_info(&lsa_pol, info_level, &lsa_info);

		if (NT_STATUS_IS_OK(status))
		{
			display_lsa_info(lsa_info, info_level);
			lsa_free_info_union(lsa_info);
		}
	}
	lsa_close(&lsa_pol);
	return status;
}

/****************************************************************************
lookup names
****************************************************************************/
NTSTATUS cmd_lsa_lookup_names(struct client_info *info, int argc, char *argv[])
{
	fstring srv_name;
	int num_names;
	char **names;
	uint32 *types = NULL;
	DOM_SID *sids = NULL;
	uint32 num_sids = 0;
	NTSTATUS ret;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	DEBUG(4, ("cmd_lsa_lookup_names: server: %s\n", srv_name));

	num_names = argc-1;
	names = argv+1;

	if (num_names <= 0)
	{
		report(out_hnd, "lookupnames <name> [<name> ...]\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	ret = lookup_lsa_names(srv_name,
			       num_names, names, &num_sids, &sids, &types);

	if (!NT_STATUS_IS_OK(ret))
	{
		report(out_hnd, "cmd_lsa_lookup_names: FAILED: %s\n",
		       get_nt_error_msg(ret));
	}

	if (sids != NULL)
	{
		uint32 i;
		fstring temp;

		report(out_hnd, "Lookup Names:\n");
		for (i = 0; i < num_sids; i++)
		{
			sid_to_string(temp, &sids[i]);
			report(out_hnd, "SID: %s -> %s (%d: %s)\n",
			       names[i], temp, types[i],
			       get_sid_name_use_str(types[i]));
#if 0
			safe_free(sids[i]);
#endif
		}
		free(sids);
	}
	safe_free(types);
	return ret;
}

/****************************************************************************
lookup sids
****************************************************************************/
NTSTATUS cmd_lsa_lookup_sids(struct client_info *info, int argc, char *argv[])
{
	POLICY_HND lsa_pol;
	fstring srv_name;
	DOM_SID **sids = NULL;
	uint32 num_sids = 0;
	char **names = NULL;
	uint32 *types = NULL;
	int num_names = 0;

	NTSTATUS ret = NT_STATUS_UNSUCCESSFUL;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	DEBUG(4, ("cmd_lsa_lookup_sids: server: %s\n", srv_name));

	argv++;
	argc--;

	while (argc > 0)
	{
		DOM_SID sid;
		fstring sid_name;
		if (strnequal("S-", argv[0], 2))
		{
			fstrcpy(sid_name, argv[0]);
		}
		else
		{
			sid_to_string(sid_name, &info->dom.level5_sid);

			if (sid_name[0] == 0)
			{
				report(out_hnd,
				       "please use lsaquery first or specify a complete SID\n");
				return NT_STATUS_UNSUCCESSFUL;
			}

			fstrcat(sid_name, "-");
			fstrcat(sid_name, argv[0]);
		}
		string_to_sid(&sid, sid_name);

		add_sid_to_array(&num_sids, &sids, &sid);

		argc--;
		argv++;
	}

	if (num_sids == 0)
	{
		report(out_hnd, "lookupsids <SID or RID> [<SID or RID>...]\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	if (lsa_open_policy(srv_name, &lsa_pol, True,
			    SEC_RIGHTS_MAXIMUM_ALLOWED) &&
	    lsa_lookup_sids(&lsa_pol, num_sids, sids,
			    &names, &types, &num_names) &&
	    lsa_close(&lsa_pol))
		ret = NT_STATUS_OK;

	if (names != NULL)
	{
		int i;
		report(out_hnd, "Lookup SIDS:\n");
		for (i = 0; i < num_names; i++)
		{
			fstring temp;
			sid_to_string(temp, sids[i]);
			report(out_hnd, "SID: %s -> %s (%d: %s)\n",
			       temp, names[i], types[i],
			       get_sid_name_use_str(types[i]));
			safe_free(names[i]);
		}
		free(names);
	}

	safe_free(types);

	free_sid_array(num_sids, sids);

	DEBUG(5, ("cmd_lsa_lookup_sids: %s\n",
		  NT_STATUS_IS_OK(ret) ? "query succeeded" : "query failed"));

	return ret;
}


/*
 * "setsecret":
 */
NTSTATUS cmd_lsa_set_secret(struct client_info *info, int argc, char *argv[])
{
	const char *secret_name;
	fstring srv_name;
	const char *data;
	size_t len, ucs2_len;
	void *data_ucs2 = NULL;
	BOOL ret;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	if (argc != 3)
	{
		cmd_help_current(info);
		return NT_STATUS_INVALID_PARAMETER;
	}

	secret_name = argv[1];
	data = argv[2];
	len = strlen(data);

	ucs2_len = convert_string_allocate(CH_UNIX, CH_UCS2,
					   data, len,
					   &data_ucs2);

	if (ucs2_len == (size_t)-1 || !data_ucs2)
		return NT_STATUS_NO_MEMORY;

	ret = msrpc_lsa_set_secret(srv_name, secret_name,
				   data_ucs2, ucs2_len);

	safe_free(data_ucs2);

	return (ret ? NT_STATUS_OK : NT_STATUS_UNSUCCESSFUL);
}


NTSTATUS cmd_lsa_create_secret(struct client_info *info, int argc, char *argv[])
{
	char *secret_name;
	fstring srv_name;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	if (argc != 2)
	{
		report(out_hnd, "createsecret <secret name>\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	secret_name = argv[1];

	if (msrpc_lsa_create_secret(srv_name, secret_name, 0x020003))
	{
		report(out_hnd, "LSA Create Secret: OK\n");
		return NT_STATUS_OK;
	}
	report(out_hnd, "LSA Query Secret: failed\n");
	return NT_STATUS_UNSUCCESSFUL;
}


static NTSTATUS lsa_open_object(const POLICY_HND *lsa_hnd,
				int argc, char *argv[],
				uint32 des_access, POLICY_HND *obj_hnd)
{
	if (argc<=1)
	{
		report(out_hnd,
		       "\tpossible objects:\n"
		       "\t\t-connect\tThe lsa handle itself\n"
		       "\t\t-secret NAME\tAn LSA secret\n"
		       "\t\t-trust SID\tA Trust\n"
		      );
		return NT_STATUS_INVALID_PARAMETER;
	}

	if (lsa_hnd==NULL || obj_hnd==NULL)
		return NT_STATUS_INVALID_PARAMETER;

	if (strequal(argv[1], "-connect"))
	{
		*obj_hnd = *lsa_hnd;
		return NT_STATUS_OK;
	}

	if (strequal(argv[1], "-secret"))
	{
		if (argc < 3)
		{
			report(out_hnd, "-secret SECRETNAME\n");
			return NT_STATUS_INVALID_PARAMETER;
		}
		if (!lsa_open_secret(lsa_hnd, argv[2],
				     des_access,
				     obj_hnd))
		{
			return NT_STATUS_UNSUCCESSFUL;
		}
		return NT_STATUS_OK;
	}

	if (strequal(argv[1], "-trust"))
	{
		DOM_SID sid;
		if (argc < 3)
		{
			report(out_hnd, "-trust SID\n");
			return NT_STATUS_INVALID_PARAMETER;
		}
		if (! string_to_sid(&sid, argv[2]))
			return NT_STATUS_INVALID_PARAMETER;
		return lsa_open_trust_dom(lsa_hnd, &sid,
					  des_access, obj_hnd);
	}

	report(out_hnd, "unknown/unsupported object: %s\n", argv[1]);
	return NT_STATUS_INVALID_PARAMETER;
}

/****************************************************************************
lsa show security descriptor for object
****************************************************************************/
NTSTATUS cmd_lsa_show_sd(struct client_info *info, int argc, char *argv[])
{
	fstring srv_name;

	NTSTATUS status;

	POLICY_HND obj_hnd;
	POLICY_HND lsa_pol;
	SEC_DESC_BUF buf;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	ZERO_STRUCT(buf);

	if (argc < 2)
	{
		report(out_hnd, "lsashowsd OBJECT\n");
		lsa_open_object(NULL, 0, NULL, 0, NULL);
		return NT_STATUS_INVALID_PARAMETER;
	}

	/* lookup domain controller; receive a policy handle */
	if (!lsa_open_policy(srv_name, &lsa_pol, False,
			     SEC_RIGHTS_MAXIMUM_ALLOWED))
	{
		report(out_hnd, "Couldn't connect to lsa\n");
		return NT_STATUS_UNSUCCESSFUL;
	}

	status = lsa_open_object(&lsa_pol, argc, argv,
				 SEC_RIGHTS_MAXIMUM_ALLOWED, &obj_hnd);
	if (!NT_STATUS_IS_OK(status))
		goto out;

	if (lsa_query_sec_obj(&obj_hnd, 0x07, &buf) &&
	    buf.sec != NULL)
	{
		display_sec_desc(out_hnd, ACTION_ENUMERATE, buf.sec);
		free_sec_desc_buf(&buf);
	}
	else
	{
		report(out_hnd, "LSA Get Security Descriptor failed\n");
		status = NT_STATUS_UNSUCCESSFUL;
	}

	lsa_close(&obj_hnd);
out:
	lsa_close(&lsa_pol);

	return status;
}


/*
 * "querysecret":
 */
NTSTATUS cmd_lsa_query_secret(struct client_info *info, int argc, char *argv[])
{
	char *secret_name;
	STRING2 secret;
	NTTIME last_update;
	fstring srv_name;
	uint32 len;
	pstring data;
	uchar nt_owf[16];

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	ZERO_STRUCT(secret);

	if (argc != 2)
	{
		report(out_hnd, "querysecret <secret name>\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	secret_name = argv[1];

	if (!msrpc_lsa_query_secret(srv_name, secret_name, &secret,
				    &last_update))
	{
		report(out_hnd, "LSA Query Secret: failed\n");
		return NT_STATUS_UNSUCCESSFUL;
	}		

	report(out_hnd, "\tValue:\n");
	out_data(out_hnd, secret.buffer, secret.str_str_len, 8, "\t\t");

	if (secret_get_data(&secret, data, &len))
	{
		report(out_hnd, "\tPlaintext:\n");
		out_data(out_hnd, data, len, 8, "\t\t");
	}
	if (secret_to_nt_owf(nt_owf, &secret))
	{
		report(out_hnd, "\tNT OWF:\n");
		out_data(out_hnd, nt_owf, 16, 8, "\t\t");
	}

	display_nttime(out_hnd, "\tLast Updated: ", &last_update);

	return NT_STATUS_OK;
}


static void show_priv_dispname(const POLICY_HND *lsa_hnd,
			       const UNISTR2 *privname, uint16 lang_id)
{
	UNISTR2 *uni_desc = NULL;
	uint16 ret_lang_id = 0xffff;
	char *desc;
	lsa_priv_get_dispname(lsa_hnd, privname,
			      lang_id,
			      &uni_desc, &ret_lang_id);
	desc = unistr2_to_ascii(NULL, uni_desc, 0);
	report(out_hnd, "\t\t%s (0x%x)\n",
	       desc, lang_id);
	safe_free(desc);
	unistr2_free(uni_desc);
}


/*
 * "enumprivs":
 */
NTSTATUS cmd_lsa_enum_privs(struct client_info *info, int argc, char *argv[])
{
	NTSTATUS ret = NT_STATUS_UNSUCCESSFUL;
	fstring srv_name;
	uint32 unk0 = 0, unk1 = 0x1000;
	BOOL do_info = False;
	POLICY_HND lsa_pol;
	uint32 count, i;
	LSA_PRIV_ENTRY *privs;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	DEBUG(4, ("cmd_lsa_enum_privileges: server:%s\n", srv_name));

	while (argc >= 2)
	{
		if (strequal(argv[1], "-i"))
		{
			do_info = True;
			argc--; argv++;
		}
		else
		{
			cmd_help_current(info);
			return NT_STATUS_INVALID_PARAMETER;
		}
	}

	/* lookup domain controller; receive a policy handle */
	if (!lsa_open_policy(srv_name, &lsa_pol, False,
			     SEC_RIGHTS_MAXIMUM_ALLOWED))
		return ret;

	if (!lsa_enum_privs(&lsa_pol, unk0, unk1, &count, &privs))
		goto out1;

	ret = NT_STATUS_OK;

	for (i = 0; i < count; i++)
	{
		char *name = unistr2_to_ascii(NULL, &privs[i].name, 0);
		report(out_hnd, "\t0x%x/0x%02x  %s\n",
		       privs[i].luid_high, privs[i].luid_low, name);
		if (do_info)
		{
			show_priv_dispname(&lsa_pol, &privs[i].name, 0x400);
		}
		safe_free(name);
	}
out1:
	lsa_close(&lsa_pol);
	return ret;
}


NTSTATUS cmd_lsa_priv_info(struct client_info *info, int argc, char *argv[])
{
	NTSTATUS ret;
	fstring srv_name;
	uint16 lang_in = 0x407;
	POLICY_HND lsa_pol;
	UNISTR2 name;
	UNISTR2 *uni_desc = NULL;
	uint16 lang_out = 0;

	if (argc < 2 || argc > 3)
	{
		report(out_hnd, "privinfo <priv-name> [lang_id]\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	unistr2_assign_ascii_str(&name, argv[1]);

	if (argc == 3)
		lang_in = atoi(argv[2]);

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	DEBUG(4, ("cmd_lsa_enum_privileges: server:%s\n", srv_name));

	/* lookup domain controller; receive a policy handle */
	if (!lsa_open_policy(srv_name, &lsa_pol, False,
			     SEC_RIGHTS_MAXIMUM_ALLOWED))
		return NT_STATUS_UNSUCCESSFUL;

	ret = lsa_priv_get_dispname(&lsa_pol, &name, lang_in, &uni_desc, &lang_out);

	if (NT_STATUS_IS_OK(ret))
	{
		char *desc;
		desc = unistr2_to_ascii(NULL, uni_desc, 0);
		report(out_hnd, "\t%s (0x%x)\n",
		       desc, lang_out);
		safe_free(desc);
		unistr2_free(uni_desc);
	}
	lsa_close(&lsa_pol);
	return ret;
}


static void show_privs_for_sid(const POLICY_HND *lsa_hnd, const DOM_SID *sid)
{
	uint32 num_privs;
	UNISTR2 **privs;
	uint32 i;

	lsa_sid_get_privs(lsa_hnd, sid,
			  &num_privs, &privs);

	if (privs == NULL)
		num_privs = 0;

	for (i = 0; i < num_privs; i++)
	{
		char *priv_asc = unistr2_to_ascii(NULL, privs[i], 0);

		report(out_hnd, "\t\t%s\n", priv_asc);
		show_priv_dispname(lsa_hnd, privs[i], 0x400);
		safe_free(priv_asc);
		safe_free(privs[i]);
	}

	safe_free(privs);
}


NTSTATUS cmd_lsa_enum_sids(struct client_info *info, int argc, char *argv[])
{
	NTSTATUS ret;
	int opt;
	fstring srv_name;
	uint32 unk0 = 0, unk1 = 0x1000;
	POLICY_HND lsa_pol;
	uint32 count, num_names;
	DOM_SID **sids;
	char **names;
	BOOL show_privs = False;

	while ((opt = getopt(argc, argv, "p")) != EOF)
	{
		switch (opt)
		{
			case 'p':
			{
				show_privs = True;
			}
		}
	}

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	DEBUG(4, ("cmd_lsa_enum_sids: server:%s\n", srv_name));

	/* lookup domain controller; receive a policy handle */
	if (!lsa_open_policy(srv_name, &lsa_pol, False,
			     SEC_RIGHTS_MAXIMUM_ALLOWED))
		return NT_STATUS_UNSUCCESSFUL;

	ret = lsa_enum_sids(&lsa_pol, &unk0, unk1, &count, &sids);
	if (!NT_STATUS_IS_OK(ret))
		goto out;
	if (!lsa_lookup_sids(&lsa_pol, count, sids, &names, NULL, &num_names))
	{
		ret = NT_STATUS_UNSUCCESSFUL;
		goto out;
	}

	if (names)
	{
		uint32 i;

		for (i = 0; i < count; i++)
		{
			fstring sid_str;

			sid_to_string(sid_str, sids[i]);
			report(out_hnd, "\t%s -> %s\n", sid_str, names[i]);
			safe_free(names[i]);
			if (show_privs)
				show_privs_for_sid(&lsa_pol, sids[i]);
		}
		safe_free(names);
	}
out:
	lsa_close(&lsa_pol);
	return ret;
}

NTSTATUS cmd_lsa_query_trust_info(struct client_info *info, int argc, char *argv[])
{
	fstring srv_name;
	NTSTATUS status;
	POLICY_HND lsa_pol;
	DOM_SID sid;
	uint16 info_level = 2;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	if (argc < 2 || argc > 3)
	{
		report(out_hnd, "lsatrustinfo SID [level]\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	if (argc == 3)
	{
		info_level = atol(argv[2]);

		if (info_level==0)
		{
			report(out_hnd, "lsaquery [infolevel]\n");
			return NT_STATUS_INVALID_PARAMETER;
		}
	}

	if (!string_to_sid(&sid, argv[1]))
	{
		report(out_hnd, "Invalid SID\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	DEBUG(4, ("cmd_lsa_query_trust_info: server:%s\n", srv_name));

	/* lookup domain controller; receive a policy handle */
	if (!lsa_open_policy(srv_name, &lsa_pol,
			     False, SEC_RIGHTS_MAXIMUM_ALLOWED))
	{
		report(out_hnd, "lsa_open_policy failed\n");
		return NT_STATUS_UNSUCCESSFUL;
	}

	status = lsa_get_trust_info2(&lsa_pol, &sid, info_level, NULL);

	lsa_close(&lsa_pol);

	return status;
}
