/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   Samba utility functions
   Copyright (C) Andrew Tridgell 1992-1999
   Copyright (C) Luke Kenneth Casson Leighton 1996 - 1999
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#include "libsamba.h"
#include "tng_misc.h"
#include "rpcclient.h"
#include "rpc_client.h"
#include "rpc_reg.h"


static BOOL lookup_one_sid(const POLICY_HND *lsa_hnd, const DOM_SID *sid,
			   char **name, uint32 *type)
{
	DOM_SID *sids[1];
	char **names = NULL;
	uint32 *types = NULL;
	int ret_names;

	/* STRIP_CONST is currently ok - lsa_lookup_sids() doesn't modify the array of sids */
	sids[0] = STRIP_CONST(sid);

	if (!lsa_lookup_sids(lsa_hnd, 1, sids,
			     &names, &types, &ret_names))
	{
		if (name)
			*name = strdup("(lsa_lookup_sids failed)");
		return False;
	}

	if (ret_names != 1)
	{
		safe_free(names);
		safe_free(types);
		return False;
	}
	if (type)
	{
		*type = types[0];
	}
	if (name && names)
	{
		*name = names[0];
		names[0] = NULL;
	}
	safe_free(types);
	safe_free(names);

	return True;
}

static void print_sid(FILE *out_hnd, const POLICY_HND *lsa_hnd,
		      const DOM_SID *sid)
{
	fstring sid_str;
	char *name = NULL;
	uint32 type;

	sid_to_string(sid_str, sid);

	if (!lsa_hnd || !lookup_one_sid(lsa_hnd, sid, &name, &type))
	{
		report(out_hnd, "%s", sid_str);
		return;
	}
	report(out_hnd, "%s (%s) (%s)", name,
	       get_sid_name_use_str(type), sid_str);
	safe_free(name);
}


typedef struct {
	int obj_type;
	uint32 mask;
	const char *name;
} NTAccessDef;

const NTAccessDef access_table[] = {
	{ SECOBJ_SERVICE, SEC_SERVICE_QUERY_CONFIG, "QueryConfig" },
	{ SECOBJ_SERVICE, SEC_SERVICE_CHANGE_CONFIG, "ChangeConfig" },
	{ SECOBJ_SERVICE, SEC_SERVICE_QUERY_STATUS, "QueryStatus" },
	{ SECOBJ_SERVICE, SEC_SERVICE_ENUMERATE_DEPENDENTS, "EnumDeps" },
	{ SECOBJ_SERVICE, SEC_SERVICE_START, "Start" },
	{ SECOBJ_SERVICE, SEC_SERVICE_STOP, "Stop" },
	{ SECOBJ_SERVICE, SEC_SERVICE_PAUSE_CONTINUE, "Pause/Continue" },
	{ SECOBJ_SERVICE, SEC_SERVICE_INTERROGATE, "Interrogate" },
	{ SECOBJ_SERVICE, SEC_SERVICE_USER_DEFINED_CONTROL, "UserDefControl" },

	{ SECOBJ_REGKEY, SEC_KEY_FULL_CONTROL, "FullControl" },
	{ SECOBJ_REGKEY, SEC_KEY_QUERY_VALUE, "QueryValue" },
	{ SECOBJ_REGKEY, SEC_KEY_SET_VALUE, "SetValue" },
	{ SECOBJ_REGKEY, SEC_KEY_CREATE_SUBKEY, "CreateSubkey" },
	{ SECOBJ_REGKEY, SEC_KEY_ENUM_SUBKEYS, "EnumSubkeys" },
	{ SECOBJ_REGKEY, SEC_KEY_NOTIFY, "Notify" },
	{ SECOBJ_REGKEY, SEC_KEY_CREATE_LINK, "CreateLink" },

	{ SECOBJ_ANY, SEC_RIGHTS_MAXIMUM_ALLOWED, "MaximumAllowed" },
	{ SECOBJ_ANY, SEC_RIGHTS_DELETE, "Delete" },
	{ SECOBJ_ANY, SEC_RIGHTS_READ_CONTROL, "ReadControl" },
	{ SECOBJ_ANY, SEC_RIGHTS_WRITE_DAC, "WriteDAC" },
	{ SECOBJ_ANY, SEC_RIGHTS_WRITE_OWNER, "WriteOwner" },

	{ 0, 0, NULL }
};


static void display_sec_access(FILE *out_hnd, const char *ind_str,
			       int object_type, void *obj_masks,
			       const SEC_ACCESS *info)
{
	uint32 mask, handled;
	const NTAccessDef *p;

	mask = info->mask;
	handled = 0;

	report(out_hnd, "%s\t", ind_str);

	for (p = access_table; p->name; p++)
	{
		uint32 test_mask;
		if ((p->obj_type != object_type) && (p->obj_type != 0))
		        continue;

		test_mask = p->mask;

		if (IS_BITS_SET_ALL(handled, test_mask))
			continue;
		if (! IS_BITS_SET_ALL(mask, test_mask))
			continue;

		report(out_hnd, "%s ", p->name);
		handled |= test_mask;
	}

	mask ^= handled; /* remove handled bits */
	if (mask)
		report(out_hnd, "[%04x]", mask);
	report(out_hnd, "\n");
}


static struct field_info ace_flags[] = {
	{ SEC_ACE_FLAG_OBJECT_INHERIT, "(OI)", "OBJECT_INHERIT" },
	{ SEC_ACE_FLAG_CONTAINER_INHERIT, "(CI)", "CONTAINER_INHERIT" },
	{ SEC_ACE_FLAG_NO_PROPAGATE_INHERIT, "(NP)", "NO_PROPAGATE_INHERIT" },
	{ SEC_ACE_FLAG_INHERIT_ONLY, "(IO)", "INHERIT_ONLY" },
	{ SEC_ACE_FLAG_INHERITED_ACE, "(IA)", "INHERITED_ACE" },
	{ SEC_ACE_FLAG_SUCCESSFUL_ACCESS, "SUCC", "SUCCESSFUL_ACCESS" },
	{ SEC_ACE_FLAG_FAILED_ACCESS, "FAIL", "FAILED_ACCESS" },
	{ 0, "", NULL },
	{ 0, NULL, NULL }
};


/****************************************************************************
 display sec_ace structure
 ****************************************************************************/
static void display_sec_ace(FILE *out_hnd, const char *ind_str,
			    int object_type, void *obj_masks,
			    const POLICY_HND *lsa_hnd,
			    const SEC_ACE *const ace)
{
	report(out_hnd, "%s", ind_str);
	print_sid(out_hnd, lsa_hnd, &ace->sid);
	report(out_hnd, "\n");

	report(out_hnd,
	       "%s\tType:0x%x  Perms:0x%08x\n",
	       ind_str,
	       ace->type, (uint32) ace->info.mask);
	report(out_hnd, "%s\tFlags: %s (0x%x)\n",
	       ind_str,
	       bit_field_to_str(ace->flags, ace_flags, NULL),
	       ace->flags);

	display_sec_access(out_hnd, ind_str, 
			   object_type, obj_masks,
			   &ace->info);
}

/****************************************************************************
 display sec_acl structure
 ****************************************************************************/
static void display_sec_acl(FILE *out_hnd, const char *ind_str,
			    const char *acl_name,
			    int object_type, void *obj_masks,
			    const POLICY_HND *lsa_hnd,
			    const SEC_ACL *sec_acl)
{
	if (sec_acl == NULL)
	{
		return;
	}

	report(out_hnd, "%s%s ACL   Num ACEs: %d   Revision: %x\n", 
	       ind_str, acl_name, sec_acl->num_aces, sec_acl->revision); 
	report(out_hnd, "%s---\n", ind_str);

	if (sec_acl->size != 0 && sec_acl->num_aces != 0)
	{
		int i;
		for (i = 0; i < sec_acl->num_aces; i++)
		{
			display_sec_ace(out_hnd, ind_str, 
					object_type, obj_masks,
					lsa_hnd, &sec_acl->ace[i]);
		}
	}

	report(out_hnd, "\n");
}


void print_sec_desc(const SEC_DESC *sd,
		    FILE *out_hnd, const char *ind_str,
		    int object_type, void *obj_masks,
		    const char *srv_name, const POLICY_HND *lsa_hnd)
{
	POLICY_HND *my_lsa_hnd = NULL;

	if (!lsa_hnd && srv_name)
	{
		my_lsa_hnd = g_new(POLICY_HND, 1);
		if (lsa_open_policy(srv_name, my_lsa_hnd, True,
				    SEC_RIGHTS_MAXIMUM_ALLOWED))
		{
			lsa_hnd = my_lsa_hnd;
		}
		else
		{
			safe_free(my_lsa_hnd);
		}
	}

	if (!ind_str)
	{
		ind_str = "\t";
	}

	fprintf(out_hnd,
		"%sSecurity Descriptor\n"
		"%s-------------------\n",
		ind_str, ind_str);

	if (!sd)
	{
		fprintf(out_hnd, "%s(Null)\n", ind_str);
		goto clean_up;
	}

	fprintf(out_hnd,
		"%sRevision: 0x%x\ttype: 0x%x\n",
		ind_str, sd->revision, sd->type); 

	if (sd->off_owner_sid != 0)
	{
		report(out_hnd, "%sOwner: ", ind_str);
		print_sid(out_hnd, lsa_hnd, sd->owner_sid);
		report(out_hnd, "\n");
	}
	if (sd->off_grp_sid != 0)
	{
		report(out_hnd, "%sGroup: ", ind_str);
		print_sid(out_hnd, lsa_hnd, sd->grp_sid);
		report(out_hnd, "\n");
	}
	if (sd->off_owner_sid || sd->off_grp_sid)
	{
		report(out_hnd, "%s\n", ind_str);
	}
	if (sd->off_sacl)
	{
		display_sec_acl(out_hnd, ind_str, "System ",
				object_type, obj_masks,
				lsa_hnd, sd->sacl);
	}
	if (sd->off_dacl)
	{
		display_sec_acl(out_hnd, ind_str, "D",
				object_type, obj_masks,
				lsa_hnd, sd->dacl);
	}


    clean_up:
	if (my_lsa_hnd)
	{
		lsa_close(my_lsa_hnd);
		safe_free(my_lsa_hnd);
		my_lsa_hnd = NULL;
	}
}



/****************************************************************************
 display sec_desc structure
 ****************************************************************************/
void display_sec_desc(FILE *out_hnd, enum action_type action, SEC_DESC *const sec)
{
	switch (action)
	{
		case ACTION_HEADER:
		{
			break;
		}
		case ACTION_ENUMERATE:
		{
			print_sec_desc(sec, out_hnd, NULL, 0, NULL,
				       NULL, NULL);

			break;
		}
		case ACTION_FOOTER:
		{
			break;
		}
	}
}

