/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   Samba utility functions
   Copyright (C) Andrew Tridgell 1992-1999
   Copyright (C) Luke Kenneth Casson Leighton 1996 - 1999
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#include "libsamba.h"
#include "rpcclient.h"


/****************************************************************************
convert a security permissions into a string
****************************************************************************/
static char *get_svc_start_type_str(uint32 type)
{
	static fstring typestr;

	switch (type)
	{
		case 0x00: fstrcpy(typestr, "Boot"    ); return typestr;
		case 0x01: fstrcpy(typestr, "System"  ); return typestr;
		case 0x02: fstrcpy(typestr, "Auto"    ); return typestr;
		case 0x03: fstrcpy(typestr, "Manual"  ); return typestr;
		case 0x04: fstrcpy(typestr, "Disabled"); return typestr;
		default  : break;
	}
	slprintf(typestr, sizeof(typestr)-1, "[%d]", type);
	return typestr;
}

static const char *get_svc_state_str(uint32 state)
{
	switch(state)
	{
		case SERVICE_STOPPED:		return "Stopped";
		case SERVICE_START_PENDING:	return "Start Pending";
		case SERVICE_STOP_PENDING:	return "Stop Pending";
		case SERVICE_RUNNING:		return "Running";
		case SERVICE_CONTINUE_PENDING:	return "Continue Pending";
		case SERVICE_PAUSE_PENDING:	return "Pause Pending";
		case SERVICE_PAUSED:		return "Paused";
	}
	return "Unknown";
}

static struct field_info service_type_bits[] = {
	{ 0x1, "KERNEL_DRIVER", NULL },
	{ 0x2, "FILE_SYSTEM_DRIVER", NULL },
	{ 0x10, "WIN32_OWN_PROCESS", NULL },
	{ 0x20, "WIN32_SHARE_PROCESS", NULL },
	{ 0x100, " (interactive)", NULL },
	{ 0, NULL, NULL }
};


/****************************************************************************
 display structure
 ****************************************************************************/
void display_query_svc_cfg(FILE *out_hnd, enum action_type action, 
				const QUERY_SERVICE_CONFIG *const cfg)
{
	switch (action)
	{
		case ACTION_HEADER:
		{
			fstring service;

			unistr2_to_ascii(service, &cfg->uni_display_name, sizeof(service)-1);
			report(out_hnd, "\tService:\t%s\n", service);
			report(out_hnd, "\t-------\n");
			break;
		}
		case ACTION_ENUMERATE:
		{
			fstring temp;

			unistr2_to_ascii(temp, &cfg->uni_bin_path_name, sizeof(temp)-1);
			report(out_hnd, "\tPath:\t%s\n", temp);

			unistr2_to_ascii(temp, &cfg->uni_load_order_grp, sizeof(temp)-1);
			report(out_hnd, "\tLoad Order:\t%s\n", temp);

			unistr2_to_ascii(temp, &cfg->uni_dependencies, sizeof(temp)-1);
			report(out_hnd, "\tDependencies:\t%s\n", temp);

			unistr2_to_ascii(temp, &cfg->uni_service_start_name, sizeof(temp)-1);
			report(out_hnd, "\tService Start:\t%s\n", temp);

			report(out_hnd, "\tService Type:\t0x%x  %s\n",
				cfg->service_type,
				bit_field_to_str(cfg->service_type, service_type_bits, NULL));
			report(out_hnd, "\tStart Type:\t%s\n" , get_svc_start_type_str(cfg->start_type));
			report(out_hnd, "\tError Control:\t%d\n" , cfg->error_control);
			report(out_hnd, "\tTag Id:\t%d\n" , cfg->tag_id);
			break;

		}
		case ACTION_FOOTER:
		{
			report(out_hnd, "\n");
			break;
		}
	}
}

/****************************************************************************
 display structure
 ****************************************************************************/
static void display_svc_status(FILE *out_hnd, const SVC_STATUS *svcstat)
{
	report(out_hnd, "\t\tsvc_type:\t\t0x%x  %s\n", svcstat->svc_type,
	       bit_field_to_str(svcstat->svc_type, service_type_bits, NULL));
	report(out_hnd, "\t\tcurrent_state:\t\t%d (%s)\n",
	       svcstat->current_state,
	       get_svc_state_str(svcstat->current_state));
	report(out_hnd, "\t\tcontrols_accepted:\t0x%x\n", svcstat->controls_accepted);
	report(out_hnd, "\t\twin32_exit_code:\t0x%x\n", svcstat->win32_exit_code);
	report(out_hnd, "\t\tsvc_specific_exit_code:\t0x%x\n", svcstat->svc_specific_exit_code);
	report(out_hnd, "\t\tcheck_point:\t\t0x%x\n", svcstat->check_point);
	report(out_hnd, "\t\twait_hint:\t\t0x%x\n", svcstat->wait_hint);
}

void display_svc_info(FILE *out_hnd, enum action_type action,
				const ENUM_SRVC_STATUS *const svc)
{
	switch (action)
	{
		case ACTION_HEADER:
		{
			break;
		}
		case ACTION_ENUMERATE:
		{
			display_svc_status(out_hnd, &(svc->status));
			break;
		}
		case ACTION_FOOTER:
		{
			break;
		}
	}
}

void display_svc_name(FILE *out_hnd, enum action_type action,
				const ENUM_SRVC_STATUS *const svc)
{
	switch (action)
	{
		case ACTION_HEADER:
		{
			break;
		}
		case ACTION_ENUMERATE:
		{
			fstring name;

			unistr_to_ascii(name, svc->uni_srvc_name.buffer, 
					sizeof(name)-1); /* service name */
			report(out_hnd, "\t%s:", name);

			unistr_to_ascii(name, svc->uni_disp_name.buffer, 
					sizeof(name)-1); /* display name */
			report(out_hnd, "\t%s\n", name);
			break;
		}
		case ACTION_FOOTER:
		{
			break;
		}
	}
}

