/* 
   Unix SMB/CIFS implementation.
   Blocking Locking functions
   Copyright (C) Jeremy Allison 1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: smbd_proto.h		(do not remove this line) */

#include "includes.h"
#include <errno.h>
#include "libsamba.h"
#include "dlinklist.h"
#include "byteorder.h"
#include "libsmb.h"
#include "nterr.h"
#include "doserr.h"
#include "smbd.h"
#include "tng_misc.h"

extern char *OutBuffer;

/****************************************************************************
 This is the structure to queue to implement blocking locks.
 notify. It consists of the requesting SMB and the expiry time.
*****************************************************************************/

typedef struct _blocking_lock_record {
	struct _blocking_lock_record *next;
	struct _blocking_lock_record *prev;
  int com_type;
  files_struct *fsp;
  time_t expire_time;
  int lock_num;
  char *inbuf;
  int length;
} blocking_lock_record;

static blocking_lock_record *blocking_lock_queue;

/****************************************************************************
 Destructor for the above structure.
****************************************************************************/

static void free_blocking_lock_record(blocking_lock_record *blr)
{
	DLIST_REMOVE(blocking_lock_queue, blr);
	SAFE_FREE(blr->inbuf);
	SAFE_FREE(blr);
}

/****************************************************************************
 Get the files_struct given a particular queued SMB.
*****************************************************************************/


/****************************************************************************
 Determine if this is a secondary element of a chained SMB.
  **************************************************************************/

static BOOL in_chained_smb(void)
{
  return (chain_size != 0);
}


/****************************************************************************
 Function to push a blocking lock request onto the lock queue.
****************************************************************************/

BOOL push_blocking_lock_request(files_struct *fsp, char *inbuf, int length, int lock_timeout, int lock_num)
{
  blocking_lock_record *blr, *tmp;

  if(in_chained_smb() ) {
    DEBUG(0,("push_blocking_lock_request: cannot queue a chained request (currently).\n"));
    return False;
  }

  /*
   * Now queue an entry on the blocking lock queue. We setup
   * the expiration time here.
   */

  if((blr = g_new1(blocking_lock_record)) == NULL) {
    DEBUG(0,("push_blocking_lock_request: Malloc fail !\n" ));
    return False;
  }

  if((blr->inbuf = g_new(char, length)) == NULL) {
    DEBUG(0,("push_blocking_lock_request: Malloc fail (2)!\n" ));
    free((char *)blr);
    return False;
  }

  blr->com_type = CVAL(inbuf,smb_com);
  blr->fsp = fsp;
  blr->expire_time = (lock_timeout == -1) ? (time_t)-1 : time(NULL) + (time_t)lock_timeout;
  blr->lock_num = lock_num;
  memcpy(blr->inbuf, inbuf, length);
  blr->length = length;



  DLIST_ADD_END(blocking_lock_queue, blr, tmp);


  DEBUG(3,("push_blocking_lock_request: lock request length=%d blocked with expiry time %d (+%d) \
for fnum = %d, name = %s\n", length, (int)blr->expire_time, lock_timeout,
        blr->fsp->fnum, file_debug_name(blr->fsp) ));

  /* Push the MID of this packet on the signing queue. */

  return True;
}

/****************************************************************************
 Return a smd with a given size.
*****************************************************************************/

static void send_blocking_reply(char *outbuf, int outsize)
{
  if(outsize > 4)
    smb_setlen(outbuf,outsize - 4);

  if (! send_smb(smbd_server_fd(),outbuf))
    exit_server("send_blocking_reply: send_smb failed.");
}

/****************************************************************************
 Return a lockingX success SMB.
*****************************************************************************/

static void reply_lockingX_success(blocking_lock_record *blr)
{
  char *outbuf = OutBuffer;
  int bufsize = BUFFER_SIZE;
  char *inbuf = blr->inbuf;
  int outsize = 0;

  construct_reply_common(inbuf, outbuf);
  set_message(outbuf,2,0,True);

  /*
   * As this message is a lockingX call we must handle
   * any following chained message correctly.
   * This is normally handled in construct_reply(),
   * but as that calls switch_message, we can't use
   * that here and must set up the chain info manually.
   */

  outsize = chain_reply(inbuf,outbuf,blr->length,bufsize);

  outsize += chain_size;

  send_blocking_reply(outbuf,outsize);
}

/****************************************************************************
 Return a generic lock fail error blocking call.
*****************************************************************************/

static void generic_blocking_lock_error(blocking_lock_record *blr, int eclass, int32 ecode)
{
  char *outbuf = OutBuffer;
  char *inbuf = blr->inbuf;
  construct_reply_common(inbuf, outbuf);

  if(eclass == 0) /* NT Error. */
    SSVAL(outbuf,smb_flg2, SVAL(outbuf,smb_flg2) | FLAGS2_32_BIT_ERROR_CODES);

  ERROR_DOS(eclass,ecode);
  if (! send_smb(smbd_server_fd(),outbuf))
    exit_server("generic_blocking_lock_error: send_smb failed.");
}

/****************************************************************************
 Return a lock fail error for a lockingX call. Undo all the locks we have 
 obtained first.
*****************************************************************************/

static void reply_lockingX_error(blocking_lock_record *blr, int eclass, int32 ecode)
{
  char *inbuf = blr->inbuf;
  files_struct *fsp = blr->fsp;
  uint16 num_ulocks = SVAL(inbuf,smb_vwv6);
  SMB_BIG_UINT count = (SMB_BIG_UINT)0, offset = (SMB_BIG_UINT) 0;
  uint16 lock_pid;
  unsigned char locktype = CVAL(inbuf,smb_vwv3);
  BOOL large_file_format = (locktype & LOCKING_ANDX_LARGE_FILES);
  char *data;
  int i;

  data = smb_buf(inbuf) + ((large_file_format ? 20 : 10)*num_ulocks);

  /* 
   * Data now points at the beginning of the list
   * of smb_lkrng structs.
   */

  /*
   * Ensure we don't do a remove on the lock that just failed,
   * as under POSIX rules, if we have a lock already there, we
   * will delete it (and we shouldn't) .....
   */

  for(i = blr->lock_num - 1; i >= 0; i--) {
    BOOL err;

    lock_pid = get_lock_pid( data, i, large_file_format);
    count = get_lock_count( data, i, large_file_format);
    offset = get_lock_offset( data, i, large_file_format, &err);

    /*
     * We know err cannot be set as if it was the lock
     * request would never have been queued. JRA.
     */

    do_unlock(fsp, fsp->conn, lock_pid,count,offset);
  }

  generic_blocking_lock_error(blr, eclass, ecode);
}

/****************************************************************************
 Return a lock fail error.
*****************************************************************************/

static void blocking_lock_reply_error(blocking_lock_record *blr, int eclass, int32 ecode)
{
  switch(blr->com_type) {
  case SMBlock:
  case SMBlockread:
    generic_blocking_lock_error(blr, eclass, ecode);
    break;
  case SMBlockingX:
    reply_lockingX_error(blr, eclass, ecode);
    break;
  default:
    DEBUG(0,("blocking_lock_reply_error: PANIC - unknown type on blocking lock queue - exiting.!\n"));
    exit_server("PANIC - unknown type on blocking lock queue");
  }
}

/****************************************************************************
 Attempt to finish off getting all pending blocking locks for a lockread call.
 Returns True if we want to be removed from the list.
*****************************************************************************/

static BOOL process_lockread(blocking_lock_record *blr)
{
  char *outbuf = OutBuffer;
  char *inbuf = blr->inbuf;
  ssize_t nread = -1;
  char *data;
  int outsize = 0;
  SMB_OFF_T startpos;
  size_t numtoread;
  int eclass;
  uint32 ecode;
  files_struct *fsp = blr->fsp;

  numtoread = SVAL(inbuf,smb_vwv1);
  startpos = IVAL(inbuf,smb_vwv2);
 
  numtoread = MIN(BUFFER_SIZE-outsize,numtoread);
  data = smb_buf(outbuf) + 3;

  if(!do_lock(fsp, SVAL(inbuf, smb_pid),
  	      (SMB_BIG_UINT)numtoread, (SMB_BIG_UINT)startpos, READ_LOCK, &eclass, &ecode))
  {
    /* This errno stuff looks funny... */
    if((errno != EACCES) && (errno != EAGAIN)) {
      /*
       * We have other than a "can't get lock" POSIX
       * error. Send an error.
       * Return True so we get dequeued.
       */

      generic_blocking_lock_error(blr, eclass, ecode);
      return True;
    }

    /*
     * Still waiting for lock....
     */

    DEBUG(10,("process_lockread: failed to get lock for file = %s. Still waiting....\n",
          file_debug_name(fsp)));
    return False;
  }

  nread = read_file(fsp,data,startpos,numtoread);

  if (nread < 0) {
    generic_blocking_lock_error(blr,ERRDOS,ERRnoaccess);
    return True;
  }

  construct_reply_common(inbuf, outbuf);
  outsize = set_message(outbuf,5,3,True);

  outsize += nread;
  SSVAL(outbuf,smb_vwv0,nread);
  SSVAL(outbuf,smb_vwv5,nread+3);
  SSVAL(smb_buf(outbuf),1,nread);

  DEBUG(3, ( "process_lockread file = %s, fnum=%d num=%d nread=%d\n",
        file_debug_name(fsp), fsp->fnum, (int)numtoread, (int)nread ) );

  send_blocking_reply(outbuf,outsize);
  return True;
}

/****************************************************************************
 Attempt to finish off getting all pending blocking locks for a lock call.
 Returns True if we want to be removed from the list.
*****************************************************************************/

static BOOL process_lock(blocking_lock_record *blr)
{
  char *outbuf = OutBuffer;
  char *inbuf = blr->inbuf;
  int outsize;
  SMB_OFF_T count = 0, offset = 0;
  int eclass;
  uint32 ecode;
  files_struct *fsp = blr->fsp;

  count = IVAL(inbuf,smb_vwv1);
  offset = IVAL(inbuf,smb_vwv3);

  /* XXX - This errno stuff looks funny... */
  errno = 0;
  if (!do_lock(fsp, SVAL(inbuf, smb_pid),
	       (SMB_BIG_UINT)count, (SMB_BIG_UINT)offset, WRITE_LOCK, &eclass, &ecode))
  {
    if((errno != EACCES) && (errno != EAGAIN)) {

      /*
       * We have other than a "can't get lock" POSIX
       * error. Send an error.
       * Return True so we get dequeued.
       */

      blocking_lock_reply_error(blr, eclass, ecode);
      return True;
    }

    /*
     * Still can't get the lock - keep waiting.
     */

    DEBUG(10,("process_lock: failed to get lock for file = %s. Still waiting....\n",
          file_debug_name(fsp)));
    return False;
  }

  /*
   * Success - we got the lock.
   */

  DEBUG(3,("process_lock : file=%s fnum=%d offset=%.0f count=%.0f\n",
       file_debug_name(fsp), fsp->fnum, (double)offset, (double)count));

  construct_reply_common(inbuf, outbuf);
  outsize = set_message(outbuf,0,0,True);
  send_blocking_reply(outbuf,outsize);
  return True;
}

/****************************************************************************
 Attempt to finish off getting all pending blocking locks for a lockingX call.
 Returns True if we want to be removed from the list.
*****************************************************************************/

static BOOL process_lockingX(blocking_lock_record *blr)
{
  char *inbuf = blr->inbuf;
  unsigned char locktype = CVAL(inbuf,smb_vwv3);
  files_struct *fsp = blr->fsp;
  uint16 num_ulocks = SVAL(inbuf,smb_vwv6);
  uint16 num_locks = SVAL(inbuf,smb_vwv7);
  SMB_BIG_UINT count = (SMB_BIG_UINT)0, offset = (SMB_BIG_UINT)0;
  uint16 lock_pid;
  BOOL large_file_format = (locktype & LOCKING_ANDX_LARGE_FILES);
  char *data;
  int eclass=0;
  uint32 ecode=0;

  data = smb_buf(inbuf) + ((large_file_format ? 20 : 10)*num_ulocks);

  /* 
   * Data now points at the beginning of the list
   * of smb_lkrng structs.
   */

  for(; blr->lock_num < num_locks; blr->lock_num++) {
    BOOL err;

    lock_pid = get_lock_pid( data, blr->lock_num, large_file_format);
    count = get_lock_count( data, blr->lock_num, large_file_format);
    offset = get_lock_offset( data, blr->lock_num, large_file_format, &err);

    /*
     * We know err cannot be set as if it was the lock
     * request would never have been queued. JRA.
     */
    errno = 0;
    if(!do_lock(fsp, lock_pid, count, offset,
		((locktype & 1) ? READ_LOCK : WRITE_LOCK),
                &eclass, &ecode))
      break;
  }

  if(blr->lock_num == num_locks) {

    /*
     * Success - we got all the locks.
     */

    DEBUG(3,("process_lockingX file = %s, fnum=%d type=%d num_locks=%d\n",
          file_debug_name(fsp), fsp->fnum, (unsigned int)locktype, num_locks) );

    reply_lockingX_success(blr);
    return True;
  }

	if(! ((eclass == ERRDOS) && (ecode == ERRlock)))
	{
		/* Something else went wrong, get out */
		DEBUG(2, ("process_lockingX: Something else went wrong: "
			  "0x%x/0x%x\n", eclass, ecode));
		blocking_lock_reply_error(blr, eclass, ecode);
		return True;
	}

  /*
   * Still can't get all the locks - keep waiting.
   */

  DEBUG(10,("process_lockingX: only got %d locks of %d needed for file %s, fnum = %d. \
Waiting....\n", blr->lock_num, num_locks, file_debug_name(fsp), fsp->fnum));

  return False;
}

/****************************************************************************
 Process a blocking lock SMB.
 Returns True if we want to be removed from the list.
*****************************************************************************/

static BOOL blocking_lock_record_process(blocking_lock_record *blr)
{
  switch(blr->com_type) {
  case SMBlock:
    return process_lock(blr);
  case SMBlockread:
    return process_lockread(blr);
  case SMBlockingX:
    return process_lockingX(blr);
  default:
    DEBUG(0,("blocking_lock_record_process: PANIC - unknown type on blocking lock queue - exiting.!\n"));
    exit_server("PANIC - unknown type on blocking lock queue");
  }
  return False; /* Keep compiler happy. */
}

/****************************************************************************
 Delete entries by fnum from the blocking lock pending queue.
*****************************************************************************/

void remove_pending_lock_requests_by_fid(files_struct *fsp)
{
	blocking_lock_record *blr, *next = NULL;

	for(blr = blocking_lock_queue; blr; blr = next) {
		next = blr->next;
		if(blr->fsp->fnum == fsp->fnum) {

			DEBUG(10,("remove_pending_lock_requests_by_fid - removing request type %d for "
				  "file %s fnum = %d\n", blr->com_type, file_debug_name(fsp), fsp->fnum ));


			free_blocking_lock_record(blr);
		}
	}
}

/****************************************************************************
 Delete entries by mid from the blocking lock pending queue. Always send reply.
*****************************************************************************/

void remove_pending_lock_requests_by_mid(int mid)
{
	blocking_lock_record *blr, *next = NULL;

	for(blr = blocking_lock_queue; blr; blr = next) {
		next = blr->next;
		if(SVAL(blr->inbuf,smb_mid) == mid) {
			files_struct *fsp = blr->fsp;

			DEBUG(10,("remove_pending_lock_requests_by_mid - removing request type %d for \
file %s fnum = %d\n", blr->com_type, file_debug_name(fsp), fsp->fnum ));

			blocking_lock_reply_error(blr,0,NT_STATUS_CANCELLED);
			free_blocking_lock_record(blr);
		}
	}
}

/****************************************************************************
  Set a flag as an unlock request affects one of our pending locks.
*****************************************************************************/


/****************************************************************************
 Return the number of seconds to the next blocking locks timeout, or default_timeout
*****************************************************************************/


/****************************************************************************
 Process the blocking lock queue. Note that this is only called as root.
*****************************************************************************/

void process_blocking_lock_queue(time_t t)
{
  blocking_lock_record *blr, *next = NULL;

  /*
   * Go through the queue and see if we can get any of the locks.
   */

	DEBUG(5, ("process_blocking_lock_queue for t=%d\n", (int)t));

  for (blr = blocking_lock_queue; blr; blr = next) {
    connection_struct *conn = NULL;
    uint16 vuid;
    files_struct *fsp = NULL;

    next = blr->next;

    /*
     * Ensure we don't have any old chain_fsp values
     * sitting around....
     */
    chain_size = 0;
    file_chain_reset();
    fsp = blr->fsp;

    conn = fsp->conn;
    vuid = (lp_security() == SEC_SHARE) ? UID_FIELD_INVALID :
                  SVAL(blr->inbuf,smb_uid);

    DEBUG(5,("process_blocking_lock_queue: examining pending lock fnum = %d for file %s\n",
          fsp->fnum, file_debug_name(fsp) ));
    
    DEBUG(6, ("process_blocking_locking_queue: expire time = %d, time = %d\n",
	      (int)blr->expire_time, (int)t));

    if((blr->expire_time != -1) && (blr->expire_time < t)) {
      /*
       * Lock expired - throw away all previously
       * obtained locks and return lock error.
       */
      DEBUG(5,("process_blocking_lock_queue: pending lock fnum = %d for file %s timed out.\n",
          fsp->fnum, file_debug_name(fsp) ));


      blocking_lock_reply_error(blr, 0, STATUS_FILE_LOCK_CONFLICT);
      free_blocking_lock_record(blr);
      continue;
    }

    if(!become_user(conn,vuid)) {
      DEBUG(0,("process_blocking_lock_queue: Unable to become user vuid=%d.\n",
            vuid ));
      /*
       * Remove the entry and return an error to the client.
       */
      blocking_lock_reply_error(blr,ERRSRV,ERRaccess);


      free_blocking_lock_record(blr);
      continue;
    }

    if(!become_service(conn,True)) {
      DEBUG(0,("process_blocking_lock_queue: Unable to become service Error was %s.\n", strerror(errno) ));
      /*
       * Remove the entry and return an error to the client.
       */
      blocking_lock_reply_error(blr,ERRSRV,ERRaccess);


      free_blocking_lock_record(blr);
      unbecome_user();
      continue;
    }

    /*
     * Go through the remaining locks and try and obtain them.
     * The call returns True if all locks were obtained successfully
     * and False if we still need to wait.
     */

    if(blocking_lock_record_process(blr))
    {
      DEBUG(6,("process_blocking_locking_queue: Kick the lock out the list\n"));

      free_blocking_lock_record(blr);
    }

    unbecome_user();

  }
}


/**
 * \brief Return True if the blocking lock queue has entries.
 */
BOOL blocking_locks_pending(void)
{
	blocking_lock_record *blr = blocking_lock_queue;
	return (blr != NULL);
}
