/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   SMB transaction2 handling
   Copyright (C) Jeremy Allison 1994-1998

   Extensively modified by Andrew Tridgell, 1995

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: smbd_proto.h		(do not remove this line) */

#include "includes.h"
#include <errno.h>
#include "libsamba.h"
#include "tng_misc.h"
#include "byteorder.h"
#include "libsmb.h"
#include "trans2.h"
#include "msdfs.h"
#include "doserr.h"
#include "nterr.h"
#include "winerror.h"
#include "smbprofile.h"
#include "smbd.h"
#include "vfs.h"

extern int Protocol;
extern int smb_read_error;

#define DIR_ENTRY_SAFETY_MARGIN 4096

/****************************************************************************
  Send the required number of replies back.
  We assume all fields other than the data fields are
  set correctly for the type of call.
  HACK ! Always assumes smb_setup field is zero.
****************************************************************************/
static int send_trans2_replies(connection_struct *conn,
			       char *outbuf, int bufsize, char *params,
			       int paramsize, const char *pdata, int datasize)
{
  /* As we are using a protocol > LANMAN1 then the max_send
     variable must have been set in the sessetupX call.
     This takes precedence over the max_xmit field in the
     global struct. These different max_xmit variables should
     be merged as this is now too confusing */

  extern int max_send;
  int data_to_send = datasize;
  int params_to_send = paramsize;
  int useable_space;
  char *pp = params;
  const char *pd = pdata;
  int params_sent_thistime, data_sent_thistime, total_sent_thistime;
  int alignment_offset = 3;
  int data_alignment_offset = 0;

  /* Initially set the wcnt area to be 10 - this is true for all
     trans2 replies */
  set_message(outbuf,10,0,True);

  /* If there genuinely are no parameters or data to send just send
     the empty packet */
  if(params_to_send == 0 && data_to_send == 0)
  {
    smbd_conn_send_smb(conn, outbuf);
    return 0;
  }

  /* When sending params and data ensure that both are nicely aligned */
  /* Only do this alignment when there is also data to send - else
     can cause NT redirector problems. */
  if (((params_to_send % 4) != 0) && (data_to_send != 0))
    data_alignment_offset = 4 - (params_to_send % 4);

  /* Space is bufsize minus Netbios over TCP header minus SMB header */
  /* The alignment_offset is to align the param bytes on an even byte
     boundary. NT 4.0 Beta needs this to work correctly. */
  useable_space = bufsize - ((smb_buf(outbuf)+
                    alignment_offset+data_alignment_offset) - 
                    outbuf);

  /* useable_space can never be more than max_send minus the
     alignment offset. */
  useable_space = MIN(useable_space, 
                      max_send - (alignment_offset+data_alignment_offset));


  while (params_to_send || data_to_send)
  {
    /* Calculate whether we will totally or partially fill this packet */
    total_sent_thistime = params_to_send + data_to_send + 
                            alignment_offset + data_alignment_offset;
    /* We can never send more than useable_space */
    /*
     * Note that 'useable_space' does not include the alignment offsets,
     * but we must include the alignment offsets in the calculation of
     * the length of the data we send over the wire, as the alignment offsets
     * are sent here. Fix from Marc_Jacobsen@hp.com.
     */
    total_sent_thistime = MIN(total_sent_thistime, useable_space+
			        alignment_offset + data_alignment_offset);

    set_message(outbuf, 10, total_sent_thistime, True);

    /* Set total params and data to be sent */
    SSVAL(outbuf,smb_tprcnt,paramsize);
    SSVAL(outbuf,smb_tdrcnt,datasize);

    /* Calculate how many parameters and data we can fit into
       this packet. Parameters get precedence */

    params_sent_thistime = MIN(params_to_send,useable_space);
    data_sent_thistime = useable_space - params_sent_thistime;
    data_sent_thistime = MIN(data_sent_thistime,data_to_send);

    SSVAL(outbuf,smb_prcnt, params_sent_thistime);

    /* smb_proff is the offset from the start of the SMB header to the
       parameter bytes, however the first 4 bytes of outbuf are
       the Netbios over TCP header. Thus use smb_base() to subtract
       them from the calculation */

    SSVAL(outbuf,smb_proff,((smb_buf(outbuf)+alignment_offset) - smb_base(outbuf)));

    if(params_sent_thistime == 0)
      SSVAL(outbuf,smb_prdisp,0);
    else
      /* Absolute displacement of param bytes sent in this packet */
      SSVAL(outbuf,smb_prdisp,pp - params);

    SSVAL(outbuf,smb_drcnt, data_sent_thistime);
    if(data_sent_thistime == 0)
    {
      SSVAL(outbuf,smb_droff,0);
      SSVAL(outbuf,smb_drdisp, 0);
    }
    else
    {
      /* The offset of the data bytes is the offset of the
         parameter bytes plus the number of parameters being sent this time */
      SSVAL(outbuf,smb_droff,((smb_buf(outbuf)+alignment_offset) - 
            smb_base(outbuf)) + params_sent_thistime + data_alignment_offset);
      SSVAL(outbuf,smb_drdisp, pd - pdata);
    }

    /* Copy the param bytes into the packet */
    if(params_sent_thistime)
      memcpy((smb_buf(outbuf)+alignment_offset),pp,params_sent_thistime);
    /* Copy in the data bytes */
    if(data_sent_thistime)
      memcpy(smb_buf(outbuf)+alignment_offset+params_sent_thistime+
             data_alignment_offset,pd,data_sent_thistime);

    DEBUG(9,("t2_rep: params_sent_thistime = %d, data_sent_thistime = %d, useable_space = %d\n",
          params_sent_thistime, data_sent_thistime, useable_space));
    DEBUG(9,("t2_rep: params_to_send = %d, data_to_send = %d, paramsize = %d, datasize = %d\n",
          params_to_send, data_to_send, paramsize, datasize));

    /* Send the packet */
    smbd_conn_send_smb(conn, outbuf);

    pp += params_sent_thistime;
    pd += data_sent_thistime;

    params_to_send -= params_sent_thistime;
    data_to_send -= data_sent_thistime;

    /* Sanity check */
    if(params_to_send < 0 || data_to_send < 0)
    {
      DEBUG(0,("send_trans2_replies failed sanity check pts = %d, dts = %d\n!!!",
            params_to_send, data_to_send));
      return -1;
    }
  }

  return 0;
}


/****************************************************************************
 Reply to a TRANSACT2_OPEN.
****************************************************************************/

static int call_trans2open(connection_struct *conn, char *inbuf, char *outbuf,
			   int length, int bufsize,  
			   char **pparams, int total_params, char **ppdata, int total_data)
{
  char *params = *pparams;
  int16 open_mode;
  int16 open_attr;
  BOOL oplock_request;
#if 0
  BOOL return_additional_info;
  int16 open_sattr;
  time_t open_time;
#endif
  int16 open_ofun;
  int32 open_size;
  char *pname;

  pstring fname;
  SMB_OFF_T size=0;
  int fmode=0,mtime=0,rmode;
  SMB_INO_T inode = 0;
  SMB_STRUCT_STAT sbuf;
  int smb_action = 0;
  BOOL bad_path = False;
  files_struct *fsp;

	/*
	 * Ensure we have enough parameters to perform the operation.
	 */

	if (total_params < 29)
		return(ERROR_DOS(ERRDOS,ERRinvalidparam));

	open_mode = SVAL(params, 2);
	open_attr = SVAL(params,6);
	oplock_request = (((SVAL(params,0)|(1<<1))>>1) | ((SVAL(params,0)|(1<<2))>>1));
#if 0
	return_additional_info = BITSETW(params,0);
	open_sattr = SVAL(params, 4);
	open_time = make_unix_date3(params+8);
#endif
	open_ofun = SVAL(params,12);
	open_size = IVAL(params,14);
	pname = &params[28];

	pstrcpy(fname,pname);

  DEBUG(3,("trans2open %s mode=%d attr=%d ofun=%d size=%d\n",
	   fname,open_mode, open_attr, open_ofun, open_size));

  if (IS_IPC(conn)) {
		return(ERROR_DOS(ERRSRV,ERRaccess));
  }

  /* XXXX we need to handle passed times, sattr and flags */

  unix_convert(fname,conn,0,&bad_path,&sbuf);
    
  if (!check_name(fname,conn))
  {
    set_bad_path_error(errno, bad_path);
    return(UNIXERROR(ERRDOS,ERRnoaccess));
  }

  fsp = open_file_shared(conn,fname,open_mode,open_ofun,
			 unix_mode(conn,open_attr | aARCH, fname),
			 oplock_request, &rmode,&smb_action,
			 smbd_open_des_access(open_mode),
			 SVAL(inbuf, smb_pid));
      
  if (!fsp)
  {
    set_bad_path_error(errno, bad_path);
    return(UNIXERROR(ERRDOS,ERRnoaccess));
  }

  if (fsp_stat(fsp, &sbuf) != 0) {
    close_file(fsp,False);
    return(UNIXERROR(ERRDOS,ERRnoaccess));
  }
    
  size = sbuf.st_size;
  fmode = dos_mode(conn,fname,&sbuf);
  mtime = sbuf.st_mtime;
  inode = sbuf.st_ino;
  if (fmode & aDIR) {
    close_file(fsp,False);
    return(ERROR_DOS(ERRDOS,ERRnoaccess));
  }

  /* Realloc the size of parameters and data we will return */
  if (!g_xrenew(char, pparams, 28)) {
    return(ERROR_DOS(ERRDOS,ERRnomem));
  }
  params = *pparams;

  bzero(params, 28);
  SSVAL(params,0,fsp->fnum);
  SSVAL(params,2,fmode);
  put_dos_date2(params,4, mtime);
  SIVAL(params,8, (uint32)size);
  SSVAL(params,12,rmode);

  if (oplock_request && lp_fake_oplocks(SNUM(conn))) {
    smb_action |= EXTENDED_OPLOCK_GRANTED;
  }

  SSVAL(params,18,smb_action);
  /*
   * WARNING - this may need to be changed if SMB_INO_T <> 4 bytes.
   */
  SIVAL(params,20,inode);
 
  /* Send the required number of replies */
  send_trans2_replies(conn, outbuf, bufsize, params, 28, *ppdata, 0);

  return -1;
}

/*********************************************************
* Routine to check if a given string matches exactly.
* as a special case a mask of "." does NOT match. That
* is required for correct wildcard semantics
* Case can be significant or not.
**********************************************************/
static BOOL exact_match(const char *str, const char *mask, BOOL case_sig) 
{
	if (mask[0] == '.' && mask[1] == 0) return False;
	if (case_sig) return strcmp(str,mask)==0;
	return strcasecmp(str,mask) == 0;
}

/****************************************************************************
  get a level dependent lanman2 dir entry.
****************************************************************************/
static BOOL get_lanman2_dir_entry(connection_struct *conn,
				  void *inbuf, void *outbuf,
				 char *path_mask,int dirtype,int info_level,
				 int requires_resume_key,
				 BOOL dont_descend,char **ppdata, 
				 char *base_data, int space_remaining, 
				 BOOL *out_of_space, BOOL *got_exact_match,
				 int *last_name_off)
{
  const char *dname;
  BOOL found = False;
  SMB_STRUCT_STAT sbuf;
  pstring mask;
  pstring fname;
  const char *unix_fname;
  char *p, *pdata = *ppdata;
  uint32 reskey=0;
  int prev_dirpos=0;
  int mode=0;
  SMB_OFF_T size = 0;
  uint32 len;
  char *nameptr, *pnamelen;
  BOOL was_8_3;
  NTFileStatInfo nti;
  BOOL needslash = ( conn->dirpath[strlen(conn->dirpath) -1] != '/');

  *fname = 0;
  *out_of_space = False;
  *got_exact_match = False;

  if (!conn->dirptr)
    return(False);

  p = strrchr(path_mask,'/');
  if(p != NULL)
  {
    if(p[1] == '\0')
      pstrcpy(mask,"*.*");
    else
      pstrcpy(mask, p+1);
  }
  else
    pstrcpy(mask, path_mask);

  while (!found)
  {
    BOOL got_match;

    /* Needed if we run out of space */
    prev_dirpos = TellDir(conn->dirptr);
    dname = ReadDirName(conn->dirptr);
    DEBUG(10, ("get_lanman2_dir_entry: Looking at %s\n", dname));

    /*
     * Due to bugs in NT client redirectors we are not using
     * resume keys any more - set them to zero.
     * Check out the related comments in findfirst/findnext.
     * JRA.
     */

    reskey = 0;

    DEBUG(8,("get_lanman2_dir_entry:readdir on dirptr 0x%lx now at offset %d\n",
      (long)conn->dirptr,TellDir(conn->dirptr)));
      
    if (!dname) 
      return(False);

    pstrcpy(fname,dname);      

    if(!(got_match = *got_exact_match = exact_match(fname, mask, conn->case_sensitive))) {
      got_match = mask_match(fname, mask, conn->case_sensitive);
    }

    if(!got_match && !is_8_3(fname, False)) {

      /*
       * It turns out that NT matches wildcards against
       * both long *and* short names. This may explain some
       * of the wildcard wierdness from old DOS clients
       * that some people have been seeing.... JRA.
       */

      pstring newname;
      pstrcpy( newname, fname);
      name_map_mangle(conn, newname, True, False);
      if(!(got_match = *got_exact_match = exact_match(newname, mask, conn->case_sensitive)))
        got_match = mask_match(newname, mask, conn->case_sensitive);
    }

    if(got_match)
    {
      pstring pathreal;
      BOOL isdots = (strequal(fname,"..") || strequal(fname,"."));
      if (dont_descend && !isdots)
        continue;
	  
      pstrcpy(pathreal,conn->dirpath);
      if(needslash)
        pstrcat(pathreal,"/");
      pstrcat(pathreal,dname);
      if (vfs_stat(conn, dos_to_unix_static(pathreal), &sbuf) != 0)
      {
        /* Needed to show the msdfs symlinks as directories */
        if(!lp_host_msdfs() || !lp_msdfs_root(SNUM(conn))
           || !is_msdfs_link(conn, dos_to_unix_static(pathreal), NULL, NULL, NULL))
          {
            DEBUG(3,("get_lanman2_dir_entry:Couldn't stat [%s] (%s)\n",
                     pathreal,strerror(errno)));
            continue;
          }
        else
          {
            DEBUG(5,("get_lanman2_dir_entry: Masquerading msdfs link %s as a directory\n", pathreal));
            sbuf.st_mode = (sbuf.st_mode & 0xFFF) | S_IFDIR;
          }
      }
 
      vfs_fill_ntinfo(conn, &nti, pathreal, &sbuf);
      mode = nti.FileAttributes & SAMBA_ATTRIBUTES_MASK;

      if (!dir_check_ftype(conn,mode,&sbuf,dirtype)) {
        DEBUG(5,("[%s] attribs didn't match %x\n",fname,dirtype));
        continue;
      }

      size = sbuf.st_size;

      DEBUG(5,("get_lanman2_dir_entry found %s fname=%s\n",pathreal,fname));
	  
      found = True;
    }
  }

  name_map_mangle(conn, fname, False, True);
  unix_fname = dos_to_unix_static(fname);

  p = pdata;
  nameptr = p;

  switch (info_level)
  {
    case 1:
      if(requires_resume_key) {
        SIVAL(p,0,reskey);
        p += 4;
      }
      nameptr = p + l1_achName;
      cifs_put_dos_date2(p + l1_fdateCreation, &nti.CreationTime);
      cifs_put_dos_date2(p + l1_fdateLastAccess, &nti.LastAccessTime);
      cifs_put_dos_date2(p + l1_fdateLastWrite, &nti.LastWriteTime);
      SIVAL(p,l1_cbFile,(uint32)size);
      SIVAL(p,l1_cbFileAlloc, nti.AllocationSize);
      SSVAL(p,l1_attrFile,mode);
      nameptr += cifs_align_string(outbuf, pdata, nameptr, 0);
      len = srvstr_push(outbuf, nameptr, unix_fname, -1, STR_NOALIGN);
      SCVAL(p,l1_cchName, len);
      p = nameptr + len;
      p += srvstr_push(outbuf, p, "", -1, STR_TERMINATE | STR_NOALIGN);
      break;

    case 2:
      /* info_level 2 */
      if(requires_resume_key) {
        SIVAL(p,0,reskey);
        p += 4;
      }
      nameptr = p + l2_achName;
      cifs_put_dos_date2(p + l2_fdateCreation, &nti.CreationTime);
      cifs_put_dos_date2(p + l2_fdateLastAccess, &nti.LastAccessTime);
      cifs_put_dos_date2(p + l2_fdateLastWrite, &nti.LastWriteTime);
      SIVAL(p,l2_cbFile,(uint32)size);
      SIVAL(p,l2_cbFileAlloc, nti.AllocationSize);
      SSVAL(p,l2_attrFile,mode);
      SIVAL(p,l2_cbList,0); /* No extended attributes */
      p += l2_cchName;
      len = srvstr_push(outbuf, nameptr, unix_fname, -1, STR_NOALIGN);
      SCVAL(p, 0, len);
      p = nameptr + len;
      SCVAL(p, 0, 0); p++;
      break;

    case 3:
      SIVAL(p,0,reskey);
      cifs_put_dos_date2(p + 4, &nti.CreationTime);
      cifs_put_dos_date2(p + 8, &nti.LastAccessTime);
      cifs_put_dos_date2(p + 12, &nti.LastWriteTime);
      SIVAL(p,16,(uint32)size);
      SIVAL(p,20, nti.AllocationSize);
      SSVAL(p,24,mode);
      SIVAL(p,26,4);
      CVAL(p,30) = strlen(fname);
      pstrcpy(p+31, fname);
      nameptr = p+31;
      p += 31 + strlen(fname) + 1;
      break;

    case 4:
      if(requires_resume_key) {
        SIVAL(p,0,reskey);
        p += 4;
      }
      SIVAL(p,0,33+strlen(fname)+1);
      cifs_put_dos_date2(p + 4, &nti.CreationTime);
      cifs_put_dos_date2(p + 8, &nti.LastAccessTime);
      cifs_put_dos_date2(p + 12, &nti.LastWriteTime);
      SIVAL(p,16,(uint32)size);
      SIVAL(p,20, nti.AllocationSize);
      SSVAL(p,24,mode);
      CVAL(p,32) = strlen(fname);
      pstrcpy(p + 33, fname);
      nameptr = p+33;
      p += 33 + strlen(fname) + 1;
      break;

    case SMB_FIND_FILE_BOTH_DIRECTORY_INFO:
      was_8_3 = is_8_3(fname, True);
      p += 4;
      SIVAL(p,0,reskey); p += 4;
      cifs_put_nttime(p, &nti.CreationTime); p += 8;
      cifs_put_nttime(p, &nti.LastAccessTime); p += 8;
      cifs_put_nttime(p, &nti.LastWriteTime); p += 8;
      cifs_put_nttime(p, &nti.ChangeTime); p += 8;
      SOFF_T(p,0,size);
      SOFF_T(p,8, nti.AllocationSize);
      p += 16;
      SIVAL(p,0, nti.FileAttributes); p += 4;
      pnamelen = p; p += 4;
      SIVAL(p,0,0); p += 4; /* Extended attributes */
      bzero(p, 26);
      if (!was_8_3) {
		const char *mangled_name = mangle_from_dos_to_unix(conn, fname, True, True);
		len = srvstr_push(outbuf, p+2, mangled_name, 24, STR_NOALIGN);
		SSVAL(p, 0, len);
      }
      p += 2 + 24;
      /* nameptr = p;  */
      len = srvstr_push(outbuf, p, unix_fname, -1, STR_NOALIGN);
      p += len;
      SIVAL(pnamelen, 0, len);
      len = PTR_DIFF(p, pdata);
      len = (len + 3) & ~3;
      SIVAL(pdata,0,len);
      p = pdata + len;
      break;

    case SMB_FIND_FILE_DIRECTORY_INFO:
      p += 4;
      SIVAL(p,0,reskey); p += 4;
      cifs_put_nttime(p, &nti.CreationTime); p += 8;
      cifs_put_nttime(p, &nti.LastAccessTime); p += 8;
      cifs_put_nttime(p, &nti.LastWriteTime); p += 8;
      cifs_put_nttime(p, &nti.ChangeTime); p += 8;
      SOFF_T(p,0,size);
      SOFF_T(p,8, nti.AllocationSize);
      p += 16;
      SIVAL(p,0, nti.FileAttributes); p += 4;
      len = srvstr_push(outbuf, p+4, unix_fname, -1, STR_NOALIGN);
      SIVAL(p, 0, len);
      p += 4;
      p += len;
      p = ALIGN4(p, pdata);
      len = PTR_DIFF(p, pdata);
      SIVAL(pdata,0,len);
      break;

    case SMB_FIND_FILE_FULL_DIRECTORY_INFO:
      p += 4; /* len */
      SIVAL(p,0,reskey); p += 4;
      cifs_put_nttime(p, &nti.CreationTime); p += 8;
      cifs_put_nttime(p, &nti.LastAccessTime); p += 8;
      cifs_put_nttime(p, &nti.LastWriteTime); p += 8;
      cifs_put_nttime(p, &nti.ChangeTime); p += 8;
      SOFF_T(p,0,size); 
      SOFF_T(p,8, nti.AllocationSize);
      p += 16;
      SIVAL(p,0, nti.FileAttributes); p += 4;
      pnamelen = p; p += 4;
      SIVAL(p,0,0); p += 4;
      len = srvstr_push(outbuf, p, unix_fname, -1, STR_NOALIGN);
      SIVAL(pnamelen, 0, len);
      p += len;
      p = ALIGN4(p, pdata);
      len = PTR_DIFF(p, pdata);
      SIVAL(pdata,0,len);
      break;

    case SMB_FIND_FILE_NAMES_INFO:
      p += 4; /* len */
      SIVAL(p,0,reskey); p += 4;
      pnamelen = p; p += 4;
      len = srvstr_push(outbuf, p, unix_fname, -1, STR_NOALIGN);
      SIVAL(pnamelen, 0, len);
      p += len;
      p = ALIGN4(p, pdata);
      len = PTR_DIFF(p, pdata);
      SIVAL(pdata,0,len);
      break;

    default:      
      return(False);
    }


  if (PTR_DIFF(p,pdata) > space_remaining) {
    /* Move the dirptr back to prev_dirpos */
    SeekDir(conn->dirptr, prev_dirpos);
    *out_of_space = True;
    DEBUG(9,("get_lanman2_dir_entry: out of space\n"));
    return False; /* Not finished - just out of space */
  }

  /* Setup the last_filename pointer, as an offset from base_data */
  *last_name_off = PTR_DIFF(nameptr,base_data);
  /* Advance the data pointer to the next slot */
  *ppdata = p;

  return(found);
}
  

/****************************************************************************
 Reply to a TRANS2_FINDFIRST.
****************************************************************************/

static int call_trans2findfirst(connection_struct *conn, char *inbuf, char *outbuf,
				int length, int bufsize,
				char **pparams, int total_params, char **ppdata, int total_data)
{
  /* We must be careful here that we don't return more than the
     allowed number of data bytes. If this means returning fewer than
     maxentries then so be it. We assume that the redirector has
     enough room for the fixed number of parameter bytes it has
     requested. */
  uint32 max_data_bytes = SVAL(inbuf, smb_mdrcnt);
  char *params = *pparams;
  char *pdata = *ppdata;
  int dirtype = SVAL(params,0);
  int maxentries = SVAL(params,2);
  BOOL close_after_first = BITSETW(params+4,0);
  BOOL close_if_end = BITSETW(params+4,1);
  BOOL requires_resume_key = BITSETW(params+4,2);
  int info_level = SVAL(params,6);
  pstring directory;
  pstring mask;
  char *p, *wcard;
  int last_name_off=0;
  int dptr_num = -1;
  int numentries = 0;
  int i;
  BOOL finished = False;
  BOOL dont_descend = False;
  BOOL out_of_space = False;
  int space_remaining;
  BOOL bad_path = False;
  NTSTATUS status;

	if (total_params < 12)
		return(ERROR_DOS(ERRDOS,ERRinvalidparam));

  *directory = *mask = 0;

  DEBUG(3,("call_trans2findfirst: dirtype = %d, maxentries = %d, close_after_first=%d, close_if_end = %d requires_resume_key = %d level = %d, max_data_bytes = %d\n",
	   dirtype, maxentries, close_after_first, close_if_end, requires_resume_key,
	   info_level, max_data_bytes));

	/* windows 2000+ treats 0 as being 1. Very stupid */
	if (! maxentries)
		maxentries = 1;

  switch (info_level) 
    {
    case 1:
    case 2:
    case 3:
    case 4:
    case SMB_FIND_FILE_DIRECTORY_INFO:
    case SMB_FIND_FILE_FULL_DIRECTORY_INFO:
    case SMB_FIND_FILE_NAMES_INFO:
    case SMB_FIND_FILE_BOTH_DIRECTORY_INFO:
      break;
    default:
      return(ERROR_DOS(ERRDOS,ERRunknownlevel));
    }

  /* Complete directory path with wildcard mask appended */
  srvstr_pull(inbuf, directory, params+12, sizeof(directory), total_params - 12, STR_TERMINATE|STR_NOALIGN);

	/* The last component (mask) needs to be ignored for this */
	status = path_check_syntax_wo_lastcomp(directory);
	if (!NT_STATUS_IS_OK(status))
		return ERROR_NT(status);

  unix_to_dos(directory);

  RESOLVE_FINDFIRST_DFSPATH(directory, conn, inbuf, outbuf);

  DEBUG(5,("path=%s\n",directory));

  unix_convert(directory,conn,0,&bad_path,NULL);
  if(!check_name(directory,conn)) {
    set_bad_path_error(errno, bad_path);

#if 0
    /* Ugly - NT specific hack - maybe not needed ? (JRA) */
    if((errno == ENOTDIR) && (Protocol >= PROTOCOL_NT1) && 
       (get_remote_arch() == RA_WINNT))
    {
      unix_ERR_class = ERRDOS;
      unix_ERR_code = ERRbaddirectory;
    }
#endif 

    return(UNIXERROR(ERRDOS,ERRbadpath));
  }

  p = strrchr(directory,'/');
  if(p == NULL) {
    pstrcpy(mask,directory);
    pstrcpy(directory,"./");
  } else {
    pstrcpy(mask,p+1);
    *p = 0;
  }

  DEBUG(5,("dir=%s, mask = %s\n",directory, mask));

  if (!g_xrenew(char, ppdata, max_data_bytes + DIR_ENTRY_SAFETY_MARGIN)) {
    return(ERROR_DOS(ERRDOS,ERRnomem));
  }
  pdata = *ppdata;
  bzero(pdata, max_data_bytes + DIR_ENTRY_SAFETY_MARGIN);

  /* Realloc the params space */
  if (!g_xrenew(char, pparams, 10)) {
    return(ERROR_DOS(ERRDOS,ERRnomem));
  }
  params = *pparams;

  dptr_num = dptr_create(conn,directory, False, True ,SVAL(inbuf,smb_pid));
  if (dptr_num < 0)
    return(UNIXERROR(ERRDOS,ERRbadfile));

  /* Save the wildcard match and attribs we are using on this directory - 
     needed as lanman2 assumes these are being saved between calls */

  if(!(wcard = strdup(mask))) {
    dptr_close(&dptr_num);
    return(ERROR_DOS(ERRDOS,ERRnomem));
  }

  dptr_set_wcard(dptr_num, wcard);
  dptr_set_attr(dptr_num, dirtype);

  DEBUG(4,("dptr_num is %d, wcard = %s, attr = %d\n",dptr_num, wcard, dirtype));

  /* We don't need to check for VOL here as this is returned by 
     a different TRANS2 call. */
  
  DEBUG(8,("dirpath=<%s> dontdescend=<%s>\n",
	   conn->dirpath,lp_dontdescend(SNUM(conn))));
  if (in_list(conn->dirpath,lp_dontdescend(SNUM(conn)),conn->case_sensitive))
    dont_descend = True;
    
  p = pdata;
  space_remaining = max_data_bytes;
  out_of_space = False;

  for (i=0;(i<maxentries) && !finished && !out_of_space;i++)
  {
    BOOL got_exact_match = False;

    /* this is a heuristic to avoid seeking the dirptr except when 
       absolutely necessary. It allows for a filename of about 40 chars */
    if (space_remaining < DIRLEN_GUESS && numentries > 0)
    {
      out_of_space = True;
      finished = False;
    }
    else
    {
      finished = !get_lanman2_dir_entry(conn,
					inbuf, outbuf,
					mask,dirtype,info_level,
                   requires_resume_key,dont_descend,
                   &p,pdata,space_remaining, &out_of_space, &got_exact_match,
                   &last_name_off);
    }

    if (finished && out_of_space)
      finished = False;

    if (!finished && !out_of_space)
      numentries++;

    /*
     * As an optimisation if we know we aren't looking
     * for a wildcard name (ie. the name matches the wildcard exactly)
     * then we can finish on any (first) match.
     * This speeds up large directory searches. JRA.
     */

    if(got_exact_match)
      finished = True;

    space_remaining = max_data_bytes - PTR_DIFF(p,pdata);
  }
  
  /* Check if we can close the dirptr */
  if(close_after_first || (finished && close_if_end))
  {
    DEBUG(5,("call_trans2findfirst - (2) closing dptr_num %d\n", dptr_num));
    dptr_close(&dptr_num);
  }

  /* 
   * If there are no matching entries we must return ERRDOS/ERRbadfile - 
   * from observation of NT.
   */

  if(numentries == 0)
  {
    dptr_close(&dptr_num);
    return ERROR_NT(STATUS_NO_SUCH_FILE);
  }

  /* At this point pdata points to numentries directory entries. */

  /* Set up the return parameter block */
  SSVAL(params,0,dptr_num);
  SSVAL(params,2,numentries);
  SSVAL(params,4,finished);
  SSVAL(params,6,0); /* Never an EA error */
  SSVAL(params,8,last_name_off);

  send_trans2_replies(conn, outbuf, bufsize, params, 10, pdata, PTR_DIFF(p,pdata));

  if ((! *directory) && dptr_path(dptr_num))
    slprintf(directory,sizeof(directory)-1, "(%s)",dptr_path(dptr_num));

  DEBUG( 4, ( "%s mask=%s directory=%s dirtype=%d numentries=%d\n",
	    smb_fn_name(CVAL(inbuf,smb_com)), 
	    mask, directory, dirtype, numentries ) );

  /* 
   * Force a name mangle here to ensure that the
   * mask as an 8.3 name is top of the mangled cache.
   * The reasons for this are subtle. Don't remove
   * this code unless you know what you are doing
   * (see PR#13758). JRA.
   */

  if(!is_8_3( mask, False))
    name_map_mangle(conn, mask, True, True);

  return(-1);
}


/****************************************************************************
 Reply to a TRANS2_FINDNEXT.
****************************************************************************/

static int call_trans2findnext(connection_struct *conn, char *inbuf, char *outbuf, int length, int bufsize,
					char **pparams, int total_params, char **ppdata, int total_data)
{
  /* We must be careful here that we don't return more than the
     allowed number of data bytes. If this means returning fewer than
     maxentries then so be it. We assume that the redirector has
     enough room for the fixed number of parameter bytes it has
     requested. */
  int max_data_bytes = SVAL(inbuf, smb_mdrcnt);
  char *params = *pparams;
  char *pdata = *ppdata;
  int dptr_num = SVAL(params,0);
  int maxentries = SVAL(params,2);
  uint16 info_level = SVAL(params,4);
  uint32 resume_key = IVAL(params,6);
  BOOL close_after_request = BITSETW(params+10,0);
  BOOL close_if_end = BITSETW(params+10,1);
  BOOL requires_resume_key = BITSETW(params+10,2);
  BOOL continue_bit = BITSETW(params+10,3);
  pstring resume_name;
  pstring mask;
  pstring directory;
  char *p;
  uint16 dirtype;
  int numentries = 0;
  int i, last_name_off=0;
  BOOL finished = False;
  BOOL dont_descend = False;
  BOOL out_of_space = False;
  int space_remaining;

	if (total_params < 12)
		return(ERROR_DOS(ERRDOS,ERRinvalidparam));

  *mask = *directory = *resume_name = 0;

  srvstr_pull(inbuf, resume_name, params+12, sizeof(resume_name), total_params - 12, STR_TERMINATE|STR_NOALIGN);

  DEBUG(3,("call_trans2findnext: dirhandle = %d, max_data_bytes = %d, maxentries = %d, "
	   "close_after_request=%d, close_if_end = %d requires_resume_key = %d "
	   "resume_key = %d resume name = %s continue=%d level = %d\n",
	   dptr_num, max_data_bytes, maxentries, close_after_request, close_if_end, 
	   requires_resume_key, resume_key, resume_name, continue_bit, info_level));

  switch (info_level) 
    {
    case 1:
    case 2:
    case 3:
    case 4:
    case SMB_FIND_FILE_DIRECTORY_INFO:
    case SMB_FIND_FILE_FULL_DIRECTORY_INFO:
    case SMB_FIND_FILE_NAMES_INFO:
    case SMB_FIND_FILE_BOTH_DIRECTORY_INFO:
      break;
    default:
      return(ERROR_DOS(ERRDOS,ERRunknownlevel));
    }

  if (!g_xrenew(char, ppdata, max_data_bytes + DIR_ENTRY_SAFETY_MARGIN)) {
    return(ERROR_DOS(ERRDOS,ERRnomem));
  }
  pdata = *ppdata;
  bzero(pdata, max_data_bytes + DIR_ENTRY_SAFETY_MARGIN);

  /* Realloc the params space */
  if (!g_xrenew(char, pparams, 6*SIZEOFWORD)) {
    return(ERROR_DOS(ERRDOS,ERRnomem));
  }
  params = *pparams;

  /* Check that the dptr is valid */
  if(!(conn->dirptr = dptr_fetch_lanman2(dptr_num)))
    return(ERROR_DOS(ERRDOS,ERRnofiles));

  string_set(&conn->dirpath,dptr_path(dptr_num));

  /* Get the wildcard mask from the dptr */
  if((p = dptr_wcard(dptr_num))== NULL) {
    DEBUG(2,("dptr_num %d has no wildcard\n", dptr_num));
    return (ERROR_DOS(ERRDOS,ERRnofiles));
  }
  pstrcpy(mask, p);
  pstrcpy(directory,conn->dirpath);

  /* Get the attr mask from the dptr */
  dirtype = dptr_attr(dptr_num);

  DEBUG(3,("dptr_num is %d, mask = %s, attr = %x, dirptr=(0x%lX,%d)\n",
	   dptr_num, mask, dirtype, 
	   (long)conn->dirptr,
	   TellDir(conn->dirptr)));

  /* We don't need to check for VOL here as this is returned by 
     a different TRANS2 call. */

  DEBUG(8,("dirpath=<%s> dontdescend=<%s>\n",conn->dirpath,lp_dontdescend(SNUM(conn))));
  if (in_list(conn->dirpath,lp_dontdescend(SNUM(conn)),conn->case_sensitive))
    dont_descend = True;
    
  p = pdata;
  space_remaining = max_data_bytes;
  out_of_space = False;

  /* 
   * Seek to the correct position. We no longer use the resume key but
   * depend on the last file name instead.
   */
  if(*resume_name && !continue_bit)
  {
    /*
     * Fix for NT redirector problem triggered by resume key indexes
     * changing between directory scans. We now return a resume key of 0
     * and instead look for the filename to continue from (also given
     * to us by NT/95/smbfs/smbclient). If no other scans have been done between the
     * findfirst/findnext (as is usual) then the directory pointer
     * should already be at the correct place. Check this by scanning
     * backwards looking for an exact (ie. case sensitive) filename match. 
     * If we get to the beginning of the directory and haven't found it then scan
     * forwards again looking for a match. JRA.
     */

    int current_pos, start_pos;
    const char *dname = NULL;
    void *dirptr = conn->dirptr;
    start_pos = TellDir(dirptr);
    for(current_pos = start_pos; current_pos >= 0; current_pos--)
    {
      DEBUG(7,("call_trans2findnext: seeking to pos %d\n", current_pos));

      SeekDir(dirptr, current_pos);
      dname = ReadDirName(dirptr);

      /*
       * Remember, name_map_mangle is called by
       * get_lanman2_dir_entry(), so the resume name
       * could be mangled. Ensure we do the same
       * here.
       */

      if(dname != NULL)
        dname = mangle_from_dos_to_unix(conn, dname, False, True);

      if(dname && strcsequal( resume_name, dname))
      {
        SeekDir(dirptr, current_pos+1);
        DEBUG(7,("call_trans2findnext: got match at pos %d\n", current_pos+1 ));
        break;
      }
    }

    /*
     * Scan forward from start if not found going backwards.
     */

    if(current_pos < 0)
    {
      DEBUG(7,("call_trans2findnext: notfound: seeking to pos %d\n", start_pos));
      SeekDir(dirptr, start_pos);
      for(current_pos = start_pos; (dname = ReadDirName(dirptr)) != NULL; SeekDir(dirptr,++current_pos))
      {
        /*
         * Remember, name_map_mangle is called by
         * get_lanman2_dir_entry(), so the resume name
         * could be mangled. Ensure we do the same
         * here.
         */

        if(dname != NULL)
          dname = mangle_from_dos_to_unix(conn, dname, False, True);

        if(dname && strcsequal( resume_name, dname))
        {
          SeekDir(dirptr, current_pos+1);
          DEBUG(7,("call_trans2findnext: got match at pos %d\n", current_pos+1 ));
          break;
        }
      } /* end for */
    } /* end if current_pos */
    /* Can't find the name. Just resume from where we were... */
    if (!dname)
    {
      DEBUG(3, ("call_trans2findnext: resume_name[%s] not found... weird\n",
		resume_name));
      SeekDir(dirptr, start_pos);
    }
  } /* end if requires_resume_key && !continue_bit */

  for (i=0;(i<(int)maxentries) && !finished && !out_of_space ;i++)
  {
    BOOL got_exact_match = False;

    /* this is a heuristic to avoid seeking the dirptr except when 
       absolutely necessary. It allows for a filename of about 40 chars */
    if (space_remaining < DIRLEN_GUESS && numentries > 0)
    {
      out_of_space = True;
      finished = False;
    }
    else
    {
      finished = !get_lanman2_dir_entry(conn,
					inbuf, outbuf,
					mask,dirtype,info_level,
                   requires_resume_key,dont_descend,
                   &p,pdata,space_remaining, &out_of_space, &got_exact_match,
                   &last_name_off);
    }

    if (finished && out_of_space)
      finished = False;

    if (!finished && !out_of_space)
      numentries++;

    /*
     * As an optimisation if we know we aren't looking
     * for a wildcard name (ie. the name matches the wildcard exactly)
     * then we can finish on any (first) match.
     * This speeds up large directory searches. JRA.
     */

    if(got_exact_match)
      finished = True;

    space_remaining = max_data_bytes - PTR_DIFF(p,pdata);
  }
  
  /* Check if we can close the dirptr */
  if(close_after_request || (finished && close_if_end))
  {
    DEBUG(5,("call_trans2findnext: closing dptr_num = %d\n", dptr_num));
    dptr_close(&dptr_num); /* This frees up the saved mask */
  }


  /* Set up the return parameter block */
  SSVAL(params,0,numentries);
  SSVAL(params,2,finished);
  SSVAL(params,4,0); /* Never an EA error */
  SSVAL(params,6,last_name_off);

  send_trans2_replies(conn, outbuf, bufsize, params, 8, pdata, PTR_DIFF(p,pdata));

  if ((! *directory) && dptr_path(dptr_num))
    slprintf(directory,sizeof(directory)-1, "(%s)",dptr_path(dptr_num));

  DEBUG( 3, ( "%s mask=%s directory=%s dirtype=%d numentries=%d\n",
	    smb_fn_name(CVAL(inbuf,smb_com)), 
	    mask, directory, dirtype, numentries ) );

  return(-1);
}

/****************************************************************************
 Reply to a TRANS2_QFSINFO (query filesystem info).
****************************************************************************/

static int call_trans2qfsinfo(connection_struct *conn, char *inbuf, char *outbuf, 
			      int length, int bufsize,
			char **pparams, int total_params, char **ppdata, int total_data)
{
  int max_data_bytes = SVAL(inbuf, smb_mdrcnt);
  char *pdata = *ppdata;
  char *params = *pparams;
  uint16 info_level = SVAL(params,0);
  int data_len, len;
  SMB_STRUCT_STAT st;
  char *vname = volume_label(SNUM(conn));
  int snum = SNUM(conn);
  char *fstype = lp_fstype(SNUM(conn));

  DEBUG(3,("call_trans2qfsinfo: level = %d\n", info_level));

  if(vfs_stat(conn, dos_to_unix_static("."), &st) != 0) {
    DEBUG(2,("call_trans2qfsinfo: stat of . failed (%s)\n", strerror(errno)));
    return (ERROR_DOS(ERRSRV,ERRinvdevice));
  }

  if (!g_xrenew(char, ppdata, max_data_bytes + DIR_ENTRY_SAFETY_MARGIN)) {
    return(ERROR_DOS(ERRDOS,ERRnomem));
  }
  pdata = *ppdata;
  bzero(pdata, max_data_bytes + DIR_ENTRY_SAFETY_MARGIN);

  switch (info_level) 
  {
    case SMB_INFO_ALLOCATION:
    {
      SMB_BIG_UINT dfree,dsize,bsize;
      data_len = 18;
      conn->vfs_ops->disk_free(".",False,&bsize,&dfree,&dsize);	
      SIVAL(pdata,l1_idFileSystem,st.st_dev);
      SIVAL(pdata,l1_cSectorUnit,bsize/512);
      SIVAL(pdata,l1_cUnit,dsize);
      SIVAL(pdata,l1_cUnitAvail,dfree);
      SSVAL(pdata,l1_cbSector,512);
      DEBUG(5,("call_trans2qfsinfo : bsize=%u, id=%x, cSectorUnit=%u, cUnit=%u, cUnitAvail=%u, cbSector=%d\n",
		 (unsigned int)bsize, (unsigned int)st.st_dev, ((unsigned int)bsize)/512, (unsigned int)dsize,
         (unsigned int)dfree, 512));
      break;
    }
    case SMB_INFO_VOLUME:
    { 
      /* Return volume name */
      SIVAL(pdata,0, lp_volume_serial(snum));
      len = srvstr_push(outbuf, pdata+l2_vol_szVolLabel, vname, 22, STR_NOALIGN);
      SCVAL(pdata, l2_vol_cch, len);
      data_len = l2_vol_szVolLabel + len;
      break;
    }

    case SMB_QUERY_FS_ATTRIBUTE_INFO:
    case SMB_QUERY_FS_ATTRIBUTE_INFORMATION:
    {
      SIVAL(pdata,0,FILE_CASE_PRESERVED_NAMES|FILE_CASE_SENSITIVE_SEARCH|
            (lp_nt_acl_support() ? FILE_PERSISTENT_ACLS : 0)); /* FS ATTRIBUTES */

      SIVAL(pdata,4,128); /* Max filename component length */
      len = unix_PutUniCode(pdata+12, fstype, sizeof(pstring), False);
      SIVAL(pdata,8,len);
      data_len = 12 + len;
      SSVAL(outbuf,smb_flg2,SVAL(outbuf,smb_flg2)|FLAGS2_UNICODE_STRINGS);
      break;
    }

    case SMB_QUERY_FS_LABEL_INFO:
    case SMB_QUERY_FS_LABEL_INFORMATION:
      len = srvstr_push(outbuf, pdata+4, vname, 22, STR_NOALIGN);
      SIVAL(pdata, 0, len);
      data_len = 4 + len;
      break;

    case SMB_QUERY_FS_VOLUME_INFO:      
    case SMB_QUERY_FS_VOLUME_INFORMATION: 
      SIVAL(pdata,8, lp_volume_serial(snum));

      len = srvstr_push(outbuf, pdata+18, vname,
			max_data_bytes - 18, STR_UNICODE|STR_NOALIGN);
      data_len = 18 + len;
      SIVAL(pdata,12, len);

      DEBUG(5,("call_trans2qfsinfo : SMB_QUERY_FS_VOLUME_INFO namelen = %d, vol = %s\n", 
	       (int)strlen(vname),vname));
      break;

    case SMB_QUERY_FS_SIZE_INFO:
    case SMB_QUERY_FS_SIZE_INFORMATION:
    {
      SMB_BIG_UINT dfree,dsize,bsize;
      data_len = 24;
      conn->vfs_ops->disk_free(".",False,&bsize,&dfree,&dsize);	
      SBIG_UINT(pdata,0,dsize);
      SBIG_UINT(pdata,8,dfree);
      SIVAL(pdata,16,bsize/512);
      SIVAL(pdata,20,512);
      break;
    }

    case SMB_QUERY_FS_DEVICE_INFO:
    case SMB_QUERY_FS_DEVICE_INFORAMTION:
      data_len = 8;
      SIVAL(pdata,0,0); /* dev type */
      SIVAL(pdata,4,0); /* characteristics */
      break;

    case SMB_QUERY_FS_UNIX_INFO:
      data_len = 12;
      SSVAL(pdata,0,CIFS_UNIX_MAJOR_VERSION); /* CIFS major server version support */
      SSVAL(pdata,2,CIFS_UNIX_MINOR_VERSION); /* CIFS minor server version support */
      SIVAL(pdata,4,0); /* capabilities */
      SIVAL(pdata,8,0);
      break;

    case SMB_MAC_QUERY_FS_INFO:
	    /*
	     * Thursby MAC extension... ONLY on NTFS filesystems
	     * once we do streams then we don't need this
	     */
	    if (strequal(lp_fstype(SNUM(conn)),"NTFS")) {
		    data_len = 88;
		    SIVAL(pdata,84,0x100); /* Don't support mac... */
		    break;
	    }
	    /* drop through */
    default:
      DEBUG(2, ("call_trans2qfsinfo: Info level %u (0x%x) not known\n",
		info_level, info_level));
      return(ERROR_DOS(ERRDOS,ERRunknownlevel));
  }


  send_trans2_replies(conn, outbuf, bufsize, params, 0, pdata, data_len);

  DEBUG( 4, ( "%s info_level = %d\n",
            smb_fn_name(CVAL(inbuf,smb_com)), info_level) );

  return -1;
}

/****************************************************************************
  reply to a TRANS2_SETFSINFO (set filesystem info)
****************************************************************************/
static int call_trans2setfsinfo(connection_struct *conn,
				char *inbuf, char *outbuf,
				int length, int bufsize,
				char **pparams, int total_params, char **ppdata, int total_data)
{
  /* Just say yes we did it - there is nothing that
     can be set here so it doesn't matter. */
  int outsize;
  DEBUG(3,("call_trans2setfsinfo\n"));

  if (!CAN_WRITE(conn))
    return(ERROR_DOS(ERRSRV,ERRaccess));

  outsize = set_message(outbuf,10,0,True);

  return outsize;
}

/****************************************************************************
  Reply to a TRANS2_QFILEPATHINFO or TRANSACT2_QFILEINFO (query file info by
  file name or file id).
****************************************************************************/

static int call_trans2qfilepathinfo(connection_struct *conn,
				    char *inbuf, char *outbuf, int length, 
				    int bufsize,
					char **pparams, int total_params, char **ppdata, int total_data)
{
  int max_data_bytes = SVAL(inbuf, smb_mdrcnt);
  char *params = *pparams;
  char *pdata = *ppdata;
  uint16 tran_call = SVAL(inbuf, smb_setup0);
  uint16 info_level;
  int mode=0;
  SMB_OFF_T size=0;
  unsigned int data_size;
  size_t param_size = 2;
  SMB_STRUCT_STAT sbuf;
  NTFileStatInfo nti;
  pstring fname1;
  char *fname;
  char *p;
  pstring long_fname_unix;
  SMB_OFF_T pos = 0;
  BOOL bad_path = False;
  BOOL delete_pending = False;
  int len;
  time_t c_time;

	if (!params)
		return ERROR_NT(NT_STATUS_INVALID_PARAMETER);

  if (tran_call == TRANSACT2_QFILEINFO) {
    files_struct *fsp;

		if (total_params < 4)
			return(ERROR_DOS(ERRDOS,ERRinvalidparam));

    fsp = file_fsp(params,0);
    info_level = SVAL(params,2);

	  DEBUG(3,("call_trans2qfilepathinfo: TRANSACT2_QFILEINFO: level = %d\n", 
		   info_level));

    if(fsp && (fsp->is_directory || fsp->stat_open)) {
      /*
       * This is actually a QFILEINFO on a directory
       * handle (returned from an NT SMB). NT5.0 seems
       * to do this call. JRA.
       */
      fname = fsp->dos_fsp_name;
      unix_convert(fname,conn,0,&bad_path,&sbuf);
      if (!check_name(fname,conn) || 
          (!VALID_STAT(sbuf) && vfs_stat(conn, dos_to_unix_static(fname), &sbuf))) {
        DEBUG(3,("fileinfo of %s failed (%s)\n",fname,strerror(errno)));
        set_bad_path_error(errno, bad_path);
        return(UNIXERROR(ERRDOS,ERRbadpath));
      }

      delete_pending = fsp->directory_delete_on_close;
    } else {
      /*
       * Original code - this is an open file.
       */
      CHECK_FSP(fsp,conn);
      CHECK_ERROR(fsp);

      fname = fsp->dos_fsp_name;
      if (fsp_stat(fsp, &sbuf) != 0) {
        DEBUG(3,("fstat of fnum %d failed (%s)\n",
		 fsp->fnum, strerror(errno)));
        return(UNIXERROR(ERRDOS,ERRbadfid));
      }
      if((pos = fsp->conn->vfs_ops->lseek(fsp->fd,0,SEEK_CUR)) == -1)
        return(UNIXERROR(ERRDOS,ERRnoaccess));

      delete_pending = fsp->delete_on_close;
    }
  } else {
    /* qpathinfo */
		if (total_params < 6)
			return(ERROR_DOS(ERRDOS,ERRinvalidparam));

    info_level = SVAL(params,0);

    DEBUG(3,("call_trans2qfilepathinfo: TRANSACT2_QPATHINFO: level = %d\n",
		info_level));

    srvstr_pull(inbuf, fname1, params+6, sizeof(fname1), total_params - 6, STR_TERMINATE|STR_NOALIGN);
    unix_to_dos(fname1);
    fname = &fname1[0];

    RESOLVE_DFSPATH(fname, conn, inbuf, outbuf);

    unix_convert(fname,conn,0,&bad_path,&sbuf);
    if (!check_name(fname,conn) || 
        (!VALID_STAT(sbuf) && vfs_stat(conn, dos_to_unix_static(fname), &sbuf))) {
      DEBUG(3,("fileinfo of %s failed (%s)\n",fname,strerror(errno)));
      set_bad_path_error(errno, bad_path);
      return(UNIXERROR(ERRDOS,ERRbadpath));
    }
  }


  DEBUG(3,("call_trans2qfilepathinfo %s level=%d call=%d total_data=%d\n",
	   fname,info_level,tran_call,total_data));

	pstrcpy(long_fname_unix, "\\");
	if (!strcsequal(fname, "."))
	{
		pstrcat(long_fname_unix, dos_to_unix_static(fname));
		string_replace(long_fname_unix, '/', '\\');
	}

  p = strrchr(fname,'/'); 
  if (!p) 
    p = fname;
  else
    p++;
  len = strlen(p);  

  vfs_fill_ntinfo(conn, &nti, fname, &sbuf);
  mode = nti.FileAttributes & SAMBA_ATTRIBUTES_MASK;

  size = sbuf.st_size;
  if (mode & aDIR) size = 0;

  /* from now on we only want the part after the / */
  fname = p;

  if (!g_xrenew(char, pparams, 2)) {
	  return(ERROR_DOS(ERRDOS,ERRnomem));
  }
  params = *pparams;
  bzero(params, 2);
  data_size = max_data_bytes + DIR_ENTRY_SAFETY_MARGIN;
  if (!g_xrenew(char, ppdata, data_size)) {
    return(ERROR_DOS(ERRDOS,ERRnomem));
  }
  pdata = *ppdata;

  if (total_data >= 4 && IVAL(pdata,0) == total_data) {
    /* uggh, EAs for OS2 */
    DEBUG(1, ("Warning: Possible EA stuff for OS/2, continueing\n"));
  }

  bzero(pdata, data_size);

  c_time = get_create_time(&sbuf,lp_fake_dir_create_times(SNUM(conn)));

  if (lp_dos_filetime_resolution(SNUM(conn))) {
    c_time &= ~1;
    sbuf.st_atime &= ~1;
    sbuf.st_mtime &= ~1;
  }

  if (delete_pending && sbuf.st_nlink)
    sbuf.st_nlink--;

  switch (info_level) 
    {
    case SMB_INFO_STANDARD:
    case SMB_INFO_QUERY_EA_SIZE:
      data_size = (info_level==1?22:26);
      put_dos_date2(pdata,l1_fdateCreation,c_time);
      put_dos_date2(pdata,l1_fdateLastAccess,sbuf.st_atime);
      put_dos_date2(pdata,l1_fdateLastWrite,sbuf.st_mtime); /* write time */
      SIVAL(pdata,l1_cbFile,(uint32)size);
      SIVAL(pdata,l1_cbFileAlloc,SMB_ROUNDUP(size,1024));
      SSVAL(pdata,l1_attrFile,mode);
      SIVAL(pdata,l1_attrFile+2,4); /* this is what OS2 does */
      break;

    case SMB_INFO_QUERY_EAS_FROM_LIST:
      data_size = 24;
      put_dos_date2(pdata,0,c_time);
      put_dos_date2(pdata,4,sbuf.st_atime);
      put_dos_date2(pdata,8,sbuf.st_mtime);
      SIVAL(pdata,12,(uint32)size);
      SIVAL(pdata,16,SMB_ROUNDUP(size,1024));
      SIVAL(pdata,20,mode);
      break;

    case SMB_INFO_QUERY_ALL_EAS:
      data_size = 4;
      SIVAL(pdata,0,data_size);
      break;

    case SMB_INFO_IS_NAME_VALID:
      if (tran_call == TRANSACT2_QFILEINFO)
      {
	/* os/2 needs this */
	return(ERROR_DOS(ERRDOS,ERRbadfunc)); 
      }
      data_size = 0;
      param_size = 0;
      break;

    case SMB_QUERY_FILE_BASIC_INFO:
    case SMB_FILE_BASIC_INFORMATION:
      data_size = 36; /* w95 returns 40 bytes not 36 - why ?. */
      put_long_date(pdata,c_time);
      put_long_date(pdata+8,sbuf.st_atime);
      put_long_date(pdata+16,sbuf.st_mtime); /* write time */
      put_long_date(pdata+24,sbuf.st_mtime); /* change time */
      SIVAL(pdata,32,mode);

      DEBUG(5,("SMB_QFBI - "));
      {
        time_t create_time = c_time;
        DEBUG(5,("create: %s ", ctime(&create_time)));
      }
      DEBUG(5,("access: %s ", ctime(&sbuf.st_atime)));
      DEBUG(5,("write: %s ", ctime(&sbuf.st_mtime)));
      DEBUG(5,("change: %s ", ctime(&sbuf.st_mtime)));
      DEBUG(5,("mode: %x\n", mode));

      break;

    case SMB_QUERY_FILE_STANDARD_INFO:
    case SMB_FILE_STANDARD_INFORMATION:
      data_size = 24;
      SOFF_T(pdata,0, SMB_ROUNDUP(size, 1024));
      SOFF_T(pdata,8,size);
      SIVAL(pdata,16,sbuf.st_nlink);
      CVAL(pdata,20) = 0;
      CVAL(pdata,21) = (mode&aDIR)?1:0;
      break;

    case SMB_QUERY_FILE_EA_INFO:
      data_size = 4;
      break;

    /* Get the 8.3 name - used if NT SMB was negotiated. */
    case SMB_QUERY_FILE_ALT_NAME_INFO:
    case SMB_FILE_ALTERNATE_NAME_INFORMATION:
      {
	pstring short_name;

        pstrcpy(short_name,p);
        /* Mangle if not already 8.3 */
        if(!is_8_3(short_name, True))
        {
          if(!name_map_mangle(conn, short_name, True, True))
            *short_name = '\0';
        }
        strupper(short_name);
        len = dos_PutUniCode(pdata + 4, short_name, sizeof(pstring), False);
        data_size = 4 + len;
        SIVAL(pdata,0, len);
      }
      break;

    case SMB_QUERY_FILE_NAME_INFO:
    case SMB_FILE_NAME_INFORMATION:
      /*
       * The first part of this code is essential
       * to get security descriptors to work on mapped
       * drives. Don't ask how I discovered this unless
       * you like hearing about me suffering.... :-). JRA.
       */
      len = srvstr_push(outbuf, pdata+4, long_fname_unix, data_size-4, STR_NOALIGN);
      data_size = 4 + len;
      SIVAL(pdata,0,len);
      break;

    case SMB_QUERY_FILE_ALLOCATION_INFO:
    case SMB_QUERY_FILE_END_OF_FILEINFO:
      data_size = 8;
      SOFF_T(pdata,0,size);
      break;

    case SMB_QUERY_FILE_ALL_INFO:
    case SMB_FILE_ALL_INFORMATION:
      put_long_date(pdata,c_time);
      put_long_date(pdata+8,sbuf.st_atime);
      put_long_date(pdata+16,sbuf.st_mtime); /* write time */
      put_long_date(pdata+24,sbuf.st_mtime); /* change time */
      SIVAL(pdata,32, nti.FileAttributes);
      pdata += 40;
      SOFF_T(pdata,0, nti.AllocationSize);
      SOFF_T(pdata,8,size);
      SIVAL(pdata,16,sbuf.st_nlink);
      CVAL(pdata,20) = delete_pending;
      CVAL(pdata,21) = (mode&aDIR)?1:0;
      pdata += 24;
      pdata += 4; /* EA info */
      len = srvstr_push(outbuf, pdata+4, long_fname_unix,
			data_size - (4+PTR_DIFF(pdata, *ppdata)), STR_NOALIGN);
      SIVAL(pdata, 0, len);
      pdata += 4 + len;
      data_size = PTR_DIFF(pdata,(*ppdata));
      break;

	/*
	 * Windows 2000 completely undocumented new SMB info levels.
	 * Thanks Microsoft.... sure you're working on making this
	 * protocol a standard.... sure you are... :-).
	 * Lying rat-bastards. JRA.
	 */

	case 1006:
		SIVAL(pdata,0,0x907); /* ??? */
		SIVAL(pdata,4,0x690000); /* ??? */
		data_size = 8;
		break;

	case 1007:
		SIVAL(pdata,0,0); /* ??? */
		data_size = 4;
		break;

	case 1008:
		SIVAL(pdata,0,0x12019F); /* ??? */
		data_size = 4;
		break;


	case 1014:
		SIVAL(pdata,0,0); /* ??? */
		SIVAL(pdata,4,0); /* ??? */
		data_size = 8;
		break;

	case 1016:
		SIVAL(pdata,0,0); /* ??? */
		data_size = 4;
		break;

	case 1017:
		SIVAL(pdata,0,0); /* ??? */
		data_size = 4;
		break;

		/*
		 * tridge says this, but I don't believe him:
		 *    NT4 server just returns "invalid query" to this - if we try to answer
		 *    it then NTws gets a BSOD! (tridge)
		 */
		/* FileStreamInformation/FILE_STREAM_INFORMATION */
		case SMB_QUERY_FILE_STREAM_INFO:
		case SMB_FILE_STREAM_INFORMATION:
			if (mode & aDIR)
				data_size = 0;
			else
			{
				len = srvstr_push(outbuf, pdata+24, "::$DATA", data_size-24, STR_NOALIGN);
				SIVAL(pdata, 0, 0); /* NextEntryOffset */
				SIVAL(pdata, 4, len); /* StreamNameLength */
				SOFF_T(pdata, 8, size); /* StreamSize */
				SOFF_T(pdata, 16, size); /* StreamAllocationSize */
				data_size = 24 + len;
			}
			break;

	case SMB_QUERY_COMPRESSION_INFO:
	case SMB_FILE_COMPRESSION_INFORMATION:
		SOFF_T(pdata,0,size);
		SIVAL(pdata,8,0); /* ??? */
		SIVAL(pdata,12,0); /* ??? */
		data_size = 16;
		break;

	case SMB_FILE_NETWORK_OPEN_INFORMATION:
		put_long_date(pdata,c_time);
		put_long_date(pdata+8,sbuf.st_atime);
		put_long_date(pdata+16,sbuf.st_mtime); /* write time */
		put_long_date(pdata+24,sbuf.st_mtime); /* change time */
		SOFF_T(pdata,32, SMB_ROUNDUP(size, 1024));
		SOFF_T(pdata,40,size);
		SIVAL(pdata,48,mode);
		SIVAL(pdata,52,0); /* ??? */
		data_size = 56;
		break;

	case 1035:
		SIVAL(pdata,0,mode);
		SIVAL(pdata,4,0);
		data_size = 8;
		break;

	/*
	 * End new completely undocumented info levels... JRA.
	 */

    default:
      return(ERROR_DOS(ERRDOS,ERRunknownlevel));
    }

	send_trans2_replies(conn, outbuf, bufsize,
			    params, param_size, *ppdata, data_size);

	return -1;
}


/****************************************************************************
 Reply to a TRANS2_SETFILEINFO (set file info by fileid).
****************************************************************************/

static int call_trans2setfilepathinfo(connection_struct *conn,
					char *inbuf, char *outbuf, int length, int bufsize,
					char **pparams, int total_params, char **ppdata, int total_data)
{
  char *params = *pparams;
  char *pdata = *ppdata;
  uint16 tran_call = SVAL(inbuf, smb_setup0);
  uint16 info_level;
  int mode=0;
  SMB_OFF_T size=0;
  struct utimbuf tvs;
  SMB_STRUCT_STAT sbuf;
  pstring fname1;
  char *fname;
  int fd = -1;
  BOOL bad_path = False;
  files_struct *fsp = NULL;

  if (!CAN_WRITE(conn))
    return(ERROR_DOS(ERRSRV,ERRaccess));

  if (tran_call == TRANSACT2_SETFILEINFO) {
    fsp = file_fsp(params,0);
    info_level = SVAL(params,2);    

    if(fsp && (fsp->is_directory || fsp->stat_open)) {
      /*
       * This is actually a SETFILEINFO on a directory
       * handle (returned from an NT SMB). NT5.0 seems
       * to do this call. JRA.
       */
      fname = fsp->dos_fsp_name;
      unix_convert(fname,conn,0,&bad_path,&sbuf);
      if (!check_name(fname,conn) || 
          (!VALID_STAT(sbuf) && vfs_stat(conn, dos_to_unix_static(fname), &sbuf))) {
        DEBUG(3,("fileinfo of %s failed (%s)\n",fname,strerror(errno)));
        set_bad_path_error(errno, bad_path);
        return(UNIXERROR(ERRDOS,ERRbadpath));
      }
    } else {
      /*
       * Original code - this is an open file.
       */
      CHECK_FSP(fsp,conn);
      CHECK_ERROR(fsp);

      fname = fsp->dos_fsp_name;
      fd = fsp->fd;

      if (fsp_stat(fsp, &sbuf) != 0) {
        DEBUG(3,("fstat of fnum %d failed (%s)\n",fsp->fnum, strerror(errno)));
        return(UNIXERROR(ERRDOS,ERRbadfid));
      }
    }
  } else {
    /* set path info */
		if (total_params < 6)
			return(ERROR_DOS(ERRDOS,ERRinvalidparam));

    info_level = SVAL(params,0);    

    srvstr_pull(inbuf, fname1, params+6, sizeof(fname1), total_params - 6, STR_TERMINATE|STR_NOALIGN);
    unix_to_dos(fname1);
    fname = fname1;

    unix_convert(fname1, conn,0,&bad_path,&sbuf);
    if(!check_name(fname, conn))
    {
      set_bad_path_error(errno, bad_path);
      return(UNIXERROR(ERRDOS,ERRbadpath));
    }
 
    if(!VALID_STAT(sbuf) && vfs_stat(conn, dos_to_unix_static(fname), &sbuf)!=0) {
      DEBUG(3,("stat of %s failed (%s)\n", fname, strerror(errno)));
      set_bad_path_error(errno, bad_path);
      return(UNIXERROR(ERRDOS,ERRbadpath));
    }    
  }

  if (!CAN_WRITE(conn))
    return(ERROR_DOS(ERRSRV,ERRaccess));

  DEBUG(3,("call_trans2setfilepathinfo(%d) %s info_level=%d totdata=%d\n",
	   tran_call,fname,info_level,total_data));

  /* Realloc the parameter and data sizes */
  if (!g_xrenew(char, pparams, 2)) {
    return(ERROR_DOS(ERRDOS,ERRnomem));
  }
  params = *pparams;

  SSVAL(params,0,0);

  size = sbuf.st_size;
  tvs.modtime = sbuf.st_mtime;
  tvs.actime = sbuf.st_atime;
  mode = dos_mode(conn,fname,&sbuf);

  if (total_data >= 4 && IVAL(pdata,0) == total_data) {
    /* uggh, EAs for OS2 */
    DEBUG(2, ("Warning: Possible EA stuff for OS/2, continueing\n"));
  }

  switch (info_level)
  {
    case SMB_INFO_STANDARD:
    {
      /* access time */
      tvs.actime = make_unix_date2(pdata+l1_fdateLastAccess);

      /* write time */
      tvs.modtime = make_unix_date2(pdata+l1_fdateLastWrite);

      mode = SVAL(pdata,l1_attrFile);
      size = IVAL(pdata,l1_cbFile);
      break;
    }

    case SMB_INFO_SET_EA:
    {
      DEBUG(2, ("Setting extended attributes not supported\n"));
      return ERROR_NT(STATUS_EAS_NOT_SUPPORTED);
      break;
    }

    /* XXXX um, i don't think this is right.
       it's also not in the cifs6.txt spec.
     */
    case SMB_INFO_QUERY_EAS_FROM_LIST:
			if (total_data < 28)
				return(ERROR_DOS(ERRDOS,ERRinvalidparam));

      tvs.actime = make_unix_date2(pdata+8);
      tvs.modtime = make_unix_date2(pdata+12);
      size = IVAL(pdata,16);
      mode = IVAL(pdata,24);
      break;

    /* XXXX nor this.  not in cifs6.txt, either. */
    case SMB_INFO_QUERY_ALL_EAS:
			if (total_data < 28)
				return(ERROR_DOS(ERRDOS,ERRinvalidparam));

      tvs.actime = make_unix_date2(pdata+8);
      tvs.modtime = make_unix_date2(pdata+12);
      size = IVAL(pdata,16);
      mode = IVAL(pdata,24);
      break;

    case SMB_SET_FILE_BASIC_INFO:
    case SMB_FILE_BASIC_INFORMATION:
    {
      /* Patch to do this correctly from Paul Eggert <eggert@twinsun.com>. */
      time_t write_time;
      time_t changed_time;

			if (total_data < 36)
				return(ERROR_DOS(ERRDOS,ERRinvalidparam));

      /* Ignore create time at offset pdata. */

      /* access time */
      tvs.actime = interpret_long_date(pdata+8);

      write_time = interpret_long_date(pdata+16);
      changed_time = interpret_long_date(pdata+24);

      tvs.modtime = MIN(write_time, changed_time);

      /* Prefer a defined time to an undefined one. */
      if (tvs.modtime == (time_t)0 || tvs.modtime == (time_t)-1)
       tvs.modtime = (write_time == (time_t)0 || write_time == (time_t)-1
                      ? changed_time
                      : write_time);

#if 0 /* Needs more testing... */
      /* Test from Luke to prevent Win95 from
         setting incorrect values here.
       */
      if (tvs.actime < tvs.modtime)
        return(ERROR_DOS(ERRDOS,ERRnoaccess));
#endif /* Needs more testing... */

      /* attributes */
      mode = IVAL(pdata,32);
      break;
    }

    /*
     * NT seems to use this call with a size of zero
     * to mean truncate the file. JRA.
     */

    case SMB_SET_FILE_ALLOCATION_INFO:
    case SMB_FILE_ALLOCATION_INFORMATION:
    {
      SMB_OFF_T newsize = IVAL(pdata,0);

      if (total_data < 8)
	return ERROR_NT(STATUS_INVALID_PARAMETER);

#ifdef LARGE_SMB_OFF_T
      newsize |= (((SMB_OFF_T)IVAL(pdata,4)) << 32);
#else /* LARGE_SMB_OFF_T */
      if (IVAL(pdata,4) != 0)	/* more than 32 bits? */
         return(ERROR_DOS(ERRDOS,ERRunknownlevel));
#endif /* LARGE_SMB_OFF_T */
      DEBUG(10,("call_trans2setfilepathinfo: Set file allocation info for file %s to %.0f\n",
                fname, (double)newsize));
      if(newsize == 0)
        size = 0;
      break;
    }

    case SMB_SET_FILE_END_OF_FILE_INFO:
    case SMB_FILE_END_OF_FILE_INFORMATION:
    {
      size = IVAL(pdata,0);
#ifdef LARGE_SMB_OFF_T
      size |= (((SMB_OFF_T)IVAL(pdata,4)) << 32);
#else /* LARGE_SMB_OFF_T */
      if (IVAL(pdata,4) != 0)	/* more than 32 bits? */
         return(ERROR_DOS(ERRDOS,ERRunknownlevel));
#endif /* LARGE_SMB_OFF_T */
      DEBUG(10,("call_trans2setfilepathinfo: Set end of file info for file %s to %.0f\n", fname, (double)size ));
      break;
    }

    case SMB_SET_FILE_DISPOSITION_INFO: /* Set delete on close for open file. */
    case SMB_FILE_DISPOSITION_INFORMATION:
    {
      BOOL delete_on_close;
      NTSTATUS status;

      if (total_data < 1)
        return ERROR_NT(STATUS_INVALID_PARAMETER);

      if (tran_call != TRANSACT2_SETFILEINFO)
        return(ERROR_DOS(ERRDOS,ERRunknownlevel));

      delete_on_close = (CVAL(pdata,0) ? True : False);
      status = fsp_set_delete_on_close(fsp, delete_on_close);
      if (! NT_STATUS_IS_OK(status))
        return ERROR_NT(status);

      break;
    }

		case SMB_FILE_POSITION_INFORMATION:
		{
			SMB_BIG_UINT position_information;
			
			if (total_data < 8)
				return ERROR_NT(STATUS_INVALID_PARAMETER);

			position_information = IVAL(pdata,0);
#ifdef LARGE_SMB_OFF_T
			position_information |= (((SMB_OFF_T)IVAL(pdata,4)) << 32);
#else /* LARGE_SMB_OFF_T */
			if (IVAL(pdata,4) != 0)	/* more than 32 bits? */
				return(ERROR_DOS(ERRDOS,ERRunknownlevel));
#endif /* LARGE_SMB_OFF_T */
			DEBUG(7, ("Set file position information for file %s to %.0f\n",
				  fname, (double)position_information));
			return ERROR_NT(STATUS_INVALID_LEVEL);
			break;
		}

		case SMB_FILE_MODE_INFORMATION:
		{
			uint32 new_open_mode;
			if (total_data < 4)
				return ERROR_NT(STATUS_INVALID_PARAMETER);

			new_open_mode = IVAL(pdata, 0);
			DEBUG(7, ("Intended new open_mode 0x%x\n",
				  new_open_mode));
			return ERROR_NT(STATUS_INVALID_LEVEL);
			break;
		}

		case SMB_FILE_RENAME_INFORMATION:
		{
			uint32 len;
			pstring newname;

			if (total_data <= 12)
				return ERROR_NT(STATUS_INVALID_PARAMETER);
			/* CVAL(pdata,0) -> BOOL overwrite */
			/* IVAL(pdata,4) -> uint32 root_fid */
			len = IVAL(pdata,8);
			srvstr_pull(inbuf, newname, pdata+12, sizeof(newname),
				    MIN(len, total_data-12), STR_NOALIGN);
			DEBUG(7, ("Intended rename of %s to %s\n",
				  fname, newname));
			return ERROR_NT(STATUS_INVALID_LEVEL);
		}

    default:
    {
      return(ERROR_DOS(ERRDOS,ERRunknownlevel));
    }
  }

  /* get some defaults (no modifications) if any info is zero or -1. */
  if (tvs.actime == (time_t)0 || tvs.actime == (time_t)-1)
    tvs.actime = sbuf.st_atime;

  if (tvs.modtime == (time_t)0 || tvs.modtime == (time_t)-1)
    tvs.modtime = sbuf.st_mtime;

  DEBUG(6,("actime: %s " , ctime(&tvs.actime)));
  DEBUG(6,("modtime: %s ", ctime(&tvs.modtime)));
  DEBUG(6,("size: %.0f ", (double)size));
  DEBUG(6,("mode: %x\n"  , mode));

  if(!((info_level == SMB_SET_FILE_END_OF_FILE_INFO)
       || (info_level == SMB_SET_FILE_ALLOCATION_INFO)
       || (info_level == SMB_FILE_ALLOCATION_INFORMATION))) {
    /*
     * Only do this test if we are not explicitly
     * changing the size of a file.
     */
    if (!size)
      size = sbuf.st_size;
  }

  /* Try and set the times, size and mode of this file -
     if they are different from the current values
   */
  if (sbuf.st_mtime != tvs.modtime || sbuf.st_atime != tvs.actime) {
    if(fsp != NULL) {
      /*
       * This was a setfileinfo on an open file.
       * NT does this a lot. It's actually pointless
       * setting the time here, as it will be overwritten
       * on the next write, so we save the request
       * away and will set it on file code. JRA.
       */

       if (tvs.modtime != (time_t)0 && tvs.modtime != (time_t)-1) {
         DEBUG(10,("call_trans2setfilepathinfo: setting pending modtime to %s\n",
            ctime(&tvs.modtime) ));
         fsp->pending_modtime = tvs.modtime;
       }

    } else {

      DEBUG(10,("call_trans2setfilepathinfo: setting utimes to modified values.\n"));

      if(file_utime(conn, fname, &tvs)!=0)
        return(UNIXERROR(ERRDOS,ERRnoaccess));
    }
  }

  /* check the mode isn't different, before changing it */
  if ((mode != 0) && (mode != dos_mode(conn, fname, &sbuf))) {

    DEBUG(10,("call_trans2setfilepathinfo: file %s : setting dos mode %x\n",
          fname, mode ));

    if(file_chmod(conn, fname, mode, NULL)) {
      DEBUG(2,("chmod of %s failed (%s)\n", fname, strerror(errno)));
      return(UNIXERROR(ERRDOS,ERRnoaccess));
    }
  }

  if(size != sbuf.st_size) {

    DEBUG(10,("call_trans2setfilepathinfo: file %s : setting new size to %.0f\n",
          fname, (double)size ));

    if (fd == -1) {
      fd = conn->vfs_ops->open(dos_to_unix_static(fname),O_RDWR,0);
      if (fd == -1)
        return(UNIXERROR(ERRDOS,ERRbadpath));
      set_filelen(fd, size); /* tpot vfs */
      conn->vfs_ops->close(NULL, fd);
    } else {
        set_filelen(fd, size); /* tpot vfs */
    }

    if(fsp)
      set_filelen_write_cache(fsp, size);
  }

  SSVAL(params,0,0);

  send_trans2_replies(conn, outbuf, bufsize, params, 2, *ppdata, 0);
  
  return(-1);
}

/****************************************************************************
 Reply to a TRANS2_MKDIR (make directory with extended attributes).
****************************************************************************/

static int call_trans2mkdir(connection_struct *conn,
			    char *inbuf, char *outbuf, int length, int bufsize,
				char **pparams, int total_params, char **ppdata, int total_data)
{
  char *params = *pparams;
  pstring directory;
  int ret = -1;
  BOOL bad_path = False;

  if (!CAN_WRITE(conn))
    return(ERROR_DOS(ERRSRV,ERRaccess));

	if (total_params < 4)
		return(ERROR_DOS(ERRDOS,ERRinvalidparam));

	srvstr_pull(inbuf, directory, params + 4, sizeof(directory), total_params - 4, STR_NOALIGN);

  DEBUG(3,("call_trans2mkdir : name = %s\n", directory));

	unix_to_dos(directory);

  unix_convert(directory,conn,0,&bad_path,NULL);
  if (check_name(directory,conn))
    ret = conn->vfs_ops->mkdir(dos_to_unix_static(directory),
			      unix_mode(conn,aDIR,directory));
  
  if(ret < 0)
    {
      DEBUG(5,("call_trans2mkdir error (%s)\n", strerror(errno)));
      set_bad_path_error(errno, bad_path);
      return(UNIXERROR(ERRDOS,ERRnoaccess));
    }

  /* Realloc the parameter and data sizes */
  if (!g_xrenew(char, pparams, 2)) {
    return(ERROR_DOS(ERRDOS,ERRnomem));
  }
  params = *pparams;

  SSVAL(params,0,0);

  send_trans2_replies(conn, outbuf, bufsize, params, 2, *ppdata, 0);
  
  return(-1);
}

/****************************************************************************
 Reply to a TRANS2_FINDNOTIFYFIRST (start monitoring a directory for changes).
  We don't actually do this - we just send a null response.
****************************************************************************/

static int call_trans2findnotifyfirst(connection_struct *conn,
					char *inbuf, char *outbuf, int length, int bufsize,
					char **pparams, int total_params, char **ppdata, int total_data)
{
  static uint16 fnf_handle = 257;
  char *params = *pparams;
  uint16 info_level;

	if (total_params < 6)
		return(ERROR_DOS(ERRDOS,ERRinvalidparam));

	info_level = SVAL(params,4);

  DEBUG(3,("call_trans2findnotifyfirst - info_level %d\n", info_level));

  switch (info_level) 
    {
    case 1:
    case 2:
      break;
    default:
      return(ERROR_DOS(ERRDOS,ERRunknownlevel));
    }

  /* Realloc the parameter and data sizes */
  if (!g_xrenew(char, pparams, 6)) {
    return(ERROR_DOS(ERRDOS,ERRnomem));
  }
  params = *pparams;

  SSVAL(params,0,fnf_handle);
  SSVAL(params,2,0); /* No changes */
  SSVAL(params,4,0); /* No EA errors */

  fnf_handle++;

  if(fnf_handle == 0)
    fnf_handle = 257;

  send_trans2_replies(conn, outbuf, bufsize, params, 6, *ppdata, 0);
  
  return(-1);
}

/****************************************************************************
  reply to a TRANS2_FINDNOTIFYNEXT (continue monitoring a directory for 
  changes). Currently this does nothing.
****************************************************************************/
static int call_trans2findnotifynext(connection_struct *conn,
				     char *inbuf, char *outbuf, 
				     int length, int bufsize,
				     char **pparams, int total_params, char **ppdata, int total_data)
{
  DEBUG(3,("call_trans2findnotifynext\n"));

  /* Realloc the parameter and data sizes */
  if (!g_xrenew(char, pparams, 4)) {
    return(ERROR_DOS(ERRDOS,ERRnomem));
  }

  SSVAL(*pparams,0,0); /* No changes */
  SSVAL(*pparams,2,0); /* No EA errors */

  send_trans2_replies(conn, outbuf, bufsize, *pparams, 4, *ppdata, 0);
  
  return(-1);
}

/****************************************************************************
 Reply to a TRANS2_GET_DFS_REFERRAL - Shirish Kalele <kalele@veritas.com>.
****************************************************************************/

static int call_trans2getdfsreferral(connection_struct *conn, char* inbuf,
				     char* outbuf, int length, int bufsize,
					char **pparams, int total_params, char **ppdata, int total_data)
{
  char *params = *pparams;
  enum remote_arch_types ra_type = get_remote_arch();
  BOOL NT_arch = ((ra_type == RA_WINNT) || (ra_type == RA_WIN2K));
  pstring pathname;
  int reply_size = 0;
  int max_referral_level;

  DEBUG(10,("call_trans2getdfsreferral\n"));

	if (total_params < 2)
		return(ERROR_DOS(ERRDOS,ERRinvalidparam));

	max_referral_level = SVAL(params,0);

  if(!lp_host_msdfs())
    return(ERROR_DOS(ERRDOS,ERRbadfunc));

  /* if pathname is in UNICODE, convert to DOS */
  /* NT always sends in UNICODE, may not set UNICODE flag */
  if(NT_arch || (SVAL(inbuf,smb_flg2) & FLAGS2_UNICODE_STRINGS))
    {
      unistr_to_dos(pathname, &params[2], sizeof(pathname));
      DEBUG(10,("UNICODE referral for %s\n",pathname));
    }
  else
    pstrcpy(pathname,&params[2]);

  if((reply_size = setup_dfs_referral(pathname,max_referral_level,ppdata)) < 0)
    return(ERROR_DOS(ERRDOS,ERRbadfile));
    
  SSVAL(outbuf,smb_flg2,SVAL(outbuf,smb_flg2) | FLAGS2_UNICODE_STRINGS | 
	FLAGS2_DFS_PATHNAMES);
  send_trans2_replies(conn, outbuf,bufsize,0,0,*ppdata,reply_size);

  return(-1);
}


/****************************************************************************
 Reply to a SMBfindclose (stop trans2 directory search).
****************************************************************************/

int reply_findclose(connection_struct *conn,
		    char *inbuf,char *outbuf,int length,int bufsize)
{
	int outsize = 0;
	int dptr_num=SVALS(inbuf,smb_vwv0);

	DEBUG(3,("reply_findclose, dptr_num = %d\n", dptr_num));

	dptr_close(&dptr_num);

	outsize = set_message(outbuf,0,0,True);

	DEBUG(3,("SMBfindclose dptr_num = %d\n", dptr_num));

	return(outsize);
}

/****************************************************************************
  reply to a SMBfindnclose (stop FINDNOTIFYFIRST directory search)
****************************************************************************/
int reply_findnclose(connection_struct *conn, 
		     char *inbuf,char *outbuf,int length,int bufsize)
{
	int outsize = 0;
	int dptr_num= -1;
	START_PROFILE(SMBfindnclose);
	
	dptr_num = SVAL(inbuf,smb_vwv0);

	DEBUG(3,("reply_findnclose, dptr_num = %d\n", dptr_num));

	/* We never give out valid handles for a 
	   findnotifyfirst - so any dptr_num is ok here. 
	   Just ignore it. */

	outsize = set_message(outbuf,0,0,True);

	DEBUG(3,("SMB_findnclose dptr_num = %d\n", dptr_num));

	END_PROFILE(SMBfindnclose);
	return(outsize);
}


/****************************************************************************
  reply to a SMBtranss2 - just ignore it!
****************************************************************************/
int reply_transs2(connection_struct *conn,
		  char *inbuf,char *outbuf,int length,int bufsize)
{
	START_PROFILE(SMBtranss2);
	DEBUG(4,("Ignoring transs2 of length %d\n",length));
	END_PROFILE(SMBtranss2);
	return(-1);
}


typedef int (* Trans2Handler) (connection_struct *conn,
			       char *inbuf, char *outbuf,
			       int length, int bufsize,
			       char **pparams, int total_params,
			       char **ppdata, int total_data);

#define X(a,b) {a, #a, b}

static const struct {
	unsigned int tran_call;
	const char *name;
	Trans2Handler fn;
} trans2fn_table[] = {
	X(TRANSACT2_OPEN, call_trans2open),
	X(TRANSACT2_FINDFIRST, call_trans2findfirst),
	X(TRANSACT2_FINDNEXT, call_trans2findnext),
	X(TRANSACT2_QFSINFO, call_trans2qfsinfo),
	X(TRANSACT2_SETFSINFO, call_trans2setfsinfo),
	X(TRANSACT2_QPATHINFO, call_trans2qfilepathinfo),
	X(TRANSACT2_QFILEINFO, call_trans2qfilepathinfo),
	X(TRANSACT2_SETPATHINFO, call_trans2setfilepathinfo),
	X(TRANSACT2_SETFILEINFO, call_trans2setfilepathinfo),
	X(TRANSACT2_FINDNOTIFYFIRST, call_trans2findnotifyfirst),
	X(TRANSACT2_FINDNOTIFYNEXT, call_trans2findnotifynext),
	X(TRANSACT2_MKDIR, call_trans2mkdir),
	X(TRANSACT2_GET_DFS_REFERRAL, call_trans2getdfsreferral),
	{ 0, NULL, NULL }
};
#undef X

/****************************************************************************
  reply to a SMBtrans2
****************************************************************************/
int reply_trans2(connection_struct *conn,
		 char *inbuf,char *outbuf,int length,int bufsize)
{
	int outsize = 0;
	unsigned int total_params = SVAL(inbuf, smb_tpscnt);
	unsigned int total_data =SVAL(inbuf, smb_tdscnt);
#if 0
	unsigned int max_param_reply = SVAL(inbuf, smb_mprcnt);
	unsigned int max_data_reply = SVAL(inbuf, smb_mdrcnt);
	unsigned int max_setup_fields = SVAL(inbuf, smb_msrcnt);
	BOOL close_tid = BITSETW(inbuf+smb_flags,0);
	BOOL no_final_response = BITSETW(inbuf+smb_flags,1);
	int32 timeout = IVALS(inbuf,smb_timeout);
#endif
	unsigned int suwcnt = SVAL(inbuf, smb_suwcnt);
	unsigned int tran_call = SVAL(inbuf, smb_setup0);
	char *params = NULL, *data = NULL;
	size_t num_params, num_params_sofar, num_data, num_data_sofar;
	size_t i;
	START_PROFILE(SMBtrans2);

	if(global_oplock_break && (tran_call == TRANSACT2_OPEN)) {
		/* Queue this open message as we are the process of an
		 * oplock break.  */

		DEBUG(2,("reply_trans2: queueing message trans2open due to being "));
		DEBUGADD(2,( "in oplock break state.\n"));

		push_oplock_pending_smb_message(inbuf, length);
		return -1;
	}
	
	if (IS_IPC(conn) && (tran_call != TRANSACT2_OPEN)) 
		return(ERROR_DOS(ERRSRV,ERRaccess));

	outsize = set_message(outbuf,0,0,True);

	/* All trans2 messages we handle have smb_sucnt == 1 - ensure this
	   is so as a sanity check */
	if (suwcnt != 1) {
		DEBUG(2,("Invalid smb_sucnt in trans2 call\n"));
		return(ERROR_DOS(ERRSRV,ERRerror));
	}
    
	/* Allocate the space for the maximum needed parameters and data */
	if (total_params > 0)
		params = g_malloc(total_params);
	if (total_data > 0)
		data = g_malloc(total_data);
  
	if ((total_params && !params)  || (total_data && !data)) {
		DEBUG(2,("Out of memory in reply_trans2\n"));
		safe_free(params);
		safe_free(data);
		return(ERROR_DOS(ERRDOS,ERRnomem));
	}

	/* Copy the param and data bytes sent with this request into
	   the params buffer */
	num_params = num_params_sofar = SVAL(inbuf,smb_pscnt);
	num_data = num_data_sofar = SVAL(inbuf, smb_dscnt);

	if (num_params > total_params || num_data > total_data)
		exit_server("invalid params in reply_trans2");

	if(params)
	{
		size_t psoff = SVAL(inbuf, smb_psoff);
		if (length < SMB_BASE_OFFSET ||
		    SIZE_OVERFLOW3(psoff, num_params, length - SMB_BASE_OFFSET))
			goto bad_param;
		memcpy(params, smb_base(inbuf) + psoff, num_params);
	}
	if(data)
	{
		size_t dsoff = SVAL(inbuf, smb_dsoff);
		if (length < SMB_BASE_OFFSET ||
		    SIZE_OVERFLOW3(dsoff, num_data, length - SMB_BASE_OFFSET))
			goto bad_param;
		memcpy(data, smb_base(inbuf) + dsoff, num_data);
	}

	if(num_data_sofar < total_data || num_params_sofar < total_params)  {
		/* We need to send an interim response then receive the rest
		   of the parameter/data bytes */
		outsize = set_message(outbuf,0,0,True);
		smbd_conn_send_smb(conn, outbuf);

		while (num_data_sofar < total_data || 
		       num_params_sofar < total_params) {
			BOOL ret;

			ret = receive_next_smb(inbuf,bufsize,SMB_SECONDARY_WAIT);
			
			if ((ret && 
			     (CVAL(inbuf, smb_com) != SMBtranss2)) || !ret) {
				outsize = set_message(outbuf,0,0,True);
				if(ret)
					DEBUG(0,("reply_trans2: Invalid secondary trans2 packet\n"));
				else
					DEBUG(0,("reply_trans2: %s in getting secondary trans2 response.\n",
						 (smb_read_error == READ_ERROR) ? "error" : "timeout" ));
				if(params)
					free(params);
				if(data)
					free(data);
				return(ERROR_DOS(ERRSRV,ERRerror));
			}
      
			/* Revise total_params and total_data in case
                           they have changed downwards */
			total_params = MIN(total_params, SVAL(inbuf, smb_tpscnt));
			total_data = MIN(total_data, SVAL(inbuf, smb_tdscnt));
			num_params_sofar += (num_params = SVAL(inbuf,smb_spscnt));
			num_data_sofar += ( num_data = SVAL(inbuf, smb_sdscnt));
			if (num_params_sofar > total_params || num_data_sofar > total_data)
				goto bad_param;

			if (num_params)
			{
				size_t displace = SVAL(inbuf, smb_spsdisp);
				size_t offset = SVAL(inbuf, smb_spsoff);
				if (PTR_OVERFLOW(params, displace) ||
				    SIZE_OVERFLOW3(displace, num_params,
						   total_params-1) ||
				    SIZE_OVERFLOW3(offset, num_params,
						   bufsize-1-SMB_BASE_OFFSET))
					goto bad_param;

				memcpy(params + displace,
				       smb_base(inbuf) + offset, num_params);
			}
			if (num_data)
			{
				size_t displace = SVAL(inbuf, smb_sdsdisp);
				size_t offset = SVAL(inbuf, smb_sdsoff);
				if (PTR_OVERFLOW(data, displace) ||
				    SIZE_OVERFLOW3(displace, num_data,
						   total_data-1) ||
				    SIZE_OVERFLOW3(offset, num_data,
						   bufsize-1-SMB_BASE_OFFSET))
					goto bad_param;
				memcpy(data + displace,
				       smb_base(inbuf) + offset, num_data);
			}
		}
	}
	
	if (Protocol >= PROTOCOL_NT1) {
		uint16 flg2 = SVAL(outbuf,smb_flg2);
		SSVAL(outbuf,smb_flg2,flg2 | 0x40); /* IS_LONG_NAME */
	}

	/* Now we must call the relevant TRANS2 function */
	for (i = 0; trans2fn_table[i].fn; i++)
	{
		if (trans2fn_table[i].tran_call == tran_call)
			break;
	}
	if (! trans2fn_table[i].fn)
	{
		/* Error in request */
		DEBUG(2,("Unknown request %d in trans2 call\n", tran_call));
		safe_free(params);
		safe_free(data);
		return ERROR_DOS(ERRSRV,ERRerror);
	}
	DEBUG(3, ("reply_trans2: Calling %s subcall\n",
		  trans2fn_table[i].name));
	outsize = trans2fn_table[i].fn(conn,
				       inbuf, outbuf, length, bufsize,
				       &params, total_params,
				       &data, total_data);
	
	/* As we do not know how many data packets will need to be
	   returned here the various call_trans2xxxx calls
	   must send their own. Thus a call_trans2xxx routine only
	   returns a value other than -1 when it wants to send
	   an error packet. 
	*/
	
	safe_free(params);
	safe_free(data);
	END_PROFILE(SMBtrans2);
	return outsize; /* If a correct response was needed the
			   call_trans2xxx calls have already sent
			   it. If outsize != -1 then it is returning */

bad_param:
	SAFE_FREE(params);
	SAFE_FREE(data);
	END_PROFILE(SMBtrans2);
	return ERROR_NT(NT_STATUS_INVALID_PARAMETER);
}
