/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   VFS initialisation and support functions
   Copyright (C) Tim Potter 1999
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: smbd_proto.h		(do not remove this line) */

#include "includes.h"
#include <errno.h>
#include "libsamba.h"
#include "tng_misc.h"
#include "smbd.h"
#include "vfs.h"


/****************************************************************************
  initialise default vfs hooks
****************************************************************************/
BOOL vfs_init_default(connection_struct *conn)
{
    DEBUG(3, ("Initialising default vfs hooks\n"));

    conn->vfs_ops = g_new1(struct vfs_ops);
    vfs_ops_setup_default(conn->vfs_ops);
    return True;
}


BOOL vfs_directory_exist(connection_struct *conn, char *dname,
                         SMB_STRUCT_STAT *st)
{
  SMB_STRUCT_STAT st2;
  BOOL ret;

  if (!st) st = &st2;

  if (vfs_stat(conn, dos_to_unix_static(dname), st) != 0) 
    return(False);

  ret = S_ISDIR(st->st_mode);
  if(!ret)
    errno = ENOTDIR;

  return ret;
}

/*******************************************************************
  check if a vfs file exists
********************************************************************/
BOOL vfs_file_exist(connection_struct *conn, const char *fname,SMB_STRUCT_STAT *sbuf)
{
  SMB_STRUCT_STAT st;
  if (!sbuf) sbuf = &st;
  
  if (vfs_stat(conn, fname, sbuf) != 0)
    return(False);

  return(S_ISREG(sbuf->st_mode));
}

/****************************************************************************
  write data to a fd on the vfs
****************************************************************************/
ssize_t vfs_write_data(files_struct *fsp,char *buffer,size_t N)
{
  size_t total=0;
  ssize_t ret;

  while (total < N)
  {
    ret = fsp->conn->vfs_ops->write(fsp->fd,buffer + total,N - total);

    if (ret == -1) return -1;
    if (ret == 0) return total;

    total += ret;
  }
  return (ssize_t)total;
}

/****************************************************************************
transfer some data between two file_struct's
****************************************************************************/
SMB_OFF_T vfs_transfer_file(int in_fd, files_struct *in_fsp, 
			    int out_fd, files_struct *out_fsp,
			    SMB_OFF_T n, char *header, int headlen, int align)
{
  static char *buf=NULL;  
  static int size=0;
  char *buf1,*abuf;
  SMB_OFF_T total = 0;

  DEBUG(4,("vfs_transfer_file n=%.0f  (head=%d) called\n",(double)n,headlen));

  /* Check we have at least somewhere to read from */

  SMB_ASSERT((in_fd != -1) || (in_fsp != NULL));

  if (size == 0) {
    size = lp_readsize();
    size = MAX(size,1024);
  }

  while (!buf && size>0) {
    buf = g_new(char, size + 8);
    if (!buf)
      size /= 2;
  }

  if (!buf) {
    DEBUG(0,("Can't allocate transfer buffer!\n"));
    exit(1);
  }

  abuf = buf + (align%8);

  if (header)
    n += headlen;

  while (n > 0)
  {
    int s = (int)MIN(n,(SMB_OFF_T)size);
    int ret,ret2=0;

    ret = 0;

    if (header && (headlen >= MIN(s,1024))) {
      buf1 = header;
      s = headlen;
      ret = headlen;
      headlen = 0;
      header = NULL;
    } else {
      buf1 = abuf;
    }

    if (header && headlen > 0)
    {
      ret = MIN(headlen,size);
      memcpy(buf1,header,ret);
      headlen -= ret;
      header += ret;
      if (headlen <= 0) header = NULL;
    }

    if (s > ret) {
      if (in_fsp)
	ret += in_fsp->conn->vfs_ops->read(in_fsp->fd,buf1+ret,s-ret);
      else
	ret += read_socket_data(in_fd,buf1+ret,s-ret);
    }

    if (ret > 0)
    {
	if (out_fsp) {
	    ret2 = out_fsp->conn->vfs_ops->write(out_fsp->fd,buf1,ret);
	} else {
	    ret2= (out_fd != -1) ? write_socket_data(out_fd,buf1,ret) : ret;
	}
    }

      if (ret2 > 0) total += ret2;
      /* if we can't write then dump excess data */
      if (ret2 != ret)
        vfs_transfer_file(in_fd, in_fsp, -1,NULL,n-(ret+headlen),NULL,0,0);

    if (ret <= 0 || ret2 != ret)
      return(total);
    n -= ret;
  }
  return(total);
}

/*******************************************************************
a vfs_readdir wrapper which just returns the file name
********************************************************************/
const char *vfs_readdirname(connection_struct *conn, void *p)
{
	struct dirent *ptr;
	char *dname;

	if (!p) return(NULL);
  
	ptr = (struct dirent *)conn->vfs_ops->readdir(p);
	if (!ptr) return(NULL);

	dname = ptr->d_name;

#ifdef NEXT2
	if (telldir(p) < 0)
		return(NULL);
#endif

#ifdef HAVE_BROKEN_READDIR
	/* using /usr/ucb/cc is BAD */
	dname = dname - 2;
#endif

	return(dname);
}


/*******************************************************************
 Getwd - takes a UNIX directory name and returns the name
 in dos format.
********************************************************************/

static
char *dos_getwd(char *unix_path)
{
	char *wd;
	wd = sys_getwd(unix_path);
	if (wd)
		unix_to_dos(wd);
	return wd;
}

/**
 * \brief Do chdir() on a conn.
 */
int vfs_ChDir(connection_struct *conn, const char *path)
{
  int res;
  static pstring LastDir="";

  path = dos_to_unix_static(path);

  if (strcsequal(path,"."))
    return(0);

  if (*path == '/' && strcsequal(LastDir,path))
    return(0);

  DEBUG(3, ("vfs_ChDir to [%s]\n",path));

  res = chdir(path);
  if (!res)
    pstrcpy(LastDir,path);
  return(res);
}

/* number of list structures for a caching GetWd function. */
#define MAX_GETWDCACHE (50)

static struct
{
  SMB_DEV_T dev; /* These *must* be compatible with the types returned in a stat() call. */
  SMB_INO_T inode; /* These *must* be compatible with the types returned in a stat() call. */
  char *dos_path; /* The pathname in DOS format. */
  BOOL valid;
} ino_list[MAX_GETWDCACHE];

extern BOOL use_getwd_cache;

/****************************************************************************
 Prompte a ptr (to make it recently used)
****************************************************************************/

static void array_promote(char *array,int elsize,int element)
{
  char *p;
  if (element == 0)
    return;

  p = g_new(char, elsize);

	if (!p)
	{
		DEBUG(5,("array_promote: malloc fail\n"));
		return;
	}

  memcpy(p,array + element * elsize, elsize);
  memmove(array + elsize,array,elsize*element);
  memcpy(array,p,elsize);
  free(p);
}

/*******************************************************************
 Return the absolute current directory path - given a UNIX pathname.
 Note that this path is returned in DOS format, not UNIX
 format.
********************************************************************/

char *vfs_GetWd(connection_struct *conn, char *path)
{
  pstring s;
  static BOOL getwd_cache_init = False;
  SMB_STRUCT_STAT st, st2;
  int i;

  *s = 0;

  if (!use_getwd_cache)
    return(dos_getwd(path));

  /* init the cache */
  if (!getwd_cache_init)
  {
    getwd_cache_init = True;
    for (i=0;i<MAX_GETWDCACHE;i++)
    {
      string_set(&ino_list[i].dos_path,"");
      ino_list[i].valid = False;
    }
  }

  /*  Get the inode of the current directory, if this doesn't work we're
      in trouble :-) */

  if (sys_stat(".",&st) == -1)
  {
    DEBUG(0,("Very strange, couldn't stat \".\" path=%s\n", path));
    return(dos_getwd(path));
  }


  for (i=0; i<MAX_GETWDCACHE; i++)
    if (ino_list[i].valid)
    {

      /*  If we have found an entry with a matching inode and dev number
          then find the inode number for the directory in the cached string.
          If this agrees with that returned by the stat for the current
          directory then all is o.k. (but make sure it is a directory all
          the same...) */

      if (st.st_ino == ino_list[i].inode &&
          st.st_dev == ino_list[i].dev)
      {
        if (vfs_stat(conn, dos_to_unix_static(ino_list[i].dos_path), &st2) == 0)
        {
          if (st.st_ino == st2.st_ino &&
              st.st_dev == st2.st_dev &&
              (st2.st_mode & S_IFMT) == S_IFDIR)
          {
            pstrcpy (path, ino_list[i].dos_path);

            /* promote it for future use */
            array_promote((char *)&ino_list[0],sizeof(ino_list[0]),i);
            return (path);
          }
          else
          {
            /*  If the inode is different then something's changed,
                scrub the entry and start from scratch. */
            ino_list[i].valid = False;
          }
        }
      }
    }


  /*  We don't have the information to hand so rely on traditional methods.
      The very slow getcwd, which spawns a process on some systems, or the
      not quite so bad getwd. */

  if (!dos_getwd(s))
  {
    DEBUG(0,("vfs_GetWd: dos_getwd call failed, errno %s\n",strerror(errno)));
    return (NULL);
  }

  pstrcpy(path,s);

  DEBUG(5,("vfs_GetWd %s, inode %.0f, dev %.0f\n",s,(double)st.st_ino,(double)st.st_dev));

  /* add it to the cache */
  i = MAX_GETWDCACHE - 1;
  string_set(&ino_list[i].dos_path,s);
  ino_list[i].dev = st.st_dev;
  ino_list[i].inode = st.st_ino;
  ino_list[i].valid = True;

  /* put it at the top of the list */
  array_promote((char *)&ino_list[0],sizeof(ino_list[0]),i);

  return (path);
}

/*******************************************************************
 Reduce a file name, removing .. elements and checking that 
 it is below dir in the heirachy. This uses vfs_GetWd() and so must be run
 on the system that has the referenced file system.
 Widelinks are allowed if widelinks is true.
********************************************************************/

BOOL reduce_name(connection_struct *conn, char *s, const char *dir, BOOL widelinks)
{
#ifndef REDUCE_PATHS
	return True;
#else
	pstring dir2;
	pstring wd;
	pstring base_name;
	pstring newname;
	char *p = NULL;
	BOOL relative = (*s != '/');

	DEBUG(3, ("reduce_name: [%s] for dir [%s]\n", s, dir));

	*dir2 = *wd = *base_name = *newname = 0;

	if (widelinks)
	{
		/* can't have a leading .. */
		if (strncmp(s, "..", 2) == 0 && (s[2] == 0 || s[2] == '/'))
		{
			DEBUG(3, ("Illegal file name? (%s)\n", s));
			return (False);
		}

		if (strlen(s) == 0)
			pstrcpy(s, "./");

		DEBUGADD(3, ("      after: [%s]\n", s));

		return (True);
	}

	/* remove any double slashes */
	all_string_sub_g(s, "//", "/", 0);

	pstrcpy(base_name, s);
	p = strrchr(base_name, '/');

	if (!p)
		return (True);

	if (!vfs_GetWd(conn, wd))
	{
		DEBUG(0, ("couldn't getwd for %s %s\n", s, dir));
		return (False);
	}

	if (vfs_ChDir(conn, dir) != 0)
	{
		DEBUG(0, ("couldn't chdir to %s\n", dir));
		return (False);
	}

	if (!vfs_GetWd(conn, dir2))
	{
		DEBUG(0, ("couldn't getwd for %s\n", dir));
		vfs_ChDir(conn, wd);
		return (False);
	}

	if (p && (p != base_name))
	{
		*p = 0;
		if (strcmp(p + 1, ".") == 0)
			p[1] = 0;
		if (strcmp(p + 1, "..") == 0)
			*p = '/';
	}

	if (vfs_ChDir(conn, base_name) != 0)
	{
		vfs_ChDir(conn, wd);
		DEBUG(3,
		      ("couldn't chdir for %s %s basename=%s\n", s, dir,
		       base_name));
		return (False);
	}

	if (!vfs_GetWd(conn, newname))
	{
		vfs_ChDir(conn, wd);
		DEBUG(2, ("couldn't get wd for %s %s\n", s, dir2));
		return (False);
	}

	if (p && (p != base_name))
	{
		pstrcat(newname, "/");
		pstrcat(newname, p + 1);
	}

	{
		size_t l = strlen(dir2);
		if (dir2[l - 1] == '/')
			l--;

		if (strncmp(newname, dir2, l) != 0)
		{
			vfs_ChDir(conn, wd);
			DEBUG(2,
			      ("Bad access attempt? s=%s dir=%s newname=%s l=%d\n",
			       s, dir2, newname, (int)l));
			return (False);
		}

		if (relative)
		{
			if (newname[l] == '/')
				pstrcpy(s, newname + l + 1);
			else
				pstrcpy(s, newname + l);
		}
		else
			pstrcpy(s, newname);
	}

	vfs_ChDir(conn, wd);

	if (strlen(s) == 0)
		pstrcpy(s, "./");

	DEBUGADD(3, ("      after: [%s]\n", s));

	return (True);
#endif
}


/*******************************************************************
reduce a file name, removing .. elements. 
********************************************************************/
void unix_clean_name(char *s)
{
	char *p = NULL;

	DEBUG(3, ("unix_clean_name [%s]\n", s));

	/* remove any double slashes */
	all_string_sub_g(s, "//", "/", 0);

	/* Remove leading ./ characters */
	if (strncmp(s, "./", 2) == 0)
	{
		trim_string(s, "./", NULL);
		if (*s == 0)
			pstrcpy(s, "./");
	}

	while ((p = strstr(s, "/../")) != NULL)
	{
		pstring s1;

		*p = 0;
		pstrcpy(s1, p + 3);

		if ((p = strrchr(s, '/')) != NULL)
			*p = 0;
		else
			*s = 0;
		pstrcat(s, s1);
	}

	trim_string(s, NULL, "/..");

	DEBUGADD(3, ("         after: [%s]\n", s));
}


void vfs_fill_ntinfo(connection_struct *conn, NTFileStatInfo *nti,
		     const char *dos_fname,
		     SMB_STRUCT_STAT *psbuf)
{
	time_t unixtime;
	BOOL fat_times = lp_dos_filetime_resolution(SNUM(conn));
	uint32 file_attr;

	nti->EndOfFile = psbuf->st_size;
	nti->AllocationSize = SMB_ROUNDUP(psbuf->st_size, 1024);

	unixtime = get_create_time(psbuf, lp_fake_dir_create_times(SNUM(conn)));
	if (fat_times)
		unixtime &= ~1;
	unix_to_nt_time(&nti->CreationTime, unixtime);

	unixtime = psbuf->st_atime;
	if (fat_times)
		unixtime &= ~1;
	unix_to_nt_time(&nti->LastAccessTime, unixtime);

	unixtime = psbuf->st_mtime;
	if (fat_times)
		unixtime &= ~1;
	unix_to_nt_time(&nti->LastWriteTime, unixtime);

	nti->ChangeTime = nti->LastWriteTime;

	file_attr = dos_mode(conn, dos_fname, psbuf);
	if (! file_attr)
		file_attr = FILE_ATTRIBUTE_NORMAL;
	nti->FileAttributes = file_attr;
}
