
/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Andrew Tridgell              1992-2001,
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-2001,
 *  Copyright (C) Sander Striker                    2001,
 *  Copyright (C) Mark Cave-Ayland <mca198@ecs.soton.ac.uk>  2001,
 *  Copyright (C) Elrond <elrond@samba-tng.org>     2001.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "includes.h"
#include "libsamba.h"
#include "rpc_parse.h"
#include "nterr.h"
#include "doserr.h"
#include "winerror.h"
#include "services.h"
#include "others.h"


extern SERVICE_TABLE srv_list[MAX_SERVICES];
extern int srv_index;

/****************************************************************************
 get_policy_svc_name 
****************************************************************************/
static BOOL get_policy_svc_name(struct policy_cache *cache, 
					  const POLICY_HND *hnd,
					  fstring name)
{
	char *dev;
	dev = (char *)get_policy_state_info(cache, hnd);

	if (dev != NULL)
	{
		fstrcpy(name, dev);
		DEBUG(5,("getting policy svc name=%s\n", name));
		return True;
	}

	DEBUG(3,("Error getting policy svc name\n"));
	return False;
}

/****************************************************************************
 set_policy_svc_name 
****************************************************************************/
static BOOL set_policy_svc_name(struct policy_cache *cache, POLICY_HND *hnd,
				fstring name)
{
	char *dev = strdup(name);
	if (dev != NULL)
	{
		if (set_policy_state(cache, hnd, NULL, (void*)dev))
		{
			DEBUG(3,("Service setting policy name=%s\n", name));
			return True;
		}
		free(dev);
		return True;
	}

	DEBUG(3,("Error setting policy name=%s\n", name));
	return False;
}


static SERVICE_TABLE *find_servinfo_by_name(const char *svc_name)
{
	int svc_num = -1;
	int i;

	/* Find the index into the service table */
	for (i=0; i<=srv_index; i++)
	{
		if (strequal(srv_list[i].name, svc_name))
		{
			svc_num=i;
			break;
		}
	}

	if (svc_num==-1)
	{
		DEBUG(3, ("find_servinfo_by_name: Service %s not found\n",
			  svc_name));
		return NULL;
	}

	return &srv_list[svc_num];
}


static SERVICE_TABLE *find_servinfo_by_hnd(const POLICY_HND *hnd)
{
	fstring svc_name;

	if (find_policy_by_hnd(get_global_hnd_cache(), hnd) == -1 ||
		!get_policy_svc_name(get_global_hnd_cache(), hnd, svc_name))
	{
		return NULL;
	}

	return find_servinfo_by_name(svc_name);
}

static void fill_svc_status(SVC_STATUS *svc_status,
			    const SERVICE_TABLE *servinfo)
{
	svc_status->svc_type = 0x10; /* process, non-interactive */

	/* Report running if possible? */
	if (servinfo->status==1) {
		svc_status->current_state = 4; /* running */
	} else {
		svc_status->current_state = 1; /* stopped */
	}

	/* stoppable, pausable, accepts_shutdown */
	svc_status->controls_accepted = 7;
}


/*******************************************************************
 _svc_close
 ********************************************************************/
uint32 _svc_close(POLICY_HND *pol)
{

	/* close the policy handle */
	if (!close_policy_hnd(get_global_hnd_cache(), pol))
	{
		return NT_STATUS_OBJECT_NAME_INVALID;
	}

	return ERROR_SUCCESS;
}

/*******************************************************************
 _svc_open_service
 ********************************************************************/
uint32 _svc_open_service(const POLICY_HND *scman_pol,
				 const UNISTR2* uni_svc_name,
				 uint32 des_access,
				 POLICY_HND *pol)
{
	fstring name;

	if (find_policy_by_hnd(get_global_hnd_cache(), scman_pol) == -1)
	{
		return NT_STATUS_INVALID_HANDLE;
	}

	unistr2_to_ascii(name, uni_svc_name, sizeof(name)-1);

	DEBUG(5,("svc_open_service: %s\n", name));

	if (! find_servinfo_by_name(name))
		return ERROR_SERVICE_DOES_NOT_EXIST;

	if (!open_policy_hnd_link(get_global_hnd_cache(),
		scman_pol, pol, des_access))
	{
		return NT_STATUS_TOO_MANY_SECRETS; /* ha ha very droll */
	}

	if (!set_policy_svc_name(get_global_hnd_cache(), pol, name))
	{
		return NT_STATUS_TOO_MANY_SECRETS; /* ha ha very droll */
	}

	return ERROR_SUCCESS;
}

/*******************************************************************
 _svc_stop_service
 ********************************************************************/
uint32 _svc_stop_service(const POLICY_HND *hnd, uint32 unk0,
			 SVC_STATUS *svc_status)
{
	SERVICE_TABLE *servinfo = find_servinfo_by_hnd(hnd);

	if (! servinfo)
	{
		return NT_STATUS_INVALID_HANDLE;
	}

	DEBUG(4,("stop_service: %s\n", servinfo->stop));

	/* stop the service here */
	if (smbrun(servinfo->stop, "/tmp/foo", False) != 0)
	{
		return NT_STATUS_ACCESS_DENIED;
	}

	/* Service now stopped */
	/* XXX - This status is lost, when this process dies ); */
	servinfo->status=0;

	return ERROR_SUCCESS;
}

/*******************************************************************
 _svc_start_service
 ********************************************************************/
uint32 _svc_start_service(const POLICY_HND *hnd,
				  uint32 argc,
				  uint32 argc2,
				  const UNISTR2 *argv)
{
	SERVICE_TABLE *servinfo = find_servinfo_by_hnd(hnd);

	if (! servinfo)
	{
		return NT_STATUS_INVALID_HANDLE;
	}

	DEBUG(4,("svc_start_service: %s\n", servinfo->start));

	/* start the service here */
	if (smbrun(servinfo->start, "/tmp/foo", False) != 0)
	{
		return NT_STATUS_ACCESS_DENIED;
	}

	/* Service now running */
	/* XXX - This status is lost, when this process dies ); */
	servinfo->status = 1;

	return ERROR_SUCCESS;
}

/*******************************************************************
 _svc_open_sc_man
 ********************************************************************/
uint32 _svc_open_sc_man(PRS_HANDLE *hnd, 
			const UNISTR2 *uni_srv_name,
			const UNISTR2 *uni_db_name,
			uint32 des_access,
			POLICY_HND *pol)
{
	fstring name;

	if (!open_policy_hnd(get_global_hnd_cache(),
		get_sec_ctx(), pol, des_access))
	{
		return NT_STATUS_TOO_MANY_SECRETS; /* ha ha very droll */
	}

	unistr2_to_ascii(name, uni_srv_name, sizeof(name)-1);

	DEBUG(5,("svc_open_sc_man: %s\n", name));
	/* lkcl XXXX do a check on the name, here */

	if (!set_policy_svc_name(get_global_hnd_cache(), pol, name))
	{
		return NT_STATUS_TOO_MANY_SECRETS; /* ha ha very droll */
	}

	return ERROR_SUCCESS;
}

/*******************************************************************
 _svc_enum_svcs_status
 ********************************************************************/
uint32 _svc_enum_svcs_status(const POLICY_HND *pol,
				     uint32 service_type,
				     uint32 service_state,
				     uint32 *buf_size,
				     ENUM_HND *resume_hnd,
				     ENUM_SRVC_STATUS *svcs,
				     uint32 *more_buf_size,
				     uint32 *num_svcs)
{
	uint32 dos_status = 0;
	int i = get_enum_hnd(resume_hnd);
	uint32 local_resume_hnd = 0;
	uint32 local_buf_size = 0;

	*more_buf_size = 0x10000;
	*num_svcs = 0;

	if (find_policy_by_hnd(get_global_hnd_cache(), pol) == -1)
	{
		return NT_STATUS_INVALID_HANDLE;
	}

	for (i = 0; i <= srv_index; i++)
	{
		ENUM_SRVC_STATUS *svc = NULL;
		fstring svc_name, svc_fname;
		int len;

		fstrcpy(svc_name, srv_list[i].name);
		len = strlen(srv_list[i].name);
		fstrcpy(svc_fname, srv_list[i].fullname);

		local_buf_size += (len+1) * 2;
		local_buf_size += 9 * sizeof(uint32);

		DEBUG(10,("r_buf_size: %d q_buf_size: %d\n",
			   local_buf_size, *buf_size));

		if (local_buf_size >= *more_buf_size)
		{
			local_resume_hnd = i;
			break;
		}

		if (local_buf_size > *buf_size)
		{
			dos_status = ERRmoredata;
			break;
		}

		(*num_svcs)++;
		if (*num_svcs > MAX_SERVICES)
		{
			dos_status = ERRnomem;
			*num_svcs = 0;
			break;
		}

		svc = &svcs[(*num_svcs) - 1];
		ZERO_STRUCTP(svc);

		make_unistr(&svc->uni_srvc_name, svc_name);
		make_unistr(&svc->uni_disp_name, svc_fname);

		fill_svc_status(&svc->status, &srv_list[i]);

		DEBUG(10,("show service: %s\n", svc_name));
	}

	/*
	 * check for finished condition: no resume handle and last buffer fits
	 */

	if (local_resume_hnd == 0 && local_buf_size <= *buf_size)
	{
		/* this indicates, along with resume_hnd of 0, an end. */
		*more_buf_size = 0;
	}

	make_enum_hnd(resume_hnd, local_resume_hnd);

	return dos_status;
}

/*******************************************************************
 _svc_query_disp_name
 ********************************************************************/
uint32 _svc_query_disp_name(const POLICY_HND *scman_pol,
				    const UNISTR2 *uni_svc_name,
				    uint32 buf_size,
				    UNISTR2 *uni_disp_name,
				    uint32 *pbuf_size)
{
	if (find_policy_by_hnd(get_global_hnd_cache(), scman_pol) == -1)
	{
		return NT_STATUS_INVALID_HANDLE;
	}

	/* for now display name = service name */
	copy_unistr2(uni_disp_name, uni_svc_name);
	/* and thus the length of the strings is the same */
	*pbuf_size = buf_size;

	return ERROR_SUCCESS;
}

/*******************************************************************
 ********************************************************************/
uint32 _svc_lock(const POLICY_HND *scman_hnd, POLICY_HND *hnd)
{
	if (find_policy_by_hnd(get_global_hnd_cache(), scman_hnd) == -1)
	{
		return NT_STATUS_INVALID_HANDLE;
	}

	if (!open_policy_hnd_link(get_global_hnd_cache(),
				  scman_hnd, hnd, 0x0))
	{
		return NT_STATUS_TOO_MANY_SECRETS; /* ha ha very droll */
	}

	policy_hnd_set_name(get_global_hnd_cache(), hnd, "svc_lock");

	return ERROR_SUCCESS;
}

uint32 _svc_query_svc_config(const POLICY_HND *hnd, QUERY_SERVICE_CONFIG *cfg,
			     uint32 *buf_size)
{
	SERVICE_TABLE *servinfo = find_servinfo_by_hnd(hnd);

	if (! servinfo)
	{
		return NT_STATUS_INVALID_HANDLE;
	}

	make_svc_query_svc_cfg(cfg,
			    0x10, /* process, non-interactive */
			    3,
			    1,
			    servinfo->start,
			    "load_order_grp",
			    0,
			    "dependencies",
			    "LocalSystem",
			    servinfo->fullname);

	return 0x0;
}

uint32 _svc_query_svc_status(const POLICY_HND *hnd, SVC_STATUS *status)
{
	SERVICE_TABLE *servinfo = find_servinfo_by_hnd(hnd);

	if (! servinfo)
		return NT_STATUS_INVALID_HANDLE;

	if (! status)
		return NT_STATUS_INVALID_PARAMETER;

	fill_svc_status(status, servinfo);

	return ERROR_SUCCESS;
}
